/**
 * Copyright 2019-2020 Huawei Technologies Co., Ltd
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <gtest/gtest.h>

#include "common/formats/format_transfers/format_transfer_fracz_nchw.h"

#include "common/formats/format_transfers/format_transfer.h"
#include "common/fp16_t.h"

namespace ge {
namespace formats {
class UtestFormatTransferFraczNchw : public testing::Test {
 protected:
  void SetUp() {}
  void TearDown() {}
};

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_invalid_data_type) {
  uint8_t data[16 * 1 * 32 * 16] = {1};

  TransArgs args{data, FORMAT_FRACTAL_Z, FORMAT_NCHW, {16, 1, 32, 16}, {1, 1, 4, 4}, DT_UNDEFINED};
  TransResult result;

  FormatTransferFracZNchw transfer;
  EXPECT_EQ(transfer.TransFormat(args, result), PARAM_INVALID);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_Invalid_src_format_reserved) {
  float data[16 * 1 * 16 * 16] = {1};

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_RESERVED, FORMAT_NCHW, {16, 1, 16, 16}, {1, 1, 4, 4}, DT_FLOAT};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), PARAM_INVALID);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_invalid_dst_format_reserved) {
  float data[16 * 1 * 16 * 16] = {1};

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_FRACTAL_Z, FORMAT_RESERVED, {16, 1, 16, 16}, {1, 1, 4, 4}, DT_FLOAT};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), PARAM_INVALID);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_invalid_src_shape) {
  float data[16 * 1 * 16 * 16] = {1};

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_FRACTAL_Z, FORMAT_NCHW, {16, 1, 1, 16, 16}, {1, 1, 4, 4}, DT_FLOAT};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), PARAM_INVALID);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_invalid_src_shape2) {
  float data[16 * 1 * 16 * 16] = {1};

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_FRACTAL_Z, FORMAT_NCHW, {16, 1, -16, 16}, {1, 1, 4, 4}, DT_FLOAT};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), PARAM_INVALID);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_invalid_dst_shape) {
  float data[16 * 1 * 16 * 16] = {1};

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_FRACTAL_Z, FORMAT_NCHW, {16, 1, 16, 16}, {1, 4, 4}, DT_FLOAT};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), PARAM_INVALID);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_invalid_dst_shape2) {
  float data[16 * 1 * 16 * 16] = {1};

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_FRACTAL_Z, FORMAT_NCHW, {16, 1, 16, 16}, {1, -1, 4, 4}, DT_FLOAT};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), PARAM_INVALID);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_invalid_src_dst_shape_relation1) {
  float data[16 * 1 * 16 * 16] = {1};

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_FRACTAL_Z, FORMAT_NCHW, {16, 1, 16, 16}, {1, 17, 4, 4}, DT_FLOAT};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), PARAM_INVALID);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_invalid_src_dst_shape_relation2) {
  float data[16 * 1 * 16 * 16] = {1};

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_FRACTAL_Z, FORMAT_NCHW, {16, 1, 16, 16}, {17, 1, 4, 4}, DT_FLOAT};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), PARAM_INVALID);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_fp16_success_lt_cube) {
  uint16_t data_4d[1 * 1 * 16 * 16] = {
      13040, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,     0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  };
  uint16_t data[1 * 1 * 1 * 1] = {
      13040,
  };

  TransArgs args{
      reinterpret_cast<uint8_t *>(data_4d), FORMAT_FRACTAL_Z, FORMAT_NCHW, {1, 1, 16, 16}, {1, 1, 1, 1}, DT_FLOAT16};
  TransResult result;

  FormatTransferFracZNchw transfer;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(data));
  for (int i = 0; i < sizeof(data) / sizeof(data[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<uint16_t *>(result.data.get()))[i], data[i]);
  }
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_fp16_success_eq_cube) {
  uint16_t data_4d[4 * 1 * 16 * 16] = {
      15321, 14501, 14483, 13936, 15043, 13673, 12307, 13885, 4847,  15263, 13716, 15066, 14419, 15319, 9266,  14930,
      10867, 15258, 15241, 11237, 15191, 15087, 14477, 10812, 15069, 9464,  7729,  14554, 15221, 14669, 14987, 14285,
      15201, 12849, 15322, 15137, 15088, 14748, 13775, 11884, 15230, 14420, 13994, 15304, 14805, 13879, 14491, 14404,
      14787, 15313, 9650,  15062, 12020, 13837, 14360, 14635, 12925, 14525, 14668, 15016, 14857, 13496, 15269, 13929,
      15157, 12904, 14142, 14221, 12612, 15184, 14286, 14212, 14339, 13084, 13186, 14730, 11482, 14712, 14820, 14965,
      13685, 8037,  14072, 13687, 15179, 14509, 14434, 15260, 14824, 13738, 13709, 15131, 11415, 14541, 13876, 14083,
      14844, 15291, 14663, 14954, 14046, 12994, 12355, 12498, 14382, 14003, 13924, 13914, 15342, 13374, 14012, 13714,
      15217, 14495, 9197,  13551, 13981, 11611, 14250, 12013, 14498, 14992, 13532, 11597, 15354, 15321, 14483, 14470,
      14032, 11620, 14527, 15108, 15105, 14942, 13826, 9073,  14412, 14109, 14718, 13055, 13486, 15285, 14326, 13160,
      13669, 13807, 14516, 13405, 15020, 14336, 13866, 14850, 14987, 11233, 13271, 15030, 13129, 10303, 15201, 12672,
      15010, 11951, 14837, 13946, 10496, 14971, 14106, 13516, 12637, 14084, 13097, 12852, 14811, 12480, 14914, 15073,
      11546, 13817, 14617, 13540, 12857, 14473, 13375, 14729, 13752, 14813, 14286, 12731, 14300, 10109, 14881, 13741,
      14706, 13214, 14164, 14549, 13435, 14038, 15090, 15058, 14382, 14118, 14606, 13358, 14086, 14881, 14289, 14537,
      15347, 10719, 11996, 13391, 14272, 15135, 14728, 15134, 11692, 12862, 12646, 14786, 14944, 14416, 15092, 13003,
      8681,  15041, 14187, 11620, 11693, 14667, 14583, 14792, 14642, 14493, 13267, 13506, 14712, 15291, 15079, 15337,
      14735, 11768, 12961, 14219, 14846, 14494, 14729, 13060, 14464, 15269, 14425, 15333, 13674, 11749, 14606, 11590,
      12886, 13850, 14716, 13941, 14643, 14744, 15163, 15141, 13983, 14858, 12830, 13025, 14988, 14285, 14740, 13352,
      11972, 12316, 13555, 15197, 13984, 15069, 13775, 14626, 12878, 14771, 12878, 15137, 14383, 12753, 13520, 14902,
      13924, 11361, 10655, 9100,  13999, 15220, 15166, 13336, 12293, 10088, 11778, 15081, 14045, 14406, 13509, 9504,
      14371, 13225, 15034, 14703, 14686, 14808, 13976, 9695,  14616, 14631, 14859, 13826, 14998, 15101, 13471, 14673,
      8286,  14988, 13555, 14705, 15105, 14937, 10691, 15195, 7636,  13144, 15128, 14778, 15130, 13607, 13974, 14622,
      13249, 14119, 12877, 14659, 14075, 13603, 13973, 14976, 14296, 11888, 14241, 13299, 15292, 12780, 14819, 15081,
      14631, 14631, 13379, 14471, 14757, 14252, 14349, 15024, 13475, 15320, 13755, 14478, 15228, 13543, 9409,  11631,
      15184, 14076, 15043, 14265, 15241, 11449, 15078, 14865, 12988, 14643, 14339, 14504, 6814,  13461, 15347, 15053,
      13509, 12315, 12104, 14685, 15141, 15253, 14752, 13715, 13940, 13548, 14359, 12752, 7951,  14404, 13046, 14376,
      15334, 14907, 15135, 13383, 13520, 13990, 13881, 12601, 14975, 13665, 13909, 13901, 12795, 14043, 15240, 15324,
      14427, 14493, 14279, 13421, 14165, 10574, 15358, 13503, 12607, 13488, 13919, 14707, 12374, 15050, 14569, 14277,
      14198, 12343, 13527, 14970, 14988, 14855, 12149, 15119, 13368, 15227, 13476, 14857, 13668, 14744, 13224, 13701,
      13270, 14597, 13880, 13364, 12825, 13980, 10073, 15330, 12665, 14870, 14886, 14702, 14577, 15333, 14759, 15003,
      13233, 11785, 15345, 10633, 15260, 14557, 15030, 9786,  14747, 11463, 11585, 14625, 13729, 14911, 15189, 14083,
      13764, 14552, 10020, 12569, 14805, 11953, 15147, 14240, 13118, 12816, 14696, 13327, 13627, 15095, 14043, 13298,
      14615, 15027, 13456, 14340, 12897, 14207, 13463, 14715, 14580, 13762, 14954, 11439, 15224, 13949, 14756, 14388,
      10292, 12758, 14301, 14076, 15130, 15224, 15297, 13788, 14156, 14302, 14879, 13440, 13880, 11505, 13511, 14961,
      14366, 14674, 14709, 14618, 15082, 9169,  15350, 13438, 12408, 14954, 13990, 15256, 12514, 13144, 13893, 11435,
      13429, 14448, 14499, 12315, 15284, 14809, 13124, 9087,  14867, 14483, 12832, 13040, 14751, 14554, 14490, 14826,
      14355, 15221, 15122, 11546, 14906, 14486, 15225, 10278, 13774, 14724, 11395, 11465, 15244, 13935, 15322, 9695,
      14966, 14081, 13138, 13616, 13736, 12922, 13494, 13992, 15297, 12437, 11768, 14151, 15332, 13332, 14841, 15236,
      14435, 14608, 13683, 15143, 14395, 15319, 14938, 15110, 15236, 11931, 12694, 11947, 14570, 14369, 15144, 14535,
      15048, 14735, 15217, 12554, 13548, 14092, 14719, 15086, 14539, 13947, 11368, 14496, 13421, 12350, 13215, 12610,
      15316, 9328,  14670, 15166, 14062, 15338, 14535, 12646, 13907, 13461, 14725, 14683, 13344, 13222, 14474, 14745,
      14084, 14972, 15212, 12449, 14218, 11659, 15286, 12451, 9711,  12833, 13730, 12460, 13748, 11887, 9794,  15355,
      13291, 11506, 14543, 13526, 12386, 14689, 13129, 12716, 14726, 14135, 12690, 15035, 13825, 15024, 14735, 13333,
      14906, 14092, 14458, 14576, 11812, 14960, 14741, 13067, 15348, 14798, 15101, 9753,  14040, 14780, 12358, 15206,
      15144, 13346, 13120, 13470, 14845, 13907, 15229, 13879, 12299, 14952, 15061, 14391, 15084, 14760, 13896, 14320,
      15338, 14404, 15097, 14815, 13002, 13366, 15333, 14624, 12032, 11321, 14102, 12735, 13433, 12332, 13779, 15070,
      15248, 11753, 12877, 13162, 13740, 14906, 14746, 14618, 13075, 13339, 15016, 14879, 15117, 12455, 14950, 14084,
      12651, 15233, 12770, 12456, 12289, 15170, 12421, 14159, 13430, 14361, 14422, 13700, 14693, 14835, 14234, 14830,
      15197, 14830, 13318, 15217, 13361, 9601,  14396, 14383, 14724, 12701, 12424, 15183, 10020, 13572, 15196, 14493,
      15259, 14687, 13772, 15168, 14343, 12708, 15149, 11266, 13343, 11713, 13761, 14513, 15063, 9882,  14773, 15235,
      13125, 14347, 14993, 14868, 14425, 14803, 14950, 15123, 14736, 11347, 13960, 12695, 13338, 10399, 10862, 15249,
      14968, 15201, 11690, 14989, 14946, 15144, 14526, 14773, 15027, 8041,  12886, 15298, 14655, 13168, 14068, 14810,
      10837, 14681, 12314, 12583, 14510, 14758, 14750, 14545, 15168, 14825, 14803, 11993, 13811, 15269, 13376, 14697,
      13312, 14764, 13857, 14524, 14852, 14172, 13180, 13348, 14413, 11386, 14733, 14822, 14770, 12511, 13369, 14240,
      14831, 15141, 13868, 13562, 15162, 14840, 15129, 13387, 14974, 13710, 14834, 14922, 13966, 13712, 15194, 13610,
      14951, 13562, 14678, 13369, 13551, 12896, 14704, 10316, 14428, 14880, 15264, 13866, 12114, 12500, 13385, 10455,
      14679, 14471, 12361, 15320, 14386, 13844, 14864, 11922, 10933, 13356, 14782, 14598, 13236, 14829, 15231, 14350,
      14550, 14964, 14845, 11692, 15325, 14383, 14090, 14873, 12926, 15343, 13661, 14803, 12968, 13741, 14004, 14699,
      12099, 13531, 14968, 14644, 15170, 15256, 12070, 14853, 11500, 15343, 14004, 12168, 11066, 13855, 10086, 14083,
      14107, 14870, 8556,  12951, 10031, 10387, 14534, 11951, 14015, 10938, 14463, 15349, 14973, 12431, 15119, 13704,
      14527, 12773, 14853, 15148, 13012, 14921, 14187, 14027, 10103, 15266, 13372, 12727, 15100, 14009, 13683, 15234,
      12717, 14650, 14601, 14453, 13799, 14819, 13656, 12366, 15349, 11187, 9283,  15028, 14000, 14023, 14979, 14105,
      14899, 13435, 11447, 11158, 14662, 12884, 15025, 9574,  15064, 15027, 13888, 15105, 11581, 13353, 13919, 14064,
      12278, 13483, 13835, 13839, 9470,  15330, 14400, 13076, 11430, 13613, 14875, 14824, 14792, 12630, 14557, 14563,
      13672, 14934, 15327, 13292, 15152, 12363, 12970, 14119, 14671, 15097, 14815, 12299, 15154, 15122, 14722, 13551,
  };
  uint16_t data[16 * 16 * 2 * 2] = {
      15321, 12886, 10292, 15259, 14501, 13850, 12758, 14687, 14483, 14716, 14301, 13772, 13936, 13941, 14076, 15168,
      15043, 14643, 15130, 14343, 13673, 14744, 15224, 12708, 12307, 15163, 15297, 15149, 13885, 15141, 13788, 11266,
      4847,  13983, 14156, 13343, 15263, 14858, 14302, 11713, 13716, 12830, 14879, 13761, 15066, 13025, 13440, 14513,
      14419, 14988, 13880, 15063, 15319, 14285, 11505, 9882,  9266,  14740, 13511, 14773, 14930, 13352, 14961, 15235,
      10867, 11972, 14366, 13125, 15258, 12316, 14674, 14347, 15241, 13555, 14709, 14993, 11237, 15197, 14618, 14868,
      15191, 13984, 15082, 14425, 15087, 15069, 9169,  14803, 14477, 13775, 15350, 14950, 10812, 14626, 13438, 15123,
      15069, 12878, 12408, 14736, 9464,  14771, 14954, 11347, 7729,  12878, 13990, 13960, 14554, 15137, 15256, 12695,
      15221, 14383, 12514, 13338, 14669, 12753, 13144, 10399, 14987, 13520, 13893, 10862, 14285, 14902, 11435, 15249,
      15201, 13924, 13429, 14968, 12849, 11361, 14448, 15201, 15322, 10655, 14499, 11690, 15137, 9100,  12315, 14989,
      15088, 13999, 15284, 14946, 14748, 15220, 14809, 15144, 13775, 15166, 13124, 14526, 11884, 13336, 9087,  14773,
      15230, 12293, 14867, 15027, 14420, 10088, 14483, 8041,  13994, 11778, 12832, 12886, 15304, 15081, 13040, 15298,
      14805, 14045, 14751, 14655, 13879, 14406, 14554, 13168, 14491, 13509, 14490, 14068, 14404, 9504,  14826, 14810,
      14787, 14371, 14355, 10837, 15313, 13225, 15221, 14681, 9650,  15034, 15122, 12314, 15062, 14703, 11546, 12583,
      12020, 14686, 14906, 14510, 13837, 14808, 14486, 14758, 14360, 13976, 15225, 14750, 14635, 9695,  10278, 14545,
      12925, 14616, 13774, 15168, 14525, 14631, 14724, 14825, 14668, 14859, 11395, 14803, 15016, 13826, 11465, 11993,
      14857, 14998, 15244, 13811, 13496, 15101, 13935, 15269, 15269, 13471, 15322, 13376, 13929, 14673, 9695,  14697,
      15157, 8286,  14966, 13312, 12904, 14988, 14081, 14764, 14142, 13555, 13138, 13857, 14221, 14705, 13616, 14524,
      12612, 15105, 13736, 14852, 15184, 14937, 12922, 14172, 14286, 10691, 13494, 13180, 14212, 15195, 13992, 13348,
      14339, 7636,  15297, 14413, 13084, 13144, 12437, 11386, 13186, 15128, 11768, 14733, 14730, 14778, 14151, 14822,
      11482, 15130, 15332, 14770, 14712, 13607, 13332, 12511, 14820, 13974, 14841, 13369, 14965, 14622, 15236, 14240,
      13685, 13249, 14435, 14831, 8037,  14119, 14608, 15141, 14072, 12877, 13683, 13868, 13687, 14659, 15143, 13562,
      15179, 14075, 14395, 15162, 14509, 13603, 15319, 14840, 14434, 13973, 14938, 15129, 15260, 14976, 15110, 13387,
      14824, 14296, 15236, 14974, 13738, 11888, 11931, 13710, 13709, 14241, 12694, 14834, 15131, 13299, 11947, 14922,
      11415, 15292, 14570, 13966, 14541, 12780, 14369, 13712, 13876, 14819, 15144, 15194, 14083, 15081, 14535, 13610,
      14844, 14631, 15048, 14951, 15291, 14631, 14735, 13562, 14663, 13379, 15217, 14678, 14954, 14471, 12554, 13369,
      14046, 14757, 13548, 13551, 12994, 14252, 14092, 12896, 12355, 14349, 14719, 14704, 12498, 15024, 15086, 10316,
      14382, 13475, 14539, 14428, 14003, 15320, 13947, 14880, 13924, 13755, 11368, 15264, 13914, 14478, 14496, 13866,
      15342, 15228, 13421, 12114, 13374, 13543, 12350, 12500, 14012, 9409,  13215, 13385, 13714, 11631, 12610, 10455,
      15217, 15184, 15316, 14679, 14495, 14076, 9328,  14471, 9197,  15043, 14670, 12361, 13551, 14265, 15166, 15320,
      13981, 15241, 14062, 14386, 11611, 11449, 15338, 13844, 14250, 15078, 14535, 14864, 12013, 14865, 12646, 11922,
      14498, 12988, 13907, 10933, 14992, 14643, 13461, 13356, 13532, 14339, 14725, 14782, 11597, 14504, 14683, 14598,
      15354, 6814,  13344, 13236, 15321, 13461, 13222, 14829, 14483, 15347, 14474, 15231, 14470, 15053, 14745, 14350,
      14032, 13509, 14084, 14550, 11620, 12315, 14972, 14964, 14527, 12104, 15212, 14845, 15108, 14685, 12449, 11692,
      15105, 15141, 14218, 15325, 14942, 15253, 11659, 14383, 13826, 14752, 15286, 14090, 9073,  13715, 12451, 14873,
      14412, 13940, 9711,  12926, 14109, 13548, 12833, 15343, 14718, 14359, 13730, 13661, 13055, 12752, 12460, 14803,
      13486, 7951,  13748, 12968, 15285, 14404, 11887, 13741, 14326, 13046, 9794,  14004, 13160, 14376, 15355, 14699,
      13669, 15334, 13291, 12099, 13807, 14907, 11506, 13531, 14516, 15135, 14543, 14968, 13405, 13383, 13526, 14644,
      15020, 13520, 12386, 15170, 14336, 13990, 14689, 15256, 13866, 13881, 13129, 12070, 14850, 12601, 12716, 14853,
      14987, 14975, 14726, 11500, 11233, 13665, 14135, 15343, 13271, 13909, 12690, 14004, 15030, 13901, 15035, 12168,
      13129, 12795, 13825, 11066, 10303, 14043, 15024, 13855, 15201, 15240, 14735, 10086, 12672, 15324, 13333, 14083,
      15010, 14427, 14906, 14107, 11951, 14493, 14092, 14870, 14837, 14279, 14458, 8556,  13946, 13421, 14576, 12951,
      10496, 14165, 11812, 10031, 14971, 10574, 14960, 10387, 14106, 15358, 14741, 14534, 13516, 13503, 13067, 11951,
      12637, 12607, 15348, 14015, 14084, 13488, 14798, 10938, 13097, 13919, 15101, 14463, 12852, 14707, 9753,  15349,
      14811, 12374, 14040, 14973, 12480, 15050, 14780, 12431, 14914, 14569, 12358, 15119, 15073, 14277, 15206, 13704,
      11546, 14198, 15144, 14527, 13817, 12343, 13346, 12773, 14617, 13527, 13120, 14853, 13540, 14970, 13470, 15148,
      12857, 14988, 14845, 13012, 14473, 14855, 13907, 14921, 13375, 12149, 15229, 14187, 14729, 15119, 13879, 14027,
      13752, 13368, 12299, 10103, 14813, 15227, 14952, 15266, 14286, 13476, 15061, 13372, 12731, 14857, 14391, 12727,
      14300, 13668, 15084, 15100, 10109, 14744, 14760, 14009, 14881, 13224, 13896, 13683, 13741, 13701, 14320, 15234,
      14706, 13270, 15338, 12717, 13214, 14597, 14404, 14650, 14164, 13880, 15097, 14601, 14549, 13364, 14815, 14453,
      13435, 12825, 13002, 13799, 14038, 13980, 13366, 14819, 15090, 10073, 15333, 13656, 15058, 15330, 14624, 12366,
      14382, 12665, 12032, 15349, 14118, 14870, 11321, 11187, 14606, 14886, 14102, 9283,  13358, 14702, 12735, 15028,
      14086, 14577, 13433, 14000, 14881, 15333, 12332, 14023, 14289, 14759, 13779, 14979, 14537, 15003, 15070, 14105,
      15347, 13233, 15248, 14899, 10719, 11785, 11753, 13435, 11996, 15345, 12877, 11447, 13391, 10633, 13162, 11158,
      14272, 15260, 13740, 14662, 15135, 14557, 14906, 12884, 14728, 15030, 14746, 15025, 15134, 9786,  14618, 9574,
      11692, 14747, 13075, 15064, 12862, 11463, 13339, 15027, 12646, 11585, 15016, 13888, 14786, 14625, 14879, 15105,
      14944, 13729, 15117, 11581, 14416, 14911, 12455, 13353, 15092, 15189, 14950, 13919, 13003, 14083, 14084, 14064,
      8681,  13764, 12651, 12278, 15041, 14552, 15233, 13483, 14187, 10020, 12770, 13835, 11620, 12569, 12456, 13839,
      11693, 14805, 12289, 9470,  14667, 11953, 15170, 15330, 14583, 15147, 12421, 14400, 14792, 14240, 14159, 13076,
      14642, 13118, 13430, 11430, 14493, 12816, 14361, 13613, 13267, 14696, 14422, 14875, 13506, 13327, 13700, 14824,
      14712, 13627, 14693, 14792, 15291, 15095, 14835, 12630, 15079, 14043, 14234, 14557, 15337, 13298, 14830, 14563,
      14735, 14615, 15197, 13672, 11768, 15027, 14830, 14934, 12961, 13456, 13318, 15327, 14219, 14340, 15217, 13292,
      14846, 12897, 13361, 15152, 14494, 14207, 9601,  12363, 14729, 13463, 14396, 12970, 13060, 14715, 14383, 14119,
      14464, 14580, 14724, 14671, 15269, 13762, 12701, 15097, 14425, 14954, 12424, 14815, 15333, 11439, 15183, 12299,
      13674, 15224, 10020, 15154, 11749, 13949, 13572, 15122, 14606, 14756, 15196, 14722, 11590, 14388, 14493, 13551,
  };

  TransArgs args{
      reinterpret_cast<uint8_t *>(data_4d), FORMAT_FRACTAL_Z, FORMAT_NCHW, {4, 1, 16, 16}, {16, 16, 2, 2}, DT_FLOAT16};
  TransResult result;

  FormatTransferFracZNchw transfer;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(data));
  for (int i = 0; i < sizeof(data) / sizeof(data[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<uint16_t *>(result.data.get()))[i], data[i]);
  }
  Status status =
      transfer.TransShape(args.src_format, args.src_shape, args.src_data_type, args.dst_format, args.dst_shape);
  EXPECT_EQ(status, UNSUPPORTED);
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_fp16_success_gt_cube) {
  uint16_t data_4d[8 * 2 * 16 * 16] = {
      14154, 14684, 14727, 13722, 14690, 14935, 14421, 13953, 12940, 12530, 14800, 14640, 14945, 13325, 9403,  14385,
      13906, 13803, 13931, 15357, 14369, 14980, 14535, 5279,  15317, 13421, 14772, 14603, 10823, 14745, 15225, 15067,
      12558, 15054, 10046, 14110, 13129, 11599, 9248,  14644, 10818, 14568, 14743, 14885, 14629, 12260, 12808, 15242,
      11947, 14247, 12952, 11328, 13163, 15182, 14328, 14944, 12718, 13768, 15312, 12814, 15297, 14043, 14478, 14790,
      14411, 14373, 14325, 14025, 14205, 15186, 13849, 15307, 14734, 13643, 15281, 14473, 5762,  13974, 6694,  15220,
      15079, 13535, 14741, 8934,  14764, 13322, 14812, 14932, 14580, 14064, 15027, 14373, 14620, 15236, 15246, 12821,
      8629,  7304,  14729, 10175, 13538, 14521, 11603, 13960, 14852, 15138, 14708, 11553, 13516, 15351, 14635, 14532,
      11791, 14493, 13742, 15049, 13081, 14433, 13545, 14912, 14482, 15320, 13218, 15265, 13532, 12787, 13425, 14713,
      14760, 15347, 13096, 11612, 14251, 15237, 12356, 15276, 14359, 14245, 13778, 14509, 15278, 14521, 15065, 14704,
      13425, 12037, 14441, 15194, 14650, 14131, 14811, 14373, 10475, 15121, 11375, 9445,  11907, 14355, 14583, 15262,
      14939, 15243, 15056, 12781, 15003, 13862, 13329, 14617, 11336, 14337, 14778, 14716, 8625,  13549, 13190, 15293,
      15098, 14455, 15180, 15082, 14121, 12136, 13707, 13597, 13483, 15292, 15171, 15220, 13349, 13013, 15070, 14766,
      14707, 13540, 13779, 14457, 13684, 13347, 14087, 12587, 12677, 11717, 15282, 12870, 14710, 14482, 14167, 14831,
      14669, 15125, 13674, 15151, 14981, 13730, 14962, 13548, 15286, 15083, 13917, 13323, 14693, 14461, 15077, 15229,
      14385, 14464, 13659, 11526, 14506, 14463, 14427, 13615, 11809, 14998, 15025, 15115, 13023, 14517, 13778, 14345,
      13024, 13770, 15184, 13291, 13454, 14736, 14465, 13678, 14183, 11796, 14389, 13549, 15297, 14870, 11984, 15230,
      14956, 12907, 15241, 13404, 14466, 15211, 13506, 5438,  14443, 13097, 15153, 14254, 13329, 14648, 15295, 12227,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11800, 12358, 14316, 12598, 13583, 8102,  14411, 14479, 14557, 15017, 15153, 11356, 14136, 8613,  15125, 13679,
      11053, 13226, 11435, 15073, 15220, 14361, 13055, 12462, 15119, 13677, 11813, 14653, 9952,  13571, 13645, 14280,
      14501, 13499, 14900, 15211, 14420, 15268, 15025, 14933, 13993, 15303, 14487, 12617, 13585, 14460, 12358, 13565,
      12293, 13924, 14311, 14576, 13357, 12506, 11957, 15203, 13746, 15329, 14796, 12237, 14906, 15050, 12760, 13392,
      12750, 11969, 14920, 14600, 15327, 14775, 7403,  13519, 15348, 15103, 11304, 11048, 14523, 13798, 13774, 14755,
      13771, 14988, 11779, 14298, 14563, 14384, 13986, 15323, 14718, 14860, 14085, 12370, 14686, 14467, 15030, 13812,
      13382, 12618, 14118, 14452, 11174, 11473, 14109, 14605, 12979, 14736, 15207, 14737, 15193, 13449, 13957, 14396,
      13852, 1878,  14919, 13765, 14563, 14505, 13745, 14634, 15087, 14368, 13795, 14069, 15353, 14643, 12714, 12792,
      14056, 14923, 14819, 11435, 14751, 14441, 14160, 10889, 9986,  7205,  14862, 14598, 14741, 14325, 12374, 14987,
      13673, 14869, 11179, 14934, 14436, 10938, 14941, 8729,  13266, 14625, 13982, 8387,  14026, 12480, 14168, 8390,
      14696, 8866,  13506, 14369, 14029, 14936, 14969, 15066, 13177, 15282, 13286, 15056, 14178, 13413, 15308, 6845,
      14532, 12612, 10677, 13312, 11024, 14404, 14585, 13696, 13914, 14397, 12951, 15074, 14445, 14755, 15024, 15270,
      14463, 14953, 13485, 14672, 14036, 12445, 12533, 13715, 14323, 14783, 11897, 12776, 11327, 14698, 14690, 13010,
      13831, 13249, 12859, 10272, 13878, 14753, 14122, 12725, 10978, 15320, 14607, 14884, 14836, 15070, 13231, 9284,
      14372, 14985, 13348, 14939, 14916, 15302, 15311, 14139, 13058, 14226, 13932, 11407, 12343, 14710, 5722,  8516,
      15113, 14820, 14658, 12285, 13870, 15056, 15086, 13957, 15162, 13755, 15124, 14350, 15099, 14817, 14727, 13388,
      11990, 14651, 15344, 11849, 14892, 12450, 12593, 13974, 14782, 15105, 13778, 13788, 15151, 13586, 14046, 14400,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13512, 14640, 14461, 10825, 13981, 12378, 13832, 14227, 13920, 12225, 10915, 14182, 14444, 13064, 15072, 14836,
      14973, 14923, 15184, 14532, 14095, 9147,  14710, 12578, 15043, 14200, 13632, 14806, 14299, 13382, 14640, 14747,
      14917, 12732, 14266, 12630, 12611, 14319, 11974, 15103, 13543, 14535, 13446, 15240, 13313, 14928, 14465, 12020,
      13141, 13175, 14724, 14817, 14639, 13378, 14696, 12602, 14982, 12295, 13923, 12940, 12651, 11972, 15324, 14715,
      14990, 15083, 14750, 8474,  13334, 14199, 11237, 11631, 13509, 15045, 14897, 9206,  13220, 14569, 15204, 7383,
      14189, 14549, 12888, 15104, 13199, 11908, 14612, 13819, 13813, 12517, 10472, 12990, 15176, 14045, 13634, 15000,
      9699,  14619, 13175, 14292, 15082, 14888, 13390, 12366, 14212, 10265, 14337, 14022, 13839, 13570, 14608, 14819,
      11608, 12516, 11311, 14933, 15186, 14829, 15227, 10570, 15091, 14722, 10043, 12823, 14933, 15355, 14393, 13497,
      13333, 14904, 14567, 14895, 12612, 13737, 11376, 13772, 14523, 10706, 12092, 13318, 14824, 13816, 14837, 11674,
      14373, 12621, 14442, 14424, 14388, 13065, 14722, 14355, 15340, 12170, 12517, 15266, 14767, 14043, 14630, 14440,
      11406, 14385, 14981, 14970, 12614, 12085, 15304, 12188, 12739, 14061, 11147, 14850, 11490, 14949, 13713, 7462,
      14444, 15019, 14319, 13325, 5092,  10389, 15093, 14802, 14628, 14786, 13637, 13629, 12970, 10894, 12485, 11866,
      14129, 14570, 15027, 14723, 15047, 14604, 14319, 14614, 13108, 13319, 13199, 14501, 14840, 11856, 13794, 14902,
      13609, 13467, 14848, 12820, 14397, 14863, 14685, 15285, 14472, 14739, 14583, 13708, 15077, 14041, 14212, 13567,
      14074, 14695, 13370, 9930,  14304, 15315, 14821, 13994, 13469, 12706, 12675, 14357, 14028, 13777, 14463, 13905,
      13754, 15008, 14980, 13714, 13490, 15058, 13770, 14504, 14316, 14734, 14575, 14604, 11725, 11796, 12995, 15170,
      14396, 14583, 14839, 15070, 14791, 14068, 10384, 10719, 12585, 14690, 14779, 13934, 15314, 12019, 15067, 14448,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10308, 14741, 12058, 14995, 14428, 14722, 15321, 15183, 14091, 13848, 13365, 12050, 14716, 13000, 14552, 12681,
      11216, 15252, 13499, 12581, 15181, 14955, 13063, 13897, 14016, 15309, 13825, 15228, 14840, 13324, 13516, 13078,
      15144, 15016, 14905, 14395, 12703, 14347, 13033, 11565, 14654, 14098, 14281, 14443, 13940, 9779,  14714, 12614,
      14573, 12248, 14439, 12143, 14548, 7240,  14180, 14096, 15004, 14926, 14948, 11338, 13864, 15138, 14975, 14767,
      13601, 13063, 15104, 13401, 11289, 15045, 14177, 13620, 11498, 15131, 12756, 14399, 11377, 15101, 12778, 12935,
      13883, 15294, 14741, 14532, 13800, 9177,  14953, 15249, 13365, 14995, 14838, 13245, 12433, 13683, 11853, 14642,
      13543, 15183, 12583, 14399, 15031, 14006, 10820, 14875, 13137, 13705, 13612, 15317, 15242, 15101, 15248, 14775,
      14142, 13980, 14424, 15318, 12946, 10844, 15148, 13020, 12949, 14393, 15021, 15026, 14548, 14016, 13323, 12955,
      14877, 12928, 14541, 14669, 14096, 12731, 14645, 14491, 12676, 13304, 15165, 13506, 14748, 13327, 13579, 14431,
      14351, 13225, 12845, 12674, 14701, 14949, 12380, 13545, 13713, 12862, 14987, 13983, 12618, 15124, 15326, 15170,
      14030, 13322, 14443, 14960, 10340, 14820, 9438,  14923, 11185, 10851, 14603, 14517, 14264, 14507, 13530, 14177,
      12951, 15351, 11045, 15065, 11293, 14979, 14865, 14845, 14958, 14534, 11270, 14390, 14793, 13774, 12444, 13114,
      13725, 14625, 12422, 13912, 13766, 13048, 14569, 10447, 13123, 14987, 7305,  13692, 15024, 12935, 11960, 14928,
      14260, 13770, 14945, 14897, 13331, 14123, 9608,  14570, 14778, 14530, 13492, 15281, 15088, 11678, 12623, 14531,
      13388, 12533, 10379, 14162, 15100, 12815, 8714,  14044, 12414, 15087, 13566, 11750, 13354, 13371, 13660, 14537,
      15021, 15228, 14674, 14424, 12422, 14996, 13398, 14737, 14667, 15152, 14512, 14338, 14463, 14789, 13383, 13544,
      15161, 14430, 15167, 13795, 14443, 7770,  14507, 13644, 14576, 11515, 13661, 15097, 15248, 15031, 12337, 15007,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12041, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14750, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14240, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13541, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14571, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14750, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15295, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14736, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13354, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10879, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14898, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14673, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12841, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14384, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14768, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14181, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13577, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10846, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14315, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14129, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12467, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14351, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14377, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14213, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14975, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14173, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14416, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14626, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14876, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11506, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14122, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14363, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13801, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14442, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10269, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14404, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14554, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14322, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13762, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13481, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14396, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14777, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14808, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13164, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13393, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15099, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13082, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13968, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14128, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14766, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14684, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13635, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14036, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12037, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14479, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13976, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11516, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14177, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15172, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14726, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13388, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13888, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14665, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14364, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14499, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12528, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15072, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10487, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
  };
  uint16_t data[17 * 17 * 2 * 2] = {
      14154, 11800, 13512, 10308, 14684, 12358, 14640, 14741, 14727, 14316, 14461, 12058, 13722, 12598, 10825, 14995,
      14690, 13583, 13981, 14428, 14935, 8102,  12378, 14722, 14421, 14411, 13832, 15321, 13953, 14479, 14227, 15183,
      12940, 14557, 13920, 14091, 12530, 15017, 12225, 13848, 14800, 15153, 10915, 13365, 14640, 11356, 14182, 12050,
      14945, 14136, 14444, 14716, 13325, 8613,  13064, 13000, 9403,  15125, 15072, 14552, 14385, 13679, 14836, 12681,
      12041, 10846, 10269, 13635, 13906, 11053, 14973, 11216, 13803, 13226, 14923, 15252, 13931, 11435, 15184, 13499,
      15357, 15073, 14532, 12581, 14369, 15220, 14095, 15181, 14980, 14361, 9147,  14955, 14535, 13055, 14710, 13063,
      5279,  12462, 12578, 13897, 15317, 15119, 15043, 14016, 13421, 13677, 14200, 15309, 14772, 11813, 13632, 13825,
      14603, 14653, 14806, 15228, 10823, 9952,  14299, 14840, 14745, 13571, 13382, 13324, 15225, 13645, 14640, 13516,
      15067, 14280, 14747, 13078, 14750, 14315, 14404, 14036, 12558, 14501, 14917, 15144, 15054, 13499, 12732, 15016,
      10046, 14900, 14266, 14905, 14110, 15211, 12630, 14395, 13129, 14420, 12611, 12703, 11599, 15268, 14319, 14347,
      9248,  15025, 11974, 13033, 14644, 14933, 15103, 11565, 10818, 13993, 13543, 14654, 14568, 15303, 14535, 14098,
      14743, 14487, 13446, 14281, 14885, 12617, 15240, 14443, 14629, 13585, 13313, 13940, 12260, 14460, 14928, 9779,
      12808, 12358, 14465, 14714, 15242, 13565, 12020, 12614, 14240, 14129, 14554, 12037, 11947, 12293, 13141, 14573,
      14247, 13924, 13175, 12248, 12952, 14311, 14724, 14439, 11328, 14576, 14817, 12143, 13163, 13357, 14639, 14548,
      15182, 12506, 13378, 7240,  14328, 11957, 14696, 14180, 14944, 15203, 12602, 14096, 12718, 13746, 14982, 15004,
      13768, 15329, 12295, 14926, 15312, 14796, 13923, 14948, 12814, 12237, 12940, 11338, 15297, 14906, 12651, 13864,
      14043, 15050, 11972, 15138, 14478, 12760, 15324, 14975, 14790, 13392, 14715, 14767, 13541, 12467, 14322, 14479,
      14411, 12750, 14990, 13601, 14373, 11969, 15083, 13063, 14325, 14920, 14750, 15104, 14025, 14600, 8474,  13401,
      14205, 15327, 13334, 11289, 15186, 14775, 14199, 15045, 13849, 7403,  11237, 14177, 15307, 13519, 11631, 13620,
      14734, 15348, 13509, 11498, 13643, 15103, 15045, 15131, 15281, 11304, 14897, 12756, 14473, 11048, 9206,  14399,
      5762,  14523, 13220, 11377, 13974, 13798, 14569, 15101, 6694,  13774, 15204, 12778, 15220, 14755, 7383,  12935,
      14571, 14351, 13762, 13976, 15079, 13771, 14189, 13883, 13535, 14988, 14549, 15294, 14741, 11779, 12888, 14741,
      8934,  14298, 15104, 14532, 14764, 14563, 13199, 13800, 13322, 14384, 11908, 9177,  14812, 13986, 14612, 14953,
      14932, 15323, 13819, 15249, 14580, 14718, 13813, 13365, 14064, 14860, 12517, 14995, 15027, 14085, 10472, 14838,
      14373, 12370, 12990, 13245, 14620, 14686, 15176, 12433, 15236, 14467, 14045, 13683, 15246, 15030, 13634, 11853,
      12821, 13812, 15000, 14642, 14750, 14377, 13481, 11516, 8629,  13382, 9699,  13543, 7304,  12618, 14619, 15183,
      14729, 14118, 13175, 12583, 10175, 14452, 14292, 14399, 13538, 11174, 15082, 15031, 14521, 11473, 14888, 14006,
      11603, 14109, 13390, 10820, 13960, 14605, 12366, 14875, 14852, 12979, 14212, 13137, 15138, 14736, 10265, 13705,
      14708, 15207, 14337, 13612, 11553, 14737, 14022, 15317, 13516, 15193, 13839, 15242, 15351, 13449, 13570, 15101,
      14635, 13957, 14608, 15248, 14532, 14396, 14819, 14775, 15295, 14213, 14396, 14177, 11791, 13852, 11608, 14142,
      14493, 1878,  12516, 13980, 13742, 14919, 11311, 14424, 15049, 13765, 14933, 15318, 13081, 14563, 15186, 12946,
      14433, 14505, 14829, 10844, 13545, 13745, 15227, 15148, 14912, 14634, 10570, 13020, 14482, 15087, 15091, 12949,
      15320, 14368, 14722, 14393, 13218, 13795, 10043, 15021, 15265, 14069, 12823, 15026, 13532, 15353, 14933, 14548,
      12787, 14643, 15355, 14016, 13425, 12714, 14393, 13323, 14713, 12792, 13497, 12955, 14736, 14975, 14777, 15172,
      14760, 14056, 13333, 14877, 15347, 14923, 14904, 12928, 13096, 14819, 14567, 14541, 11612, 11435, 14895, 14669,
      14251, 14751, 12612, 14096, 15237, 14441, 13737, 12731, 12356, 14160, 11376, 14645, 15276, 10889, 13772, 14491,
      14359, 9986,  14523, 12676, 14245, 7205,  10706, 13304, 13778, 14862, 12092, 15165, 14509, 14598, 13318, 13506,
      15278, 14741, 14824, 14748, 14521, 14325, 13816, 13327, 15065, 12374, 14837, 13579, 14704, 14987, 11674, 14431,
      13354, 14173, 14808, 14726, 13425, 13673, 14373, 14351, 12037, 14869, 12621, 13225, 14441, 11179, 14442, 12845,
      15194, 14934, 14424, 12674, 14650, 14436, 14388, 14701, 14131, 10938, 13065, 14949, 14811, 14941, 14722, 12380,
      14373, 8729,  14355, 13545, 10475, 13266, 15340, 13713, 15121, 14625, 12170, 12862, 11375, 13982, 12517, 14987,
      9445,  8387,  15266, 13983, 11907, 14026, 14767, 12618, 14355, 12480, 14043, 15124, 14583, 14168, 14630, 15326,
      15262, 8390,  14440, 15170, 10879, 14416, 13164, 13388, 14939, 14696, 11406, 14030, 15243, 8866,  14385, 13322,
      15056, 13506, 14981, 14443, 12781, 14369, 14970, 14960, 15003, 14029, 12614, 10340, 13862, 14936, 12085, 14820,
      13329, 14969, 15304, 9438,  14617, 15066, 12188, 14923, 11336, 13177, 12739, 11185, 14337, 15282, 14061, 10851,
      14778, 13286, 11147, 14603, 14716, 15056, 14850, 14517, 8625,  14178, 11490, 14264, 13549, 13413, 14949, 14507,
      13190, 15308, 13713, 13530, 15293, 6845,  7462,  14177, 14898, 14626, 13393, 13888, 15098, 14532, 14444, 12951,
      14455, 12612, 15019, 15351, 15180, 10677, 14319, 11045, 15082, 13312, 13325, 15065, 14121, 11024, 5092,  11293,
      12136, 14404, 10389, 14979, 13707, 14585, 15093, 14865, 13597, 13696, 14802, 14845, 13483, 13914, 14628, 14958,
      15292, 14397, 14786, 14534, 15171, 12951, 13637, 11270, 15220, 15074, 13629, 14390, 13349, 14445, 12970, 14793,
      13013, 14755, 10894, 13774, 15070, 15024, 12485, 12444, 14766, 15270, 11866, 13114, 14673, 14876, 15099, 14665,
      14707, 14463, 14129, 13725, 13540, 14953, 14570, 14625, 13779, 13485, 15027, 12422, 14457, 14672, 14723, 13912,
      13684, 14036, 15047, 13766, 13347, 12445, 14604, 13048, 14087, 12533, 14319, 14569, 12587, 13715, 14614, 10447,
      12677, 14323, 13108, 13123, 11717, 14783, 13319, 14987, 15282, 11897, 13199, 7305,  12870, 12776, 14501, 13692,
      14710, 11327, 14840, 15024, 14482, 14698, 11856, 12935, 14167, 14690, 13794, 11960, 14831, 13010, 14902, 14928,
      12841, 11506, 13082, 14364, 14669, 13831, 13609, 14260, 15125, 13249, 13467, 13770, 13674, 12859, 14848, 14945,
      15151, 10272, 12820, 14897, 14981, 13878, 14397, 13331, 13730, 14753, 14863, 14123, 14962, 14122, 14685, 9608,
      13548, 12725, 15285, 14570, 15286, 10978, 14472, 14778, 15083, 15320, 14739, 14530, 13917, 14607, 14583, 13492,
      13323, 14884, 13708, 15281, 14693, 14836, 15077, 15088, 14461, 15070, 14041, 11678, 15077, 13231, 14212, 12623,
      15229, 9284,  13567, 14531, 14384, 14122, 13968, 14499, 14385, 14372, 14074, 13388, 14464, 14985, 14695, 12533,
      13659, 13348, 13370, 10379, 11526, 14939, 9930,  14162, 14506, 14916, 14304, 15100, 14463, 15302, 15315, 12815,
      14427, 15311, 14821, 8714,  13615, 14139, 13994, 14044, 11809, 13058, 13469, 12414, 14998, 14226, 12706, 15087,
      15025, 13932, 12675, 13566, 15115, 11407, 14357, 11750, 13023, 12343, 14028, 13354, 14517, 14710, 13777, 13371,
      13778, 5722,  14463, 13660, 14345, 8516,  13905, 14537, 14768, 14363, 14128, 12528, 13024, 15113, 13754, 15021,
      13770, 14820, 15008, 15228, 15184, 14658, 14980, 14674, 13291, 12285, 13714, 14424, 13454, 13870, 13490, 12422,
      14736, 15056, 15058, 14996, 14465, 15086, 13770, 13398, 13678, 13957, 14504, 14737, 14183, 15162, 14316, 14667,
      11796, 13755, 14734, 15152, 14389, 15124, 14575, 14512, 13549, 14350, 14604, 14338, 15297, 15099, 11725, 14463,
      14870, 14817, 11796, 14789, 11984, 14727, 12995, 13383, 15230, 13388, 15170, 13544, 14181, 13801, 14766, 15072,
      14956, 11990, 14396, 15161, 12907, 14651, 14583, 14430, 15241, 15344, 14839, 15167, 13404, 11849, 15070, 13795,
      14466, 14892, 14791, 14443, 15211, 12450, 14068, 7770,  13506, 12593, 10384, 14507, 5438,  13974, 10719, 13644,
      14443, 14782, 12585, 14576, 13097, 15105, 14690, 11515, 15153, 13778, 14779, 13661, 14254, 13788, 13934, 15097,
      13329, 15151, 15314, 15248, 14648, 13586, 12019, 15031, 15295, 14046, 15067, 12337, 12227, 14400, 14448, 15007,
      13577, 14442, 14684, 10487,
  };

  TransArgs args{
      reinterpret_cast<uint8_t *>(data_4d), FORMAT_FRACTAL_Z, FORMAT_NCHW, {8, 2, 16, 16}, {17, 17, 2, 2}, DT_FLOAT16};
  TransResult result;

  FormatTransferFracZNchw transfer;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(data));
  for (int i = 0; i < sizeof(data) / sizeof(data[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<uint16_t *>(result.data.get()))[i], data[i]);
  }
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_fp32_success_lt_cube) {
  float data_4d[1 * 1 * 16 * 16] = {
      0.09754133710736013,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
  };
  float data[1 * 1 * 1 * 1] = {
      0.09754133710736013,
  };

  TransArgs args{
      reinterpret_cast<uint8_t *>(data_4d), FORMAT_FRACTAL_Z, FORMAT_NCHW, {1, 1, 16, 16}, {1, 1, 1, 1}, DT_FLOAT};
  TransResult result;

  FormatTransferFracZNchw transfer;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(data));
  for (int i = 0; i < sizeof(data) / sizeof(data[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<float *>(result.data.get()))[i], data[i]);
  }
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_fp32_success_eq_cube) {
  float data_4d[4 * 1 * 16 * 16] = {
      0.8168560091594788,    0.9509842896547284,   0.5833092887893205,   0.6589822147276274,    0.47045074330618286,
      0.9281547983684513,    0.600092769881412,    0.17898594440529647,  0.8576318058687813,    0.938720727628625,
      0.8033558006015729,    0.24200277760696154,  0.6186692952513718,   0.835520484071766,     0.4311907295387437,
      0.617409884383429,     0.814295189333189,    0.9794132570935358,   0.06066024635820566,   0.7862193019409862,
      0.15602380534563587,   0.5061528494608266,   0.023818815576188035, 0.8383178948092223,    0.006568502107600405,
      0.10855668820302111,   0.6390430740654722,   0.7105474488440885,   0.18442298188011697,   0.39633873084569105,
      0.02203410245430748,   0.9585879381347081,   0.6253312439146503,   0.8797792374489929,    0.564915192256268,
      0.17059305879396147,   0.13449603025595858,  0.5132140444074369,   0.888281188122505,     0.7458979631897087,
      0.20784801035172418,   0.5316210502378305,   0.264930003499788,    0.822790801630084,     0.1275310727535136,
      0.17621014057762763,   0.20121209254119576,  0.9629309059771185,   0.16152061993669253,   0.9184351717655141,
      0.5768371846556524,    0.09692577674404401,  0.19365802143801403,  0.6713471682467593,    0.4464779971247871,
      0.35841838325964526,   0.8838362049508657,   0.14644641430223437,  0.020600348804988733,  0.5830140247131442,
      0.9120723994416113,    0.2890771263757931,   0.6731885002986142,   0.05978784194051445,   0.49836605146120294,
      0.5361904204459131,    0.7037203835218252,   0.2337470724618712,   0.9359869249295187,    0.6511806581115431,
      0.16960732523446453,   0.6331176515629879,   0.29776350367890303,  0.898087610056731,     0.7050336566805279,
      0.1779294951964464,    0.8407454580002686,   0.9030176530428079,   0.3962354901081948,    0.8480860206951764,
      0.44641475102114236,   0.6189982722879348,   0.5677669260248176,   0.8756562316364245,    0.9218814825285827,
      0.3385841911868406,    0.43452691499150753,  0.43495420630177895,  0.2167458539368593,    0.2747435364989407,
      0.44742375825894276,   0.28845466817065657,  0.8351637336690974,   0.08351581783753559,   0.22109616153818434,
      0.007403305376558755,  0.3096894526703965,   0.07174528411870695,  0.8392748235221433,    0.10064316912490501,
      0.854298828885103,     0.33412876368740774,  0.6625524883774276,   0.5441341649229283,    0.0046000314503452655,
      0.8961701205958728,    0.4738158753347098,   0.6638762364422152,   0.6280238370222847,    0.15693613434765008,
      0.647995408143734,     0.7489801614723965,   0.00681896126863879,  0.3314307419400613,    0.4745661412858506,
      0.26383036751434474,   0.4052284090198067,   0.852589324987285,    0.9548352483736139,    0.13954018638569377,
      0.8944096051223135,    0.4668506648602152,   0.6601070226329898,   0.36285239316259754,   0.17661350678122167,
      0.8132552504971018,    0.4667529963466772,   0.469826848277653,    0.858475203494622,     0.6680567128563278,
      0.5990851321154904,    0.4999661773857067,   0.9946626628204948,   0.8020650217923826,    0.6303835339592258,
      0.976235178750195,     0.5329535915098538,   0.3159470095831284,   0.2252686632477826,    0.8962128673665215,
      0.5509642305707612,    0.6420467128229859,   0.21302728758817935,  0.46137147849437576,   0.24492029694593542,
      0.3933137970579751,    0.5590786322601159,   0.1663466717888198,   0.3789353856819866,    0.7835470327519002,
      0.5172370168586898,    0.1893090477523638,   0.8628182031087973,   0.48556423350072164,   0.3184362891300201,
      0.3405437685721455,    0.30773451388643114,  0.3310473758050274,   0.26617106280235514,   0.8881836032717928,
      0.0019748406404110908, 0.11411778617269408,  0.8813082291176062,   0.6974321018129322,    0.7107020765056907,
      0.5959190816016348,    0.7285173323071417,   0.11388295655175429,  0.36364630195494363,   0.430570120947953,
      0.794418360007892,     0.9589303369282219,   0.8049303691996378,   0.20783858732450966,   0.813758326094692,
      0.42831396861230087,   0.8338790183483511,   0.6240323008565776,   0.7998340910323554,    0.9920403620580774,
      0.5709015334284396,    0.05223128642784547,  0.03675975391122588,  0.8923908481198983,    0.38875135912006153,
      0.15633456851623195,   0.47645536361462515,  0.4524629303911627,   0.16560337215822885,   0.20511084091555487,
      0.24422253257338689,   0.5399578206643209,   0.5895995758457725,   0.49512837325997494,   0.9866682191699911,
      0.1565576231389576,    0.11031857761037567,  0.7053415600250933,   0.6070071446106846,    0.5611511345020163,
      0.8381768551013316,    0.24164170248232963,  0.49323005501992434,  0.03687616857370135,   0.3376426574069038,
      0.5267449154082644,    0.8576270193692666,   0.6450736154582144,   0.819101464517182,     0.1144351292725918,
      0.4017280714518162,    0.5556116757224207,   0.9221926500998648,   0.19774518884072245,   0.8122964973474701,
      0.3262859226363999,    0.9107762918587479,   0.7931491228346562,   0.17300525008891987,   0.8896454505060492,
      0.6975214852271043,    0.11085715794712914,  0.5406368884178097,   0.24236996044129,      0.005040450959806875,
      0.5307190997595231,    0.21917947859334652,  0.12836684746170424,  0.04341533031285172,   0.40282978335635267,
      0.10689728015561661,   0.7062597786898408,   0.10060052958019916,  0.7519581440272585,    0.33898387421578013,
      0.7197323638730856,    0.10818828151427162,  0.257128306617908,    0.9431399649274476,    0.3427575441649615,
      0.027590449104072157,  0.344080675836019,    0.9188712333199375,   0.7774798184435455,    0.4703208696983341,
      0.4877224153511128,    0.9744804142777878,   0.3865392856063704,   0.9086582770947371,    0.8118173580897312,
      0.7127504119865766,    0.5428116981162028,   0.6567208354036572,   0.9941028268677842,    0.7089451172899908,
      0.1291678032399346,    0.8143848013695553,   0.08371796772497964,  0.7572744073194271,    0.7355714422434593,
      0.6474000922419452,    0.3543097445054675,   0.9775931220354745,   0.39195795313344217,   0.734977049152991,
      0.40992200643773524,   0.8956803193572815,   0.5304486284524556,   0.6917120158692178,    0.16698225613025464,
      0.25305871430218674,   0.2773774239009309,   0.2585681607191772,   0.34706972398955294,   0.03412426914877709,
      0.5866199298906816,    0.8611266571199963,   0.5615543838156779,   0.3107543580607147,    0.49795816525650705,
      0.8863226516778276,    0.2299424424942591,   0.43831819986329346,  0.4518885193253971,    0.6383902848395534,
      0.6430345264842787,    0.9609476845062578,   0.8590313057683825,   0.946593498563966,     0.2983608041225252,
      0.7282427899662679,    0.7739583717212988,   0.5164712358293232,   0.3829975311218565,    0.49259771833928856,
      0.164872639082684,     0.7908460033382655,   0.37086685026871635,  0.2832773609739344,    0.5560741120852402,
      0.423183068847399,     0.7638596377247268,   0.7000310278304417,   0.3259451790595508,    0.8949307563709777,
      0.45246027501617325,   0.4285428045299762,   0.4446071658359174,   0.3628421665222693,    0.5001169820074454,
      0.35067001891108707,   0.36640616781084445,  0.1319510785443646,   0.6374188461865751,    0.9619721858079823,
      0.450648239278473,     0.9458594644503746,   0.8368954310764464,   0.5842754845864907,    0.8095185597678237,
      0.8024953167276908,    0.05684876774993508,  0.04398365551438299,  0.3654011515745126,    0.6548055931649895,
      0.3431922859529116,    0.7930336989599233,   0.9661409956880415,   0.9404276716436277,    0.12139618629958815,
      0.8684061415265188,    0.03925836062942434,  0.24546224115198678,  0.4418476578509971,    0.3092537970650777,
      0.6166894942612399,    0.8914552592267662,   0.7601940155928462,   0.46273264284275795,   0.9910667860131801,
      0.989613199284938,     0.1350031504055178,   0.6601097097467896,   0.707338878990174,     0.35275396575269313,
      0.26552339318174023,   0.03836960346510854,  0.8575456787092489,   0.5490190619411595,    0.7267212712570938,
      0.2762364592057298,    0.701780668337549,    0.2774126338983015,   0.7886402903198428,    0.3112319568588495,
      0.9914504745956655,    0.23751441228513104,  0.7672314578717838,   0.5243418879999135,    0.020762148251404766,
      0.24795198841784238,   0.14033068437521923,  0.9820365012566946,   0.6543938822239167,    0.9640711664489289,
      0.8966714624583073,    0.5377330391683879,   0.31007202504736653,  0.8106485186894752,    0.20714445608926513,
      0.9437903821587161,    0.5964319454038916,   0.5451424408949661,   0.6477227012675425,    0.542586343018901,
      0.06833388853078093,   0.32606464238339183,  0.5581681811335306,   0.3275977545031692,    0.28340841019953833,
      0.7481897117282391,    0.30369871162505524,  0.47988765582449644,  0.7199247952470547,    0.7740570642269338,
      0.3792168788107044,    0.08724745594330285,  0.10057488876028953,  0.9035165171151448,    0.6383129952890263,
      0.5618959248529879,    0.9325368374505515,   0.08110775144153692,  0.0711648088478326,    0.16467109290939597,
      0.14915808899876637,   0.4351685924488591,   0.09918991161233881,  0.26787331851893303,   0.3752180188092137,
      0.9570321603275969,    0.018174267700791447, 0.5354470030366117,   0.08197310494768051,   0.1698909520507832,
      0.36465305001153236,   0.9686532192504956,   0.018478913704299882, 0.7490592279789233,    0.66958227563015,
      0.8065219848142055,    0.04008517538279088,  0.872515125292373,    0.4860298654033053,    0.18322168851434328,
      0.6048670557593928,    0.5013006566885004,   0.9184904088200172,   0.11545698833942053,   0.9388634472496009,
      0.5090817884855765,    0.4259104411681456,   0.6001774322912643,   0.7998639326761543,    0.3517388371715816,
      0.22736693792043783,   0.3089134680838309,   0.80724353471677,     0.3305659100205217,    0.7622798133203901,
      0.548658338677721,     0.41969457789024045,  0.8220890665327201,   0.7478312908945212,    0.8507041436520468,
      0.871258831698886,     0.10663552610035198,  0.9838857866221765,   0.7750037566395708,    0.2589287012836461,
      0.5509778371334433,    0.12181367833394996,  0.7187267185588344,   0.23212767811353918,   0.08317782184394085,
      0.8583710732117895,    0.6114511863557442,   0.9301399797123336,   0.919567831488458,     0.8688510005455988,
      0.4514271103179398,    0.03260164380485031,  0.7919825170425685,   0.8840906648864124,    0.18755046657366148,
      0.3532735787477914,    0.0316474703213836,   0.1128557788135679,   0.8453920161239551,    0.9442933089890817,
      0.7275184011552345,    0.5187615618800065,   0.547374563712834,    0.9316544035727097,    0.17710067883829017,
      0.8461317931422729,    0.6975272247451271,   0.6638162363778531,   0.43962725522043367,   0.9722150986994611,
      0.9924930481370859,    0.48465595076579016,  0.3638622689991824,   0.29202054581835457,   0.836306784595344,
      0.9599966862466585,    0.24014743456248733,  0.057907977353944706, 0.4052971131253651,    0.409528724694306,
      0.016921739543765635,  0.9944957404740558,   0.497678370549514,    0.3109661172605419,    0.65494173043691,
      0.8700685824205098,    0.28898664019111053,  0.1656176601583519,   0.6720371646498426,    0.5068669682523493,
      0.45612210493017646,   0.12612039229185712,  0.6216751159917858,   0.229399433957176,     0.39090320483261953,
      0.42480641493443805,   0.30915922531455675,  0.6718290412296023,   0.6662306327726546,    0.38675267059285845,
      0.5278348214068792,    0.928656700953216,    0.3404487585426379,   0.08665641437850102,   0.7659937015592914,
      0.2481889944626371,    0.0820735670409255,   0.6591059848854349,   0.8575519848132668,    0.4312086049658347,
      0.40734998866870453,   0.9127160055196917,   0.17831086075491986,  0.4987696353380683,    0.24969463334542352,
      0.25387278255025425,   0.45260915183568373,  0.9489936531210268,   0.25731793764593913,   0.549497380000946,
      0.04987042645173634,   0.6176062723559267,   0.4974628172419382,   0.22002087852372565,   0.8478032105864682,
      0.9771214815295913,    0.25763265280560277,  0.9468669882728353,   0.8788003530387016,    0.009612198576430142,
      0.6015103951914454,    0.07840696980709017,  0.8574066041211744,   0.8512521760763662,    0.6294403926212478,
      0.31422988517015304,   0.03260665549833752,  0.5459645373661592,   0.566558385068647,     0.4394158960380766,
      0.20460252793714095,   0.6301246345403493,   0.959197925681073,    0.13161493994917695,   0.25354386299396436,
      0.3435784686761524,    0.30447875804616154,  0.2641596508968739,   0.18034575029023447,   0.9141084597985908,
      0.2686654682752302,    0.2423135345770029,   0.1369363683392847,   0.9046537057606677,    0.08651204329137108,
      0.587229796493613,     0.14998627678593768,  0.9951223899404943,   0.05284484108795895,   0.7069841606969918,
      0.3457427661753919,    0.5677182815844276,   0.8867620993871848,   0.15609582018314616,   0.8248984097469988,
      0.5215435683614805,    0.7622504327833796,   0.42547271680893395,  0.5963613199856507,    0.07376145722251282,
      0.7431532306702398,    0.8719344267165662,   0.7967791529178645,   0.41533986772266474,   0.655829342011151,
      0.5547342350977146,    0.7346706935343904,   0.31834800355814097,  0.6917315704923358,    0.7974655877028941,
      0.7856595575225411,    0.33875107001435634,  0.2234786808748691,   0.6521145402408668,    0.6726573436211514,
      0.8733807978796556,    0.9621610691368013,   0.34279676512041746,  0.6716812009164921,    0.1403220922375994,
      0.6971779423432034,    0.600741255229763,    0.8852669468778349,   0.635908978571998,     0.17265456191779038,
      0.5999955102379257,    0.6672321014924975,   0.45626209843205523,  0.9458311544541305,    0.6570898489991635,
      0.9047193195927493,    0.7631557896257812,   0.2717021832489478,   0.2372662172825153,    0.17562859770047534,
      0.7519080171095838,    0.545009922910364,    0.061896113732203495, 0.32781141126662117,   0.38143075565845896,
      0.7499092284286264,    0.5997967230277158,   0.5919689972495351,   0.9749273043906194,    0.7129627435131244,
      0.8212988182775148,    0.5283680741464317,   0.9175043929909029,   0.02808909882978694,   0.5160217262690808,
      0.15792767510896566,   0.6904311507714506,   0.5623044553967744,   0.3686699379521433,    0.9022023913778509,
      0.5747619290732661,    0.018612257619440964, 0.36118967316669814,  0.8403688184533271,    0.09695520335025176,
      0.5400152625182123,    0.9407961067118134,   0.2255517082156655,   0.7439862596864373,    0.3371838848204518,
      0.05583253517461406,   0.8674952421961661,   0.15771469591732024,  0.9328305900348385,    0.6064305145965554,
      0.24440769376582883,   0.9138651868178408,   0.17029307748314682,  0.2627206088303473,    0.4039569145649755,
      0.27976198708603306,   0.9081095529075438,   0.7341822172341462,   0.6535583901663519,    0.9037218781902312,
      0.4132853954269039,    0.8677647022630963,   0.1363785587788613,   0.031726056012278736,  0.6967913663146085,
      0.251338912609961,     0.4504258736016229,   0.7939957881108681,   0.14373999152171169,   0.1345170110145837,
      0.22939821164356122,   0.2783262247792664,   0.5065783963181841,   0.2983062713783934,    0.09420519718693543,
      0.7678426568123674,    0.547205839024813,    0.9173821772312987,   0.9147495466711926,    0.6808591135362148,
      0.9433708812876275,    0.1888338795032064,   0.806439515054873,    0.5478245245285066,    0.06243129077708354,
      0.6738377942022481,    0.39313272333964333,  0.7107772272634887,   0.37140531406236,      0.9752231781906548,
      0.15530983015174504,   0.5083889494703363,   0.45787573647433233,  0.719369064985418,     0.695487795959419,
      0.648910517283219,     0.37255976835344096,  0.5128959126867926,   0.8462728472107226,    0.07161197909954886,
      0.5971577610206942,    0.3474139774062083,   0.17309257680758672,  0.4557553537522939,    0.48640839826996807,
      0.556565593151529,     0.909822162078459,    0.7815653980073078,   0.5558710994447176,    0.939412569745768,
      0.22458732015840954,   0.8854274720474302,   0.6612317372691434,   0.2353120266565194,    0.5431618026688126,
      0.8080683403154149,    0.9649378581633239,   0.5911987082154285,   0.4014005464684842,    0.31309623122426067,
      0.9209060982654941,    0.35675947143505937,  0.997932692240173,    0.05977641223564234,   0.4906372581279175,
      0.12540566557242627,   0.8113504889733711,   0.03439967972502145,  0.6996834889783037,    0.8580431256846641,
      0.29550733460934664,   0.8647558928256841,   0.17166159986929208,  0.2988438217337216,    0.30685222281881586,
      0.8740648760980403,    0.5733339597519914,   0.30588789631733104,  0.32034957581661505,   0.40447880890518706,
      0.05643993675567216,   0.9841350296823745,   0.0235057813209415,   0.5907496794641183,    0.9938255962978501,
      0.666736772901298,     0.09142543413958681,  0.5867777814038379,   0.10954805879898033,   0.5192786459845283,
      0.577284845201679,     0.4663457531259627,   0.9661498415023864,   0.7418291449321575,    0.6919757941011632,
      0.938162816686544,     0.2574986744080595,   0.5743986035644238,   0.3722004061266849,    0.48731682130765364,
      0.48804003791575845,   0.26568686568088584,  0.3024462081682625,   0.4927729450565037,    0.8208388909384069,
      0.3524966116601721,    0.7699774350342129,   0.3699307222802808,   0.6854571865797977,    0.05342232186562068,
      0.2568140216774778,    0.07986634967315454,  0.42206069298473226,  0.503112433651799,     0.2709998780180116,
      0.8540634740401672,    0.804771765143114,    0.08307842440728763,  0.5855745330669161,    0.20132581514653336,
      0.4733824279238499,    0.5397079569901397,   0.22332659849685244,  0.5681449614592287,    0.34305090560502416,
      0.19045524744723108,   0.14510857090406903,  0.8994852388900417,   0.3606955413718713,    0.03434264356844052,
      0.8425487893633322,    0.14471220280317265,  0.795235579857456,    0.9928577361495664,    0.8944565820128865,
      0.5951646624476978,    0.3387207739399991,   0.5254294787876147,   0.7528483715664833,    0.7724822370664601,
      0.06341949098827515,   0.1104781806479469,   0.2871131014826662,   0.9763926620716555,    0.8365936991760566,
      0.6894468038940823,    0.5516079036449986,   0.46459727852607835,  0.29698554757837436,   0.8641332259921782,
      0.2903577124285951,    0.5222172363827591,   0.2781390218294064,   0.6782975009525524,    0.43142531465001865,
      0.5482135764861943,    0.7958468590852784,   0.4601983363144355,   0.8293797711344699,    0.8218089614461451,
      0.2840468119059041,    0.5435124947402741,   0.292138895069197,    0.7093564706598441,    0.8634789517279025,
      0.7425214561915762,    0.4281558028893979,   0.23760087848913258,  0.12788063731046162,   0.4256176878159583,
      0.879074694521102,     0.9422873213978823,   0.1972547796962546,   0.7242073256517022,    0.33232344535504643,
      0.5537318465846828,    0.05869543998802962,  0.018360338257800812, 0.11030166295839527,   0.932707044606341,
      0.8123473545874034,    0.5693378326924512,   0.11500464247603825,  0.28180799331414497,   0.8696676356158933,
      0.34191489224832605,   0.8143563939660721,   0.010614668169477537, 0.7342136789417802,    0.8532271966112469,
      0.5748001835729629,    0.259112776070359,    0.5624154624891504,   0.09445036349110147,   0.3181656968422939,
      0.8939438952061383,    0.36760093868600874,  0.12440825480994533,  0.8224276403148182,    0.9440369737584494,
      0.8535593738819319,    0.012696538588551909, 0.39742245686374533,  0.3981006126038902,    0.12670159963685623,
      0.2225579855362756,    0.3639484396143289,   0.0736550465308492,   0.10747766849305518,   0.772278857461047,
      0.6434712254626346,    0.5946129067876548,   0.575837578716988,    0.6174587426513194,    0.9867408326366033,
      0.7658708170703061,    0.06043464109847374,  0.013078794647001901, 0.3219023249140547,    0.6928581061373904,
      0.29883540042062473,   0.6351799737621933,   0.2450520777737294,   0.7173739254073681,    0.2686491080989024,
      0.709877147040917,     0.17955632096592133,  0.5365934363287246,   0.692273457387911,     0.33888954272328764,
      0.2998404098945733,    0.13370452878408368,  0.8849589530269457,   0.43138580638305324,   0.8546451329124286,
      0.3457307101229713,    0.13674533003654288,  0.5925656225040199,   0.37366051705050807,   0.39713227198955114,
      0.8874356560086295,    0.8448286162871339,   0.2999165142924891,   0.6272603347680122,    0.9780875560006471,
      0.19514027598932304,   0.15829426674672442,  0.3554952077139394,   0.5987084697855811,    0.13497880870204337,
      0.9370079887875561,    0.9741398726037899,   0.9947348781247561,   0.5404091154422912,    0.6175609477753102,
      0.6131714181668053,    0.002477181737273959, 0.562280412091289,    0.7526435019202964,    0.3886096650631513,
      0.7821589398562553,    0.4695756803811453,   0.33999744031557955,  0.026417766980760926,  0.425174218763559,
      0.9712645643794963,    0.9696591839902201,   0.5373884989988515,   0.2875715965653579,    0.5325921358773241,
      0.9375649075751361,    0.7495572291868463,   0.2845140573246496,   0.46222375013708306,   0.84766692046905,
      0.6096220988114378,    0.15281616741547543,  0.8976612157167602,   0.9598524638502606,    0.23176953671336242,
      0.35578130428870847,   0.9120284563249282,   0.7833430085511944,   0.4083339394119787,    0.15267251534727655,
      0.6987791734238211,    0.318356175762636,    0.30599782277384735,  0.4823778703838705,    0.34246199799786725,
      0.45326212230640783,   0.9298832251551966,   0.80478120813917,     0.3842837068723949,    0.33483707922857575,
      0.5349291719254071,    0.6969575869879915,   0.7640239512132223,   0.19254059499782594,   0.0015221041677935254,
      0.411096175108574,     0.07717616484569123,  0.26137439637971194,  0.4751277053825842,    0.1870851529008637,
      0.7035527103294765,    0.22997764922641362,  0.6298971009845201,   0.41996421105921633,   0.24160305460943277,
      0.20894095942263502,   0.34448120238498403,  0.4793234914333593,   0.6012743927518429,    0.7412751836227155,
      0.23117503944651796,   0.7061590551786392,   0.6047964625652399,   0.26804051988359034,   0.20076795111638046,
      0.4676035290247943,    0.1882473105335125,   0.9180400475167111,   0.9989722501677417,    0.1895142045817545,
      0.3578897315238895,    0.7435437908679559,   0.5298733651365005,   0.2304246678026204,    0.6011480070982935,
      0.10039562955799841,   0.8701814345396093,   0.0940865963515104,   0.0003793943320034021, 0.5157733659631437,
      0.8546444054631211,    0.45404164701093686,  0.25421124457554,     0.8613953032846574,    0.7864680405695876,
      0.8470290022128667,    0.9414818442785001,   0.7626297442507252,   0.09357238052156414,   0.12743456850506396,
      0.5938704185634249,    0.7419615280910089,   0.621961090816814,    0.9067310219255523,    0.1360738049917236,
      0.5398831308323078,    0.1684504073334684,   0.7202214965416577,   0.7983453971630412,    0.33506042716085815,
      0.884016368776341,     0.6997701510258648,   0.5994184647706472,   0.24249478773185507,   0.19198624273422238,
      0.14966846053861182,   0.8872570608294355,   0.3925258780570219,   0.6037222439415791,    0.4108783367719028,
      0.032474872356025974,  0.15007040476252542,  0.5703765565370817,   0.03564049082735243,   0.9251552343127925,
      0.333522021214367,     0.6463190763686415,   0.544050804224951,    0.1428602387810044,
  };
  float data[16 * 16 * 2 * 2] = {
      0.8168560091594788,    0.8143848013695553,    0.17831086075491986,   0.5855745330669161,    0.9509842896547284,
      0.08371796772497964,   0.4987696353380683,    0.20132581514653336,   0.5833092887893205,    0.7572744073194271,
      0.24969463334542352,   0.4733824279238499,    0.6589822147276274,    0.7355714422434593,    0.25387278255025425,
      0.5397079569901397,    0.47045074330618286,   0.6474000922419452,    0.45260915183568373,   0.22332659849685244,
      0.9281547983684513,    0.3543097445054675,    0.9489936531210268,    0.5681449614592287,    0.600092769881412,
      0.9775931220354745,    0.25731793764593913,   0.34305090560502416,   0.17898594440529647,   0.39195795313344217,
      0.549497380000946,     0.19045524744723108,   0.8576318058687813,    0.734977049152991,     0.04987042645173634,
      0.14510857090406903,   0.938720727628625,     0.40992200643773524,   0.6176062723559267,    0.8994852388900417,
      0.8033558006015729,    0.8956803193572815,    0.4974628172419382,    0.3606955413718713,    0.24200277760696154,
      0.5304486284524556,    0.22002087852372565,   0.03434264356844052,   0.6186692952513718,    0.6917120158692178,
      0.8478032105864682,    0.8425487893633322,    0.835520484071766,     0.16698225613025464,   0.9771214815295913,
      0.14471220280317265,   0.4311907295387437,    0.25305871430218674,   0.25763265280560277,   0.795235579857456,
      0.617409884383429,     0.2773774239009309,    0.9468669882728353,    0.9928577361495664,    0.814295189333189,
      0.2585681607191772,    0.8788003530387016,    0.8944565820128865,    0.9794132570935358,    0.34706972398955294,
      0.009612198576430142,  0.5951646624476978,    0.06066024635820566,   0.03412426914877709,   0.6015103951914454,
      0.3387207739399991,    0.7862193019409862,    0.5866199298906816,    0.07840696980709017,   0.5254294787876147,
      0.15602380534563587,   0.8611266571199963,    0.8574066041211744,    0.7528483715664833,    0.5061528494608266,
      0.5615543838156779,    0.8512521760763662,    0.7724822370664601,    0.023818815576188035,  0.3107543580607147,
      0.6294403926212478,    0.06341949098827515,   0.8383178948092223,    0.49795816525650705,   0.31422988517015304,
      0.1104781806479469,    0.006568502107600405,  0.8863226516778276,    0.03260665549833752,   0.2871131014826662,
      0.10855668820302111,   0.2299424424942591,    0.5459645373661592,    0.9763926620716555,    0.6390430740654722,
      0.43831819986329346,   0.566558385068647,     0.8365936991760566,    0.7105474488440885,    0.4518885193253971,
      0.4394158960380766,    0.6894468038940823,    0.18442298188011697,   0.6383902848395534,    0.20460252793714095,
      0.5516079036449986,    0.39633873084569105,   0.6430345264842787,    0.6301246345403493,    0.46459727852607835,
      0.02203410245430748,   0.9609476845062578,    0.959197925681073,     0.29698554757837436,   0.9585879381347081,
      0.8590313057683825,    0.13161493994917695,   0.8641332259921782,    0.6253312439146503,    0.946593498563966,
      0.25354386299396436,   0.2903577124285951,    0.8797792374489929,    0.2983608041225252,    0.3435784686761524,
      0.5222172363827591,    0.564915192256268,     0.7282427899662679,    0.30447875804616154,   0.2781390218294064,
      0.17059305879396147,   0.7739583717212988,    0.2641596508968739,    0.6782975009525524,    0.13449603025595858,
      0.5164712358293232,    0.18034575029023447,   0.43142531465001865,   0.5132140444074369,    0.3829975311218565,
      0.9141084597985908,    0.5482135764861943,    0.888281188122505,     0.49259771833928856,   0.2686654682752302,
      0.7958468590852784,    0.7458979631897087,    0.164872639082684,     0.2423135345770029,    0.4601983363144355,
      0.20784801035172418,   0.7908460033382655,    0.1369363683392847,    0.8293797711344699,    0.5316210502378305,
      0.37086685026871635,   0.9046537057606677,    0.8218089614461451,    0.264930003499788,     0.2832773609739344,
      0.08651204329137108,   0.2840468119059041,    0.822790801630084,     0.5560741120852402,    0.587229796493613,
      0.5435124947402741,    0.1275310727535136,    0.423183068847399,     0.14998627678593768,   0.292138895069197,
      0.17621014057762763,   0.7638596377247268,    0.9951223899404943,    0.7093564706598441,    0.20121209254119576,
      0.7000310278304417,    0.05284484108795895,   0.8634789517279025,    0.9629309059771185,    0.3259451790595508,
      0.7069841606969918,    0.7425214561915762,    0.16152061993669253,   0.8949307563709777,    0.3457427661753919,
      0.4281558028893979,    0.9184351717655141,    0.45246027501617325,   0.5677182815844276,    0.23760087848913258,
      0.5768371846556524,    0.4285428045299762,    0.8867620993871848,    0.12788063731046162,   0.09692577674404401,
      0.4446071658359174,    0.15609582018314616,   0.4256176878159583,    0.19365802143801403,   0.3628421665222693,
      0.8248984097469988,    0.879074694521102,     0.6713471682467593,    0.5001169820074454,    0.5215435683614805,
      0.9422873213978823,    0.4464779971247871,    0.35067001891108707,   0.7622504327833796,    0.1972547796962546,
      0.35841838325964526,   0.36640616781084445,   0.42547271680893395,   0.7242073256517022,    0.8838362049508657,
      0.1319510785443646,    0.5963613199856507,    0.33232344535504643,   0.14644641430223437,   0.6374188461865751,
      0.07376145722251282,   0.5537318465846828,    0.020600348804988733,  0.9619721858079823,    0.7431532306702398,
      0.05869543998802962,   0.5830140247131442,    0.450648239278473,     0.8719344267165662,    0.018360338257800812,
      0.9120723994416113,    0.9458594644503746,    0.7967791529178645,    0.11030166295839527,   0.2890771263757931,
      0.8368954310764464,    0.41533986772266474,   0.932707044606341,     0.6731885002986142,    0.5842754845864907,
      0.655829342011151,     0.8123473545874034,    0.05978784194051445,   0.8095185597678237,    0.5547342350977146,
      0.5693378326924512,    0.49836605146120294,   0.8024953167276908,    0.7346706935343904,    0.11500464247603825,
      0.5361904204459131,    0.05684876774993508,   0.31834800355814097,   0.28180799331414497,   0.7037203835218252,
      0.04398365551438299,   0.6917315704923358,    0.8696676356158933,    0.2337470724618712,    0.3654011515745126,
      0.7974655877028941,    0.34191489224832605,   0.9359869249295187,    0.6548055931649895,    0.7856595575225411,
      0.8143563939660721,    0.6511806581115431,    0.3431922859529116,    0.33875107001435634,   0.010614668169477537,
      0.16960732523446453,   0.7930336989599233,    0.2234786808748691,    0.7342136789417802,    0.6331176515629879,
      0.9661409956880415,    0.6521145402408668,    0.8532271966112469,    0.29776350367890303,   0.9404276716436277,
      0.6726573436211514,    0.5748001835729629,    0.898087610056731,     0.12139618629958815,   0.8733807978796556,
      0.259112776070359,     0.7050336566805279,    0.8684061415265188,    0.9621610691368013,    0.5624154624891504,
      0.1779294951964464,    0.03925836062942434,   0.34279676512041746,   0.09445036349110147,   0.8407454580002686,
      0.24546224115198678,   0.6716812009164921,    0.3181656968422939,    0.9030176530428079,    0.4418476578509971,
      0.1403220922375994,    0.8939438952061383,    0.3962354901081948,    0.3092537970650777,    0.6971779423432034,
      0.36760093868600874,   0.8480860206951764,    0.6166894942612399,    0.600741255229763,     0.12440825480994533,
      0.44641475102114236,   0.8914552592267662,    0.8852669468778349,    0.8224276403148182,    0.6189982722879348,
      0.7601940155928462,    0.635908978571998,     0.9440369737584494,    0.5677669260248176,    0.46273264284275795,
      0.17265456191779038,   0.8535593738819319,    0.8756562316364245,    0.9910667860131801,    0.5999955102379257,
      0.012696538588551909,  0.9218814825285827,    0.989613199284938,     0.6672321014924975,    0.39742245686374533,
      0.3385841911868406,    0.1350031504055178,    0.45626209843205523,   0.3981006126038902,    0.43452691499150753,
      0.6601097097467896,    0.9458311544541305,    0.12670159963685623,   0.43495420630177895,   0.707338878990174,
      0.6570898489991635,    0.2225579855362756,    0.2167458539368593,    0.35275396575269313,   0.9047193195927493,
      0.3639484396143289,    0.2747435364989407,    0.26552339318174023,   0.7631557896257812,    0.0736550465308492,
      0.44742375825894276,   0.03836960346510854,   0.2717021832489478,    0.10747766849305518,   0.28845466817065657,
      0.8575456787092489,    0.2372662172825153,    0.772278857461047,     0.8351637336690974,    0.5490190619411595,
      0.17562859770047534,   0.6434712254626346,    0.08351581783753559,   0.7267212712570938,    0.7519080171095838,
      0.5946129067876548,    0.22109616153818434,   0.2762364592057298,    0.545009922910364,     0.575837578716988,
      0.007403305376558755,  0.701780668337549,     0.061896113732203495,  0.6174587426513194,    0.3096894526703965,
      0.2774126338983015,    0.32781141126662117,   0.9867408326366033,    0.07174528411870695,   0.7886402903198428,
      0.38143075565845896,   0.7658708170703061,    0.8392748235221433,    0.3112319568588495,    0.7499092284286264,
      0.06043464109847374,   0.10064316912490501,   0.9914504745956655,    0.5997967230277158,    0.013078794647001901,
      0.854298828885103,     0.23751441228513104,   0.5919689972495351,    0.3219023249140547,    0.33412876368740774,
      0.7672314578717838,    0.9749273043906194,    0.6928581061373904,    0.6625524883774276,    0.5243418879999135,
      0.7129627435131244,    0.29883540042062473,   0.5441341649229283,    0.020762148251404766,  0.8212988182775148,
      0.6351799737621933,    0.0046000314503452655, 0.24795198841784238,   0.5283680741464317,    0.2450520777737294,
      0.8961701205958728,    0.14033068437521923,   0.9175043929909029,    0.7173739254073681,    0.4738158753347098,
      0.9820365012566946,    0.02808909882978694,   0.2686491080989024,    0.6638762364422152,    0.6543938822239167,
      0.5160217262690808,    0.709877147040917,     0.6280238370222847,    0.9640711664489289,    0.15792767510896566,
      0.17955632096592133,   0.15693613434765008,   0.8966714624583073,    0.6904311507714506,    0.5365934363287246,
      0.647995408143734,     0.5377330391683879,    0.5623044553967744,    0.692273457387911,     0.7489801614723965,
      0.31007202504736653,   0.3686699379521433,    0.33888954272328764,   0.00681896126863879,   0.8106485186894752,
      0.9022023913778509,    0.2998404098945733,    0.3314307419400613,    0.20714445608926513,   0.5747619290732661,
      0.13370452878408368,   0.4745661412858506,    0.9437903821587161,    0.018612257619440964,  0.8849589530269457,
      0.26383036751434474,   0.5964319454038916,    0.36118967316669814,   0.43138580638305324,   0.4052284090198067,
      0.5451424408949661,    0.8403688184533271,    0.8546451329124286,    0.852589324987285,     0.6477227012675425,
      0.09695520335025176,   0.3457307101229713,    0.9548352483736139,    0.542586343018901,     0.5400152625182123,
      0.13674533003654288,   0.13954018638569377,   0.06833388853078093,   0.9407961067118134,    0.5925656225040199,
      0.8944096051223135,    0.32606464238339183,   0.2255517082156655,    0.37366051705050807,   0.4668506648602152,
      0.5581681811335306,    0.7439862596864373,    0.39713227198955114,   0.6601070226329898,    0.3275977545031692,
      0.3371838848204518,    0.8874356560086295,    0.36285239316259754,   0.28340841019953833,   0.05583253517461406,
      0.8448286162871339,    0.17661350678122167,   0.7481897117282391,    0.8674952421961661,    0.2999165142924891,
      0.8132552504971018,    0.30369871162505524,   0.15771469591732024,   0.6272603347680122,    0.4667529963466772,
      0.47988765582449644,   0.9328305900348385,    0.9780875560006471,    0.469826848277653,     0.7199247952470547,
      0.6064305145965554,    0.19514027598932304,   0.858475203494622,     0.7740570642269338,    0.24440769376582883,
      0.15829426674672442,   0.6680567128563278,    0.3792168788107044,    0.9138651868178408,    0.3554952077139394,
      0.5990851321154904,    0.08724745594330285,   0.17029307748314682,   0.5987084697855811,    0.4999661773857067,
      0.10057488876028953,   0.2627206088303473,    0.13497880870204337,   0.9946626628204948,    0.9035165171151448,
      0.4039569145649755,    0.9370079887875561,    0.8020650217923826,    0.6383129952890263,    0.27976198708603306,
      0.9741398726037899,    0.6303835339592258,    0.5618959248529879,    0.9081095529075438,    0.9947348781247561,
      0.976235178750195,     0.9325368374505515,    0.7341822172341462,    0.5404091154422912,    0.5329535915098538,
      0.08110775144153692,   0.6535583901663519,    0.6175609477753102,    0.3159470095831284,    0.0711648088478326,
      0.9037218781902312,    0.6131714181668053,    0.2252686632477826,    0.16467109290939597,   0.4132853954269039,
      0.002477181737273959,  0.8962128673665215,    0.14915808899876637,   0.8677647022630963,    0.562280412091289,
      0.5509642305707612,    0.4351685924488591,    0.1363785587788613,    0.7526435019202964,    0.6420467128229859,
      0.09918991161233881,   0.031726056012278736,  0.3886096650631513,    0.21302728758817935,   0.26787331851893303,
      0.6967913663146085,    0.7821589398562553,    0.46137147849437576,   0.3752180188092137,    0.251338912609961,
      0.4695756803811453,    0.24492029694593542,   0.9570321603275969,    0.4504258736016229,    0.33999744031557955,
      0.3933137970579751,    0.018174267700791447,  0.7939957881108681,    0.026417766980760926,  0.5590786322601159,
      0.5354470030366117,    0.14373999152171169,   0.425174218763559,     0.1663466717888198,    0.08197310494768051,
      0.1345170110145837,    0.9712645643794963,    0.3789353856819866,    0.1698909520507832,    0.22939821164356122,
      0.9696591839902201,    0.7835470327519002,    0.36465305001153236,   0.2783262247792664,    0.5373884989988515,
      0.5172370168586898,    0.9686532192504956,    0.5065783963181841,    0.2875715965653579,    0.1893090477523638,
      0.018478913704299882,  0.2983062713783934,    0.5325921358773241,    0.8628182031087973,    0.7490592279789233,
      0.09420519718693543,   0.9375649075751361,    0.48556423350072164,   0.66958227563015,      0.7678426568123674,
      0.7495572291868463,    0.3184362891300201,    0.8065219848142055,    0.547205839024813,     0.2845140573246496,
      0.3405437685721455,    0.04008517538279088,   0.9173821772312987,    0.46222375013708306,   0.30773451388643114,
      0.872515125292373,     0.9147495466711926,    0.84766692046905,      0.3310473758050274,    0.4860298654033053,
      0.6808591135362148,    0.6096220988114378,    0.26617106280235514,   0.18322168851434328,   0.9433708812876275,
      0.15281616741547543,   0.8881836032717928,    0.6048670557593928,    0.1888338795032064,    0.8976612157167602,
      0.0019748406404110908, 0.5013006566885004,    0.806439515054873,     0.9598524638502606,    0.11411778617269408,
      0.9184904088200172,    0.5478245245285066,    0.23176953671336242,   0.8813082291176062,    0.11545698833942053,
      0.06243129077708354,   0.35578130428870847,   0.6974321018129322,    0.9388634472496009,    0.6738377942022481,
      0.9120284563249282,    0.7107020765056907,    0.5090817884855765,    0.39313272333964333,   0.7833430085511944,
      0.5959190816016348,    0.4259104411681456,    0.7107772272634887,    0.4083339394119787,    0.7285173323071417,
      0.6001774322912643,    0.37140531406236,      0.15267251534727655,   0.11388295655175429,   0.7998639326761543,
      0.9752231781906548,    0.6987791734238211,    0.36364630195494363,   0.3517388371715816,    0.15530983015174504,
      0.318356175762636,     0.430570120947953,     0.22736693792043783,   0.5083889494703363,    0.30599782277384735,
      0.794418360007892,     0.3089134680838309,    0.45787573647433233,   0.4823778703838705,    0.9589303369282219,
      0.80724353471677,      0.719369064985418,     0.34246199799786725,   0.8049303691996378,    0.3305659100205217,
      0.695487795959419,     0.45326212230640783,   0.20783858732450966,   0.7622798133203901,    0.648910517283219,
      0.9298832251551966,    0.813758326094692,     0.548658338677721,     0.37255976835344096,   0.80478120813917,
      0.42831396861230087,   0.41969457789024045,   0.5128959126867926,    0.3842837068723949,    0.8338790183483511,
      0.8220890665327201,    0.8462728472107226,    0.33483707922857575,   0.6240323008565776,    0.7478312908945212,
      0.07161197909954886,   0.5349291719254071,    0.7998340910323554,    0.8507041436520468,    0.5971577610206942,
      0.6969575869879915,    0.9920403620580774,    0.871258831698886,     0.3474139774062083,    0.7640239512132223,
      0.5709015334284396,    0.10663552610035198,   0.17309257680758672,   0.19254059499782594,   0.05223128642784547,
      0.9838857866221765,    0.4557553537522939,    0.0015221041677935254, 0.03675975391122588,   0.7750037566395708,
      0.48640839826996807,   0.411096175108574,     0.8923908481198983,    0.2589287012836461,    0.556565593151529,
      0.07717616484569123,   0.38875135912006153,   0.5509778371334433,    0.909822162078459,     0.26137439637971194,
      0.15633456851623195,   0.12181367833394996,   0.7815653980073078,    0.4751277053825842,    0.47645536361462515,
      0.7187267185588344,    0.5558710994447176,    0.1870851529008637,    0.4524629303911627,    0.23212767811353918,
      0.939412569745768,     0.7035527103294765,    0.16560337215822885,   0.08317782184394085,   0.22458732015840954,
      0.22997764922641362,   0.20511084091555487,   0.8583710732117895,    0.8854274720474302,    0.6298971009845201,
      0.24422253257338689,   0.6114511863557442,    0.6612317372691434,    0.41996421105921633,   0.5399578206643209,
      0.9301399797123336,    0.2353120266565194,    0.24160305460943277,   0.5895995758457725,    0.919567831488458,
      0.5431618026688126,    0.20894095942263502,   0.49512837325997494,   0.8688510005455988,    0.8080683403154149,
      0.34448120238498403,   0.9866682191699911,    0.4514271103179398,    0.9649378581633239,    0.4793234914333593,
      0.1565576231389576,    0.03260164380485031,   0.5911987082154285,    0.6012743927518429,    0.11031857761037567,
      0.7919825170425685,    0.4014005464684842,    0.7412751836227155,    0.7053415600250933,    0.8840906648864124,
      0.31309623122426067,   0.23117503944651796,   0.6070071446106846,    0.18755046657366148,   0.9209060982654941,
      0.7061590551786392,    0.5611511345020163,    0.3532735787477914,    0.35675947143505937,   0.6047964625652399,
      0.8381768551013316,    0.0316474703213836,    0.997932692240173,     0.26804051988359034,   0.24164170248232963,
      0.1128557788135679,    0.05977641223564234,   0.20076795111638046,   0.49323005501992434,   0.8453920161239551,
      0.4906372581279175,    0.4676035290247943,    0.03687616857370135,   0.9442933089890817,    0.12540566557242627,
      0.1882473105335125,    0.3376426574069038,    0.7275184011552345,    0.8113504889733711,    0.9180400475167111,
      0.5267449154082644,    0.5187615618800065,    0.03439967972502145,   0.9989722501677417,    0.8576270193692666,
      0.547374563712834,     0.6996834889783037,    0.1895142045817545,    0.6450736154582144,    0.9316544035727097,
      0.8580431256846641,    0.3578897315238895,    0.819101464517182,     0.17710067883829017,   0.29550733460934664,
      0.7435437908679559,    0.1144351292725918,    0.8461317931422729,    0.8647558928256841,    0.5298733651365005,
      0.4017280714518162,    0.6975272247451271,    0.17166159986929208,   0.2304246678026204,    0.5556116757224207,
      0.6638162363778531,    0.2988438217337216,    0.6011480070982935,    0.9221926500998648,    0.43962725522043367,
      0.30685222281881586,   0.10039562955799841,   0.19774518884072245,   0.9722150986994611,    0.8740648760980403,
      0.8701814345396093,    0.8122964973474701,    0.9924930481370859,    0.5733339597519914,    0.0940865963515104,
      0.3262859226363999,    0.48465595076579016,   0.30588789631733104,   0.0003793943320034021, 0.9107762918587479,
      0.3638622689991824,    0.32034957581661505,   0.5157733659631437,    0.7931491228346562,    0.29202054581835457,
      0.40447880890518706,   0.8546444054631211,    0.17300525008891987,   0.836306784595344,     0.05643993675567216,
      0.45404164701093686,   0.8896454505060492,    0.9599966862466585,    0.9841350296823745,    0.25421124457554,
      0.6975214852271043,    0.24014743456248733,   0.0235057813209415,    0.8613953032846574,    0.11085715794712914,
      0.057907977353944706,  0.5907496794641183,    0.7864680405695876,    0.5406368884178097,    0.4052971131253651,
      0.9938255962978501,    0.8470290022128667,    0.24236996044129,      0.409528724694306,     0.666736772901298,
      0.9414818442785001,    0.005040450959806875,  0.016921739543765635,  0.09142543413958681,   0.7626297442507252,
      0.5307190997595231,    0.9944957404740558,    0.5867777814038379,    0.09357238052156414,   0.21917947859334652,
      0.497678370549514,     0.10954805879898033,   0.12743456850506396,   0.12836684746170424,   0.3109661172605419,
      0.5192786459845283,    0.5938704185634249,    0.04341533031285172,   0.65494173043691,      0.577284845201679,
      0.7419615280910089,    0.40282978335635267,   0.8700685824205098,    0.4663457531259627,    0.621961090816814,
      0.10689728015561661,   0.28898664019111053,   0.9661498415023864,    0.9067310219255523,    0.7062597786898408,
      0.1656176601583519,    0.7418291449321575,    0.1360738049917236,    0.10060052958019916,   0.6720371646498426,
      0.6919757941011632,    0.5398831308323078,    0.7519581440272585,    0.5068669682523493,    0.938162816686544,
      0.1684504073334684,    0.33898387421578013,   0.45612210493017646,   0.2574986744080595,    0.7202214965416577,
      0.7197323638730856,    0.12612039229185712,   0.5743986035644238,    0.7983453971630412,    0.10818828151427162,
      0.6216751159917858,    0.3722004061266849,    0.33506042716085815,   0.257128306617908,     0.229399433957176,
      0.48731682130765364,   0.884016368776341,     0.9431399649274476,    0.39090320483261953,   0.48804003791575845,
      0.6997701510258648,    0.3427575441649615,    0.42480641493443805,   0.26568686568088584,   0.5994184647706472,
      0.027590449104072157,  0.30915922531455675,   0.3024462081682625,    0.24249478773185507,   0.344080675836019,
      0.6718290412296023,    0.4927729450565037,    0.19198624273422238,   0.9188712333199375,    0.6662306327726546,
      0.8208388909384069,    0.14966846053861182,   0.7774798184435455,    0.38675267059285845,   0.3524966116601721,
      0.8872570608294355,    0.4703208696983341,    0.5278348214068792,    0.7699774350342129,    0.3925258780570219,
      0.4877224153511128,    0.928656700953216,     0.3699307222802808,    0.6037222439415791,    0.9744804142777878,
      0.3404487585426379,    0.6854571865797977,    0.4108783367719028,    0.3865392856063704,    0.08665641437850102,
      0.05342232186562068,   0.032474872356025974,  0.9086582770947371,    0.7659937015592914,    0.2568140216774778,
      0.15007040476252542,   0.8118173580897312,    0.2481889944626371,    0.07986634967315454,   0.5703765565370817,
      0.7127504119865766,    0.0820735670409255,    0.42206069298473226,   0.03564049082735243,   0.5428116981162028,
      0.6591059848854349,    0.503112433651799,     0.9251552343127925,    0.6567208354036572,    0.8575519848132668,
      0.2709998780180116,    0.333522021214367,     0.9941028268677842,    0.4312086049658347,    0.8540634740401672,
      0.6463190763686415,    0.7089451172899908,    0.40734998866870453,   0.804771765143114,     0.544050804224951,
      0.1291678032399346,    0.9127160055196917,    0.08307842440728763,   0.1428602387810044,
  };

  TransArgs args{
      reinterpret_cast<uint8_t *>(data_4d), FORMAT_FRACTAL_Z, FORMAT_NCHW, {4, 1, 16, 16}, {16, 16, 2, 2}, DT_FLOAT};
  TransResult result;

  FormatTransferFracZNchw transfer;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(data));
  for (int i = 0; i < sizeof(data) / sizeof(data[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<float *>(result.data.get()))[i], data[i]);
  }
}

TEST_F(UtestFormatTransferFraczNchw, fracz_to_nchw_fp32_success_gt_cube) {
  float data_4d[8 * 2 * 16 * 16] = {
      0.13861991091475812,
      0.10525121122814463,
      0.14870114967919246,
      0.37158545264352527,
      0.08879297931832297,
      0.06760373986806234,
      0.003849344223073925,
      0.7561138729196587,
      0.0028097578732378947,
      0.7353701611370006,
      0.6933417527357065,
      0.2656723224637475,
      0.31405414074086935,
      0.9820863796986057,
      0.6698167465680883,
      0.7025073707567491,
      0.7925940267469043,
      0.9858147072404602,
      0.727796317012897,
      0.015787104612503655,
      0.27491105976633057,
      0.6369448325207008,
      0.6896774205279625,
      0.9323244549718339,
      0.6913756703206293,
      0.16754362854376104,
      0.8188111598667317,
      0.5428410503366073,
      0.5125348441029931,
      0.9913510386173713,
      0.9375861536812142,
      0.8736979003360933,
      0.7260342849793483,
      0.20120410264885236,
      0.7646333976747771,
      0.15173165326001736,
      0.39265689606394716,
      0.15243591466584205,
      0.010340535191060374,
      0.7563590156106308,
      0.9033088706649143,
      0.5709275013402282,
      0.4128767392139201,
      0.18461034076886662,
      0.917605405768162,
      0.5914222027681203,
      0.8848126150904408,
      0.02129918275230247,
      0.7696659137833982,
      0.8785872487273547,
      0.37446354377616475,
      0.7646072372892175,
      0.7758773698800284,
      0.3065250204557999,
      0.4170857632470394,
      0.044089732363661094,
      0.6890756783794824,
      0.43909351458338675,
      0.2939490692189499,
      0.2459954395643219,
      0.4618194259116881,
      0.2407026393868288,
      0.36086002695619024,
      0.8640851072532453,
      0.15128698698720322,
      0.10445646966322941,
      0.6279086302955916,
      0.30063695907830557,
      0.5803620249830622,
      0.834738368359478,
      0.9658333522412966,
      0.49084190134047334,
      0.5982948387159709,
      0.7364746708042765,
      0.1728040800468621,
      0.8960926950526985,
      0.5284394832775934,
      0.6568173944147312,
      0.7069769008278068,
      0.5579428147048966,
      0.13357691464798138,
      0.2908730982610468,
      0.5010437198187823,
      0.09963121799246955,
      0.17217086189255992,
      0.5648254121117188,
      0.053705219700942,
      0.35796688301072,
      0.20469063893497697,
      0.09371763139775569,
      0.17616094680644334,
      0.15162118086324772,
      0.690327705519752,
      0.9014654398254837,
      0.8026738672485313,
      0.2336754606791367,
      0.21401989030280077,
      0.3050069506431471,
      0.6236653000860762,
      0.713337515092525,
      0.10737739320492923,
      0.3873591518740387,
      0.8659119048707833,
      0.5925084869238261,
      0.9111021456064748,
      0.919429204772624,
      0.8384109327495086,
      0.1575614925244445,
      0.80509149343632,
      0.28370128790625326,
      0.4282317439660449,
      0.2357917781969051,
      0.011471956973760244,
      0.17339338632884005,
      0.6715959811967953,
      0.233220125983571,
      0.05314850242766944,
      0.7056619496920902,
      0.5456598510028011,
      0.13598966999855922,
      0.012529344446802426,
      0.6201816276837667,
      0.006964022371572898,
      0.3412241888432692,
      0.8049499889919485,
      0.667343472513223,
      0.4701897375183579,
      0.7418169336190366,
      0.9737001474297444,
      0.20654566517027428,
      0.17248149069917285,
      0.053157049516807664,
      0.2861359192158687,
      0.7593751161265202,
      0.06755860607493347,
      0.3852737431079606,
      0.04992360626473957,
      0.302489320076742,
      0.40132951312095344,
      0.6795152678016566,
      0.782572010580318,
      0.5746635474611804,
      0.7966442996538724,
      0.15459601428614667,
      0.3693961500059507,
      0.06411018686027592,
      0.22945572978660644,
      0.8731427819714899,
      0.5076890561556178,
      0.12930627273556972,
      0.9656927645096538,
      0.38569829114808596,
      0.12375154698912139,
      0.2509506112091433,
      0.1319370980474388,
      0.38289710756686746,
      0.25840864046822043,
      0.38571855983419834,
      0.8364680728786794,
      0.05987516981191243,
      0.42570991590805674,
      0.4119405321035826,
      0.17762967468331525,
      0.637507115987954,
      0.6450032164935743,
      0.5654207482006985,
      0.6329507004942213,
      0.33844084372877925,
      0.9049256528358469,
      0.10047293005385105,
      0.8194138858161407,
      0.6769310231737116,
      0.20440622491432292,
      0.8741638758927367,
      0.44505905665328027,
      0.21280609696170782,
      0.011781058613303919,
      0.47688438712340286,
      0.07768770074191944,
      0.00822444546154455,
      0.9681521716228816,
      0.776893063617345,
      0.17870297652809786,
      0.9918631665731272,
      0.18754660546807156,
      0.09674403006985133,
      0.7256042711333616,
      0.37613730048899263,
      0.5668621385905648,
      0.8147168836664814,
      0.7145827173057614,
      0.5419252027429404,
      0.8624241578915682,
      0.9767778331103008,
      0.33134453444767853,
      0.1511696460358204,
      0.931109609080415,
      0.26760280136617476,
      0.720504118679755,
      0.561054639152762,
      0.04403144974993356,
      0.6650312577642018,
      0.20842812187870807,
      0.9709344182765282,
      0.29749853648546964,
      0.16720208620538812,
      0.936915662933375,
      0.786663329801668,
      0.8726576278463952,
      0.9747036251493297,
      0.09534090496984604,
      0.6034024789176371,
      0.9165559738358574,
      0.17251694852573773,
      0.6518257199411946,
      0.30944313225985576,
      0.7563828055308841,
      0.253253633473531,
      0.5629789915185913,
      0.15465010252531697,
      0.6402301841568493,
      0.0795608311951832,
      0.13552737276834392,
      0.4753501403209113,
      0.12763429765746592,
      0.7976683366910478,
      0.8941181691336204,
      0.5075573010849087,
      0.18532317502693563,
      0.6848997362918184,
      0.5714156890506997,
      0.44174703642978874,
      0.3424985993076507,
      0.951509130793907,
      0.08295791787636164,
      0.11549926102263153,
      0.2749400181421586,
      0.9113280686842665,
      0.7990989912334494,
      0.14710662300775412,
      0.46955053458590934,
      0.40065138391111943,
      0.5752642927349247,
      0.07116527059062705,
      0.2712956260859689,
      0.080480044589748,
      0.6450046812325141,
      0.531319360316121,
      0.781899452578717,
      0.11552970112165872,
      0.5611500316920904,
      0.9249303707149428,
      0.3779808468902329,
      0.04917726412895995,
      0.7272035505938381,
      0.20229470167487462,
      0.6225381036572762,
      0.09665526780685885,
      0.4790754927951175,
      0.8852506071141146,
      0.6378272080403014,
      0.4302617641125852,
      0.17529425791527875,
      0.2911580454791233,
      0.5685378955963167,
      0.5996629884166993,
      0.35379997432878574,
      0.4812449632594248,
      0.9327970145025657,
      0.29959697003490815,
      0.18566739299261037,
      0.7906655716605615,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8795377572972203,
      0.1470755313991291,
      0.561247554034285,
      0.7662220587760888,
      0.2986541461597556,
      0.42411698786743934,
      0.7611568879677035,
      0.11264806170566222,
      0.440973443278477,
      0.25390038851594676,
      0.26612481125474574,
      0.3350292840944643,
      0.3005887681812899,
      0.22284371079118692,
      0.5659697003311785,
      0.7122019087559107,
      0.9567807277929921,
      0.20818145030885005,
      0.60588935414239,
      0.5040087971937715,
      0.29392139130299244,
      0.6363756825181826,
      0.7333242098434483,
      0.261729101999164,
      0.9569771739814522,
      0.6143153706299556,
      0.631409359723997,
      0.011901826035510932,
      0.2955498735301396,
      0.8816750025849788,
      0.14633316716600764,
      0.4793915962021015,
      0.5704212237455046,
      0.13150180417478674,
      0.31690962084442686,
      0.21228124048991093,
      0.7296604549037735,
      0.6650141171161462,
      0.04241446875266841,
      0.3321108610667033,
      0.9575885483833089,
      0.2436150246808676,
      0.987210551195358,
      0.6372128000142347,
      0.761113851092484,
      0.7437116188736885,
      0.45594654788563493,
      0.830052343780784,
      0.7982349711510114,
      0.004734651371140819,
      0.6690845068463606,
      0.7188268519310974,
      0.9407991624965886,
      0.1865674886269657,
      0.27113574284356157,
      0.7096913513550435,
      0.08449102565738409,
      0.9898338931888807,
      0.24247197683608523,
      0.16556601945810856,
      0.8106885120353601,
      0.4321572361784677,
      0.7113731972624318,
      0.007494933722813291,
      0.8141904135220546,
      0.7395150888511759,
      0.547581203606972,
      0.6530263306464384,
      0.007521198939880125,
      0.7474194348631291,
      0.5777144381034068,
      0.8584438346701353,
      0.16969711093899553,
      0.21749292052270808,
      0.36244880368539034,
      0.6945674442909555,
      0.04384851936719336,
      0.7586877459463743,
      0.869698410395238,
      0.9767059033674064,
      0.9446884919932719,
      0.6572566972190457,
      0.2048989668957929,
      0.8170741278189959,
      0.5547941031298822,
      0.7309773397133232,
      0.5063767684051376,
      0.2772421373998536,
      0.04711937267440225,
      0.23774692855986723,
      0.6737215431713256,
      0.30986446944527846,
      0.5491562617797402,
      0.1401554977665822,
      0.7932244713519243,
      0.7323878899162881,
      0.002396651178786957,
      0.08089773378821119,
      0.461475864945851,
      0.5224297784281214,
      0.3141596605465167,
      0.48263368203794865,
      0.2464902259465227,
      0.4728058512337572,
      0.47155205473777895,
      0.9833083681599707,
      0.03332507249957106,
      0.7163502359065218,
      0.3421635067549482,
      0.9787191868196575,
      0.8307200462203801,
      0.12284932223048928,
      0.5773682448264239,
      0.6558446451611865,
      0.680394068488586,
      0.7310433251859529,
      0.35523994189400077,
      0.6766207358391074,
      0.9060741761912895,
      0.04914274667362528,
      0.19144189303663528,
      0.4845610368675931,
      0.6541048336512872,
      0.8575012513297512,
      0.8537472449635101,
      0.7450443706280852,
      0.5803426777495934,
      0.5689408381290824,
      0.9092918118236812,
      0.17460334682241974,
      0.2340700994442091,
      0.9298928386669005,
      0.36914916933156305,
      0.12622765437152517,
      0.10689808095884157,
      0.4310771822646119,
      0.13267160640865372,
      0.14895973120547823,
      0.1397935054459304,
      0.9349783221173635,
      0.2503384778850122,
      0.31822595375881446,
      0.9913550891610706,
      0.55454466322197,
      0.6118908274540877,
      0.8779734145096334,
      0.3543109331520725,
      0.306988901348892,
      0.6726982566869614,
      0.02467090616045342,
      0.06553857351644254,
      0.22442983249751902,
      0.26395669898138985,
      0.9632237865411843,
      0.732461885956622,
      0.7317620554576368,
      0.49519103539917386,
      0.18413692212953858,
      0.7150341406645073,
      0.04522592798730041,
      0.6447408662943692,
      0.9385896789116761,
      0.24032378135246224,
      0.9682515458629827,
      0.26982819483722165,
      0.48178308190918895,
      0.5823940174941661,
      0.409359485776034,
      0.40752327445425973,
      0.1471890531783654,
      0.6242049562872195,
      0.18811560788574455,
      0.9893442282921309,
      0.42331698403095386,
      0.9302534567908484,
      0.35889687716970375,
      0.8660877509389459,
      0.928772577903708,
      0.6158032483682454,
      0.16070664756432096,
      0.9661412869875025,
      0.31898622890124084,
      0.5441507345809006,
      0.5055786252454072,
      0.8853269573231654,
      0.9931343022347453,
      0.08922742117815974,
      0.45857685361148903,
      0.7770746229113067,
      0.0887419464763255,
      0.3162106953425323,
      0.8536288391503049,
      0.28011809028393886,
      0.8521904076959731,
      0.7248634019855653,
      0.49869013399498285,
      0.4053194525003626,
      0.4735972943090352,
      0.17389059420154984,
      0.03366201552897685,
      0.9823521603852087,
      0.7808072048924388,
      0.5526270573062637,
      0.6738010436288211,
      0.5432293708851752,
      0.6914357121517429,
      0.43842610640884583,
      0.8530609582658267,
      0.11278100676094394,
      0.4941714606912455,
      0.4638603168540302,
      0.22194186056037057,
      0.6716327950358132,
      0.9166902550664012,
      0.8351982565565297,
      0.8412216162402254,
      0.2222539480109168,
      0.5991520863203645,
      0.38864001352791677,
      0.34985786042192857,
      0.15795832535816468,
      0.9661077058919707,
      0.012634417342274218,
      0.343423762898566,
      0.451146025504531,
      0.8761611666855926,
      0.8608539041919699,
      0.8903503433345549,
      0.6548981638110877,
      0.35486660560902894,
      0.08811346271007414,
      0.8110740131813556,
      0.8539951703461579,
      0.9555089087268872,
      0.9562723010283051,
      0.4681219537960889,
      0.37592691228594466,
      0.9992573696993887,
      0.664341812871904,
      0.23944640184088473,
      0.2963560937896125,
      0.508501991516788,
      0.15376590814347213,
      0.7379530846527411,
      0.7977010542015935,
      0.8915772781308771,
      0.5524912580999269,
      0.30375439556599226,
      0.6566031761406997,
      0.31858523647827175,
      0.019979751727917994,
      0.8743281776278087,
      0.3810202153747745,
      0.5203446091374957,
      0.966155481985703,
      0.37469353816237017,
      0.7352375292410388,
      0.3666447891178437,
      0.015451749872628873,
      0.7832706823237181,
      0.3999266945419052,
      0.4364277580092467,
      0.7230546551485972,
      0.157525724298896,
      0.5729599421962962,
      0.19803754933587325,
      0.8333412543498672,
      0.7867785122337513,
      0.27896398037168957,
      0.5484736202867737,
      0.32766071211385694,
      0.38499709012670014,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.5293502876783279,
      0.8291193070208391,
      0.42200289273599556,
      0.686857118791328,
      0.9305878823942679,
      0.509816187082568,
      0.41487977014592914,
      0.09554332355010453,
      0.7412162233285089,
      0.9973832859416821,
      0.23910159953071597,
      0.48004632089879007,
      0.039772947348565624,
      0.0894049198451472,
      0.01874007337969552,
      0.11961507320612019,
      0.7101967112871143,
      0.7197179842368491,
      0.578232652038329,
      0.040072115029379685,
      0.5804692341544302,
      0.6623947182585623,
      0.5438434786671921,
      0.17343133597811278,
      0.7367546132475674,
      0.13201613557039393,
      0.09897754229118771,
      0.8999983118049748,
      0.6003901478395222,
      0.32512047896820595,
      0.9128032309113029,
      0.9031362578958313,
      0.8276019986094242,
      0.16844251099981267,
      0.14661295960750242,
      0.5333431363401799,
      0.8299745503174504,
      0.5339333337288752,
      0.9106147969126901,
      0.19580399811145166,
      0.7078571833791006,
      0.48145075892895806,
      0.603131492234237,
      0.27265318902818514,
      0.8791795132792125,
      0.8411100928304986,
      0.017593478154503073,
      0.7060157268133952,
      0.7724279248361262,
      0.8861613857589661,
      0.343469114076287,
      0.017394361915610057,
      0.7883518755638007,
      0.4391735391309788,
      0.2504695401963266,
      0.15170174477220055,
      0.21393977565249778,
      0.443592550980236,
      0.5095334554083597,
      0.09938394670258921,
      0.028315604429974273,
      0.45463003066865415,
      0.8298597748035649,
      0.4532423883161699,
      0.7007554265498704,
      0.40115166637181754,
      0.1470799829764481,
      0.6294304491599311,
      0.23594389336137767,
      0.7112579137154827,
      0.10185947547633445,
      0.20139399288408033,
      0.1180404666237802,
      0.8399634215048786,
      0.7632940118923165,
      0.5303499842405933,
      0.4155261517501555,
      0.5548711377923504,
      0.025360358738519495,
      0.8232568061375535,
      0.3346104779456238,
      0.532915432070937,
      0.40015851318727125,
      0.5389460419617879,
      0.5987416086426901,
      0.6698552912716527,
      0.780388759983075,
      0.7925588085624273,
      0.45037796758318804,
      0.23963041265644203,
      0.36874366609595643,
      0.18375410637584955,
      0.5281492850328181,
      0.5460266989110848,
      0.9280948067239118,
      0.9873043353493716,
      0.03442022110787202,
      0.6291590014933156,
      0.9981627789260562,
      0.3200659468426875,
      0.44180687340667857,
      0.3349303332479078,
      0.4553140241842599,
      0.05359923274228118,
      0.03929951143652333,
      0.8592328433935742,
      0.6149351679083243,
      0.16259509931077443,
      0.24314980248993667,
      0.7595415381121309,
      0.7096499077090591,
      0.6615524634203029,
      0.6035851122706498,
      0.7343833438614558,
      0.41474303482746533,
      0.4310140395683666,
      0.5462996571286937,
      0.636139222807529,
      0.4261420957770189,
      0.1721447584349456,
      0.47147571375284814,
      0.5003705454959895,
      0.41863524425463694,
      0.1212972250835862,
      0.9945715392091543,
      0.40446452849666314,
      0.5562287852609348,
      0.46298522943000064,
      0.21327330991620574,
      0.4174038781545222,
      0.5506923312964094,
      0.08960554649117725,
      0.16821297722710693,
      0.009183189735727626,
      0.9519318593697729,
      0.7911885910393738,
      0.30243793737732894,
      0.36171955618702634,
      0.7275163676699398,
      0.6017302317686811,
      0.9211449821926627,
      0.016789744324152922,
      0.2231842692779209,
      0.7320508107990614,
      0.9172186193742928,
      0.38726858040604806,
      0.45346727152383326,
      0.9736037921150315,
      0.1415848835987198,
      0.4287144407465886,
      0.7192727473967828,
      0.44212966241115104,
      0.6075560359436376,
      0.2699313871787882,
      0.9501669443744405,
      0.6891624698228346,
      0.03594820662269016,
      0.4140665514245061,
      0.19123529007541273,
      0.6956985142613639,
      0.18608296200597518,
      0.21395244570112526,
      0.14994545679313542,
      0.7955362084226552,
      0.2908836237948339,
      0.3189721351237217,
      0.9113691555501472,
      0.16608774824486627,
      0.3652540906020737,
      0.7661105179559459,
      0.5503285960900492,
      0.6296482965637312,
      0.021618970592347186,
      0.3222338893974612,
      0.1771909927336177,
      0.36132216279703644,
      0.7031384528295931,
      0.5596365181728887,
      0.8053366142728067,
      0.8638400429864582,
      0.9085095227101969,
      0.5058802974670702,
      0.022489996885331487,
      0.16136704513340805,
      0.6783741471392065,
      0.2433432886640784,
      0.08811202713859834,
      0.7019149642340254,
      0.7074398653385896,
      0.4188458869744829,
      0.7417870718217474,
      0.09314542994725039,
      0.9526702722469567,
      0.15447695419829233,
      0.5234613373790115,
      0.4587900976769065,
      0.4666560847640946,
      0.47741158025412755,
      0.3941302129542221,
      0.8725938986300577,
      0.1868618851608026,
      0.5960817832633942,
      0.5402399164827054,
      0.6542430096117432,
      0.14093404302294743,
      0.7351008222411094,
      0.626896869991658,
      0.8153132939963299,
      0.4692693688236108,
      0.05305160105808626,
      0.2791581608581686,
      0.9228439462893766,
      0.7332915821155096,
      0.6628517982577824,
      0.04860139460088986,
      0.08951277138503844,
      0.8320411266990881,
      0.2048707134105746,
      0.8068872022870146,
      0.4948504180620279,
      0.41184033137111975,
      0.7190998979639045,
      0.8380990899424366,
      0.086092548343773,
      0.9820085929439971,
      0.32879915468366416,
      0.84204761146793,
      0.8700722120474124,
      0.48076406766202595,
      0.0732106249839285,
      0.14783939754527176,
      0.574931463245764,
      0.7079277349244375,
      0.7292094480906886,
      0.7047746244055081,
      0.7943898405857803,
      0.45038405067793374,
      0.3012144754381675,
      0.27280381214968086,
      0.12926601852738606,
      0.05313715099993188,
      0.9954587172603774,
      0.4718694982334253,
      0.4298764101970639,
      0.6681098129873408,
      0.42036967115181456,
      0.0839795459090027,
      0.5792609679185063,
      0.37605072282304686,
      0.6583979311718429,
      0.5148337803671061,
      0.385262882179612,
      0.5273305274418582,
      0.6199363284810673,
      0.16762752670302516,
      0.017492223678952867,
      0.30648818939511857,
      0.1260323700300645,
      0.7715815662819216,
      0.4481707153463781,
      0.763405910124729,
      0.09025576212085595,
      0.5033159147423758,
      0.35268716544351064,
      0.25698746611256706,
      0.8420450508984809,
      0.5982144085578488,
      0.9586114553080439,
      0.6344034099803877,
      0.5143253458855047,
      0.5249453898611923,
      0.12136272756597788,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.37446715213889414,
      0.37998233432424766,
      0.2031003471352949,
      0.12415625519029183,
      0.6085733581700042,
      0.3796334566305014,
      0.8371386332055144,
      0.012793800556977097,
      0.5833205323286814,
      0.09779108988244167,
      0.5520359287683562,
      0.4849558923188444,
      0.5282674162369237,
      0.06675479822979058,
      0.5553542608092429,
      0.39696323881119,
      0.2671520815086441,
      0.5412004706206277,
      0.9803949953224173,
      0.18482177407515232,
      0.28474674412781464,
      0.5217727387416771,
      0.27255844427862186,
      0.09233866919232392,
      0.11346891778995816,
      0.6851858317061092,
      0.0005627333655138234,
      0.9053829799883957,
      0.1141296807683726,
      0.7571659430911946,
      0.030398776009645023,
      0.8521206636214926,
      0.686751963460763,
      0.9234485268943335,
      0.4262558321277006,
      0.3659515307852249,
      0.8886441531294881,
      0.6990554941733981,
      0.8768036198178328,
      0.3319762164814921,
      0.7118863871542145,
      0.8291405634796339,
      0.7454138001482092,
      0.6496653584298466,
      0.44449868173307283,
      0.9805574793354571,
      0.62370311211052,
      0.5317392897133469,
      0.5670304904635779,
      0.007148288012197801,
      0.0857835687436056,
      0.4467901200429536,
      0.661795939990852,
      0.13366433186289184,
      0.8378440756500211,
      0.4622738238198597,
      0.9381592583428692,
      0.9596830650619335,
      0.3450364035776594,
      0.3035151186951639,
      0.4686002937786584,
      0.9453220019662685,
      0.6452195208794166,
      0.4358498394990138,
      0.5212431731505457,
      0.19932525488647057,
      0.9898306221037652,
      0.4747038016224169,
      0.82876683912034,
      0.9175597660672588,
      0.28147168770097997,
      0.25502934165477764,
      0.6415827981200575,
      0.799877067418722,
      0.5105555039916515,
      0.9146513147957873,
      0.9060172079755968,
      0.791556691837571,
      0.32491506165874695,
      0.6486302438986716,
      0.4569527670153789,
      0.9495997246091489,
      0.9396139468185104,
      0.45899358803544554,
      0.6894229831946286,
      0.270746137652799,
      0.7248139481233704,
      0.7626497479115815,
      0.75066781795124,
      0.22192123430700905,
      0.28396757247535576,
      0.9112190358293802,
      0.26330626036820004,
      0.3113617142082512,
      0.07148685551219547,
      0.7195158974755017,
      0.43477271771982284,
      0.5787097784473253,
      0.7313260977369199,
      0.9107728219751681,
      0.032906572995198724,
      0.7598530618731676,
      0.9818987602819136,
      0.4657556371687326,
      0.594695390220099,
      0.6883643754551224,
      0.2719751089503445,
      0.37581000966714084,
      0.83824915279393,
      0.06549247203877806,
      0.36632445386556045,
      0.2322214574176914,
      0.4704668571121897,
      0.14006001098754373,
      0.9681288328156152,
      0.47809152528622967,
      0.9657400895521029,
      0.47776444371478033,
      0.9687034478573067,
      0.07412119197941236,
      0.5721744237867166,
      0.6141593977164898,
      0.042054099789581856,
      0.8459542642975082,
      0.5189399037858603,
      0.2535302003984826,
      0.36275145211112614,
      0.48400839758585756,
      0.6285648741302996,
      0.6191394018077834,
      0.7749665246023152,
      0.45757180036095835,
      0.7671537731273556,
      0.4709601630937571,
      0.42941454508912114,
      0.8657921575600451,
      0.8985445127220502,
      0.9861050189074767,
      0.24757531177032255,
      0.978850701352717,
      0.7158918887968034,
      0.6831064106036061,
      0.9180071448105787,
      0.8376882158653953,
      0.9638665261744293,
      0.45042997757613545,
      0.21665688963627494,
      0.35195278013601794,
      0.05267969117623417,
      0.5389555311153045,
      0.9114851765836308,
      0.6258773970631035,
      0.7199341249146247,
      0.7923169276434796,
      0.8366591904387193,
      0.18053239868755144,
      0.272265164206111,
      0.331101485981824,
      0.8612879717194625,
      0.634731689137481,
      0.6486972068084338,
      0.33847889999990544,
      0.9967302039665963,
      0.2414374570780865,
      0.847995824351949,
      0.11948257056250688,
      0.02712104372887425,
      0.1840501737732555,
      0.9860125291135977,
      0.2193886006138287,
      0.7272713759543263,
      0.21513305795319226,
      0.5524594198562498,
      0.7513328175757993,
      0.09209983389065812,
      0.8560345875874015,
      0.37399657249679696,
      0.08589741028223441,
      0.3236249823421554,
      0.8698790447233965,
      0.1790385977070852,
      0.818425585420191,
      0.5829548995026376,
      0.984699782767251,
      0.18726766084750612,
      0.0658699939740679,
      0.8821854015107203,
      0.2788585976377317,
      0.2940684401683791,
      0.6238528546466062,
      0.9857679125734509,
      0.5518248962183375,
      0.3293403217991683,
      0.39029609682738287,
      0.6664252543793544,
      0.9743791622455666,
      0.22571944921717335,
      0.526565707851956,
      0.9769750936911633,
      0.39225171270804193,
      0.3389503252328022,
      0.30958977269472954,
      0.9213149201546954,
      0.5618971748260793,
      0.8745377717246547,
      0.5394045987877338,
      0.19551840539914045,
      0.35059087479820794,
      0.7836610133411863,
      0.29958461037233053,
      0.3938045042892838,
      0.18104328507119194,
      0.06778745965102362,
      0.9561014878015668,
      0.4551212110068099,
      0.025066292417908254,
      0.9492800499708482,
      0.47352200454570115,
      0.48246698916058783,
      0.18473893932854824,
      0.20994529454414756,
      0.33427933653414255,
      0.2864221132522502,
      0.3156841132503668,
      0.27292955984128087,
      0.4566805827078497,
      0.1462692343227675,
      0.7712660332064996,
      0.32975091626883435,
      0.5574361427043032,
      0.017003207714762714,
      0.5102761455044432,
      0.4111481217664793,
      0.4409066019171828,
      0.9072393279560211,
      0.5483245331582209,
      0.5609330056140861,
      0.6302384260721924,
      0.6742101055320844,
      0.5015443794327109,
      0.9202060801646266,
      0.4150966747261118,
      0.7791425089745136,
      0.3891997488372002,
      0.10129548673467059,
      0.36666062118419507,
      0.636617168184616,
      0.814697212596663,
      0.001881280677498931,
      0.60526396245784,
      0.721749351841282,
      0.95594196409197,
      0.7585796885360693,
      0.919887225009393,
      0.6189213811964427,
      0.2548479350612681,
      0.9162607004972148,
      0.7172637798754251,
      0.283564079660791,
      0.9552100378519058,
      0.6670310568579203,
      0.5654935550366178,
      0.24678578258378903,
      0.6118146734984746,
      0.07729267856727107,
      0.3301131217627997,
      0.24512151253567938,
      0.6908428767174394,
      0.5206197306406314,
      0.14555190477564528,
      0.3111331699242166,
      0.7052840640615085,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.3764290776930458,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.4146979809830683,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.9958462749035231,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.15849243309504923,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.9572203123547415,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.541530670814326,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.767028446649636,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.1309162637452962,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.10659462196323277,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.6772010208911962,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.43806594461121307,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.5769624715772051,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.19737388285992674,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.9971339679357473,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.647689169873493,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.3909836391230528,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.02543957664974894,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8964726438537993,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.30631061637899615,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.862029888701881,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.005941495896108373,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.556525443906726,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.6817726202638454,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8759690783659748,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.5360842728537408,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.3718155827833767,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.2659724158347494,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.7644468243071734,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.20475493002236655,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.3995961251457969,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.9932862480814585,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.34550246822699693,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.7738675159774514,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.024816608620341962,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.5665559410658447,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.9452069954015251,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8017745565146187,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.5238548480735659,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8706467667318868,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.4353807234830298,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8306436124180614,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.09064214693350137,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8685785467629582,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8345253702291049,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.11103772926010236,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.7070617835227128,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.23790479221880045,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.9264437066213292,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.45351833013599707,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8606352509445246,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.8580755998879297,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.5067469023858207,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.929885851198275,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.6782413640337536,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.7098499981352698,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.5823811990622354,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.9674534075204786,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.5798472290187143,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.7175920799518698,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.30953641019091327,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.12167573120389918,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.31084959197890827,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.34157322002831525,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.13141740731351004,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.18204780231842665,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.7288130733024423,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.4139691275664631,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.1475584382157734,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
  };
  float data[17 * 17 * 2 * 2] = {
      0.13861991091475812,  0.8795377572972203,    0.5293502876783279,    0.37446715213889414,  0.10525121122814463,
      0.1470755313991291,   0.8291193070208391,    0.37998233432424766,   0.14870114967919246,  0.561247554034285,
      0.42200289273599556,  0.2031003471352949,    0.37158545264352527,   0.7662220587760888,   0.686857118791328,
      0.12415625519029183,  0.08879297931832297,   0.2986541461597556,    0.9305878823942679,   0.6085733581700042,
      0.06760373986806234,  0.42411698786743934,   0.509816187082568,     0.3796334566305014,   0.003849344223073925,
      0.7611568879677035,   0.41487977014592914,   0.8371386332055144,    0.7561138729196587,   0.11264806170566222,
      0.09554332355010453,  0.012793800556977097,  0.0028097578732378947, 0.440973443278477,    0.7412162233285089,
      0.5833205323286814,   0.7353701611370006,    0.25390038851594676,   0.9973832859416821,   0.09779108988244167,
      0.6933417527357065,   0.26612481125474574,   0.23910159953071597,   0.5520359287683562,   0.2656723224637475,
      0.3350292840944643,   0.48004632089879007,   0.4849558923188444,    0.31405414074086935,  0.3005887681812899,
      0.039772947348565624, 0.5282674162369237,    0.9820863796986057,    0.22284371079118692,  0.0894049198451472,
      0.06675479822979058,  0.6698167465680883,    0.5659697003311785,    0.01874007337969552,  0.5553542608092429,
      0.7025073707567491,   0.7122019087559107,    0.11961507320612019,   0.39696323881119,     0.3764290776930458,
      0.8964726438537993,   0.5665559410658447,    0.5067469023858207,    0.7925940267469043,   0.9567807277929921,
      0.7101967112871143,   0.2671520815086441,    0.9858147072404602,    0.20818145030885005,  0.7197179842368491,
      0.5412004706206277,   0.727796317012897,     0.60588935414239,      0.578232652038329,    0.9803949953224173,
      0.015787104612503655, 0.5040087971937715,    0.040072115029379685,  0.18482177407515232,  0.27491105976633057,
      0.29392139130299244,  0.5804692341544302,    0.28474674412781464,   0.6369448325207008,   0.6363756825181826,
      0.6623947182585623,   0.5217727387416771,    0.6896774205279625,    0.7333242098434483,   0.5438434786671921,
      0.27255844427862186,  0.9323244549718339,    0.261729101999164,     0.17343133597811278,  0.09233866919232392,
      0.6913756703206293,   0.9569771739814522,    0.7367546132475674,    0.11346891778995816,  0.16754362854376104,
      0.6143153706299556,   0.13201613557039393,   0.6851858317061092,    0.8188111598667317,   0.631409359723997,
      0.09897754229118771,  0.0005627333655138234, 0.5428410503366073,    0.011901826035510932, 0.8999983118049748,
      0.9053829799883957,   0.5125348441029931,    0.2955498735301396,    0.6003901478395222,   0.1141296807683726,
      0.9913510386173713,   0.8816750025849788,    0.32512047896820595,   0.7571659430911946,   0.9375861536812142,
      0.14633316716600764,  0.9128032309113029,    0.030398776009645023,  0.8736979003360933,   0.4793915962021015,
      0.9031362578958313,   0.8521206636214926,    0.4146979809830683,    0.30631061637899615,  0.9452069954015251,
      0.929885851198275,    0.7260342849793483,    0.5704212237455046,    0.8276019986094242,   0.686751963460763,
      0.20120410264885236,  0.13150180417478674,   0.16844251099981267,   0.9234485268943335,   0.7646333976747771,
      0.31690962084442686,  0.14661295960750242,   0.4262558321277006,    0.15173165326001736,  0.21228124048991093,
      0.5333431363401799,   0.3659515307852249,    0.39265689606394716,   0.7296604549037735,   0.8299745503174504,
      0.8886441531294881,   0.15243591466584205,   0.6650141171161462,    0.5339333337288752,   0.6990554941733981,
      0.010340535191060374, 0.04241446875266841,   0.9106147969126901,    0.8768036198178328,   0.7563590156106308,
      0.3321108610667033,   0.19580399811145166,   0.3319762164814921,    0.9033088706649143,   0.9575885483833089,
      0.7078571833791006,   0.7118863871542145,    0.5709275013402282,    0.2436150246808676,   0.48145075892895806,
      0.8291405634796339,   0.4128767392139201,    0.987210551195358,     0.603131492234237,    0.7454138001482092,
      0.18461034076886662,  0.6372128000142347,    0.27265318902818514,   0.6496653584298466,   0.917605405768162,
      0.761113851092484,    0.8791795132792125,    0.44449868173307283,   0.5914222027681203,   0.7437116188736885,
      0.8411100928304986,   0.9805574793354571,    0.8848126150904408,    0.45594654788563493,  0.017593478154503073,
      0.62370311211052,     0.02129918275230247,   0.830052343780784,     0.7060157268133952,   0.5317392897133469,
      0.9958462749035231,   0.862029888701881,     0.8017745565146187,    0.6782413640337536,   0.7696659137833982,
      0.7982349711510114,   0.7724279248361262,    0.5670304904635779,    0.8785872487273547,   0.004734651371140819,
      0.8861613857589661,   0.007148288012197801,  0.37446354377616475,   0.6690845068463606,   0.343469114076287,
      0.0857835687436056,   0.7646072372892175,    0.7188268519310974,    0.017394361915610057, 0.4467901200429536,
      0.7758773698800284,   0.9407991624965886,    0.7883518755638007,    0.661795939990852,    0.3065250204557999,
      0.1865674886269657,   0.4391735391309788,    0.13366433186289184,   0.4170857632470394,   0.27113574284356157,
      0.2504695401963266,   0.8378440756500211,    0.044089732363661094,  0.7096913513550435,   0.15170174477220055,
      0.4622738238198597,   0.6890756783794824,    0.08449102565738409,   0.21393977565249778,  0.9381592583428692,
      0.43909351458338675,  0.9898338931888807,    0.443592550980236,     0.9596830650619335,   0.2939490692189499,
      0.24247197683608523,  0.5095334554083597,    0.3450364035776594,    0.2459954395643219,   0.16556601945810856,
      0.09938394670258921,  0.3035151186951639,    0.4618194259116881,    0.8106885120353601,   0.028315604429974273,
      0.4686002937786584,   0.2407026393868288,    0.4321572361784677,    0.45463003066865415,  0.9453220019662685,
      0.36086002695619024,  0.7113731972624318,    0.8298597748035649,    0.6452195208794166,   0.8640851072532453,
      0.007494933722813291, 0.4532423883161699,    0.4358498394990138,    0.15849243309504923,  0.005941495896108373,
      0.5238548480735659,   0.7098499981352698,    0.15128698698720322,   0.8141904135220546,   0.7007554265498704,
      0.5212431731505457,   0.10445646966322941,   0.7395150888511759,    0.40115166637181754,  0.19932525488647057,
      0.6279086302955916,   0.547581203606972,     0.1470799829764481,    0.9898306221037652,   0.30063695907830557,
      0.6530263306464384,   0.6294304491599311,    0.4747038016224169,    0.5803620249830622,   0.007521198939880125,
      0.23594389336137767,  0.82876683912034,      0.834738368359478,     0.7474194348631291,   0.7112579137154827,
      0.9175597660672588,   0.9658333522412966,    0.5777144381034068,    0.10185947547633445,  0.28147168770097997,
      0.49084190134047334,  0.8584438346701353,    0.20139399288408033,   0.25502934165477764,  0.5982948387159709,
      0.16969711093899553,  0.1180404666237802,    0.6415827981200575,    0.7364746708042765,   0.21749292052270808,
      0.8399634215048786,   0.799877067418722,     0.1728040800468621,    0.36244880368539034,  0.7632940118923165,
      0.5105555039916515,   0.8960926950526985,    0.6945674442909555,    0.5303499842405933,   0.9146513147957873,
      0.5284394832775934,   0.04384851936719336,   0.4155261517501555,    0.9060172079755968,   0.6568173944147312,
      0.7586877459463743,   0.5548711377923504,    0.791556691837571,     0.7069769008278068,   0.869698410395238,
      0.025360358738519495, 0.32491506165874695,   0.5579428147048966,    0.9767059033674064,   0.8232568061375535,
      0.6486302438986716,   0.9572203123547415,    0.556525443906726,     0.8706467667318868,   0.5823811990622354,
      0.13357691464798138,  0.9446884919932719,    0.3346104779456238,    0.4569527670153789,   0.2908730982610468,
      0.6572566972190457,   0.532915432070937,     0.9495997246091489,    0.5010437198187823,   0.2048989668957929,
      0.40015851318727125,  0.9396139468185104,    0.09963121799246955,   0.8170741278189959,   0.5389460419617879,
      0.45899358803544554,  0.17217086189255992,   0.5547941031298822,    0.5987416086426901,   0.6894229831946286,
      0.5648254121117188,   0.7309773397133232,    0.6698552912716527,    0.270746137652799,    0.053705219700942,
      0.5063767684051376,   0.780388759983075,     0.7248139481233704,    0.35796688301072,     0.2772421373998536,
      0.7925588085624273,   0.7626497479115815,    0.20469063893497697,   0.04711937267440225,  0.45037796758318804,
      0.75066781795124,     0.09371763139775569,   0.23774692855986723,   0.23963041265644203,  0.22192123430700905,
      0.17616094680644334,  0.6737215431713256,    0.36874366609595643,   0.28396757247535576,  0.15162118086324772,
      0.30986446944527846,  0.18375410637584955,   0.9112190358293802,    0.690327705519752,    0.5491562617797402,
      0.5281492850328181,   0.26330626036820004,   0.9014654398254837,    0.1401554977665822,   0.5460266989110848,
      0.3113617142082512,   0.8026738672485313,    0.7932244713519243,    0.9280948067239118,   0.07148685551219547,
      0.2336754606791367,   0.7323878899162881,    0.9873043353493716,    0.7195158974755017,   0.541530670814326,
      0.6817726202638454,   0.4353807234830298,    0.9674534075204786,    0.21401989030280077,  0.002396651178786957,
      0.03442022110787202,  0.43477271771982284,   0.3050069506431471,    0.08089773378821119,  0.6291590014933156,
      0.5787097784473253,   0.6236653000860762,    0.461475864945851,     0.9981627789260562,   0.7313260977369199,
      0.713337515092525,    0.5224297784281214,    0.3200659468426875,    0.9107728219751681,   0.10737739320492923,
      0.3141596605465167,   0.44180687340667857,   0.032906572995198724,  0.3873591518740387,   0.48263368203794865,
      0.3349303332479078,   0.7598530618731676,    0.8659119048707833,    0.2464902259465227,   0.4553140241842599,
      0.9818987602819136,   0.5925084869238261,    0.4728058512337572,    0.05359923274228118,  0.4657556371687326,
      0.9111021456064748,   0.47155205473777895,   0.03929951143652333,   0.594695390220099,    0.919429204772624,
      0.9833083681599707,   0.8592328433935742,    0.6883643754551224,    0.8384109327495086,   0.03332507249957106,
      0.6149351679083243,   0.2719751089503445,    0.1575614925244445,    0.7163502359065218,   0.16259509931077443,
      0.37581000966714084,  0.80509149343632,      0.3421635067549482,    0.24314980248993667,  0.83824915279393,
      0.28370128790625326,  0.9787191868196575,    0.7595415381121309,    0.06549247203877806,  0.4282317439660449,
      0.8307200462203801,   0.7096499077090591,    0.36632445386556045,   0.2357917781969051,   0.12284932223048928,
      0.6615524634203029,   0.2322214574176914,    0.767028446649636,     0.8759690783659748,   0.8306436124180614,
      0.5798472290187143,   0.011471956973760244,  0.5773682448264239,    0.6035851122706498,   0.4704668571121897,
      0.17339338632884005,  0.6558446451611865,    0.7343833438614558,    0.14006001098754373,  0.6715959811967953,
      0.680394068488586,    0.41474303482746533,   0.9681288328156152,    0.233220125983571,    0.7310433251859529,
      0.4310140395683666,   0.47809152528622967,   0.05314850242766944,   0.35523994189400077,  0.5462996571286937,
      0.9657400895521029,   0.7056619496920902,    0.6766207358391074,    0.636139222807529,    0.47776444371478033,
      0.5456598510028011,   0.9060741761912895,    0.4261420957770189,    0.9687034478573067,   0.13598966999855922,
      0.04914274667362528,  0.1721447584349456,    0.07412119197941236,   0.012529344446802426, 0.19144189303663528,
      0.47147571375284814,  0.5721744237867166,    0.6201816276837667,    0.4845610368675931,   0.5003705454959895,
      0.6141593977164898,   0.006964022371572898,  0.6541048336512872,    0.41863524425463694,  0.042054099789581856,
      0.3412241888432692,   0.8575012513297512,    0.1212972250835862,    0.8459542642975082,   0.8049499889919485,
      0.8537472449635101,   0.9945715392091543,    0.5189399037858603,    0.667343472513223,    0.7450443706280852,
      0.40446452849666314,  0.2535302003984826,    0.4701897375183579,    0.5803426777495934,   0.5562287852609348,
      0.36275145211112614,  0.7418169336190366,    0.5689408381290824,    0.46298522943000064,  0.48400839758585756,
      0.1309162637452962,   0.5360842728537408,    0.09064214693350137,   0.7175920799518698,   0.9737001474297444,
      0.9092918118236812,   0.21327330991620574,   0.6285648741302996,    0.20654566517027428,  0.17460334682241974,
      0.4174038781545222,   0.6191394018077834,    0.17248149069917285,   0.2340700994442091,   0.5506923312964094,
      0.7749665246023152,   0.053157049516807664,  0.9298928386669005,    0.08960554649117725,  0.45757180036095835,
      0.2861359192158687,   0.36914916933156305,   0.16821297722710693,   0.7671537731273556,   0.7593751161265202,
      0.12622765437152517,  0.009183189735727626,  0.4709601630937571,    0.06755860607493347,  0.10689808095884157,
      0.9519318593697729,   0.42941454508912114,   0.3852737431079606,    0.4310771822646119,   0.7911885910393738,
      0.8657921575600451,   0.04992360626473957,   0.13267160640865372,   0.30243793737732894,  0.8985445127220502,
      0.302489320076742,    0.14895973120547823,   0.36171955618702634,   0.9861050189074767,   0.40132951312095344,
      0.1397935054459304,   0.7275163676699398,    0.24757531177032255,   0.6795152678016566,   0.9349783221173635,
      0.6017302317686811,   0.978850701352717,     0.782572010580318,     0.2503384778850122,   0.9211449821926627,
      0.7158918887968034,   0.5746635474611804,    0.31822595375881446,   0.016789744324152922, 0.6831064106036061,
      0.7966442996538724,   0.9913550891610706,    0.2231842692779209,    0.9180071448105787,   0.15459601428614667,
      0.55454466322197,     0.7320508107990614,    0.8376882158653953,    0.10659462196323277,  0.3718155827833767,
      0.8685785467629582,   0.30953641019091327,   0.3693961500059507,    0.6118908274540877,   0.9172186193742928,
      0.9638665261744293,   0.06411018686027592,   0.8779734145096334,    0.38726858040604806,  0.45042997757613545,
      0.22945572978660644,  0.3543109331520725,    0.45346727152383326,   0.21665688963627494,  0.8731427819714899,
      0.306988901348892,    0.9736037921150315,    0.35195278013601794,   0.5076890561556178,   0.6726982566869614,
      0.1415848835987198,   0.05267969117623417,   0.12930627273556972,   0.02467090616045342,  0.4287144407465886,
      0.5389555311153045,   0.9656927645096538,    0.06553857351644254,   0.7192727473967828,   0.9114851765836308,
      0.38569829114808596,  0.22442983249751902,   0.44212966241115104,   0.6258773970631035,   0.12375154698912139,
      0.26395669898138985,  0.6075560359436376,    0.7199341249146247,    0.2509506112091433,   0.9632237865411843,
      0.2699313871787882,   0.7923169276434796,    0.1319370980474388,    0.732461885956622,    0.9501669443744405,
      0.8366591904387193,   0.38289710756686746,   0.7317620554576368,    0.6891624698228346,   0.18053239868755144,
      0.25840864046822043,  0.49519103539917386,   0.03594820662269016,   0.272265164206111,    0.38571855983419834,
      0.18413692212953858,  0.4140665514245061,    0.331101485981824,     0.8364680728786794,   0.7150341406645073,
      0.19123529007541273,  0.8612879717194625,    0.05987516981191243,   0.04522592798730041,  0.6956985142613639,
      0.634731689137481,    0.6772010208911962,    0.2659724158347494,    0.8345253702291049,   0.12167573120389918,
      0.42570991590805674,  0.6447408662943692,    0.18608296200597518,   0.6486972068084338,   0.4119405321035826,
      0.9385896789116761,   0.21395244570112526,   0.33847889999990544,   0.17762967468331525,  0.24032378135246224,
      0.14994545679313542,  0.9967302039665963,    0.637507115987954,     0.9682515458629827,   0.7955362084226552,
      0.2414374570780865,   0.6450032164935743,    0.26982819483722165,   0.2908836237948339,   0.847995824351949,
      0.5654207482006985,   0.48178308190918895,   0.3189721351237217,    0.11948257056250688,  0.6329507004942213,
      0.5823940174941661,   0.9113691555501472,    0.02712104372887425,   0.33844084372877925,  0.409359485776034,
      0.16608774824486627,  0.1840501737732555,    0.9049256528358469,    0.40752327445425973,  0.3652540906020737,
      0.9860125291135977,   0.10047293005385105,   0.1471890531783654,    0.7661105179559459,   0.2193886006138287,
      0.8194138858161407,   0.6242049562872195,    0.5503285960900492,    0.7272713759543263,   0.6769310231737116,
      0.18811560788574455,  0.6296482965637312,    0.21513305795319226,   0.20440622491432292,  0.9893442282921309,
      0.021618970592347186, 0.5524594198562498,    0.8741638758927367,    0.42331698403095386,  0.3222338893974612,
      0.7513328175757993,   0.44505905665328027,   0.9302534567908484,    0.1771909927336177,   0.09209983389065812,
      0.21280609696170782,  0.35889687716970375,   0.36132216279703644,   0.8560345875874015,   0.43806594461121307,
      0.7644468243071734,   0.11103772926010236,   0.31084959197890827,   0.011781058613303919, 0.8660877509389459,
      0.7031384528295931,   0.37399657249679696,   0.47688438712340286,   0.928772577903708,    0.5596365181728887,
      0.08589741028223441,  0.07768770074191944,   0.6158032483682454,    0.8053366142728067,   0.3236249823421554,
      0.00822444546154455,  0.16070664756432096,   0.8638400429864582,    0.8698790447233965,   0.9681521716228816,
      0.9661412869875025,   0.9085095227101969,    0.1790385977070852,    0.776893063617345,    0.31898622890124084,
      0.5058802974670702,   0.818425585420191,     0.17870297652809786,   0.5441507345809006,   0.022489996885331487,
      0.5829548995026376,   0.9918631665731272,    0.5055786252454072,    0.16136704513340805,  0.984699782767251,
      0.18754660546807156,  0.8853269573231654,    0.6783741471392065,    0.18726766084750612,  0.09674403006985133,
      0.9931343022347453,   0.2433432886640784,    0.0658699939740679,    0.7256042711333616,   0.08922742117815974,
      0.08811202713859834,  0.8821854015107203,    0.37613730048899263,   0.45857685361148903,  0.7019149642340254,
      0.2788585976377317,   0.5668621385905648,    0.7770746229113067,    0.7074398653385896,   0.2940684401683791,
      0.8147168836664814,   0.0887419464763255,    0.4188458869744829,    0.6238528546466062,   0.7145827173057614,
      0.3162106953425323,   0.7417870718217474,    0.9857679125734509,    0.5419252027429404,   0.8536288391503049,
      0.09314542994725039,  0.5518248962183375,    0.5769624715772051,    0.20475493002236655,  0.7070617835227128,
      0.34157322002831525,  0.8624241578915682,    0.28011809028393886,   0.9526702722469567,   0.3293403217991683,
      0.9767778331103008,   0.8521904076959731,    0.15447695419829233,   0.39029609682738287,  0.33134453444767853,
      0.7248634019855653,   0.5234613373790115,    0.6664252543793544,    0.1511696460358204,   0.49869013399498285,
      0.4587900976769065,   0.9743791622455666,    0.931109609080415,     0.4053194525003626,   0.4666560847640946,
      0.22571944921717335,  0.26760280136617476,   0.4735972943090352,    0.47741158025412755,  0.526565707851956,
      0.720504118679755,    0.17389059420154984,   0.3941302129542221,    0.9769750936911633,   0.561054639152762,
      0.03366201552897685,  0.8725938986300577,    0.39225171270804193,   0.04403144974993356,  0.9823521603852087,
      0.1868618851608026,   0.3389503252328022,    0.6650312577642018,    0.7808072048924388,   0.5960817832633942,
      0.30958977269472954,  0.20842812187870807,   0.5526270573062637,    0.5402399164827054,   0.9213149201546954,
      0.9709344182765282,   0.6738010436288211,    0.6542430096117432,    0.5618971748260793,   0.29749853648546964,
      0.5432293708851752,   0.14093404302294743,   0.8745377717246547,    0.16720208620538812,  0.6914357121517429,
      0.7351008222411094,   0.5394045987877338,    0.936915662933375,     0.43842610640884583,  0.626896869991658,
      0.19551840539914045,  0.786663329801668,     0.8530609582658267,    0.8153132939963299,   0.35059087479820794,
      0.19737388285992674,  0.3995961251457969,    0.23790479221880045,   0.13141740731351004,  0.8726576278463952,
      0.11278100676094394,  0.4692693688236108,    0.7836610133411863,    0.9747036251493297,   0.4941714606912455,
      0.05305160105808626,  0.29958461037233053,   0.09534090496984604,   0.4638603168540302,   0.2791581608581686,
      0.3938045042892838,   0.6034024789176371,    0.22194186056037057,   0.9228439462893766,   0.18104328507119194,
      0.9165559738358574,   0.6716327950358132,    0.7332915821155096,    0.06778745965102362,  0.17251694852573773,
      0.9166902550664012,   0.6628517982577824,    0.9561014878015668,    0.6518257199411946,   0.8351982565565297,
      0.04860139460088986,  0.4551212110068099,    0.30944313225985576,   0.8412216162402254,   0.08951277138503844,
      0.025066292417908254, 0.7563828055308841,    0.2222539480109168,    0.8320411266990881,   0.9492800499708482,
      0.253253633473531,    0.5991520863203645,    0.2048707134105746,    0.47352200454570115,  0.5629789915185913,
      0.38864001352791677,  0.8068872022870146,    0.48246698916058783,   0.15465010252531697,  0.34985786042192857,
      0.4948504180620279,   0.18473893932854824,   0.6402301841568493,    0.15795832535816468,  0.41184033137111975,
      0.20994529454414756,  0.0795608311951832,    0.9661077058919707,    0.7190998979639045,   0.33427933653414255,
      0.13552737276834392,  0.012634417342274218,  0.8380990899424366,    0.2864221132522502,   0.4753501403209113,
      0.343423762898566,    0.086092548343773,     0.3156841132503668,    0.9971339679357473,   0.9932862480814585,
      0.9264437066213292,   0.18204780231842665,   0.12763429765746592,   0.451146025504531,    0.9820085929439971,
      0.27292955984128087,  0.7976683366910478,    0.8761611666855926,    0.32879915468366416,  0.4566805827078497,
      0.8941181691336204,   0.8608539041919699,    0.84204761146793,      0.1462692343227675,   0.5075573010849087,
      0.8903503433345549,   0.8700722120474124,    0.7712660332064996,    0.18532317502693563,  0.6548981638110877,
      0.48076406766202595,  0.32975091626883435,   0.6848997362918184,    0.35486660560902894,  0.0732106249839285,
      0.5574361427043032,   0.5714156890506997,    0.08811346271007414,   0.14783939754527176,  0.017003207714762714,
      0.44174703642978874,  0.8110740131813556,    0.574931463245764,     0.5102761455044432,   0.3424985993076507,
      0.8539951703461579,   0.7079277349244375,    0.4111481217664793,    0.951509130793907,    0.9555089087268872,
      0.7292094480906886,   0.4409066019171828,    0.08295791787636164,   0.9562723010283051,   0.7047746244055081,
      0.9072393279560211,   0.11549926102263153,   0.4681219537960889,    0.7943898405857803,   0.5483245331582209,
      0.2749400181421586,   0.37592691228594466,   0.45038405067793374,   0.5609330056140861,   0.9113280686842665,
      0.9992573696993887,   0.3012144754381675,    0.6302384260721924,    0.7990989912334494,   0.664341812871904,
      0.27280381214968086,  0.6742101055320844,    0.14710662300775412,   0.23944640184088473,  0.12926601852738606,
      0.5015443794327109,   0.647689169873493,     0.34550246822699693,   0.45351833013599707,  0.7288130733024423,
      0.46955053458590934,  0.2963560937896125,    0.05313715099993188,   0.9202060801646266,   0.40065138391111943,
      0.508501991516788,    0.9954587172603774,    0.4150966747261118,    0.5752642927349247,   0.15376590814347213,
      0.4718694982334253,   0.7791425089745136,    0.07116527059062705,   0.7379530846527411,   0.4298764101970639,
      0.3891997488372002,   0.2712956260859689,    0.7977010542015935,    0.6681098129873408,   0.10129548673467059,
      0.080480044589748,    0.8915772781308771,    0.42036967115181456,   0.36666062118419507,  0.6450046812325141,
      0.5524912580999269,   0.0839795459090027,    0.636617168184616,     0.531319360316121,    0.30375439556599226,
      0.5792609679185063,   0.814697212596663,     0.781899452578717,     0.6566031761406997,   0.37605072282304686,
      0.001881280677498931, 0.11552970112165872,   0.31858523647827175,   0.6583979311718429,   0.60526396245784,
      0.5611500316920904,   0.019979751727917994,  0.5148337803671061,    0.721749351841282,    0.9249303707149428,
      0.8743281776278087,   0.385262882179612,     0.95594196409197,      0.3779808468902329,   0.3810202153747745,
      0.5273305274418582,   0.7585796885360693,    0.04917726412895995,   0.5203446091374957,   0.6199363284810673,
      0.919887225009393,    0.7272035505938381,    0.966155481985703,     0.16762752670302516,  0.6189213811964427,
      0.20229470167487462,  0.37469353816237017,   0.017492223678952867,  0.2548479350612681,   0.3909836391230528,
      0.7738675159774514,   0.8606352509445246,    0.4139691275664631,    0.6225381036572762,   0.7352375292410388,
      0.30648818939511857,  0.9162607004972148,    0.09665526780685885,   0.3666447891178437,   0.1260323700300645,
      0.7172637798754251,   0.4790754927951175,    0.015451749872628873,  0.7715815662819216,   0.283564079660791,
      0.8852506071141146,   0.7832706823237181,    0.4481707153463781,    0.9552100378519058,   0.6378272080403014,
      0.3999266945419052,   0.763405910124729,     0.6670310568579203,    0.4302617641125852,   0.4364277580092467,
      0.09025576212085595,  0.5654935550366178,    0.17529425791527875,   0.7230546551485972,   0.5033159147423758,
      0.24678578258378903,  0.2911580454791233,    0.157525724298896,     0.35268716544351064,  0.6118146734984746,
      0.5685378955963167,   0.5729599421962962,    0.25698746611256706,   0.07729267856727107,  0.5996629884166993,
      0.19803754933587325,  0.8420450508984809,    0.3301131217627997,    0.35379997432878574,  0.8333412543498672,
      0.5982144085578488,   0.24512151253567938,   0.4812449632594248,    0.7867785122337513,   0.9586114553080439,
      0.6908428767174394,   0.9327970145025657,    0.27896398037168957,   0.6344034099803877,   0.5206197306406314,
      0.29959697003490815,  0.5484736202867737,    0.5143253458855047,    0.14555190477564528,  0.18566739299261037,
      0.32766071211385694,  0.5249453898611923,    0.3111331699242166,    0.7906655716605615,   0.38499709012670014,
      0.12136272756597788,  0.7052840640615085,    0.02543957664974894,   0.024816608620341962, 0.8580755998879297,
      0.1475584382157734,
  };

  TransArgs args{
      reinterpret_cast<uint8_t *>(data_4d), FORMAT_FRACTAL_Z, FORMAT_NCHW, {8, 2, 16, 16}, {17, 17, 2, 2}, DT_FLOAT};
  TransResult result;

  FormatTransferFracZNchw transfer;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(data));
  for (int i = 0; i < sizeof(data) / sizeof(data[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<float *>(result.data.get()))[i], data[i]);
  }
}

TEST_F(UtestFormatTransferFraczNchw, int8_1) {
  uint8_t ret[1 * 1 * 1 * 1] = {70};
  uint8_t data[1 * 1 * 16 * 32] = {
      70, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
      0,  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  };

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_FRACTAL_Z, FORMAT_NCHW, {1, 1, 16, 32}, {1, 1, 1, 1}, DT_INT8};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(ret) / sizeof(ret[0]));
  for (int i = 0; i < sizeof(ret) / sizeof(ret[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<uint8_t *>(result.data.get()))[i], ret[i]);
  }
}

TEST_F(UtestFormatTransferFraczNchw, uint8_2c_2n) {
  uint8_t ret[32 * 64 * 3 * 3] = {
      116, 236, 243, 202, 173, 49,  65,  195, 129, 147, 208, 52,  95,  22,  14,  0,   158, 123, 48,  214, 35,  238, 226,
      165, 2,   152, 179, 65,  128, 243, 151, 3,   125, 102, 201, 0,   40,  143, 164, 240, 58,  38,  208, 47,  98,  100,
      174, 44,  73,  56,  234, 197, 184, 44,  194, 8,   25,  56,  21,  212, 78,  69,  123, 33,  97,  232, 197, 162, 177,
      89,  214, 158, 93,  20,  117, 48,  95,  252, 253, 244, 153, 130, 131, 63,  28,  138, 102, 148, 196, 214, 44,  92,
      184, 205, 227, 94,  54,  120, 212, 108, 3,   188, 164, 193, 192, 65,  178, 15,  251, 111, 186, 189, 219, 125, 210,
      30,  1,   49,  89,  141, 253, 167, 223, 121, 13,  161, 249, 133, 41,  254, 51,  42,  163, 46,  203, 225, 46,  92,
      28,  62,  106, 68,  232, 21,  210, 128, 190, 23,  109, 116, 98,  124, 203, 23,  3,   119, 254, 214, 50,  184, 168,
      223, 163, 208, 45,  106, 247, 202, 25,  100, 152, 80,  226, 149, 178, 136, 37,  28,  199, 53,  144, 19,  232, 201,
      218, 223, 13,  160, 213, 249, 55,  108, 96,  199, 33,  61,  157, 54,  202, 80,  28,  233, 104, 125, 167, 0,   155,
      157, 233, 95,  111, 15,  225, 2,   95,  166, 185, 205, 203, 135, 51,  40,  192, 21,  203, 38,  129, 233, 73,  184,
      145, 206, 45,  126, 167, 170, 77,  202, 98,  180, 59,  231, 130, 187, 242, 199, 40,  223, 159, 220, 206, 147, 87,
      249, 251, 189, 122, 84,  183, 31,  66,  122, 159, 249, 47,  179, 81,  88,  77,  13,  245, 122, 62,  146, 211, 198,
      145, 149, 165, 235, 132, 96,  229, 210, 28,  207, 208, 29,  197, 214, 131, 27,  2,   217, 212, 51,  233, 27,  44,
      226, 210, 139, 68,  24,  186, 9,   53,  114, 44,  173, 13,  141, 85,  60,  15,  156, 78,  197, 64,  183, 204, 217,
      46,  200, 7,   160, 132, 173, 108, 251, 110, 4,   222, 40,  254, 99,  92,  156, 5,   69,  84,  243, 4,   156, 26,
      20,  124, 0,   251, 212, 241, 151, 153, 48,  193, 249, 70,  48,  32,  214, 101, 242, 38,  21,  248, 57,  124, 161,
      104, 26,  143, 217, 25,  107, 200, 194, 191, 241, 36,  211, 16,  164, 114, 181, 106, 59,  176, 201, 199, 231, 72,
      162, 98,  158, 10,  136, 42,  192, 32,  242, 32,  162, 169, 52,  246, 174, 221, 223, 95,  162, 185, 245, 29,  129,
      140, 187, 239, 30,  193, 130, 119, 2,   39,  251, 227, 103, 94,  253, 151, 71,  208, 251, 220, 56,  74,  79,  23,
      192, 42,  56,  65,  73,  246, 29,  195, 78,  58,  120, 203, 165, 163, 201, 198, 217, 163, 116, 78,  74,  7,   195,
      198, 90,  88,  123, 71,  128, 80,  141, 24,  157, 159, 55,  250, 149, 184, 38,  139, 86,  32,  191, 24,  204, 82,
      116, 195, 146, 206, 141, 103, 221, 44,  76,  230, 13,  110, 0,   184, 186, 48,  130, 232, 34,  190, 119, 48,  167,
      116, 149, 199, 66,  22,  237, 68,  79,  194, 217, 110, 97,  32,  21,  207, 110, 23,  1,   191, 224, 97,  127, 166,
      204, 49,  223, 142, 92,  158, 199, 6,   34,  211, 73,  72,  173, 74,  30,  103, 62,  229, 192, 30,  102, 57,  222,
      212, 118, 132, 212, 190, 126, 181, 149, 158, 223, 29,  132, 167, 156, 48,  85,  246, 217, 88,  124, 131, 159, 237,
      252, 60,  81,  54,  71,  76,  82,  242, 78,  167, 192, 250, 5,   142, 9,   27,  168, 246, 75,  171, 200, 128, 119,
      49,  241, 101, 4,   222, 182, 23,  154, 39,  215, 108, 79,  40,  182, 79,  79,  49,  20,  69,  240, 105, 246, 215,
      111, 36,  108, 192, 69,  223, 179, 205, 100, 136, 13,  160, 93,  91,  132, 98,  39,  69,  246, 141, 237, 135, 55,
      151, 51,  153, 39,  129, 139, 59,  67,  90,  197, 231, 47,  129, 40,  241, 165, 191, 131, 18,  95,  147, 135, 49,
      228, 186, 84,  161, 178, 64,  123, 229, 28,  71,  112, 139, 195, 21,  99,  106, 31,  118, 101, 121, 7,   91,  128,
      136, 124, 147, 8,   241, 227, 130, 39,  151, 123, 234, 239, 240, 232, 211, 110, 176, 32,  91,  153, 230, 99,  53,
      174, 105, 231, 236, 115, 146, 97,  87,  172, 25,  204, 221, 103, 23,  208, 83,  125, 81,  87,  18,  233, 126, 188,
      244, 102, 140, 120, 37,  68,  128, 36,  177, 23,  236, 65,  70,  217, 217, 51,  189, 170, 99,  242, 77,  187, 153,
      85,  70,  111, 176, 94,  108, 100, 17,  67,  226, 89,  50,  201, 87,  106, 210, 234, 156, 51,  88,  9,   204, 4,
      132, 196, 163, 144, 232, 14,  49,  124, 77,  230, 81,  146, 128, 112, 226, 2,   230, 72,  106, 254, 126, 130, 191,
      2,   104, 232, 136, 73,  202, 174, 183, 32,  174, 27,  185, 186, 88,  173, 165, 62,  214, 23,  166, 35,  27,  9,
      106, 104, 144, 133, 112, 231, 191, 10,  8,   169, 204, 70,  149, 175, 2,   116, 26,  142, 44,  127, 123, 58,  56,
      203, 215, 226, 196, 81,  244, 147, 68,  74,  79,  122, 230, 79,  132, 121, 133, 93,  198, 240, 12,  222, 134, 141,
      84,  46,  2,   164, 39,  173, 159, 219, 24,  200, 0,   78,  41,  113, 82,  168, 214, 112, 237, 253, 149, 163, 119,
      244, 99,  218, 150, 153, 182, 109, 134, 87,  156, 175, 133, 72,  158, 166, 146, 181, 38,  135, 12,  138, 217, 157,
      187, 103, 200, 238, 239, 46,  26,  127, 20,  205, 101, 143, 79,  51,  221, 90,  204, 10,  216, 175, 95,  17,  78,
      63,  152, 111, 108, 168, 118, 150, 208, 194, 183, 23,  73,  36,  48,  11,  183, 126, 169, 222, 112, 189, 185, 157,
      59,  130, 174, 219, 216, 175, 140, 192, 69,  141, 224, 79,  32,  12,  44,  190, 199, 189, 76,  26,  103, 9,   43,
      22,  80,  218, 135, 235, 101, 253, 75,  154, 32,  216, 185, 145, 38,  243, 189, 234, 154, 180, 36,  130, 152, 138,
      22,  236, 147, 71,  135, 205, 76,  118, 25,  3,   178, 252, 222, 207, 131, 207, 181, 249, 50,  232, 51,  111, 213,
      43,  126, 43,  65,  193, 68,  214, 12,  76,  116, 135, 118, 187, 233, 123, 75,  58,  36,  251, 46,  184, 119, 105,
      226, 182, 103, 185, 57,  219, 175, 240, 251, 2,   193, 44,  173, 166, 45,  140, 185, 90,  130, 45,  110, 225, 54,
      3,   94,  175, 208, 154, 150, 146, 12,  37,  89,  253, 71,  171, 191, 30,  139, 121, 94,  51,  16,  110, 100, 72,
      27,  87,  199, 230, 73,  95,  202, 79,  90,  20,  55,  181, 157, 180, 11,  116, 238, 236, 51,  24,  251, 119, 211,
      210, 40,  107, 186, 66,  43,  148, 37,  27,  182, 217, 65,  235, 20,  9,   54,  166, 99,  48,  182, 254, 149, 212,
      188, 55,  124, 50,  163, 177, 2,   109, 98,  246, 194, 16,  250, 203, 181, 0,   31,  183, 151, 222, 168, 145, 44,
      73,  109, 28,  75,  221, 205, 106, 56,  254, 24,  25,  182, 152, 216, 45,  31,  163, 185, 134, 241, 168, 236, 74,
      212, 152, 172, 4,   206, 72,  252, 213, 156, 241, 153, 204, 205, 95,  116, 212, 237, 132, 125, 29,  166, 127, 144,
      170, 28,  183, 12,  221, 217, 9,   129, 187, 41,  112, 29,  192, 40,  40,  162, 125, 157, 32,  191, 17,  215, 57,
      55,  101, 143, 130, 160, 213, 139, 6,   24,  3,   227, 193, 142, 228, 73,  86,  205, 184, 46,  187, 95,  85,  23,
      34,  136, 76,  246, 225, 13,  26,  138, 191, 143, 183, 98,  178, 195, 172, 92,  115, 244, 239, 6,   113, 101, 74,
      237, 75,  184, 186, 36,  43,  178, 244, 150, 16,  68,  123, 46,  94,  206, 139, 153, 144, 107, 11,  139, 76,  181,
      129, 38,  225, 250, 207, 174, 181, 129, 46,  103, 145, 219, 121, 124, 167, 102, 205, 115, 148, 30,  179, 219, 144,
      209, 99,  40,  186, 92,  214, 17,  184, 36,  0,   154, 188, 100, 131, 223, 47,  110, 31,  2,   58,  210, 212, 139,
      161, 184, 48,  168, 222, 241, 15,  247, 23,  21,  3,   36,  120, 235, 50,  125, 45,  187, 187, 78,  13,  176, 123,
      94,  118, 134, 35,  143, 85,  129, 82,  129, 218, 21,  93,  71,  81,  151, 10,  68,  21,  172, 136, 146, 230, 230,
      206, 144, 158, 204, 207, 101, 65,  30,  29,  33,  246, 131, 143, 254, 12,  74,  29,  193, 251, 185, 118, 107, 143,
      107, 104, 124, 32,  109, 245, 150, 239, 106, 81,  89,  158, 181, 159, 87,  99,  112, 60,  126, 46,  204, 143, 15,
      93,  252, 62,  191, 159, 155, 232, 92,  171, 31,  30,  101, 74,  25,  94,  174, 58,  162, 166, 250, 167, 127, 23,
      203, 87,  250, 145, 60,  12,  169, 50,  174, 96,  87,  21,  114, 37,  50,  73,  131, 198, 129, 154, 187, 127, 191,
      145, 105, 65,  71,  61,  48,  45,  99,  136, 145, 240, 74,  55,  8,   199, 109, 168, 101, 252, 61,  4,   163, 57,
      26,  147, 182, 240, 87,  203, 209, 71,  123, 23,  196, 72,  246, 198, 74,  170, 40,  37,  205, 178, 190, 24,  40,
      25,  44,  224, 206, 62,  57,  27,  75,  188, 212, 246, 245, 240, 86,  65,  56,  160, 89,  169, 246, 227, 72,  3,
      51,  250, 31,  126, 78,  139, 185, 59,  230, 133, 193, 215, 224, 76,  235, 12,  64,  124, 176, 174, 173, 125, 113,
      61,  236, 43,  215, 177, 31,  175, 139, 8,   243, 76,  87,  150, 173, 229, 101, 81,  201, 71,  98,  165, 244, 137,
      89,  155, 18,  183, 59,  104, 216, 215, 13,  83,  85,  76,  80,  179, 204, 176, 150, 162, 133, 106, 123, 207, 213,
      105, 161, 243, 46,  179, 201, 82,  63,  62,  230, 226, 178, 110, 13,  147, 188, 164, 203, 10,  232, 71,  77,  179,
      39,  94,  100, 194, 81,  48,  216, 190, 242, 244, 60,  116, 150, 69,  101, 156, 231, 91,  100, 243, 236, 106, 65,
      146, 20,  126, 19,  55,  21,  221, 45,  245, 155, 233, 218, 161, 250, 138, 193, 177, 53,  146, 230, 204, 163, 144,
      148, 85,  91,  144, 227, 138, 71,  54,  165, 36,  117, 30,  149, 54,  87,  198, 177, 246, 252, 36,  63,  134, 209,
      144, 190, 111, 24,  145, 44,  158, 1,   117, 224, 223, 155, 147, 74,  10,  174, 98,  244, 29,  9,   138, 141, 93,
      27,  236, 175, 187, 131, 86,  161, 177, 153, 153, 189, 228, 132, 1,   164, 190, 14,  117, 56,  215, 68,  103, 3,
      190, 201, 127, 112, 48,  207, 36,  33,  79,  36,  194, 138, 185, 48,  72,  111, 90,  0,   148, 167, 123, 49,  164,
      86,  178, 40,  178, 66,  2,   197, 196, 108, 105, 118, 191, 205, 33,  172, 208, 162, 74,  174, 47,  166, 87,  205,
      6,   215, 68,  20,  157, 125, 241, 74,  31,  239, 161, 219, 143, 118, 29,  111, 44,  96,  23,  16,  245, 233, 164,
      201, 59,  221, 115, 238, 146, 9,   98,  36,  202, 64,  25,  144, 93,  157, 57,  192, 125, 200, 96,  137, 154, 80,
      112, 117, 254, 246, 159, 98,  180, 45,  101, 145, 189, 128, 199, 216, 176, 194, 238, 33,  4,   203, 226, 140, 50,
      196, 57,  249, 66,  229, 60,  46,  98,  202, 230, 65,  234, 169, 212, 10,  0,   68,  99,  171, 68,  1,   91,  81,
      89,  166, 209, 23,  245, 241, 150, 62,  65,  75,  183, 151, 55,  94,  160, 37,  200, 9,   111, 15,  235, 223, 179,
      200, 120, 53,  90,  145, 106, 183, 195, 37,  209, 44,  237, 77,  223, 144, 232, 86,  233, 73,  224, 60,  16,  234,
      254, 12,  45,  200, 208, 76,  13,  41,  97,  65,  140, 231, 2,   223, 224, 153, 182, 210, 54,  1,   95,  198, 204,
      232, 239, 250, 161, 204, 250, 123, 115, 53,  46,  113, 68,  0,   249, 178, 191, 48,  122, 85,  130, 154, 231, 29,
      245, 184, 184, 237, 225, 89,  22,  168, 105, 85,  162, 18,  100, 242, 39,  175, 8,   75,  202, 77,  210, 20,  124,
      70,  229, 78,  179, 6,   139, 243, 83,  92,  144, 114, 188, 67,  231, 246, 136, 220, 176, 67,  56,  186, 28,  216,
      126, 233, 104, 197, 178, 77,  204, 220, 81,  29,  124, 198, 158, 123, 175, 190, 219, 57,  47,  209, 162, 224, 216,
      121, 161, 24,  79,  48,  189, 164, 239, 232, 191, 193, 97,  21,  86,  35,  219, 236, 86,  161, 27,  119, 224, 144,
      34,  179, 250, 52,  242, 223, 102, 185, 84,  163, 245, 178, 218, 158, 124, 68,  57,  190, 9,   79,  171, 1,   45,
      223, 100, 151, 249, 205, 37,  235, 224, 135, 81,  119, 209, 183, 68,  158, 179, 30,  121, 202, 74,  206, 115, 250,
      147, 245, 182, 33,  55,  152, 46,  139, 166, 154, 186, 69,  227, 229, 85,  218, 178, 172, 110, 120, 225, 69,  212,
      101, 15,  19,  91,  228, 89,  0,   195, 62,  230, 2,   136, 143, 122, 86,  71,  108, 32,  254, 21,  106, 247, 65,
      195, 45,  225, 248, 105, 30,  225, 127, 227, 41,  232, 27,  254, 162, 104, 227, 97,  187, 227, 144, 18,  225, 6,
      242, 189, 45,  114, 153, 48,  165, 152, 174, 87,  223, 122, 176, 35,  245, 221, 195, 186, 98,  185, 119, 21,  133,
      43,  193, 113, 37,  239, 79,  97,  36,  199, 1,   253, 185, 233, 167, 199, 102, 77,  85,  151, 153, 165, 161, 112,
      246, 77,  232, 210, 82,  119, 164, 121, 235, 251, 192, 137, 194, 45,  79,  117, 8,   20,  9,   4,   71,  26,  168,
      191, 251, 126, 152, 114, 133, 24,  137, 9,   222, 189, 24,  236, 158, 192, 108, 251, 206, 190, 134, 235, 216, 219,
      116, 70,  48,  116, 136, 182, 253, 177, 59,  161, 64,  38,  31,  248, 217, 134, 17,  105, 24,  194, 243, 135, 93,
      56,  148, 203, 7,   242, 53,  28,  171, 254, 213, 38,  29,  22,  153, 156, 127, 216, 225, 28,  16,  40,  186, 146,
      46,  64,  15,  175, 213, 9,   97,  55,  82,  103, 133, 27,  167, 109, 3,   56,  195, 178, 214, 127, 52,  198, 118,
      28,  186, 49,  209, 239, 176, 211, 205, 73,  132, 111, 215, 99,  230, 85,  39,  226, 28,  222, 252, 32,  60,  155,
      92,  1,   26,  70,  157, 224, 186, 183, 247, 142, 11,  127, 254, 65,  86,  197, 98,  251, 17,  241, 99,  215, 159,
      140, 33,  252, 190, 201, 142, 33,  134, 106, 249, 91,  11,  102, 161, 221, 91,  188, 78,  127, 204, 95,  0,   249,
      159, 49,  22,  107, 180, 138, 41,  245, 145, 4,   108, 219, 32,  243, 186, 137, 104, 237, 194, 160, 171, 72,  239,
      72,  176, 10,  41,  151, 234, 74,  7,   161, 145, 203, 72,  42,  228, 162, 134, 206, 252, 18,  14,  202, 67,  55,
      185, 59,  122, 218, 1,   185, 45,  97,  181, 13,  89,  9,   8,   206, 226, 112, 2,   130, 51,  192, 105, 200, 47,
      79,  235, 146, 78,  205, 116, 21,  113, 186, 227, 32,  172, 168, 229, 74,  115, 136, 212, 218, 198, 17,  15,  238,
      118, 182, 85,  91,  143, 234, 187, 130, 237, 90,  104, 221, 97,  164, 183, 33,  99,  45,  233, 228, 89,  18,  177,
      205, 190, 119, 33,  49,  136, 1,   196, 97,  95,  33,  69,  244, 108, 58,  158, 156, 56,  170, 98,  88,  8,   75,
      186, 157, 240, 231, 145, 204, 142, 122, 160, 167, 144, 172, 251, 214, 22,  225, 8,   89,  243, 153, 204, 128, 193,
      187, 84,  231, 9,   55,  196, 198, 14,  54,  106, 121, 95,  231, 229, 48,  97,  90,  113, 124, 250, 126, 4,   111,
      211, 171, 22,  121, 42,  151, 151, 217, 51,  248, 69,  161, 245, 33,  119, 34,  54,  129, 46,  39,  42,  157, 145,
      219, 125, 194, 201, 169, 64,  106, 64,  17,  101, 138, 89,  148, 63,  2,   20,  202, 24,  109, 42,  236, 57,  95,
      250, 53,  41,  163, 5,   67,  194, 128, 253, 76,  224, 216, 100, 187, 32,  34,  6,   87,  196, 201, 151, 193, 133,
      65,  144, 53,  198, 47,  84,  9,   41,  154, 128, 169, 58,  158, 238, 178, 96,  172, 215, 37,  171, 28,  62,  9,
      107, 198, 21,  208, 157, 31,  243, 115, 240, 49,  172, 114, 235, 12,  23,  197, 231, 8,   54,  14,  16,  114, 108,
      52,  221, 155, 140, 33,  204, 132, 100, 106, 220, 130, 113, 168, 44,  193, 3,   66,  250, 79,  232, 54,  20,  108,
      83,  176, 133, 130, 236, 6,   216, 161, 101, 108, 86,  127, 98,  179, 22,  225, 210, 111, 86,  216, 66,  145, 190,
      248, 114, 117, 124, 154, 34,  133, 155, 1,   203, 23,  180, 84,  225, 140, 7,   25,  63,  64,  180, 190, 110, 215,
      181, 1,   98,  142, 183, 222, 90,  218, 114, 175, 152, 116, 34,  248, 80,  195, 54,  169, 183, 23,  115, 55,  212,
      124, 29,  211, 117, 162, 65,  111, 203, 87,  109, 176, 223, 20,  153, 249, 38,  138, 72,  104, 162, 243, 73,  52,
      116, 114, 12,  82,  184, 247, 70,  82,  196, 219, 143, 17,  200, 219, 79,  95,  18,  154, 225, 125, 86,  201, 73,
      194, 249, 37,  54,  150, 141, 35,  179, 7,   19,  13,  10,  30,  98,  224, 239, 98,  15,  159, 170, 245, 149, 39,
      6,   162, 83,  123, 104, 211, 25,  245, 249, 153, 175, 254, 181, 254, 167, 53,  86,  110, 111, 204, 5,   230, 230,
      181, 74,  223, 85,  123, 99,  118, 221, 52,  106, 214, 209, 163, 12,  143, 174, 194, 214, 17,  201, 200, 119, 87,
      112, 128, 27,  130, 220, 7,   209, 29,  164, 6,   144, 4,   225, 9,   240, 106, 108, 50,  206, 32,  123, 141, 58,
      79,  173, 180, 225, 51,  179, 177, 248, 44,  44,  55,  172, 146, 176, 166, 26,  114, 231, 79,  113, 56,  175, 149,
      213, 215, 161, 47,  149, 221, 9,   22,  239, 155, 192, 91,  9,   46,  145, 164, 248, 70,  220, 150, 62,  104, 203,
      254, 50,  123, 108, 175, 68,  54,  178, 28,  101, 189, 13,  30,  79,  65,  235, 143, 40,  54,  149, 213, 14,  107,
      103, 11,  130, 24,  174, 129, 128, 96,  137, 210, 189, 138, 222, 19,  200, 17,  188, 150, 16,  67,  28,  232, 240,
      50,  54,  86,  117, 0,   128, 163, 70,  14,  67,  39,  100, 44,  42,  24,  136, 10,  42,  237, 79,  252, 98,  233,
      245, 63,  34,  80,  172, 83,  18,  58,  32,  10,  148, 137, 12,  47,  81,  62,  76,  250, 68,  230, 90,  18,  139,
      185, 158, 7,   236, 225, 54,  50,  237, 53,  85,  48,  29,  175, 235, 76,  199, 127, 67,  123, 180, 193, 49,  218,
      83,  239, 72,  24,  174, 200, 165, 39,  169, 97,  192, 42,  20,  58,  93,  124, 253, 40,  202, 72,  39,  88,  108,
      221, 171, 196, 245, 238, 184, 209, 35,  98,  111, 218, 196, 211, 65,  44,  239, 51,  225, 137, 87,  226, 141, 186,
      225, 68,  163, 33,  152, 77,  37,  120, 236, 102, 134, 30,  252, 170, 51,  10,  36,  122, 214, 51,  162, 159, 178,
      164, 15,  137, 60,  248, 195, 115, 251, 102, 149, 90,  210, 151, 222, 175, 60,  189, 182, 179, 162, 250, 159, 91,
      233, 164, 88,  9,   51,  153, 57,  122, 177, 247, 96,  133, 186, 99,  60,  153, 178, 205, 44,  26,  117, 79,  243,
      217, 184, 3,   253, 108, 9,   218, 195, 9,   104, 247, 68,  58,  60,  253, 98,  44,  94,  31,  58,  132, 244, 180,
      13,  178, 189, 208, 31,  57,  61,  212, 127, 100, 184, 49,  64,  104, 46,  136, 78,  219, 210, 108, 181, 247, 207,
      184, 250, 182, 31,  147, 194, 118, 84,  40,  61,  41,  245, 227, 159, 56,  118, 124, 26,  203, 150, 168, 210, 78,
      105, 133, 191, 185, 170, 82,  81,  84,  179, 112, 155, 131, 238, 172, 99,  253, 201, 168, 47,  92,  148, 216, 63,
      87,  187, 27,  110, 58,  182, 196, 112, 206, 164, 71,  82,  197, 36,  157, 151, 160, 219, 113, 6,   247, 99,  139,
      79,  187, 245, 125, 36,  132, 181, 72,  120, 34,  72,  51,  33,  143, 135, 149, 30,  83,  155, 151, 177, 102, 3,
      179, 76,  144, 9,   77,  116, 6,   16,  28,  208, 194, 68,  150, 26,  38,  94,  179, 123, 108, 21,  196, 10,  2,
      164, 75,  118, 147, 231, 19,  222, 69,  224, 174, 59,  198, 80,  126, 199, 32,  224, 173, 188, 251, 7,   15,  18,
      115, 208, 228, 128, 160, 107, 115, 218, 195, 80,  217, 192, 87,  209, 182, 215, 90,  65,  52,  248, 68,  167, 18,
      120, 226, 68,  211, 246, 179, 82,  36,  3,   222, 212, 218, 39,  206, 190, 141, 245, 210, 9,   92,  126, 31,  106,
      64,  159, 14,  191, 32,  38,  246, 237, 85,  69,  48,  215, 62,  249, 14,  74,  211, 99,  209, 247, 194, 252, 9,
      210, 84,  110, 98,  169, 176, 15,  98,  110, 19,  114, 118, 160, 193, 14,  169, 249, 63,  240, 210, 170, 211, 10,
      24,  142, 70,  20,  111, 175, 251, 89,  219, 25,  56,  205, 16,  15,  52,  50,  177, 226, 81,  106, 149, 128, 87,
      153, 219, 117, 151, 195, 201, 74,  177, 128, 146, 173, 247, 160, 29,  142, 107, 155, 153, 93,  39,  67,  45,  252,
      143, 208, 48,  201, 199, 91,  183, 6,   247, 21,  57,  191, 19,  173, 212, 20,  5,   101, 147, 72,  190, 160, 132,
      108, 171, 242, 102, 50,  56,  211, 216, 129, 177, 210, 3,   247, 194, 11,  36,  116, 131, 198, 1,   148, 88,  78,
      143, 48,  175, 190, 244, 2,   84,  167, 202, 190, 27,  216, 249, 155, 36,  167, 111, 47,  64,  66,  109, 245, 113,
      228, 80,  212, 228, 60,  100, 98,  4,   176, 250, 96,  7,   40,  67,  152, 209, 226, 69,  61,  16,  71,  178, 52,
      138, 220, 77,  210, 93,  133, 168, 99,  159, 14,  61,  248, 38,  17,  195, 177, 10,  110, 12,  132, 242, 53,  193,
      185, 236, 20,  118, 3,   252, 221, 168, 132, 179, 226, 67,  133, 229, 228, 34,  89,  124, 198, 252, 79,  141, 223,
      7,   199, 201, 126, 100, 25,  27,  125, 47,  252, 79,  55,  164, 21,  176, 179, 44,  131, 234, 57,  59,  113, 179,
      11,  73,  1,   30,  252, 39,  98,  84,  167, 99,  35,  55,  65,  54,  159, 111, 15,  158, 9,   134, 61,  213, 226,
      235, 134, 61,  186, 198, 212, 183, 98,  1,   28,  16,  75,  3,   150, 193, 200, 90,  123, 69,  104, 212, 113, 170,
      106, 100, 136, 91,  19,  196, 89,  79,  129, 111, 226, 161, 59,  223, 109, 243, 199, 249, 143, 181, 230, 231, 68,
      173, 126, 138, 91,  9,   119, 81,  238, 154, 86,  0,   190, 214, 145, 71,  223, 48,  78,  150, 68,  167, 143, 189,
      87,  216, 98,  52,  94,  109, 45,  181, 2,   170, 102, 248, 95,  116, 98,  99,  91,  147, 242, 153, 121, 164, 76,
      14,  92,  75,  209, 33,  47,  17,  242, 180, 59,  6,   74,  80,  35,  136, 89,  60,  53,  79,  84,  176, 225, 52,
      249, 29,  65,  75,  147, 123, 67,  247, 115, 197, 228, 28,  160, 250, 118, 207, 67,  63,  38,  129, 63,  223, 218,
      238, 148, 9,   137, 189, 182, 14,  103, 225, 86,  21,  92,  57,  100, 197, 220, 250, 185, 41,  173, 73,  4,   9,
      165, 194, 216, 90,  59,  239, 252, 20,  99,  95,  155, 51,  60,  14,  189, 129, 127, 101, 34,  186, 31,  116, 226,
      52,  162, 250, 72,  174, 221, 218, 154, 82,  218, 176, 246, 2,   44,  97,  153, 170, 146, 216, 171, 53,  246, 70,
      248, 163, 241, 83,  9,   204, 185, 180, 40,  9,   184, 54,  198, 211, 168, 202, 63,  13,  244, 223, 216, 155, 105,
      115, 59,  123, 168, 251, 252, 137, 175, 138, 253, 16,  82,  160, 238, 140, 233, 20,  142, 173, 229, 52,  96,  137,
      98,  89,  119, 114, 128, 159, 203, 22,  204, 101, 186, 80,  118, 217, 106, 53,  120, 137, 187, 98,  251, 172, 151,
      208, 34,  230, 254, 10,  95,  210, 130, 54,  211, 243, 21,  19,  188, 30,  113, 194, 16,  201, 101, 166, 79,  192,
      81,  86,  56,  174, 122, 197, 116, 251, 224, 50,  248, 37,  51,  169, 6,   88,  3,   231, 44,  192, 78,  18,  7,
      102, 144, 53,  181, 223, 17,  147, 168, 160, 20,  22,  74,  243, 146, 141, 86,  165, 44,  17,  188, 84,  236, 2,
      110, 45,  154, 134, 152, 158, 190, 71,  130, 77,  89,  49,  213, 65,  135, 171, 185, 176, 173, 181, 169, 248, 23,
      76,  227, 241, 205, 168, 229, 91,  42,  53,  102, 95,  186, 254, 91,  155, 116, 95,  254, 70,  194, 3,   28,  221,
      137, 188, 139, 164, 103, 113, 232, 233, 16,  124, 177, 36,  152, 157, 216, 119, 230, 73,  100, 133, 143, 101, 122,
      145, 0,   56,  2,   129, 62,  103, 134, 201, 145, 187, 189, 90,  180, 159, 0,   106, 154, 130, 121, 228, 27,  43,
      145, 147, 20,  191, 66,  82,  23,  214, 242, 202, 199, 53,  143, 164, 3,   246, 100, 15,  172, 59,  227, 129, 161,
      115, 97,  239, 104, 85,  201, 167, 96,  12,  58,  102, 213, 104, 191, 64,  211, 59,  112, 35,  120, 83,  143, 240,
      106, 47,  244, 41,  189, 254, 175, 21,  181, 141, 166, 197, 133, 44,  238, 175, 134, 170, 37,  23,  90,  87,  121,
      102, 59,  54,  219, 82,  120, 194, 127, 152, 177, 63,  229, 69,  234, 235, 103, 128, 178, 97,  60,  126, 217, 245,
      246, 242, 220, 69,  167, 98,  235, 121, 76,  42,  148, 206, 199, 26,  133, 143, 56,  253, 200, 204, 154, 36,  158,
      120, 195, 145, 149, 214, 125, 193, 127, 49,  201, 87,  65,  75,  14,  199, 46,  242, 126, 179, 194, 211, 6,   34,
      176, 127, 99,  30,  101, 21,  49,  8,   7,   206, 97,  229, 63,  231, 43,  51,  33,  201, 206, 65,  82,  12,  253,
      176, 200, 182, 153, 48,  55,  80,  113, 245, 207, 237, 151, 6,   41,  135, 253, 74,  208, 170, 229, 36,  92,  26,
      24,  226, 239, 77,  141, 59,  222, 148, 78,  32,  79,  65,  195, 111, 28,  243, 193, 84,  59,  46,  22,  155, 137,
      149, 113, 186, 16,  115, 211, 18,  138, 74,  234, 20,  183, 203, 12,  140, 119, 72,  187, 188, 150, 137, 205, 20,
      233, 82,  86,  216, 127, 103, 145, 197, 57,  239, 27,  228, 69,  20,  175, 193, 229, 232, 68,  112, 120, 141, 6,
      121, 194, 70,  166, 204, 31,  134, 49,  54,  54,  199, 126, 26,  24,  237, 227, 38,  179, 31,  154, 143, 32,  22,
      105, 33,  232, 224, 156, 204, 152, 237, 38,  219, 221, 44,  73,  16,  251, 206, 56,  180, 194, 181, 2,   64,  205,
      163, 162, 194, 230, 64,  3,   6,   9,   25,  16,  115, 195, 158, 109, 78,  159, 49,  167, 150, 159, 57,  9,   144,
      71,  128, 200, 254, 136, 19,  104, 37,  36,  26,  229, 112, 139, 139, 40,  242, 5,   128, 131, 33,  43,  148, 206,
      17,  153, 116, 74,  116, 40,  40,  85,  160, 231, 240, 185, 239, 76,  66,  144, 228, 137, 253, 212, 139, 115, 62,
      137, 203, 90,  207, 53,  226, 141, 148, 198, 212, 183, 165, 42,  208, 37,  17,  71,  183, 116, 186, 198, 249, 121,
      170, 159, 215, 87,  158, 40,  110, 117, 201, 93,  216, 64,  69,  86,  109, 80,  35,  142, 125, 1,   144, 62,  21,
      185, 61,  58,  230, 74,  144, 151, 26,  112, 128, 131, 20,  100, 209, 31,  81,  100, 102, 21,  211, 182, 52,  161,
      72,  154, 171, 186, 137, 209, 212, 169, 124, 23,  214, 143, 176, 198, 4,   4,   149, 73,  151, 56,  140, 253, 17,
      68,  142, 196, 246, 15,  191, 62,  233, 33,  87,  108, 173, 7,   96,  101, 140, 9,   193, 121, 71,  88,  135, 179,
      200, 10,  60,  161, 186, 167, 235, 98,  185, 174, 229, 16,  127, 216, 104, 159, 72,  154, 238, 36,  153, 132, 240,
      98,  132, 248, 64,  91,  52,  238, 166, 172, 158, 208, 123, 119, 189, 56,  6,   226, 69,  223, 106, 59,  3,   135,
      125, 184, 237, 19,  24,  189, 25,  119, 82,  13,  11,  167, 31,  220, 163, 73,  16,  141, 239, 222, 211, 5,   63,
      237, 182, 138, 199, 24,  168, 44,  125, 177, 164, 95,  185, 68,  177, 206, 37,  177, 61,  95,  97,  181, 128, 158,
      170, 172, 155, 143, 209, 216, 218, 220, 77,  209, 74,  207, 60,  115, 254, 104, 87,  65,  89,  224, 18,  55,  180,
      239, 95,  57,  137, 195, 245, 22,  0,   164, 207, 171, 142, 237, 233, 87,  56,  214, 182, 68,  0,   1,   225, 65,
      41,  240, 124, 68,  40,  245, 57,  237, 45,  177, 221, 137, 78,  11,  176, 84,  25,  251, 203, 161, 228, 26,  138,
      212, 240, 6,   21,  143, 161, 232, 108, 240, 35,  84,  148, 124, 36,  75,  239, 135, 150, 112, 45,  1,   116, 214,
      1,   59,  214, 62,  152, 143, 154, 166, 115, 32,  113, 153, 7,   191, 147, 202, 66,  64,  157, 186, 249, 164, 162,
      170, 226, 180, 234, 104, 134, 59,  168, 92,  52,  90,  106, 76,  239, 231, 14,  237, 15,  131, 81,  92,  9,   95,
      190, 21,  28,  96,  190, 44,  26,  212, 83,  34,  8,   106, 5,   119, 66,  72,  200, 136, 252, 212, 104, 131, 93,
      150, 37,  67,  188, 26,  105, 125, 201, 41,  248, 61,  238, 176, 254, 205, 157, 37,  194, 191, 239, 3,   128, 90,
      221, 250, 223, 180, 111, 131, 174, 22,  235, 151, 229, 108, 1,   42,  55,  58,  118, 166, 28,  157, 107, 140, 243,
      84,  129, 138, 94,  122, 47,  107, 223, 26,  228, 89,  30,  143, 160, 219, 238, 153, 58,  12,  110, 180, 104, 98,
      245, 8,   138, 108, 160, 182, 252, 79,  162, 122, 83,  131, 54,  210, 120, 167, 53,  209, 54,  4,   147, 138, 244,
      169, 238, 169, 25,  2,   230, 73,  200, 163, 52,  86,  40,  25,  238, 244, 33,  116, 148, 173, 147, 135, 39,  75,
      86,  51,  82,  114, 23,  218, 161, 170, 53,  219, 211, 12,  208, 157, 169, 18,  113, 37,  99,  192, 2,   184, 8,
      171, 164, 92,  160, 112, 234, 8,   213, 83,  45,  99,  160, 60,  243, 9,   72,  62,  9,   20,  130, 217, 47,  82,
      78,  17,  57,  28,  187, 192, 196, 192, 225, 143, 87,  14,  161, 67,  75,  71,  97,  209, 62,  254, 65,  223, 239,
      46,  53,  251, 91,  237, 205, 149, 158, 5,   168, 32,  4,   10,  160, 66,  155, 1,   113, 225, 82,  148, 224, 18,
      216, 244, 134, 107, 47,  218, 142, 52,  125, 203, 81,  169, 79,  11,  33,  252, 90,  152, 47,  59,  121, 182, 150,
      200, 108, 246, 137, 103, 227, 2,   252, 120, 186, 137, 21,  192, 202, 98,  173, 49,  230, 39,  43,  178, 121, 21,
      113, 178, 244, 235, 68,  8,   97,  23,  181, 11,  166, 162, 179, 66,  10,  19,  77,  139, 38,  50,  213, 134, 242,
      151, 37,  116, 112, 185, 132, 118, 239, 18,  9,   180, 117, 229, 66,  118, 228, 109, 45,  64,  67,  177, 112, 173,
      245, 212, 25,  119, 108, 231, 206, 41,  211, 111, 111, 174, 43,  172, 225, 191, 43,  161, 118, 157, 108, 49,  214,
      215, 125, 54,  236, 242, 167, 205, 75,  35,  12,  124, 238, 163, 73,  238, 226, 235, 43,  199, 69,  147, 20,  187,
      157, 210, 250, 178, 43,  92,  20,  125, 51,  151, 183, 94,  85,  14,  23,  221, 244, 230, 22,  75,  194, 198, 213,
      65,  152, 200, 102, 79,  197, 202, 222, 228, 10,  198, 73,  150, 102, 26,  244, 41,  161, 196, 142, 52,  63,  138,
      74,  140, 121, 119, 192, 209, 183, 138, 53,  132, 29,  121, 161, 138, 249, 70,  46,  233, 57,  211, 60,  215, 232,
      54,  249, 118, 230, 200, 77,  233, 5,   115, 145, 203, 66,  248, 227, 171, 86,  125, 133, 25,  19,  147, 181, 243,
      40,  207, 58,  199, 15,  105, 244, 245, 15,  140, 244, 159, 62,  229, 80,  209, 252, 245, 154, 72,  113, 122, 80,
      160, 94,  146, 71,  83,  19,  28,  206, 174, 20,  30,  70,  234, 203, 242, 190, 243, 27,  83,  243, 145, 51,  217,
      70,  36,  72,  139, 82,  41,  94,  29,  68,  46,  68,  161, 76,  56,  170, 37,  20,  20,  21,  24,  219, 173, 28,
      54,  66,  125, 168, 8,   193, 232, 20,  220, 81,  187, 18,  37,  186, 78,  82,  86,  244, 184, 23,  120, 101, 245,
      229, 248, 192, 185, 208, 23,  136, 85,  110, 125, 203, 63,  44,  158, 155, 194, 97,  220, 229, 60,  159, 80,  8,
      175, 90,  93,  227, 69,  113, 139, 93,  220, 169, 100, 144, 27,  72,  242, 169, 247, 195, 37,  237, 223, 122, 138,
      70,  78,  68,  139, 102, 164, 228, 51,  227, 199, 1,   193, 164, 178, 126, 99,  99,  172, 20,  24,  50,  57,  5,
      69,  113, 39,  109, 77,  210, 232, 70,  80,  122, 168, 156, 238, 118, 88,  234, 190, 130, 95,  94,  195, 161, 197,
      215, 245, 236, 5,   29,  104, 233, 238, 207, 197, 155, 85,  202, 202, 195, 232, 79,  11,  43,  233, 49,  106, 121,
      60,  219, 6,   99,  174, 206, 121, 202, 245, 136, 5,   35,  121, 174, 51,  231, 40,  75,  13,  127, 143, 159, 225,
      158, 19,  66,  191, 193, 104, 96,  37,  128, 106, 92,  68,  1,   210, 198, 37,  78,  121, 197, 160, 72,  95,  109,
      84,  134, 65,  1,   37,  11,  214, 143, 101, 214, 86,  201, 218, 116, 213, 189, 57,  103, 69,  89,  102, 75,  114,
      170, 46,  139, 156, 241, 193, 86,  243, 50,  60,  223, 142, 250, 81,  21,  22,  37,  192, 102, 100, 186, 214, 146,
      147, 87,  230, 180, 77,  70,  12,  6,   150, 198, 11,  47,  149, 126, 132, 57,  204, 254, 230, 141, 246, 14,  51,
      160, 227, 167, 137, 176, 214, 186, 7,   218, 69,  250, 109, 123, 30,  150, 176, 44,  204, 61,  33,  135, 193, 172,
      233, 221, 251, 109, 154, 13,  245, 105, 245, 43,  184, 209, 46,  116, 214, 90,  131, 0,   52,  91,  165, 118, 140,
      243, 118, 81,  240, 88,  34,  72,  157, 177, 138, 101, 16,  45,  16,  98,  222, 7,   173, 80,  236, 106, 201, 80,
      1,   115, 189, 246, 141, 208, 88,  9,   167, 75,  200, 34,  210, 173, 249, 171, 219, 144, 160, 6,   231, 135, 228,
      193, 146, 16,  176, 117, 20,  170, 62,  72,  168, 41,  207, 138, 168, 88,  86,  241, 72,  205, 154, 78,  93,  168,
      229, 29,  166, 232, 149, 66,  130, 154, 236, 240, 91,  88,  155, 196, 139, 203, 76,  181, 105, 49,  67,  3,   127,
      28,  214, 233, 107, 10,  110, 60,  248, 222, 73,  110, 127, 44,  228, 87,  180, 215, 49,  146, 48,  129, 203, 227,
      197, 11,  23,  107, 189, 234, 38,  195, 42,  19,  13,  40,  254, 15,  97,  223, 251, 35,  236, 238, 29,  173, 199,
      181, 47,  185, 71,  125, 198, 121, 25,  211, 172, 75,  18,  225, 164, 161, 65,  198, 247, 125, 226, 249, 154, 48,
      207, 148, 151, 46,  45,  81,  154, 173, 208, 133, 107, 161, 0,   74,  14,  192, 209, 254, 126, 126, 187, 33,  171,
      210, 18,  227, 72,  174, 250, 23,  70,  71,  150, 152, 193, 43,  113, 150, 68,  247, 169, 113, 248, 210, 120, 118,
      198, 206, 229, 154, 232, 80,  175, 68,  247, 58,  57,  248, 230, 141, 93,  49,  197, 159, 177, 120, 11,  137, 188,
      42,  61,  110, 223, 29,  156, 166, 212, 135, 51,  53,  0,   191, 167, 223, 196, 231, 229, 30,  200, 248, 57,  192,
      87,  49,  172, 185, 90,  48,  123, 222, 244, 169, 22,  198, 196, 239, 95,  160, 196, 19,  127, 223, 224, 72,  24,
      62,  147, 80,  65,  77,  9,   1,   203, 210, 75,  51,  138, 250, 40,  114, 127, 230, 237, 88,  69,  173, 200, 56,
      0,   221, 88,  249, 104, 237, 12,  159, 182, 85,  113, 38,  55,  25,  132, 180, 167, 145, 192, 209, 167, 102, 17,
      21,  40,  23,  244, 10,  21,  73,  207, 43,  2,   159, 127, 0,   51,  70,  238, 78,  32,  238, 212, 154, 47,  143,
      60,  38,  135, 253, 133, 93,  69,  3,   249, 40,  157, 166, 138, 216, 169, 108, 142, 37,  254, 143, 48,  20,  101,
      79,  72,  198, 233, 64,  118, 116, 136, 117, 231, 51,  247, 165, 206, 24,  35,  241, 83,  55,  86,  15,  28,  96,
      40,  71,  171, 92,  129, 77,  9,   133, 159, 213, 173, 220, 104, 221, 247, 97,  208, 18,  163, 182, 199, 52,  81,
      128, 206, 100, 154, 116, 246, 86,  183, 204, 115, 117, 207, 184, 160, 76,  172, 220, 96,  225, 65,  212, 107, 223,
      148, 59,  213, 3,   204, 98,  228, 51,  68,  208, 143, 72,  61,  211, 121, 236, 238, 202, 1,   140, 154, 6,   136,
      129, 118, 57,  51,  133, 123, 225, 211, 11,  243, 103, 170, 230, 131, 222, 160, 49,  66,  251, 246, 92,  69,  126,
      145, 76,  236, 9,   3,   156, 165, 87,  175, 179, 18,  85,  241, 24,  220, 84,  182, 110, 145, 28,  142, 171, 252,
      226, 36,  72,  243, 111, 35,  157, 151, 207, 199, 239, 85,  133, 100, 157, 105, 35,  89,  17,  27,  44,  190, 136,
      19,  70,  122, 180, 163, 47,  99,  123, 95,  2,   138, 103, 21,  190, 31,  198, 125, 40,  32,  148, 36,  139, 122,
      140, 165, 139, 169, 91,  38,  96,  88,  20,  159, 165, 101, 18,  228, 53,  171, 237, 71,  53,  25,  253, 202, 250,
      187, 44,  227, 29,  105, 237, 96,  0,   11,  73,  161, 128, 105, 213, 137, 68,  72,  187, 60,  51,  10,  155, 34,
      115, 94,  246, 81,  212, 135, 156, 164, 43,  152, 220, 84,  168, 94,  151, 156, 62,  42,  10,  79,  57,  102, 233,
      58,  56,  242, 150, 14,  23,  80,  247, 186, 249, 110, 88,  18,  55,  202, 203, 83,  16,  101, 199, 234, 212, 101,
      123, 240, 49,  204, 227, 45,  45,  175, 74,  196, 34,  96,  8,   163, 248, 84,  164, 19,  73,  24,  111, 92,  216,
      121, 1,   175, 213, 171, 182, 72,  106, 11,  216, 152, 102, 37,  186, 164, 104, 85,  112, 245, 154, 24,  226, 248,
      31,  88,  90,  34,  134, 59,  139, 92,  173, 80,  144, 186, 222, 134, 229, 1,   103, 113, 124, 226, 83,  50,  4,
      226, 206, 159, 207, 137, 139, 136, 14,  105, 131, 148, 85,  46,  197, 146, 248, 156, 221, 145, 252, 36,  67,  159,
      9,   164, 252, 29,  94,  45,  190, 41,  252, 133, 79,  42,  82,  38,  15,  30,  221, 97,  224, 238, 136, 1,   210,
      128, 104, 218, 203, 204, 160, 95,  162, 207, 55,  119, 20,  215, 45,  177, 89,  18,  96,  237, 64,  137, 89,  89,
      130, 175, 188, 62,  20,  237, 104, 243, 145, 187, 87,  59,  242, 147, 178, 109, 65,  211, 100, 177, 110, 144, 175,
      34,  73,  20,  102, 202, 117, 8,   123, 171, 17,  193, 101, 142, 200, 6,   30,  214, 17,  69,  8,   38,  190, 122,
      68,  213, 5,   222, 185, 140, 16,  7,   245, 47,  221, 44,  50,  88,  99,  105, 2,   18,  59,  42,  72,  217, 34,
      34,  169, 76,  35,  148, 143, 57,  165, 150, 98,  141, 236, 176, 197, 125, 58,  15,  154, 221, 143, 233, 246, 62,
      248, 17,  15,  39,  74,  236, 46,  5,   220, 43,  217, 231, 127, 224, 94,  80,  58,  143, 140, 8,   56,  86,  246,
      68,  96,  54,  179, 105, 143, 5,   180, 60,  194, 57,  19,  218, 112, 129, 124, 7,   222, 75,  47,  91,  45,  192,
      190, 251, 88,  4,   23,  156, 113, 170, 99,  49,  110, 241, 65,  100, 51,  66,  3,   108, 106, 60,  121, 122, 71,
      43,  242, 161, 22,  249, 231, 182, 118, 237, 12,  33,  242, 134, 87,  106, 112, 0,   106, 193, 60,  133, 112, 245,
      99,  93,  147, 244, 26,  55,  90,  1,   107, 7,   29,  87,  115, 26,  243, 240, 153, 24,  220, 252, 63,  116, 162,
      76,  11,  18,  140, 173, 31,  51,  6,   197, 219, 240, 239, 21,  20,  143, 95,  32,  153, 210, 58,  127, 248, 194,
      24,  109, 31,  67,  147, 194, 30,  49,  28,  227, 240, 60,  100, 191, 3,   73,  57,  12,  241, 97,  154, 112, 88,
      82,  85,  24,  100, 150, 184, 78,  132, 146, 87,  151, 86,  196, 207, 65,  24,  88,  117, 110, 192, 203, 194, 8,
      218, 92,  117, 250, 132, 108, 83,  157, 253, 136, 238, 218, 27,  226, 51,  195, 126, 21,  45,  241, 122, 212, 87,
      250, 36,  87,  246, 233, 204, 183, 219, 95,  56,  202, 47,  181, 68,  106, 197, 192, 44,  110, 78,  219, 158, 78,
      75,  98,  53,  11,  127, 2,   220, 40,  49,  64,  104, 70,  239, 92,  9,   134, 115, 135, 166, 235, 104, 130, 106,
      129, 10,  97,  233, 251, 21,  127, 17,  48,  244, 200, 51,  165, 176, 114, 111, 175, 186, 206, 189, 209, 151, 6,
      128, 217, 16,  203, 5,   216, 169, 216, 67,  238, 22,  220, 123, 83,  174, 86,  26,  223, 15,  131, 61,  23,  151,
      50,  108, 240, 129, 229, 176, 153, 135, 129, 106, 10,  115, 204, 162, 186, 43,  102, 241, 107, 195, 221, 55,  254,
      214, 56,  218, 131, 221, 151, 48,  118, 104, 162, 166, 221, 118, 125, 142, 130, 252, 62,  33,  94,  65,  95,  80,
      108, 9,   180, 113, 128, 123, 213, 12,  237, 233, 53,  167, 2,   244, 113, 39,  157, 234, 53,  56,  154, 85,  199,
      143, 84,  216, 89,  163, 16,  179, 166, 152, 196, 210, 20,  138, 148, 109, 187, 179, 111, 136, 82,  135, 112, 97,
      210, 216, 200, 236, 216, 82,  30,  149, 246, 247, 220, 244, 226, 32,  247, 42,  250, 124, 9,   168, 47,  113, 127,
      6,   152, 73,  82,  124, 214, 72,  92,  79,  148, 197, 104, 78,  7,   68,  169, 138, 126, 196, 131, 139, 72,  219,
      205, 49,  92,  231, 32,  119, 93,  14,  218, 120, 222, 68,  141, 29,  23,  185, 210, 225, 141, 65,  157, 231, 245,
      2,   205, 239, 138, 114, 197, 191, 229, 196, 171, 249, 107, 177, 200, 188, 218, 2,   113, 231, 248, 133, 44,  90,
      157, 41,  44,  46,  205, 4,   79,  46,  30,  48,  218, 154, 157, 127, 249, 192, 4,   147, 176, 190, 107, 64,  47,
      6,   75,  30,  73,  90,  244, 200, 170, 21,  152, 37,  236, 108, 66,  184, 191, 97,  210, 3,   148, 101, 141, 189,
      189, 47,  95,  230, 63,  163, 227, 211, 235, 159, 213, 155, 207, 191, 46,  41,  24,  169, 167, 144, 99,  220, 113,
      50,  18,  49,  156, 11,  64,  159, 172, 46,  139, 115, 12,  240, 71,  70,  34,  118, 203, 32,  21,  178, 69,  92,
      37,  160, 11,  167, 38,  115, 196, 122, 62,  126, 92,  244, 122, 178, 18,  81,  248, 53,  208, 248, 21,  9,   177,
      75,  171, 244, 45,  110, 128, 179, 61,  71,  250, 242, 52,  152, 160, 104, 45,  247, 106, 122, 16,  78,  90,  191,
      180, 103, 178, 118, 223, 94,  122, 188, 36,  114, 129, 81,  104, 122, 165, 180, 120, 15,  60,  164, 127, 237, 118,
      17,  115, 150, 28,  253, 10,  51,  245, 9,   178, 13,  57,  137, 125, 75,  66,  118, 40,  95,  54,  99,  21,  184,
      52,  52,  58,  130, 136, 80,  71,  180, 27,  127, 58,  8,   33,  76,  160, 172, 109, 235, 224, 189, 174, 169, 202,
      14,  51,  168, 62,  168, 130, 223, 248, 136, 92,  22,  26,  144, 211, 212, 120, 13,  218, 199, 159, 57,  216, 218,
      188, 92,  223, 95,  138, 164, 251, 76,  185, 61,  170, 169, 173, 95,  73,  111, 54,  90,  71,  149, 197, 211, 98,
      229, 26,  44,  167, 102, 10,  46,  38,  58,  203, 36,  212, 115, 127, 162, 215, 11,  148, 116, 132, 4,   49,  94,
      31,  84,  156, 173, 250, 84,  2,   74,  157, 55,  50,  40,  72,  56,  92,  234, 129, 56,  189, 205, 32,  3,   129,
      250, 208, 184, 161, 174, 106, 142, 86,  24,  167, 103, 243, 143, 238, 150, 158, 56,  75,  222, 212, 30,  188, 37,
      155, 150, 218, 191, 162, 67,  163, 109, 53,  222, 117, 244, 24,  246, 169, 223, 119, 223, 201, 173, 63,  209, 209,
      51,  191, 246, 183, 12,  218, 60,  179, 195, 104, 199, 132, 131, 203, 82,  26,  217, 204, 215, 97,  131, 20,  63,
      137, 233, 208, 201, 198, 254, 99,  181, 254, 120, 129, 72,  167, 47,  243, 60,  96,  201, 187, 6,   40,  240, 133,
      180, 88,  238, 168, 222, 54,  129, 57,  80,  27,  190, 201, 35,  174, 129, 195, 229, 29,  251, 22,  151, 254, 59,
      30,  84,  39,  155, 150, 146, 133, 54,  54,  123, 116, 132, 218, 195, 83,  51,  171, 210, 193, 254, 235, 154, 64,
      107, 24,  228, 225, 224, 32,  198, 46,  31,  74,  100, 76,  78,  112, 63,  56,  16,  17,  163, 17,  2,   222, 227,
      160, 196, 16,  86,  68,  217, 50,  12,  117, 128, 73,  47,  83,  41,  219, 39,  150, 218, 21,  20,  243, 13,  159,
      175, 16,  183, 103, 149, 83,  16,  112, 117, 251, 56,  0,   248, 237, 14,  124, 232, 90,  63,  82,  139, 4,   125,
      242, 159, 40,  84,  70,  74,  233, 51,  80,  145, 183, 165, 0,   36,  210, 227, 149, 133, 236, 123, 233, 253, 28,
      121, 25,  46,  93,  43,  196, 4,   182, 92,  39,  128, 120, 31,  95,  200, 73,  32,  144, 92,  63,  96,  184, 67,
      28,  169, 245, 227, 209, 216, 17,  158, 180, 180, 99,  92,  219, 112, 135, 89,  193, 236, 188, 252, 177, 168, 121,
      63,  131, 122, 131, 12,  142, 149, 7,   13,  125, 25,  218, 240, 121, 186, 148, 251, 47,  145, 82,  152, 163, 133,
      133, 80,  130, 134, 170, 212, 218, 223, 134, 106, 245, 73,  244, 92,  100, 52,  96,  225, 27,  154, 81,  65,  215,
      208, 227, 113, 36,  83,  84,  14,  120, 42,  59,  76,  176, 53,  253, 130, 79,  86,  243, 217, 167, 248, 169, 123,
      134, 211, 172, 82,  115, 91,  31,  165, 210, 103, 155, 27,  252, 164, 36,  50,  207, 69,  215, 100, 155, 52,  12,
      242, 5,   103, 201, 225, 97,  115, 20,  160, 5,   218, 202, 201, 169, 216, 22,  100, 246, 196, 166, 82,  42,  11,
      200, 2,   14,  153, 75,  61,  11,  91,  202, 167, 217, 196, 109, 225, 6,   67,  16,  237, 88,  41,  114, 54,  229,
      73,  209, 26,  203, 70,  11,  253, 18,  166, 64,  63,  24,  253, 150, 108, 99,  184, 182, 171, 0,   172, 167, 190,
      51,  157, 207, 182, 155, 122, 21,  16,  5,   252, 111, 232, 185, 45,  243, 98,  54,  174, 86,  65,  63,  129, 234,
      127, 53,  224, 210, 66,  139, 157, 173, 86,  43,  0,   187, 160, 244, 198, 200, 79,  226, 39,  190, 239, 40,  26,
      150, 22,  200, 3,   142, 180, 115, 168, 251, 38,  163, 54,  127, 227, 236, 94,  111, 130, 60,  24,  152, 191, 108,
      33,  32,  132, 188, 26,  18,  135, 224, 233, 187, 166, 33,  5,   105, 161, 5,   97,  61,  56,  152, 184, 136, 70,
      206, 32,  247, 233, 244, 8,   193, 154, 58,  143, 157, 176, 69,  157, 70,  244, 232, 74,  1,   58,  13,  151, 90,
      10,  47,  22,  177, 154, 89,  199, 39,  196, 55,  133, 67,  157, 52,  68,  200, 162, 173, 92,  78,  147, 15,  12,
      172, 218, 1,   7,   234, 169, 225, 0,   212, 164, 148, 145, 25,  186, 170, 54,  47,  238, 101, 163, 61,  201, 173,
      144, 26,  76,  26,  43,  188, 105, 0,   3,   177, 221, 119, 138, 36,  170, 21,  133, 165, 138, 38,  30,  7,   134,
      163, 221, 214, 199, 228, 158, 89,  211, 126, 36,  53,  51,  175, 87,  139, 167, 112, 144, 157, 247, 84,  183, 139,
      236, 206, 147, 161, 177, 57,  79,  239, 173, 91,  252, 138, 90,  87,  3,   154, 51,  96,  141, 177, 47,  197, 44,
      108, 170, 221, 156, 243, 136, 54,  248, 228, 40,  245, 232, 4,   212, 133, 153, 215, 120, 119, 129, 112, 121, 71,
      31,  75,  173, 144, 127, 172, 1,   93,  131, 159, 117, 166, 231, 179, 34,  172, 31,  78,  167, 193, 122, 119, 184,
      201, 236, 156, 108, 174, 163, 9,   239, 229, 27,  69,  8,   192, 16,  63,  244, 58,  2,   247, 204, 154, 246, 80,
      113, 30,  104, 236, 109, 56,  210, 254, 4,   164, 136, 147, 5,   16,  135, 245, 237, 172, 207, 150, 122, 39,  84,
      244, 216, 74,  171, 9,   244, 235, 118, 211, 169, 24,  22,  241, 173, 127, 224, 143, 158, 102, 94,  37,  64,  23,
      221, 57,  227, 2,   128, 11,  158, 221, 96,  28,  188, 213, 217, 221, 31,  31,  77,  106, 65,  185, 34,  135, 19,
      202, 81,  49,  88,  252, 244, 194, 111, 180, 221, 203, 220, 54,  237, 42,  220, 209, 37,  136, 127, 191, 238, 40,
      152, 152, 9,   112, 45,  146, 79,  196, 243, 72,  54,  129, 193, 167, 194, 142, 29,  162, 107, 254, 147, 111, 14,
      210, 208, 56,  205, 221, 227, 83,  44,  123, 148, 248, 140, 13,  208, 230, 92,  166, 190, 158, 159, 182, 32,  118,
      197, 243, 79,  194, 78,  122, 89,  125, 167, 31,  179, 70,  62,  231, 110, 45,  126, 221, 77,  116, 95,  191, 170,
      205, 218, 156, 146, 113, 96,  115, 88,  122, 243, 236, 228, 139, 191, 175, 189, 254, 57,  210, 192, 244, 127, 239,
      89,  28,  19,  178, 153, 155, 208, 110, 224, 97,  141, 104, 112, 89,  6,   73,  196, 115, 24,  204, 183, 253, 112,
      191, 177, 111, 253, 225, 195, 41,  75,  19,  31,  148, 148, 171, 21,  32,  186, 228, 92,  227, 167, 133, 143, 61,
      215, 215, 218, 215, 211, 143, 6,   21,  166, 161, 24,  113, 37,  169, 18,  79,  54,  114, 24,  54,  227, 151, 210,
      144, 113, 22,  82,  94,  170, 116, 41,  43,  148, 38,  95,  235, 158, 100, 121, 1,   146, 220, 194, 59,  229, 130,
      70,  19,  146, 34,  33,  190, 5,   31,  250, 142, 68,  79,  9,   126, 105, 225, 3,   112, 4,   203, 84,  123, 182,
      189, 82,  127, 207, 247, 106, 211, 130, 222, 241, 105, 216, 75,  25,  147, 121, 206, 124, 146, 35,  134, 128, 191,
      29,  105, 161, 121, 75,  67,  7,   99,  232, 216, 194, 211, 218, 144, 154, 197, 157, 204, 242, 128, 221, 121, 87,
      102, 86,  27,  203, 18,  65,  101, 129, 208, 254, 216, 181, 55,  55,  43,  81,  36,  85,  23,  211, 57,  130, 18,
      78,  103, 190, 171, 208, 198, 0,   166, 252, 152, 220, 141, 55,  141, 226, 157, 214, 139, 104, 250, 49,  2,   109,
      62,  31,  182, 112, 223, 182, 166, 151, 174, 153, 97,  194, 123, 121, 34,  13,  28,  218, 32,  83,  217, 182, 12,
      160, 95,  85,  169, 216, 128, 188, 251, 203, 222, 223, 99,  241, 243, 173, 106, 216, 158, 134, 69,  157, 71,  217,
      40,  217, 144, 7,   2,   22,  51,  77,  95,  33,  204, 178, 218, 113, 171, 37,  160, 185, 190, 40,  179, 126, 123,
      48,  127, 12,  179, 168, 247, 24,  79,  70,  18,  197, 65,  51,  138, 225, 104, 124, 213, 83,  95,  145, 80,  210,
      250, 41,  101, 82,  211, 99,  118, 53,  87,  98,  254, 215, 75,  194, 93,  96,  180, 51,  142, 109, 152, 42,  199,
      34,  233, 125, 219, 200, 234, 20,  228, 58,  18,  165, 161, 218, 100, 74,  20,  177, 72,  74,  3,   39,  74,  68,
      90,  188, 93,  217, 55,  61,  129, 102, 228, 192, 27,  150, 234, 138, 250, 168, 59,  179, 82,  241, 43,  184, 253,
      164, 121, 247, 78,  134, 104, 162, 222, 27,  84,  81,  135, 100, 238, 168, 50,  42,  207, 136, 206, 216, 115, 220,
      143, 15,  74,  130, 16,  33,  216, 36,  182, 211, 100, 187, 33,  103, 72,  67,  167, 90,  226, 121, 159, 243, 163,
      10,  230, 192, 219, 210, 196, 202, 64,  163, 7,   169, 215, 169, 85,  19,  50,  14,  93,  182, 71,  218, 154, 25,
      134, 248, 196, 77,  156, 154, 155, 113, 199, 76,  39,  162, 180, 130, 105, 249, 33,  155, 8,   172, 200, 64,  198,
      50,  182, 240, 26,  214, 7,   104, 103, 199, 49,  98,  183, 177, 135, 116, 136, 11,  172, 197, 163, 147, 229, 165,
      151, 180, 248, 132, 250, 176, 174, 15,  73,  230, 159, 141, 29,  167, 82,  140, 61,  76,  221, 167, 177, 32,  32,
      104, 5,   94,  201, 36,  154, 123, 19,  99,  63,  226, 105, 108, 35,  215, 33,  155, 159, 181, 182, 10,  105, 112,
      81,  185, 108, 177, 94,  104, 71,  173, 118, 170, 118, 34,  131, 51,  148, 16,  113, 30,  147, 56,  200, 167, 7,
      227, 236, 235, 166, 124, 178, 48,  242, 236, 188, 109, 144, 35,  30,  68,  3,   53,  103, 58,  67,  135, 167, 219,
      138, 45,  205, 104, 9,   196, 53,  24,  16,  180, 191, 2,   208, 76,  57,  60,  24,  197, 54,  190, 31,  91,  23,
      169, 47,  135, 108, 21,  107, 169, 3,   137, 216, 129, 112, 42,  124, 229, 128, 48,  245, 112, 78,  167, 213, 133,
      229, 9,   98,  156, 243, 184, 54,  251, 37,  45,  64,  33,  189, 137, 49,  108, 241, 51,  169, 72,  123, 202, 89,
      174, 168, 238, 1,   219, 241, 231, 247, 219, 60,  75,  34,  106, 219, 126, 221, 58,  155, 223, 235, 46,  238, 160,
      153, 214, 146, 46,  243, 19,  59,  244, 253, 225, 190, 176, 135, 28,  176, 91,  226, 68,  69,  100, 125, 154, 66,
      98,  230, 206, 199, 119, 229, 176, 218, 176, 32,  8,   72,  19,  151, 21,  167, 122, 10,  51,  227, 98,  34,  214,
      139, 211, 5,   92,  123, 133, 116, 152, 127, 206, 32,  43,  195, 41,  57,  151, 112, 252, 250, 171, 151, 227, 195,
      245, 145, 136, 63,  254, 93,  59,  124, 197, 132, 202, 252, 161, 77,  220, 106, 249, 48,  88,  199, 10,  100, 233,
      204, 210, 105, 196, 210, 240, 133, 83,  112, 17,  33,  53,  211, 183, 204, 105, 192, 51,  27,  9,   93,  249, 156,
      177, 155, 218, 174, 36,  140, 153, 151, 81,  79,  99,  176, 183, 10,  126, 173, 136, 180, 164, 244, 136, 89,  16,
      90,  164, 175, 236, 142, 125, 87,  54,  38,  19,  244, 221, 141, 100, 216, 69,  36,  207, 139, 162, 184, 0,   61,
      67,  223, 231, 144, 138, 59,  121, 21,  145, 134, 109, 120, 18,  114, 242, 85,  185, 34,  241, 104, 230, 219, 47,
      251, 64,  247, 80,  106, 38,  200, 1,   13,  93,  235, 86,  21,  241, 79,  19,  247, 184, 11,  10,  243, 180, 47,
      37,  67,  92,  31,  199, 202, 245, 156, 238, 245, 206, 54,  95,  50,  166, 6,   119, 186, 172, 248, 91,  225, 165,
      51,  52,  35,  214, 88,  81,  44,  251, 18,  203, 50,  242, 89,  196, 220, 137, 60,  154, 41,  249, 104, 115, 99,
      250, 113, 100, 85,  22,  195, 202, 194, 47,  71,  50,  106, 101, 83,  174, 217, 91,  16,  73,  25,  26,  106, 202,
      240, 237, 244, 141, 119, 14,  32,  79,  80,  173, 107, 64,  207, 93,  188, 210, 5,   2,   185, 107, 203, 68,  113,
      78,  167, 222, 67,  220, 244, 111, 152, 131, 69,  182, 20,  249, 222, 148, 175, 102, 16,  70,  236, 68,  1,   160,
      111, 67,  135, 75,  172, 193, 225, 133, 45,  124, 137, 151, 197, 25,  12,  73,  11,  125, 67,  58,  141, 245, 37,
      69,  112, 58,  37,  29,  134, 178, 51,  239, 140, 145, 64,  196, 88,  193, 122, 196, 90,  80,  141, 185, 35,  19,
      55,  68,  104, 204, 160, 151, 114, 43,  189, 162, 122, 241, 142, 150, 46,  95,  180, 19,  134, 42,  219, 127, 146,
      211, 16,  251, 171, 83,  152, 63,  97,  174, 81,  53,  240, 36,  26,  119, 201, 254, 64,  48,  188, 50,  203, 85,
      179, 172, 172, 242, 21,  66,  53,  155, 191, 157, 12,  106, 231, 130, 1,   246, 203, 116, 134, 234, 222, 77,  102,
      59,  211, 176, 111, 228, 202, 165, 219, 30,  230, 191, 81,  162, 189, 80,  158, 142, 150, 194, 2,   156, 97,  189,
      6,   226, 78,  29,  167, 38,  232, 188, 239, 158, 212, 209, 182, 25,  23,  138, 198, 83,  117, 160, 245, 92,  47,
      164, 16,  163, 186, 194, 137, 203, 99,  144, 185, 233, 113, 99,  120, 23,  53,  221, 245, 179, 213, 209, 119, 99,
      140, 120, 156, 59,  124, 19,  56,  81,  13,  234, 113, 147, 192, 208, 114, 243, 146, 133, 157, 143, 7,   206, 230,
      253, 49,  203, 104, 201, 248, 49,  229, 14,  81,  125, 78,  54,  49,  55,  82,  215, 106, 97,  147, 83,  118, 31,
      142, 19,  189, 206, 18,  142, 73,  56,  39,  41,  223, 179, 43,  239, 168, 153, 200, 54,  39,  113, 108, 2,   27,
      134, 93,  75,  187, 246, 100, 114, 80,  3,   32,  203, 239, 95,  125, 92,  207, 72,  249, 137, 171, 150, 98,  175,
      185, 191, 227, 158, 44,  169, 42,  2,   249, 108, 164, 154, 70,  143, 113, 41,  20,  22,  222, 114, 85,  216, 130,
      112, 183, 130, 187, 93,  114, 224, 175, 93,  137, 85,  253, 85,  9,   20,  84,  248, 120, 185, 210, 149, 5,   189,
      174, 228, 93,  21,  137, 202, 74,  13,  161, 200, 11,  79,  207, 74,  83,  227, 140, 254, 138, 162, 175, 120, 244,
      242, 64,  54,  61,  247, 232, 241, 139, 37,  198, 4,   95,  235, 104, 160, 207, 100, 50,  221, 127, 149, 125, 52,
      212, 100, 94,  20,  50,  167, 228, 52,  47,  56,  251, 133, 229, 41,  16,  152, 149, 199, 108, 239, 248, 136, 115,
      242, 74,  173, 45,  17,  178, 85,  50,  161, 183, 208, 42,  211, 174, 12,  150, 10,  6,   126, 27,  25,  126, 165,
      80,  24,  54,  29,  4,   88,  193, 219, 233, 77,  183, 73,  155, 55,  11,  65,  103, 208, 227, 194, 78,  231, 250,
      254, 30,  93,  238, 178, 164, 167, 124, 62,  206, 154, 36,  213, 5,   149, 247, 122, 3,   164, 217, 112, 68,  111,
      46,  127, 60,  217, 169, 164, 130, 154, 19,  2,   160, 175, 214, 55,  170, 73,  142, 57,  206, 9,   198, 77,  80,
      158, 142, 179, 10,  203, 74,  228, 104, 75,  149, 177, 105, 30,  60,  248, 131, 59,  80,  177, 213, 24,  238, 148,
      98,  23,  126, 128, 236, 50,  131, 254, 230, 160, 162, 160, 86,  236, 131, 242, 165, 58,  98,  32,  243, 34,  40,
      226, 46,  192, 130, 246, 137, 229, 117, 29,  233, 11,  105, 160, 1,   201, 55,  173, 169, 87,  211, 121, 43,  71,
      123, 243, 47,  121, 242, 211, 194, 194, 176, 7,   153, 52,  221, 246, 152, 206, 252, 155, 239, 216, 236, 53,  156,
      188, 244, 95,  104, 2,   229, 65,  139, 26,  230, 149, 118, 150, 253, 130, 156, 23,  189, 34,  12,  23,  221, 2,
      75,  216, 151, 5,   80,  198, 104, 197, 219, 71,  29,  103, 140, 241, 125, 254, 76,  34,  69,  31,  107, 235, 71,
      195, 144, 150, 14,  217, 163, 45,  41,  19,  171, 204, 149, 171, 85,  124, 57,  119, 96,  235, 2,   242, 254, 53,
      180, 91,  145, 40,  38,  228, 189, 175, 217, 184, 15,  38,  222, 169, 100, 68,  11,  175, 91,  15,  175, 9,   51,
      199, 207, 40,  244, 179, 53,  193, 53,  50,  232, 127, 88,  1,   120, 177, 228, 107, 43,  188, 216, 5,   201, 243,
      29,  190, 140, 10,  178, 30,  221, 169, 190, 204, 204, 56,  213, 5,   83,  147, 97,  61,  7,   151, 182, 224, 142,
      215, 7,   187, 184, 22,  142, 170, 236, 70,  252, 223, 234, 239, 16,  228, 149, 196, 44,  124, 83,  42,  251, 209,
      47,  152, 184, 33,  78,  185, 200, 148, 116, 42,  52,  222, 72,  99,  25,  13,  206, 135, 244, 19,  131, 114, 150,
      178, 103, 162, 76,  183, 89,  221, 252, 139, 165, 80,  250, 201, 190, 181, 189, 249, 224, 170, 100, 184, 242, 215,
      149, 205, 192, 89,  108, 61,  57,  216, 225, 188, 63,  160, 159, 128, 170, 23,  43,  110, 75,  30,  232, 194, 75,
      87,  80,  88,  175, 33,  69,  48,  215, 143, 106, 192, 179, 113, 151, 203, 112, 170, 241, 199, 250, 16,  220, 143,
      179, 200, 224, 83,  226, 43,  254, 151, 105, 170, 215, 175, 84,  99,  32,  220, 181, 61,  236, 108, 234, 70,  181,
      58,  119, 146, 10,  168, 201, 84,  16,  84,  74,  70,  55,  110, 114, 129, 23,  2,   229, 252, 243, 48,  98,  26,
      33,  55,  148, 145, 46,  40,  127, 119, 34,  122, 245, 183, 213, 208, 221, 237, 113, 189, 175, 223, 41,  220, 53,
      129, 148, 98,  9,   161, 63,  71,  26,  167, 124, 113, 58,  177, 155, 126, 207, 89,  125, 110, 80,  236, 5,   187,
      179, 216, 187, 176, 187, 69,  39,  118, 121, 0,   149, 216, 169, 251, 51,  19,  6,   136, 139, 245, 210, 148, 149,
      52,  30,  94,  229, 144, 85,  206, 117, 56,  232, 252, 176, 149, 148, 25,  6,   70,  142, 118, 190, 180, 15,  7,
      166, 239, 203, 216, 247, 108, 227, 100, 210, 18,  54,  80,  44,  97,  202, 36,  196, 93,  140, 153, 58,  248, 110,
      70,  32,  238, 246, 53,  15,  223, 75,  93,  38,  194, 65,  172, 251, 133, 229, 113, 180, 139, 232, 157, 243, 123,
      140, 154, 112, 98,  26,  17,  155, 20,  199, 222, 183, 106, 214, 175, 167, 148, 158, 131, 155, 11,  225, 126, 55,
      132, 247, 79,  88,  36,  8,   150, 0,   160, 100, 31,  51,  175, 191, 253, 207, 127, 47,  3,   208, 38,  29,  76,
      238, 137, 75,  10,  43,  107, 167, 222, 167, 53,  15,  160, 201, 156, 108, 95,  236, 2,   193, 219, 178, 35,  168,
      48,  144, 68,  19,  102, 235, 220, 17,  59,  237, 57,  112, 222, 150, 149, 68,  11,  157, 240, 7,   123, 235, 164,
      139, 33,  5,   128, 59,  219, 54,  43,  38,  239, 139, 142, 247, 46,  47,  241, 44,  229, 72,  62,  192, 25,  154,
      111, 49,  97,  225, 5,   95,  234, 62,  10,  185, 93,  244, 20,  76,  64,  36,  247, 180, 254, 176, 205, 80,  209,
      156, 206, 26,  6,   34,  100, 57,  66,  245, 236, 149, 66,  224, 182, 91,  132, 173, 129, 180, 235, 13,  33,  107,
      180, 240, 117, 139, 106, 94,  111, 86,  190, 102, 92,  12,  194, 53,  96,  60,  165, 158, 0,   12,  75,  191, 198,
      37,  162, 236, 212, 162, 102, 233, 16,  194, 223, 168, 217, 98,  191, 171, 204, 46,  20,  101, 164, 153, 32,  239,
      85,  118, 2,   162, 200, 204, 28,  113, 111, 160, 97,  252, 33,  213, 185, 105, 73,  191, 48,  182, 92,  146, 101,
      47,  54,  52,  165, 94,  180, 55,  100, 36,  91,  10,  111, 239, 232, 65,  175, 86,  109, 84,  211, 241, 119, 29,
      254, 12,  134, 86,  99,  156, 253, 69,  106, 156, 10,  119, 187, 207, 38,  223, 187, 129, 114, 176, 198, 10,  87,
      211, 168, 44,  170, 148, 110, 40,  211, 65,  46,  215, 200, 160, 165, 225, 225, 60,  99,  145, 86,  201, 225, 126,
      110, 54,  66,  227, 220, 128, 60,  90,  182, 251, 19,  123, 146, 162, 160, 51,  95,  82,  113, 220, 229, 52,  216,
      169, 132, 89,  215, 4,   51,  224, 244, 5,   141, 165, 25,  192, 204, 18,  239, 122, 239, 47,  46,  248, 30,  143,
      83,  49,  32,  248, 1,   23,  45,  192, 88,  153, 133, 227, 9,   234, 47,  124, 92,  33,  248, 176, 242, 107, 223,
      230, 57,  19,  30,  246, 212, 137, 107, 37,  32,  192, 51,  234, 85,  236, 63,  184, 72,  203, 120, 171, 252, 57,
      76,  72,  248, 58,  198, 157, 213, 106, 73,  236, 3,   152, 60,  59,  27,  79,  92,  127, 84,  4,   90,  102, 196,
      128, 210, 166, 38,  217, 173, 238, 92,  222, 252, 206, 103, 99,  101, 230, 233, 142, 45,  88,  66,  113, 2,   73,
      166, 135, 123, 31,  7,   187, 89,  141, 18,  174, 6,   210, 145, 117, 242, 187, 151, 85,  197, 103, 84,  94,  222,
      124, 199, 29,  84,  88,  243, 178, 74,  28,  97,  191, 120, 6,   34,  164, 138, 92,  221, 219, 167, 147, 82,  192,
      50,  86,  154, 95,  44,  152, 158, 140, 130, 3,   18,  128, 25,  153, 181, 155, 175, 246, 45,  27,  2,   43,  221,
      140, 240, 230, 98,  193, 0,   114, 79,  103, 86,  182, 28,  56,  36,  216, 166, 43,  239, 118, 101, 231, 188, 244,
      68,  129, 191, 121, 242, 228, 89,  145, 242, 213, 28,  100, 75,  118, 23,  175, 189, 90,  69,  11,  243, 20,  97,
      242, 236, 188, 63,  58,  110, 53,  45,  182, 96,  68,  145, 152, 146, 71,  96,  245, 227, 155, 156, 145, 127, 201,
      68,  1,   183, 129, 29,  199, 99,  54,  120, 3,   195, 198, 62,  210, 125, 192, 212, 7,   127, 10,  29,  106, 181,
      65,  64,  71,  62,  105, 50,  42,  129, 33,  95,  68,  16,  73,  155, 12,  131, 126, 108, 0,   13,  247, 87,  75,
      35,  248, 173, 175, 120, 218, 185, 189, 117, 84,  215, 100, 165, 157, 134, 196, 20,  144, 51,  139, 72,  45,  173,
      34,  84,  225, 138, 101, 65,  34,  120, 212, 68,  85,  26,  165, 62,  190, 40,  1,   5,   248, 224, 2,   188, 207,
      185, 226, 164, 96,  228, 179, 90,  170, 219, 108, 174, 146, 128, 137, 191, 76,  112, 97,  203, 6,   174, 241, 227,
      126, 60,  28,  119, 125, 102, 179, 234, 133, 124, 167, 9,   27,  87,  158, 170, 76,  125, 191, 129, 166, 84,  228,
      1,   247, 2,   111, 1,   107, 14,  119, 182, 85,  151, 9,   152, 88,  5,   120, 158, 26,  183, 151, 170, 120, 32,
      23,  19,  158, 124, 181, 125, 175, 176, 31,  135, 162, 4,   116, 101, 43,  84,  54,  52,  75,  247, 127, 66,  132,
      35,  77,  110, 190, 63,  150, 6,   117, 249, 131, 130, 194, 164, 27,  92,  252, 99,  206, 76,  181, 250, 216, 217,
      65,  187, 253, 207, 130, 193, 69,  12,  3,   147, 80,  46,  73,  251, 179, 100, 16,  30,  88,  59,  114, 20,  160,
      63,  219, 1,   28,  133, 37,  42,  25,  212, 71,  243, 37,  224, 177, 217, 156, 33,  3,   83,  37,  68,  73,  194,
      182, 168, 141, 57,  24,  46,  10,  144, 37,  235, 28,  229, 145, 105, 24,  157, 62,  120, 27,  230, 37,  237, 149,
      204, 66,  32,  230, 75,  111, 219, 61,  85,  102, 176, 232, 247, 11,  80,  56,  66,  187, 115, 161, 170, 210, 25,
      51,  188, 200, 224, 206, 196, 24,  21,  109, 152, 159, 187, 170, 29,  188, 87,  40,  33,  104, 195, 45,  90,  176,
      55,  97,  56,  32,  20,  70,  236, 41,  16,  228, 108, 215, 232, 89,  157, 87,  209, 206, 216, 136, 26,  34,  75,
      49,  115, 194, 71,  213, 186, 48,  193, 227, 253, 146, 169, 53,  52,  134, 195, 43,  10,  47,  34,  132, 212, 145,
      199, 123, 50,  122, 140, 131, 208, 137, 80,  86,  65,  217, 117, 204, 165, 53,  61,  162, 184, 245, 183, 98,  148,
      196, 103, 7,   171, 44,  30,  57,  54,  134, 150, 201, 29,  197, 4,   17,  149, 130, 31,  11,  247, 97,  213, 65,
      145, 26,  4,   179, 187, 36,  171, 40,  164, 192, 77,  40,  75,  151, 205, 26,  31,  242, 208, 210, 147, 238, 223,
      246, 170, 252, 118, 97,  2,   52,  188, 35,  216, 28,  76,  250, 54,  79,  52,  227, 187, 178, 92,  246, 213, 159,
      215, 3,   217, 179, 132, 51,  124, 79,  196, 53,  228, 212, 80,  50,  115, 63,  107, 83,  53,  160, 176, 118, 247,
      126, 163, 98,  4,   26,  167, 127, 18,  157, 35,  107, 173, 64,  186, 201, 251, 64,  219, 189, 52,  76,  71,  47,
      47,  69,  153, 118, 147, 115, 110, 30,  151, 155, 68,  78,  131, 75,  134, 157, 72,  180, 254, 220, 150, 9,   100,
      124, 68,  178, 23,  15,  187, 84,  32,  149, 23,  159, 64,  180, 122, 230, 94,  237, 11,  107, 167, 100, 22,  96,
      25,  145, 68,  47,  142, 71,  8,   20,  194, 35,  165, 229, 47,  133, 87,  156, 54,  127, 175, 62,  246, 61,  216,
      11,  206, 101, 244, 233, 84,  146, 182, 67,  88,  5,   159, 209, 15,  165, 193, 103, 195, 144, 186, 150, 82,  132,
      223, 48,  145, 103, 44,  13,  7,   210, 128, 97,  128, 123, 246, 157, 59,  31,  246, 85,  85,  242, 78,  81,  92,
      240, 18,  207, 192, 53,  157, 31,  209, 133, 25,  212, 114, 94,  70,  150, 102, 150, 182, 37,  214, 64,  218, 224,
      124, 166, 132, 56,  66,  80,  189, 100, 171, 19,  89,  113, 48,  220, 49,  91,  216, 159, 57,  51,  102, 249, 169,
      207, 29,  253, 208, 198, 110, 101, 158, 190, 87,  100, 213, 35,  55,  205, 100, 157, 137, 77,  128, 254, 61,  184,
      9,   129, 145, 161, 227, 209, 198, 14,  119, 172, 250, 78,  84,  163, 157, 227, 123, 206, 150, 155, 2,   84,  7,
      180, 224, 132, 134, 40,  157, 154, 215, 80,  231, 118, 114, 110, 12,  101, 4,   222, 85,  46,  156, 181, 30,  126,
      136, 5,   91,  185, 93,  166, 33,  38,  44,  225, 144, 21,  30,  244, 172, 72,  229, 52,  35,  74,  162, 46,  61,
      47,  93,  24,  241, 60,  253, 190, 104, 11,  205, 106, 115, 74,  19,  254, 139, 229, 191, 15,  119, 211, 91,  126,
      77,  6,   84,  180, 157, 169, 73,  147, 154, 201, 132, 65,  118, 185, 156, 154, 6,   176, 67,  87,  128, 75,  244,
      78,  181, 43,  29,  72,  49,  217, 137, 132, 158, 11,  179, 103, 170, 92,  179, 231, 61,  209, 191, 83,  138, 66,
      209, 129, 31,  254, 185, 41,  36,  239, 230, 97,  83,  199, 115, 243, 20,  230, 194, 102, 103, 254, 202, 214, 253,
      85,  32,  17,  85,  211, 149, 96,  88,  37,  12,  154, 209, 176, 73,  53,  153, 132, 199, 18,  5,   80,  58,  48,
      139, 85,  212, 192, 138, 209, 88,  62,  250, 242, 46,  230, 90,  196, 36,  23,  214, 55,  157, 249, 220, 121, 162,
      4,   133, 22,  40,  180, 111, 157, 87,  200, 5,   167, 159, 63,  190, 176, 172, 65,  135, 220, 119, 78,  94,  151,
      62,  58,  161, 104, 200, 84,  191, 254, 43,  190, 92,  19,  101, 22,  25,  182, 22,  123, 54,  191, 246, 195, 139,
      246, 173, 150, 56,  147, 222, 149, 83,  181, 209, 199, 27,  117, 95,  151, 163, 167, 253, 64,  187, 240, 145, 193,
      153, 88,  94,  31,  211, 228, 154, 207, 219, 243, 213, 212, 138, 237, 216, 13,  82,  208, 113, 249, 30,  44,  251,
      239, 110, 40,  166, 96,  191, 78,  120, 96,  242, 212, 55,  13,  70,  160, 15,  71,  150, 122, 132, 111, 142, 43,
      97,  93,  41,  120, 222, 41,  171, 64,  178, 2,   37,  79,  77,  82,  108, 106, 181, 223, 57,  203, 248, 232, 138,
      76,  97,  69,  172, 98,  191, 227, 112, 58,  175, 13,  162, 147, 103, 164, 219, 93,  31,  127, 22,  65,  27,  143,
      142, 157, 106, 249, 239, 4,   145, 135, 174, 181, 132, 184, 125, 117, 193, 87,  198, 219, 241, 13,  136, 149, 11,
      221, 102, 72,  245, 34,  112, 230, 240, 66,  85,  180, 143, 147, 142, 152, 3,   78,  16,  52,  90,  144, 41,  229,
      238, 74,  79,  15,  58,  137, 251, 101, 93,  43,  4,   32,  170, 31,  156, 47,  249, 40,  231, 78,  195, 138, 139,
      138, 249, 119, 183, 118, 197, 167, 248, 229, 118, 251, 199, 203, 180, 209, 17,  227, 166, 215, 214, 20,  213, 254,
      54,  181, 83,  8,   18,  243, 42,  93,  167, 246, 113, 246, 67,  215, 36,  94,  235, 75,  113, 205, 148, 107, 92,
      148, 11,  38,  20,  186, 247, 237, 188, 234, 241, 176, 154, 44,  140, 115, 251, 55,  253, 151, 136, 175, 219, 71,
      250, 17,  248, 129, 47,  208, 68,  191, 130, 22,  57,  192, 226, 88,  253, 248, 48,  253, 47,  198, 234, 78,  210,
      106, 27,  127, 245, 113, 103, 192, 48,  48,  65,  191, 41,  36,  131, 191, 109, 29,  143, 187, 82,  84,  123, 165,
      65,  125, 122, 151, 81,  253, 14,  42,  68,  159, 234, 153, 196, 53,  154, 221, 123, 4,   19,  239, 253, 87,  249,
      123, 219, 67,  124, 19,  196, 88,  197, 37,  82,  115, 48,  72,  68,  19,  41,  118, 182, 21,  36,  55,  173, 157,
      100, 168, 98,  115, 119, 34,  66,  221, 181, 213, 58,  185, 203, 196, 68,  245, 21,  29,  154, 212, 162, 134, 175,
      66,  91,  170, 153, 147, 200, 176, 169, 142, 213, 168, 149, 133, 120, 114, 98,  32,  254, 102, 202, 226, 245, 129,
      72,  219, 153, 49,  172, 81,  43,  73,  184, 234, 8,   251, 121, 54,  16,  165, 241, 132, 50,  201, 223, 249, 150,
      193, 52,  131, 161, 128, 8,   183, 8,   186, 173, 87,  73,  117, 92,  13,  221, 31,  52,  123, 247, 165, 209, 35,
      232, 162, 227, 1,   66,  225, 82,  205, 185, 139, 211, 160, 46,  65,  179, 41,  216, 29,  162, 182, 134, 130, 28,
      65,  22,  80,  163, 137, 237, 217, 46,  251, 119, 30,  1,   96,  8,   225, 36,  143, 226, 41,  211, 159, 234, 152,
      89,  221, 181, 204, 145, 138, 233, 166, 83,  7,   161, 248, 239, 134, 115, 2,   31,  165, 225, 153, 234, 129, 64,
      13,  65,  75,  145, 57,  4,   113, 243, 131, 144, 99,  203, 13,  5,   182, 142, 93,  117, 30,  171, 55,  109, 56,
      136, 248, 156, 93,  241, 117, 16,  196, 139, 110, 43,  6,   111, 16,  209, 209, 59,  2,   208, 149, 56,  169, 3,
      84,  203, 152, 18,  233, 21,  98,  90,  82,  177, 226, 5,   212, 237, 196, 122, 112, 204, 244, 186, 115, 65,  194,
      112, 55,  141, 7,   147, 176, 164, 102, 138, 53,  233, 37,  46,  143, 142, 87,  208, 205, 35,  154, 187, 117, 81,
      42,  69,  82,  64,  207, 243, 219, 89,  141, 49,  215, 89,  222, 120, 41,  134, 23,  146, 154, 226, 9,   196, 51,
      211, 79,  33,  174, 66,  51,  34,  187, 66,  191, 204, 29,  239, 246, 238, 60,  234, 217, 170, 119, 221, 138, 7,
      222, 88,  122, 195, 111, 68,  209, 22,  202, 243, 105, 96,  148, 12,  187, 33,  30,  140, 244, 52,  204, 177, 179,
      9,   203, 118, 245, 193, 146, 18,  173, 188, 84,  207, 96,  16,  84,  0,   101, 172, 148, 102, 199, 213, 21,  238,
      138, 19,  157, 104, 95,  24,  36,  246, 41,  156, 179, 225, 233, 227, 148, 66,  53,  254, 141, 140, 23,  191, 9,
      49,  10,  67,  55,  57,  227, 213, 212, 98,  67,  207, 234, 36,  73,  180, 166, 94,  230, 175, 166, 58,  196, 234,
      182, 131, 66,  36,  108, 91,  251, 6,   39,  33,  209, 187, 199, 182, 41,  155, 65,  250, 153, 61,  33,  112, 42,
      201, 142, 148, 96,  61,  96,  160, 42,  233, 195, 115, 14,  207, 175, 197, 232, 112, 124, 41,  190, 232, 56,  104,
      164, 75,  68,  249, 144, 184, 237, 56,  129, 87,  124, 236, 63,  101, 132, 47,  148, 252, 78,  121, 0,   7,   81,
      192, 216, 67,  136, 212, 125, 191, 167, 167, 159, 233, 160, 112, 40,  14,  62,  48,  100, 104, 129, 48,  241, 152,
      67,  58,  135, 25,  99,  88,  254, 113, 89,  126, 211, 208, 15,  249, 245, 250, 233, 7,   129, 91,  157, 225, 63,
      137, 248, 196, 66,  120, 69,  73,  203, 209, 87,  163, 120, 111, 85,  68,  18,  47,  187, 148, 151, 226, 41,  99,
      56,  154, 58,  238, 24,  7,   22,  173, 133, 162, 246, 216, 112, 252, 199, 58,  224, 184, 23,  115, 209, 232, 244,
      218, 74,  63,  117, 69,  196, 198, 210, 52,  199, 150, 62,  236, 66,  47,  161, 94,  172, 4,   110, 179, 41,  172,
      139, 72,  246, 177, 237, 253, 8,   50,  135, 20,  93,  25,  21,  101, 160, 145, 159, 74,  99,  69,  132, 248, 46,
      182, 20,  199, 160, 210, 6,   45,  49,  178, 124, 128, 249, 192, 94,  121, 186, 76,  15,  149, 103, 42,  130, 247,
      242, 3,   97,  171, 32,  157, 56,  112, 202, 116, 124, 175, 49,  49,  13,  146, 94,  109, 61,  14,  186, 161, 145,
      66,  80,  41,  57,  133, 218, 79,  58,  214, 219, 170, 143, 106, 98,  82,  164, 229, 187, 121, 170, 147, 28,  126,
      49,  134, 16,  121, 104, 213, 207, 55,  31,  39,  165, 103, 61,  67,  209, 4,   15,  24,  72,  155, 150, 105, 149,
      141, 253, 245, 124, 137, 126, 219, 28,  27,  131, 66,  242, 215, 90,  51,  159, 134, 136, 223, 66,  241, 136, 2,
      25,  235, 176, 83,  13,  210, 136, 4,   233, 191, 162, 15,  101, 231, 113, 56,  155, 4,   128, 40,  0,   132, 4,
      195, 148, 210, 245, 69,  22,  5,   156, 62,  43,  254, 39,  236, 55,  178, 211, 26,  61,  95,  249, 158, 98,  58,
      126, 178, 45,  20,  4,   153, 216, 85,  218, 125, 2,   219, 11,  33,  202, 145, 46,  192, 201, 249, 197, 22,  231,
      132, 73,  181, 153, 253, 64,  209, 137, 253, 174, 100, 228, 157, 28,  32,  78,  145, 1,   188, 251, 147, 2,   105,
      174, 239, 64,  244, 189, 118, 147, 55,  37,  82,  130, 241, 200, 35,  167, 90,  181, 232, 171, 177, 5,   223, 204,
      175, 130, 121, 27,  202, 124, 20,  15,  18,  233, 247, 32,  8,   218, 100, 128, 235, 202, 196, 0,   6,   202, 97,
      14,  85,  247, 199, 155, 8,   99,  3,   1,   239, 18,  200, 131, 130, 114, 145, 111, 96,  53,  201, 164, 168, 87,
      67,  209, 194, 45,  237, 41,  65,  94,  48,  187, 239, 126, 181, 121, 87,  63,  143, 176, 50,  131, 15,  11,  132,
      242, 29,  126, 198, 51,  16,  79,  89,  207, 228, 220, 76,  97,  8,   237, 197, 192, 205, 89,  201, 224, 143, 44,
      50,  135, 189, 226, 212, 1,   252, 71,  204, 33,  24,  234, 46,  125, 85,  205, 192, 106, 241, 43,  198, 246, 55,
      193, 238, 227, 185, 103, 104, 145, 139, 224, 113, 133, 25,  136, 143, 93,  146, 44,  96,  150, 145, 202, 102, 135,
      32,  215, 241, 150, 139, 34,  202, 208, 10,  22,  82,  249, 79,  99,  189, 174, 240, 36,  254, 8,   142, 86,  148,
      159, 197, 235, 34,  146, 99,  94,  242, 45,  195, 178, 23,  164, 245, 32,  52,  46,  244, 233, 217, 79,  87,  159,
      48,  138, 69,  34,  93,  1,   171, 191, 56,  47,  90,  26,  215, 139, 248, 194, 138, 99,  138, 197, 145, 102, 230,
      180, 96,  115, 32,  156, 204, 162, 180, 82,  191, 5,   92,  199, 107, 79,  253, 207, 37,  46,  160, 158, 28,  136,
      240, 145, 103, 52,  186, 250, 170, 142, 123, 8,   101, 182, 242, 41,  29,  206, 115, 196, 62,  163, 29,  191, 58,
      19,  157, 99,  239, 228, 138, 33,  233, 177, 189, 137, 60,  6,   228, 52,  139, 213, 125, 36,  47,  165, 175, 168,
      111, 215, 212, 92,  94,  26,  195, 247, 83,  154, 66,  237, 232, 156, 250, 192, 162, 156, 113, 23,  10,  219, 72,
      195, 148, 19,  0,   56,  27,  78,  147, 62,  141, 184, 253, 244, 129, 60,  218, 69,  175, 229, 72,  199, 198, 234,
      37,  137, 38,  20,  153, 46,  216, 145, 51,  164, 141, 43,  219, 11,  151, 40,  242, 225, 116, 227, 110, 192, 132,
      97,  39,  120, 222, 133, 187, 122, 136, 67,  206, 92,  226, 184, 132, 75,  88,  194, 235, 13,  8,   10,  46,  87,
      54,  101, 36,  249, 89,  4,   164, 89,  106, 167, 141, 125, 159, 7,   14,  221, 42,  36,  60,  51,  20,  148, 145,
      2,   80,  243, 99,  60,  73,  220, 146, 217, 130, 167, 230, 235, 170, 49,  11,  169, 39,  214, 242, 44,  202, 32,
      117, 126, 20,  3,   155, 147, 142, 21,  210, 35,  192, 151, 226, 199, 14,  72,  84,  219, 49,  133, 122, 93,  186,
      38,  77,  32,  242, 12,  154, 188, 182, 215, 52,  10,  92,  149, 115, 79,  13,  131, 24,  110, 158, 76,  110, 98,
      126, 85,  241, 130, 27,  71,  40,  60,  112, 18,  201, 63,  215, 235, 42,  190, 155, 35,  204, 114, 138, 199, 175,
      169, 183, 0,   231, 193, 171, 207, 217, 58,  158, 254, 113, 32,  208, 106, 31,  223, 64,  1,   81,  16,  89,  161,
      51,  212, 158, 20,  179, 95,  248, 28,  189, 89,  89,  83,  71,  191, 186, 233, 157, 18,  171, 160, 49,  74,  145,
      155, 194, 202, 195, 86,  45,  60,  183, 156, 22,  145, 188, 182, 202, 176, 160, 78,  138, 124, 119, 25,  252, 79,
      16,  113, 97,  248, 79,  84,  60,  107, 47,  75,  197, 187, 138, 1,   2,   114, 184, 226, 10,  166, 88,  71,  93,
      10,  119, 15,  129, 151, 119, 50,  62,  142, 134, 211, 108, 0,   93,  254, 235, 178, 65,  108, 138, 184, 67,  124,
      250, 149, 45,  228, 168, 14,  237, 39,  193, 119, 35,  14,  144, 88,  225, 208, 72,  170, 182, 197, 14,  133, 117,
      197, 247, 121, 125, 134, 149, 53,  62,  148, 169, 131, 42,  193, 169, 87,  110, 192, 38,  26,  31,  133, 198, 79,
      94,  42,  233, 51,  133, 184, 155, 47,  136, 103, 15,  252, 19,  115, 97,  16,  122, 24,  56,  50,  170, 153, 22,
      126, 149, 108, 86,  124, 111, 159, 120, 206, 57,  31,  53,  0,   136, 6,   19,  234, 222, 235, 112, 209, 29,  4,
      171, 251, 56,  27,  132, 197, 137, 55,  187, 40,  50,  90,  101, 64,  169, 74,  241, 33,  73,  172, 167, 148, 183,
      59,  141, 84,  163, 215, 19,  211, 188, 108, 218, 145, 42,  240, 207, 73,  47,  104, 146, 68,  165, 127, 3,   119,
      46,  252, 193, 104, 39,  196, 64,  73,  19,  71,  61,  16,  135, 199, 47,  138, 142, 208, 74,  242, 76,  17,  84,
      61,  208, 53,  145, 190, 32,  149, 160, 35,  161, 101, 60,  59,  69,  66,  41,  193, 147, 177, 146, 227, 236, 16,
      17,  185, 104, 17,  192, 167, 82,  6,   172, 5,   87,  16,  76,  21,  102, 144, 164, 183, 69,  37,  120, 208, 20,
      78,  41,  149, 9,   176, 83,  94,  33,  144, 7,   158, 128, 86,  24,  121, 101, 37,  98,  70,  180, 214, 31,  66,
      252, 124, 54,  194, 79,  202, 126, 245, 49,  90,  32,  204, 90,  61,  171, 170, 52,  5,   1,   121, 93,  78,  221,
      134, 75,  114, 232, 176, 168, 182, 123, 210, 230, 9,   208, 225, 121, 103, 121, 58,  151, 19,  127, 231, 81,  71,
      158, 197, 55,  194, 9,   40,  121, 206, 239, 5,   64,  55,  98,  120, 183, 187, 164, 160, 138, 213, 193, 52,  67,
      72,  6,   81,  176, 42,  192, 170, 89,  231, 240, 165, 226, 178, 196, 221, 128, 39,  10,  196, 188, 181, 64,  129,
      46,  121, 5,   14,  74,  4,   164, 168, 130, 22,  174, 162, 172, 121, 155, 180, 101, 84,  54,  94,  34,  71,  87,
      138, 210, 223, 231, 126, 76,  194, 187, 37,  118, 99,  133, 77,  218, 113, 75,  198, 130, 21,  230, 4,   231, 184,
      150, 213, 251, 3,   2,   34,  10,  72,  116, 77,  201, 2,   117, 239, 99,  211, 47,  9,   97,  53,  210, 113, 248,
      218, 109, 246, 119, 254, 94,  11,  30,  121, 169, 29,  127, 62,  74,  12,  0,   117, 104, 107, 86,  241, 180, 154,
      10,  83,  31,  140, 109, 241, 31,  131, 68,  184, 95,  16,  102, 190, 248, 32,  88,  235, 54,  249, 179, 177, 59,
      38,  55,  6,   109, 173, 91,  105, 86,  117, 60,  102, 56,  3,   17,  155, 187, 133, 245, 48,  13,  122, 205, 97,
      207, 11,  221, 5,   128, 21,  253, 132, 51,  183, 15,  201, 179, 106, 206, 196, 1,   68,  89,  82,  147, 139, 212,
      1,   121, 206, 22,  149, 27,  208, 137, 74,  139, 107, 140, 197, 111, 84,  224, 94,  179, 218, 2,   217, 77,  62,
      54,  71,  145, 245, 195, 141, 31,  127, 210, 44,  167, 202, 229, 0,   167, 215, 102, 24,  244, 189, 40,  5,   182,
      167, 26,  127, 207, 45,  122, 184, 29,  178, 92,  111, 152, 97,  72,  148, 181, 75,  138, 17,  28,  215, 155, 147,
      155, 73,  67,  86,  67,  140, 177, 185, 145, 140, 14,  181, 56,  210, 234, 57,  235, 234, 137, 36,  50,  110, 251,
      117, 211, 97,  138, 174, 189, 133, 26,  216, 155, 108, 209, 249, 52,  9,   50,  112, 150, 164, 235, 105, 144, 115,
      238, 191, 69,  108, 238, 152, 79,  22,  162, 181, 41,  222, 252, 124, 241, 125, 106, 122, 166, 36,  234, 4,   86,
      202, 2,   42,  204, 139, 16,  253, 108, 180, 210, 6,   162, 108, 48,  0,   27,  184, 200, 13,  86,  181, 52,  216,
      55,  118, 124, 228, 38,  6,   252, 20,  148, 79,  130, 179, 208, 83,  8,   8,   108, 223, 240, 101, 135, 97,  94,
      168, 160, 198, 227, 53,  251, 62,  173, 186, 247, 214, 238, 109, 235, 84,  140, 52,  206, 175, 17,  46,  148, 237,
      142, 177, 149, 80,  66,  235, 229, 152, 94,  173, 1,   254, 11,  53,  180, 130, 16,  107, 214, 1,   204, 13,  197,
      123, 6,   19,  85,  204, 25,  38,  220, 231, 213, 95,  251, 123, 172, 239, 164, 19,  130, 162, 189, 117, 214, 47,
      73,  219, 152, 168, 138, 83,  48,  67,  110, 253, 112, 219, 79,  245, 76,  178, 227, 115, 83,  222, 181, 159, 131,
      150, 82,  154, 25,  18,  190, 47,  117, 110, 189, 40,  73,  53,  70,  238, 233, 69,  203, 121, 104, 125, 21,  63,
      69,  217, 80,  36,  164, 190, 165, 138, 48,  20,  188, 82,  38,  144, 214, 226, 253, 10,  229, 49,  221, 205, 141,
      15,  182, 227, 16,  4,   80,  220, 120, 225, 90,  99,  71,  73,  67,  136, 44,  46,  137, 194, 23,  113, 252, 85,
      76,  152, 239, 152, 69,  200, 243, 56,  201, 3,   6,   191, 94,  105, 232, 89,  156, 44,  118, 104, 71,  117, 2,
      43,  219, 142, 79,  229, 88,  4,   110, 170, 139, 162, 94,  223, 60,  53,  224, 131, 108, 228, 63,  251, 156, 235,
      3,   9,   194, 108, 37,  212, 161, 172, 46,  172, 178, 172, 203, 1,   180, 219, 39,  173, 2,   198, 33,  50,  199,
      20,  214, 242, 61,  21,  138, 217, 17,  101, 67,  245, 215, 16,  91,  8,   25,  222, 222, 82,  38,  113, 232, 125,
      208, 189, 71,  11,  106, 143, 127, 138, 36,  229, 58,  74,  127, 132, 162, 33,  162, 144, 45,  95,  26,  172, 83,
      247, 65,  148, 235, 243, 220, 98,  54,  88,  3,   203, 179, 196, 77,  8,   183, 58,  64,  0,   100, 122, 183, 142,
      154, 133, 62,  105, 240, 213, 172, 60,  73,  59,  109, 64,  81,  43,  81,  238, 251, 164, 171, 103, 32,  62,  50,
      75,  126, 192, 161, 188, 4,   81,  85,  12,  14,  227, 104, 143, 130, 246, 24,  109, 123, 236, 21,  197, 14,  118,
      241, 1,   137, 162, 197, 5,   117, 121, 239, 199, 163, 52,  220, 16,  180, 72,  28,  248, 25,  168, 151, 128, 15,
      235, 249, 124, 35,  90,  24,  91,  123, 49,  4,   249, 191, 32,  197, 102, 189, 55,  241, 78,  68,  67,  213, 14,
      72,  44,  10,  183, 37,  71,  248, 54,  140, 134, 243, 237, 29,  10,  212, 21,  238, 139, 37,  24,  181, 211, 221,
      88,  184, 127, 95,  56,  35,  63,  202, 129, 247, 77,  193, 192, 86,  195, 153, 226, 95,  56,  1,   149, 226, 216,
      249, 137, 129, 225, 175, 157, 122, 206, 110, 164, 241, 54,  17,  137, 56,  28,  144, 179, 80,  19,  39,  59,  6,
      245, 185, 35,  116, 211, 210, 249, 163, 73,  83,  48,  123, 63,  189, 65,  53,  156, 115, 41,  104, 22,  205, 151,
      86,  168, 3,   122, 116, 213, 147, 241, 193, 65,  173, 77,  211, 4,   192, 9,   176, 199, 217, 102, 98,  160, 158,
      57,  14,  17,  95,  190, 68,  204, 133, 15,  188, 131, 204, 109, 104, 110, 50,  41,  89,  155, 213, 98,  248, 55,
      155, 239, 24,  35,  235, 197, 170, 241, 144, 203, 133, 207, 207, 215, 219, 57,  21,  98,  33,  94,  38,  87,  194,
      26,  147, 192, 169, 51,  60,  191, 203, 26,  70,  54,  146, 55,  251, 155, 2,   120, 207, 197, 24,  32,  223, 136,
      50,  150, 162, 233, 157, 60,  226, 82,  162, 134, 129, 48,  90,  241, 253, 101, 230, 160, 94,  156, 138, 39,  221,
      190, 133, 158, 247, 106, 121, 143, 117, 20,  213, 75,  211, 187, 70,  227, 91,  8,   193, 24,  153, 47,  212, 186,
      17,  253, 75,  250, 7,   159, 213, 41,  226, 166, 58,  132, 173, 150, 102, 228, 57,  36,  66,  177, 0,   84,  212,
      114, 111, 15,  215, 223, 10,  195, 79,  227, 6,   66,  30,  35,  29,  18,  160, 150, 230, 159, 7,   204, 243, 142,
      232, 117, 138, 186, 153, 104, 182, 69,  83,  142, 173, 206, 173, 127, 91,  156, 191, 24,  13,  214, 241, 223, 96,
      222, 16,  179, 106, 152, 140, 102, 75,  91,  242, 117, 154, 206, 106, 127, 173, 203, 206, 10,  209, 146, 136, 67,
      156, 226, 59,  123, 25,  130, 209, 69,  58,  37,  230, 27,  251, 193, 74,  93,  87,  152, 221, 130, 98,  82,  77,
      189, 86,  167, 19,  201, 81,  94,  247, 247, 233, 198, 126, 172, 81,  126, 45,  178, 152, 105, 192, 227, 58,  108,
      117, 151, 9,   220, 80,  134, 233, 214, 104, 136, 153, 198, 30,  233, 33,  229, 200, 62,  248, 18,  9,   73,  60,
      33,  105, 208, 223, 18,  104, 209, 244, 197, 107, 240, 37,  197, 13,  108, 106, 56,  22,  109, 183, 59,  150, 238,
      16,  12,  149, 138, 155, 234, 219, 57,  105, 251, 86,  197, 141, 47,  170, 188, 73,  237, 73,  247, 88,  124, 184,
      67,  57,  218, 237, 174, 252, 0,   236, 31,  248, 225, 54,  222, 85,  59,  11,  64,  32,  142, 131, 44,  33,  55,
      53,  29,  109, 46,  22,  85,  238, 101, 188, 229, 249, 182, 194, 91,  106, 54,  192, 242, 66,  73,  142, 141, 117,
      5,   188, 45,  216, 112, 230, 131, 227, 126, 192, 153, 141, 83,  212, 117, 240, 33,  6,   33,  27,  203, 27,  100,
      18,  95,  150, 208, 161, 166, 30,  110, 197, 14,  94,  113, 52,  162, 159, 42,  136, 179, 152, 34,  78,  28,  170,
      146, 134, 198, 203, 52,  240, 7,   241, 130, 230, 75,  180, 228, 250, 220, 93,  170, 165, 218, 193, 181, 220, 160,
      110, 199, 206, 116, 117, 5,   107, 2,   193, 164, 184, 135, 25,  42,  45,  86,  166, 232, 134, 103, 126, 254, 177,
      63,  138, 79,  184, 208, 55,  15,  187, 49,  37,  122, 209, 3,   26,  107, 80,  207, 115, 137, 89,  26,  251, 138,
      140, 132, 127, 24,  254, 79,  124, 234, 123, 49,  16,  30,  121, 102, 99,  130, 198, 140, 75,  158, 189, 69,  31,
      223, 153, 112, 43,  24,  232, 170, 19,  239, 135, 167, 18,  250, 32,  248, 21,  46,  237, 238, 129, 194, 124, 80,
      237, 178, 186, 248, 22,  193, 145, 170, 154, 120, 185, 199, 88,  133, 238, 3,   20,  123, 21,  82,  144, 202, 66,
      100, 123, 82,  225, 48,  239, 149, 78,  157, 81,  16,  61,  59,  58,  178, 99,  163, 239, 105, 195, 224, 66,  126,
      43,  129, 178, 206, 36,  76,  132, 31,  28,  116, 229, 19,  64,  50,  100, 112, 235, 162, 101, 68,  66,  51,  236,
      17,  199, 73,  137, 149, 237, 231, 174, 134, 7,   246, 25,  74,  8,   64,  28,  230, 86,  176, 116, 89,  205, 140,
      42,  245, 17,  188, 105, 197, 40,  207, 32,  248, 31,  54,  127, 178, 171, 123, 159, 25,  106, 212, 123, 170, 158,
      220, 191, 112, 202, 225, 39,  133, 30,  92,  50,  151, 46,  147, 169, 129, 179, 72,  186, 250, 243, 3,   122, 16,
      201, 118, 8,   159, 83,  8,   45,  12,  36,  167, 249, 65,  186, 159, 209, 252, 122, 123, 71,  57,  155, 217, 244,
      185, 153, 132, 31,  195, 229, 2,   225, 141, 203, 31,  113, 79,  22,  37,  15,  36,  204, 200, 177, 65,  162, 250,
      235, 128, 134, 42,  252, 129, 253, 81,  222, 230, 237, 156, 213, 39,  186, 251, 86,  231, 87,  203, 16,  134, 98,
      157, 194, 117, 67,  231, 75,  9,   233, 236, 66,  9,   67,  116, 137, 155, 142, 194, 185, 18,  184, 150, 170, 223,
      184, 147, 170, 67,  67,  130, 67,  155, 226, 52,  76,  111, 65,  185, 15,  191, 148, 212, 145, 210, 242, 124, 165,
      42,  242, 137, 225, 110, 68,  113, 197, 213, 193, 176, 80,  12,  5,   223, 197, 227, 118, 85,  35,  144, 102, 71,
      110, 137, 94,  74,  132, 245, 156, 100, 201, 181, 205, 41,  200, 36,  34,  48,  94,  154, 228, 229, 207, 63,  9,
      248, 213, 167, 133, 189, 75,  115, 57,  29,  152, 163, 119, 188, 26,  208, 157, 32,  59,  219, 214, 8,   96,  140,
      51,  16,  251, 170, 23,  30,  26,  0,   164, 153, 45,  63,  112, 150, 121, 254, 112, 250, 168, 124, 114, 70,  50,
      178, 232, 239, 77,  23,  169, 224, 208, 41,  187, 157, 30,  38,  2,   205, 187, 86,  134, 120, 87,  207, 1,   245,
      167, 98,  6,   156, 153, 76,  228, 246, 198, 170, 142, 244, 205, 98,  240, 68,  209, 151, 12,  39,  59,  147, 157,
      155, 111, 144, 172, 213, 135, 159, 183, 32,  48,  141, 8,   126, 235, 16,  17,  95,  250, 47,  28,  24,  56,  211,
      74,  118, 40,  196, 159, 249, 151, 241, 239, 20,  2,   52,  207, 233, 46,  89,  41,  107, 11,  157, 246, 55,  27,
      25,  228, 34,  42,  106, 166, 10,  247, 193, 6,   119, 109, 42,  28,  95,  48,  101, 38,  223, 96,  29,  9,   50,
      197, 233, 191, 188, 93,  181, 36,  55,  1,   22,  205, 120, 220, 72,  26,  175, 22,  187, 240, 24,  58,  194, 102,
      214, 63,  9,   106, 120, 237, 150, 25,  143, 240, 233, 91,  220, 146, 95,  226, 33,  102, 224, 78,  151, 108, 183,
      225, 156, 181, 54,  178, 203, 177, 20,  208, 63,  57,  120, 48,  59,  19,  171, 224, 73,  139, 42,  206, 65,  95,
      43,  23,  155, 47,  252, 169, 214, 142, 221, 155, 123, 150, 164, 53,  49,  244, 228, 120, 159, 78,  0,   108, 117,
      161, 186, 192, 162, 133, 11,  195, 60,  139, 101, 149, 218, 106, 120, 235, 107, 85,  104, 25,  123, 212, 67,  55,
      191, 7,   127, 142, 223, 222, 200, 218, 5,   227, 237, 5,   90,  13,  219, 141, 11,  222, 150, 202, 169, 193, 105,
      195, 126, 134, 130, 45,  238, 58,  180, 94,  140, 252, 43,  103, 20,  238, 73,  69,  111, 134, 124, 174, 77,  111,
      5,   58,  115, 188, 27,  248, 214, 73,  23,  72,  231, 69,  61,  141, 177, 173, 124, 159, 238, 27,  236, 18,  43,
      236, 138, 116, 16,  42,  19,  34,  69,  94,  128, 210, 146, 249, 181, 157, 209, 171, 140, 158, 93,  183, 89,  145,
      239, 244, 6,   164, 245, 129, 233, 10,  187, 239, 246, 26,  56,  32,  70,  207, 3,   187, 121, 58,  69,  87,  213,
      238, 245, 102, 227, 116, 246, 192, 62,  24,  162, 25,  127, 67,  33,  74,  6,   215, 199, 87,  193, 24,  163, 39,
      48,  51,  159, 217, 27,  220, 53,  199, 170, 116, 25,  145, 223, 137, 249, 245, 2,   225, 116, 94,  20,  252, 189,
      89,  57,  94,  82,  163, 48,  119, 235, 204, 202, 33,  155, 176, 53,  131, 254, 82,  106, 160, 74,  95,  17,  157,
      128, 212, 26,  231, 225, 10,  67,  84,  139, 63,  147, 31,  61,  123, 177, 239, 66,  198, 62,  210, 19,  40,  243,
      145, 160, 245, 251, 243, 52,  134, 49,  111, 29,  135, 90,  166, 157, 89,  222, 44,  126, 29,  139, 196, 73,  13,
      126, 31,  143, 227, 84,  10,  91,  36,  43,  239, 208, 10,  108, 248, 136, 139, 127, 61,  51,  180, 75,  118, 77,
      222, 79,  226, 72,  174, 252, 25,  15,  241, 51,  168, 254, 49,  106, 167, 164, 217, 48,  203, 251, 93,  74,  179,
      51,  235, 229, 77,  242, 79,  195, 193, 207, 89,  121, 201, 179, 61,  182, 116, 188, 4,   210, 249, 56,  178, 95,
      200, 201, 184, 137, 163, 113, 28,  33,  243, 204, 243, 7,   178, 57,  239, 195, 105, 124, 162, 231, 201, 202, 9,
      78,  17,  245, 211, 85,  42,  192, 205, 212, 141, 90,  121, 237, 12,  192, 245, 46,  4,   250, 33,  221, 253, 78,
      30,  170, 196, 25,  147, 102, 143, 16,  167, 247, 79,  93,  175, 70,  243, 77,  135, 211, 127, 190, 130, 53,  230,
      81,  67,  246, 111, 84,  230, 183, 213, 14,  58,  57,  32,  67,  40,  44,  242, 235, 73,  169, 110, 46,  213, 35,
      33,  21,  62,  137, 58,  230, 227, 166, 55,  28,  158, 222, 106, 163, 64,  132, 55,  132, 244, 59,  131, 24,  160,
      249, 60,  176, 192, 21,  66,  128, 191, 177, 91,  25,  158, 168, 7,   49,  174, 182, 90,  3,   212, 70,  90,  250,
      55,  10,  234, 26,  34,  190, 219, 228, 169, 182, 130, 31,  113, 81,  3,   3,   49,  53,  38,  99,  182, 77,  173,
      148, 110, 174, 180, 180, 240, 16,  253, 223, 173, 23,  155, 29,  71,  44,  188, 227, 11,  233, 14,  58,  193, 60,
      25,  192, 187, 23,  171, 115, 14,  33,  218, 103, 135, 8,   157, 250, 218, 61,  199, 77,  77,  73,  214, 211, 191,
      165, 96,  118, 241, 113, 203, 224, 12,  47,  115, 105, 140, 15,  73,  8,   195, 119, 92,  231, 198, 161, 238, 40,
      89,  108, 30,  52,  222, 146, 189, 8,   238, 191, 243, 156, 214, 245, 107, 198, 132, 53,  7,   115, 169, 130, 67,
      40,  39,  240, 218, 225, 33,  15,  182, 52,  97,  87,  134, 19,  167, 145, 5,   139, 27,  250, 102, 160, 201, 150,
      13,  180, 51,  55,  40,  160, 212, 247, 126, 245, 114, 164, 25,  180, 174, 179, 221, 251, 226, 9,   129, 71,  79,
      140, 79,  73,  224, 78,  205, 15,  131, 174, 211, 154, 173, 171, 254, 10,  224, 77,  145, 212, 247, 132, 99,  4,
      213, 43,  68,  228, 49,  70,  142, 161, 65,  137, 113, 134, 105, 169, 144, 66,  98,  58,  183, 56,  71,  234, 211,
      199, 26,  21,  38,  189, 198, 236, 221, 163, 177, 230, 96,  67,  175, 63,  122, 209, 232, 226, 127, 224, 229, 40,
      205, 120, 235, 82,  54,  106, 179, 246, 66,  170, 139, 241, 87,  185, 12,  38,  222, 138, 181, 150, 7,   82,  16,
      36,  172, 155, 10,  68,  219, 184, 17,  243, 249, 188, 62,  228, 90,  120, 248, 73,  40,  75,  180, 46,  170, 215,
      139, 196, 59,  182, 142, 40,  151, 149, 175, 172, 38,  69,  134, 166, 196, 111, 63,  37,  152, 33,  58,  39,  34,
      63,  238, 202, 220, 4,   54,  63,  168, 144, 76,  213, 111, 185, 235, 132, 165, 122, 133, 119, 132, 238, 94,  154,
      174, 178, 38,  59,  76,  214, 53,  99,  109, 207, 9,   228, 148, 128, 101, 169, 91,  39,  60,  186, 22,  45,  56,
      22,  125, 211, 81,  14,  191, 181, 122, 76,  63,  196, 181, 128, 61,  115, 34,  155, 122, 121, 127, 247, 247, 177,
      134, 214, 4,   119, 177, 251, 5,   197, 56,  141, 252, 129, 97,  22,  87,  231, 28,  166, 104, 114, 189, 149, 89,
      226, 131, 24,  252, 150, 13,  211, 225, 106, 132, 139, 33,  196, 70,  243, 49,  64,  215, 182, 34,  200, 61,  89,
      160, 131, 2,   66,  81,  162, 174, 176, 190, 202, 236, 50,  184, 66,  136, 210, 15,  178, 12,  251, 96,  129, 208,
      42,  249, 241, 42,  26,  74,  141, 51,  155, 131, 29,  27,  173, 179, 156, 89,  51,  35,  96,  218, 140, 112, 169,
      164, 73,  221, 252, 230, 108, 92,  61,  250, 160, 165, 70,  138, 36,  38,  250, 87,  224, 201, 84,  152, 216, 132,
      215, 163, 137, 105, 81,  82,  196, 239, 14,  222, 234, 74,  233, 142, 194, 12,  163, 214, 184, 251, 162, 124, 212,
      1,   43,  187, 138, 36,  113, 217, 24,  232, 148, 140, 90,  101, 92,  31,  166, 155, 227, 58,  111, 99,  249, 244,
      136, 108, 102, 70,  10,  58,  210, 73,  133, 162, 70,  254, 38,  149, 239, 154, 222, 234, 100, 64,  200, 109, 199,
      27,  77,  141, 239, 122, 212, 165, 213, 216, 158, 192, 83,  25,  186, 239, 4,   156, 69,  104, 4,   81,  226, 148,
      61,  210, 128, 185, 226, 220, 22,  47,  215, 163, 177, 232, 228, 196, 54,  22,  33,  250, 77,  156, 78,  31,  37,
      87,  19,  114, 207, 19,  153, 45,  118, 161, 126, 117, 154, 228, 42,  70,  88,  124, 233, 142, 184, 206, 206, 190,
      151, 60,  24,  46,  185, 176, 202, 180, 87,  251, 208, 13,  9,   84,  105, 190, 186, 37,  174, 35,  2,   198, 135,
      174, 70,  253, 210, 215, 214, 79,  53,  97,  86,  94,  208, 134, 28,  89,  170, 198, 9,   103, 132, 40,  40,  215,
      242, 75,  114, 70,  8,   163, 220, 207, 181, 54,  70,  11,  10,  238, 84,  177, 151, 143, 16,  16,  197, 217, 64,
      155, 151, 197, 247, 121, 179, 235, 51,  28,  173, 99,  202, 85,  197, 116, 89,  55,  196, 223, 144, 99,  128, 1,
      246, 36,  107, 65,  179, 127, 181, 252, 164, 147, 147, 245, 136, 59,  212, 120, 171, 111, 165, 135, 172, 214, 164,
      184, 154, 183, 12,  157, 205, 190, 193, 110, 53,  95,  127, 182, 177, 183, 202, 213, 101, 19,  224, 26,  164, 86,
      211, 139, 148, 93,  237, 127, 238, 100, 237, 46,  217, 18,  57,  78,  32,  125, 167, 130, 16,  202, 44,  207, 81,
      206, 140, 183, 95,  152, 28,  76,  119, 83,  248, 230, 15,  139, 212, 91,  242, 185, 34,  173, 147, 167, 30,  133,
      110, 173, 57,  220, 202, 220, 135, 196, 163, 47,  174, 5,   151, 232, 119, 28,  37,  36,  136, 150, 91,  205, 191,
      112, 189, 128, 119, 111, 169, 104, 224, 106, 243, 198, 53,  108, 138, 221, 84,  160, 211, 232, 168, 184, 120, 231,
      16,  248, 138, 90,  107, 54,  248, 33,  54,  65,  185, 206, 26,  186, 112, 97,  113, 84,  94,  227, 233, 182, 148,
      51,  4,   14,  198, 28,  243, 4,   33,  81,  5,   146, 103, 248, 225, 50,  247, 219, 79,  248, 74,  53,  176, 201,
      214, 136, 12,  34,  147, 251, 113, 11,  118, 146, 130, 226, 106, 85,  56,  166, 40,  104, 115, 151, 124, 4,   130,
      220, 229, 84,  115, 8,   219, 197, 95,  161, 118, 94,  166, 213, 62,  46,  142, 121, 153, 79,  221, 97,  112, 131,
      66,  174, 18,  33,  154, 221, 45,  46,  105, 249, 159, 135, 107, 186, 135, 15,  184, 84,  213, 37,  33,  64,  223,
      166, 223, 35,  55,  187, 68,  23,  208, 161, 103, 12,  28,  73,  158, 248, 75,  243, 7,   81,  177, 39,  109, 155,
      66,  8,   18,  237, 119, 199, 19,  34,  188, 22,  123, 96,  86,  9,   140, 49,  216, 160, 227, 245, 27,  161, 215,
      112, 59,  149, 141, 56,  228, 41,  146, 190, 107, 146, 147, 243, 39,  228, 229, 49,  22,  175, 128, 185, 201, 62,
      190, 11,  195, 202, 127, 172, 157, 94,  30,  140, 48,  175, 41,  227, 244, 166, 97,  238, 228, 153, 137, 19,  136,
      215, 237, 41,  164, 72,  216, 85,  143, 245, 86,  149, 105, 43,  115, 251, 174, 19,  236, 158, 186, 225, 124, 38,
      167, 215, 39,  58,  218, 248, 22,  92,  251, 240, 175, 221, 132, 116, 54,  201, 106, 172, 41,  61,  221, 83,  52,
      206, 154, 90,  195, 218, 238, 16,  36,  124, 216, 75,  146, 68,  152, 192, 176, 59,  75,  62,  214, 55,  79,  225,
      129, 53,  205, 30,  52,  120, 16,  105, 245, 177, 45,  163, 102, 243, 227, 86,  12,  21,  123, 85,  85,  92,  11,
      25,  155, 198, 106, 199, 108, 35,  103, 223, 5,   153, 54,  149, 42,  152, 151, 141, 45,  79,  115, 168, 158, 74,
      1,   113, 68,  26,  63,  79,  120, 215, 192, 106, 228, 125, 169, 216, 136, 83,  189, 212, 154, 72,  18,  22,  154,
      197, 61,  18,  129, 223, 110, 74,  237, 179, 38,  146, 135, 19,  25,  197, 220, 20,  202, 26,  239, 1,   153, 101,
      22,  154, 4,   213, 36,  73,  111, 216, 190, 227, 181, 176, 39,  152, 97,  224, 124, 154, 89,  126, 73,  2,   157,
      239, 41,  246, 35,  185, 212, 187, 53,  77,  29,  16,  244, 130, 169, 6,   130, 224, 237, 88,  81,  33,  222, 222,
      186, 237, 216, 63,  1,   71,  130, 182, 137, 24,  13,  92,  109, 86,  119, 179, 129, 182, 198, 229, 35,  250, 69,
      93,  88,  133, 243, 169, 247, 141, 153, 106, 29,  102, 127, 108, 91,  230, 53,  235, 139, 14,  178, 165, 28,  102,
      44,  151, 227, 236, 59,  149, 68,  44,  65,  125, 135, 198, 222, 207, 203, 51,  188, 210, 249, 53,  229, 247, 18,
      119, 111, 232, 120, 17,  82,  22,  54,  196, 160, 182, 196, 26,  29,  186, 46,  22,  99,  220, 253, 191, 148, 140,
      121, 238, 135, 207, 129, 68,  43,  142, 212, 98,  55,  159, 141, 137, 19,  149, 58,  241, 23,  47,  73,  243, 37,
      172, 34,  138, 122, 7,   117, 83,  152, 215, 218, 202, 61,  46,  168, 79,  55,  207, 43,  168, 50,  211, 220, 204,
      194, 75,  179, 24,  132, 85,  190, 185, 123, 116, 76,  69,  228, 203, 88,  42,  241, 107, 103, 117, 222, 62,  95,
      85,  14,  42,  219, 27,  54,  32,  245, 2,   32,  37,  104, 138, 248, 31,  188, 90,  245, 200, 41,  218, 232, 24,
      106, 187, 187, 203, 212, 21,  158, 137, 147, 170, 253, 98,  214, 76,  29,  234, 110, 95,  234, 67,  89,  197, 92,
      148, 230, 171, 161, 4,   13,  188, 37,  92,  24,  244, 137, 146, 89,  116, 72,  63,  177, 59,  141, 86,  228, 222,
      99,  91,  199, 50,  192, 161, 79,  228, 70,  16,  64,  127, 41,  21,  132, 167, 209, 11,  235, 106, 32,  70,  196,
      179, 27,  40,  246, 195, 9,   89,  37,  33,  89,  92,  33,  208, 176, 144, 150, 144, 110, 102, 36,  37,  131, 143,
      32,  62,  226, 116, 4,   106, 6,   100, 156, 56,  202, 231, 83,  138, 22,  42,  147, 111, 229, 209, 99,  55,  44,
      72,  162, 254, 182, 1,   158, 132, 27,  18,  141, 208, 7,   189, 207, 244, 186, 129, 191, 47,  220, 144, 87,  151,
      240, 38,  40,  34,  168, 220, 221, 7,   30,  81,  21,  121, 162, 112, 105, 101, 234, 9,   237, 65,  199, 160, 229,
      227, 197, 217, 177, 12,  52,  20,  48,  159, 79,  239, 84,  144, 224, 67,  178, 140, 97,  100, 229, 247, 121, 68,
      248, 112, 40,  219, 25,  98,  240, 229, 87,  60,  239, 130, 138, 238, 205, 142, 221, 140, 83,  99,  166, 199, 183,
      135, 4,   36,  130, 66,  81,  171, 219, 227, 67,  237, 147, 50,  49,  223, 65,  82,  147, 9,   26,  157, 130, 112,
      93,  3,   36,  74,  106, 58,  241, 205, 91,  35,  137, 148, 201, 81,  4,   73,  17,  232, 30,  83,  175, 242, 187,
      165, 151, 164, 150, 60,  238, 196, 108, 186, 250, 96,  22,  125, 73,  28,  155, 242, 163, 173, 42,  93,  111, 197,
      18,  191, 19,  199, 180, 250, 86,  132, 243, 244, 237, 143, 144, 157, 29,  41,  182, 239, 1,   251, 152, 124, 186,
      45,  7,   88,  142, 125, 100, 82,  73,  145, 177, 201, 185, 251, 114, 131, 73,  64,  209, 111, 220, 231, 21,  60,
      8,   203, 239, 230, 103, 254, 209, 60,  57,  246, 139, 123, 109, 178, 242, 212, 138, 238, 182, 39,  181, 52,  189,
      188, 240, 49,  250, 189, 219, 184, 27,  17,  41,  253, 216, 94,  134, 98,  205, 168, 84,  105, 92,  27,  136, 181,
      140, 186, 208, 218, 119, 180, 56,  72,  0,   173, 93,  188, 111, 168, 206, 21,  249, 160, 126, 180, 21,  14,  214,
      32,  11,  183, 239, 88,  68,  120, 212, 89,  208, 109, 32,  74,  59,  48,  109, 143, 161, 187, 196, 36,  55,  89,
      111, 132, 124, 59,  112, 141, 91,  97,  254, 12,  208, 246, 190, 196, 83,  244, 223, 191, 66,  64,  236, 42,  205,
      131, 105, 75,  211, 224, 201, 169, 203, 182, 48,  36,  245, 91,  118, 227, 5,   42,  36,  183, 185, 134, 224, 196,
      84,  3,   32,  162, 200, 154, 223, 99,  215, 191, 251, 227, 47,  31,  167, 35,  210, 230, 163, 224, 39,  43,  70,
      165, 213, 168, 193, 40,  122, 199, 16,  228, 11,  233, 169, 2,   112, 135, 207, 206, 158, 24,  71,  134, 246, 207,
      202, 138, 211, 79,  123, 132, 100, 12,  2,   186, 181, 227, 9,   30,  124, 195, 52,  39,  42,  53,  182, 222, 136,
      151, 118, 154, 98,  116, 155, 228, 239, 242, 197, 236, 57,  172, 188, 229, 178, 125, 151, 162, 147, 146, 115, 86,
      118, 92,  248, 57,  31,  181, 113, 185, 36,  133, 239, 143, 77,  72,  148, 122, 9,   0,   45,  83,  29,  86,  143,
      54,  112, 106, 240, 144, 57,  183, 1,   5,   1,   58,  194, 244, 155, 132, 235, 203, 26,  138, 152, 181, 23,  218,
      244, 219, 215, 234, 43,  228, 245, 27,  201, 87,  237, 187, 15,  163, 189, 3,   209, 86,  247, 252, 58,  106, 145,
      209, 133, 229, 212, 213, 21,  224, 133, 250, 95,  145, 235, 103, 37,  84,  100, 32,  225, 19,  24,  224, 22,  111,
      14,  245, 9,   99,  0,   192, 25,  111, 16,  105, 189, 212, 142, 35,  124, 215, 198, 21,  199, 155, 40,  98,  17,
      105, 119, 236, 71,  84,  50,  189, 57,  115,
  };
  uint8_t data[18 * 2 * 16 * 32] = {
      116, 147, 48,  65,  40,  100, 194, 33,  93,  130, 44,  108, 251, 49,  249, 225, 210, 23,  163, 80,  144, 249, 202,
      157, 185, 38,  167, 187, 87,  122, 245, 235, 60,  192, 171, 182, 79,  111, 136, 246, 129, 40,  49,  28,  118, 8,
      240, 99,  87,  125, 140, 65,  77,  100, 210, 196, 81,  254, 202, 173, 106, 169, 44,  81,  124, 16,  168, 106, 31,
      152, 153, 29,  217, 40,  55,  3,   46,  225, 195, 74,  150, 144, 250, 121, 219, 184, 110, 48,  36,  13,  129, 10,
      144, 246, 185, 245, 24,  147, 141, 177, 14,  127, 138, 123, 197, 208, 215, 161, 16,  146, 157, 112, 145, 4,   229,
      212, 81,  65,  9,   90,  77,  16,  41,  182, 250, 68,  154, 22,  136, 248, 93,  254, 225, 175, 167, 198, 205, 226,
      26,  127, 99,  33,  91,  49,  108, 160, 234, 162, 185, 13,  51,  205, 229, 238, 237, 45,  33,  244, 8,   122, 141,
      224, 6,   153, 111, 123, 12,  87,  164, 50,  225, 146, 175, 22,  248, 123, 13,  213, 128, 17,  54,  39,  79,  83,
      81,  185, 85,  123, 174, 58,  108, 98,  82,  141, 159, 48,  247, 176, 14,  24,  25,  81,  195, 29,  252, 247, 101,
      102, 247, 88,  167, 111, 212, 7,   71,  168, 177, 236, 226, 252, 25,  176, 11,  99,  130, 194, 56,  37,  78,  147,
      86,  45,  89,  181, 229, 155, 137, 124, 100, 129, 180, 43,  242, 15,  104, 104, 143, 21,  134, 54,  229, 126, 235,
      143, 195, 87,  112, 102, 186, 176, 253, 233, 9,   60,  16,  153, 238, 6,   184, 11,  222, 168, 206, 170, 209, 89,
      195, 233, 65,  45,  251, 21,  124, 116, 166, 66,  180, 106, 11,  38,  185, 66,  173, 211, 161, 236, 163, 20,  125,
      244, 200, 73,  52,  183, 70,  249, 203, 19,  105, 80,  160, 20,  83,  82,  56,  28,  220, 244, 185, 44,  200, 6,
      20,  88,  229, 240, 105, 10,  228, 227, 42,  35,  71,  225, 154, 173, 209, 227, 193, 210, 175, 49,  61,  53,  200,
      48,  95,  62,  75,  88,  104, 25,  84,  57,  80,  203, 240, 34,  24,  182, 164, 31,  80,  124, 137, 197, 159, 252,
      97,  203, 215, 89,  243, 65,  20,  101, 38,  16,  105, 169, 141, 143, 236, 94,  204, 55,  118, 252, 180, 167, 154,
      179, 187, 216, 220, 168, 214, 68,  205, 120, 141, 114, 200, 90,  30,  147, 73,  108, 141, 211, 24,  49,  12,  178,
      196, 81,  174, 59,  54,  210, 225, 78,  222, 12,  150, 183, 0,   139, 233, 227, 25,  128, 63,  216, 135, 63,  125,
      145, 170, 92,  215, 42,  243, 82,  252, 52,  20,  100, 133, 31,  159, 167, 174, 16,  80,  4,   172, 171, 241, 64,
      221, 77,  49,  220, 191, 79,  142, 208, 248, 159, 122, 110, 205, 243, 210, 153, 89,  112, 19,  92,  20,  188, 27,
      241, 104, 168, 143, 211, 226, 210, 85,  25,  199, 155, 26,  177, 229, 15,  61,  94,  105, 10,  71,  16,  236, 109,
      67,  196, 57,  169, 216, 112, 60,  100, 106, 26,  32,  210, 167, 182, 236, 193, 12,  69,  140, 80,  160, 150, 146,
      174, 64,  242, 231, 77,  219, 142, 78,  209, 245, 203, 53,  120, 113, 143, 80,  128, 86,  34,  117, 173, 47,  52,
      236, 65,  156, 216, 29,  31,  163, 124, 180, 184, 91,  179, 120, 243, 190, 61,  184, 239, 251, 148, 206, 162, 250,
      184, 193, 102, 150, 164, 38,  229, 225, 20,  80,  66,  173, 117, 12,  75,  233, 204, 118, 97,  182, 180, 65,  254,
      156, 114, 148, 165, 126, 182, 82,  215, 192, 30,  139, 65,  190, 185, 108, 203, 125, 87,  228, 182, 26,  124, 116,
      66,  117, 99,  253, 46,  114, 42,  156, 168, 28,  230, 111, 80,  51,  152, 104, 20,  89,  75,  64,  189, 91,  29,
      100, 128, 209, 157, 180, 231, 46,  93,  244, 61,  11,  191, 180, 118, 75,  137, 231, 31,  199, 202, 96,  153, 85,
      46,  249, 111, 176, 62,  78,  48,  29,  122, 153, 253, 88,  41,  168, 58,  212, 200, 114, 72,  234, 50,  128, 92,
      35,  185, 29,  163, 96,  234, 166, 31,  75,  203, 55,  16,  209, 203, 63,  209, 187, 238, 112, 232, 210, 94,  246,
      25,  132, 45,  186, 3,   124, 14,  218, 82,  49,  39,  72,  137, 90,  2,   233, 4,   245, 236, 98,  85,  46,  181,
      19,  189, 36,  94,  156, 72,  62,  175, 20,  219, 192, 136, 194, 36,  125, 20,  220, 11,  126, 192, 133, 154, 79,
      126, 18,  204, 193, 208, 161, 189, 18,  195, 124, 32,  121, 168, 103, 158, 5,   138, 42,  196, 129, 130, 84,  231,
      77,  231, 72,  47,  246, 127, 241, 31,  32,  55,  102, 13,  21,  206, 1,   139, 218, 195, 227, 71,  113, 243, 89,
      219, 162, 63,  212, 180, 20,  67,  82,  106, 132, 83,  88,  64,  105, 81,  62,  85,  109, 137, 52,  151, 91,  189,
      44,  243, 24,  35,  123, 251, 82,  247, 178, 9,   198, 9,   209, 106, 12,  86,  247, 252, 59,  53,  229, 66,  112,
      212, 100, 197, 179, 203, 228, 220, 2,   166, 79,  209, 26,  124, 110, 5,   71,  201, 154, 133, 188, 96,  0,   112,
      239, 30,  207, 228, 68,  111, 141, 28,  249, 46,  25,  6,   223, 93,  72,  102, 143, 102, 54,  48,  65,  142, 252,
      167, 51,  89,  210, 163, 57,  9,   212, 4,   25,  175, 53,  213, 235, 62,  222, 131, 128, 174, 10,  130, 99,  240,
      44,  25,  103, 77,  113, 73,  40,  238, 211, 49,  131, 236, 251, 74,  156, 164, 160, 201, 81,  142, 212, 232, 227,
      70,  38,  109, 213, 156, 128, 232, 78,  45,  88,  60,  208, 35,  214, 89,  242, 54,  208, 243, 18,  96,  27,  41,
      229, 11,  48,  153, 216, 251, 167, 240, 41,  218, 152, 225, 245, 21,  106, 149, 158, 215, 189, 18,  135, 1,   111,
      224, 41,  16,  156, 111, 182, 189, 87,  7,   234, 217, 84,  247, 240, 142, 4,   237, 26,  58,  4,   165, 250, 173,
      180, 157, 186, 145, 209, 230, 109, 52,  27,  168, 208, 188, 236, 208, 214, 128, 143, 174, 8,   97,  20,  131, 92,
      3,   111, 89,  133, 46,  128, 3,   208, 226, 19,  55,  80,  233, 205, 129, 170, 242, 249, 159, 122, 132, 81,  250,
      200, 23,  79,  36,  13,  141, 139, 241, 228, 71,  101, 241, 232, 53,  172, 81,  120, 70,  187, 17,  234, 163, 146,
      126, 174, 165, 104, 204, 127, 244, 50,  250, 145, 56,  163, 172, 204, 166, 9,   162, 101, 227, 187, 13,  172, 237,
      16,  107, 207, 124, 144, 36,  31,  168, 120, 176, 82,  68,  158, 131, 118, 150, 145, 74,  93,  153, 117, 112, 185,
      49,  196, 162, 68,  219, 245, 9,   57,  117, 189, 203, 60,  10,  89,  75,  111, 145, 223, 234, 97,  210, 161, 0,
      231, 168, 182, 217, 56,  213, 28,  213, 109, 118, 73,  28,  70,  254, 215, 134, 188, 22,  219, 171, 74,  134, 59,
      89,  192, 116, 74,  118, 90,  233, 49,  108, 75,  160, 35,  239, 162, 175, 204, 99,  143, 112, 6,   206, 51,  176,
      149, 239, 70,  108, 30,  14,  96,  188, 86,  100, 252, 18,  62,  158, 48,  180, 200, 93,  221, 111, 36,  245, 14,
      215, 194, 15,  169, 142, 56,  106, 201, 142, 143, 21,  147, 50,  194, 78,  202, 47,  228, 40,  178, 99,  10,  20,
      67,  79,  27,  179, 73,  35,  54,  16,  174, 51,  18,  168, 165, 154, 49,  169, 91,  116, 188, 177, 133, 62,  159,
      145, 202, 172, 85,  191, 240, 181, 170, 219, 69,  217, 121, 56,  145, 65,  128, 21,  137, 198, 17,  33,  193, 161,
      127, 132, 166, 226, 237, 167, 211, 44,  37,  172, 74,  224, 245, 87,  41,  177, 203, 143, 36,  214, 115, 64,  234,
      76,  166, 50,  132, 118, 245, 111, 118, 242, 73,  187, 51,  230, 102, 150, 63,  138, 46,  118, 66,  147, 244, 209,
      94,  30,  243, 41,  170, 54,  81,  184, 208, 158, 34,  231, 170, 86,  29,  91,  49,  110, 87,  197, 19,  236, 125,
      164, 48,  208, 254, 72,  43,  120, 68,  197, 110, 0,   248, 123, 160, 147, 51,  69,  237, 132, 168, 102, 247, 83,
      49,  96,  111, 72,  104, 88,  144, 226, 139, 146, 9,   133, 224, 204, 45,  89,  145, 211, 102, 142, 190, 7,   2,
      76,  236, 233, 46,  80,  162, 254, 104, 62,  113, 2,   85,  166, 179, 200, 244, 47,  72,  169, 49,  222, 65,  197,
      188, 157, 48,  176, 90,  66,  189, 235, 169, 156, 240, 69,  122, 248, 129, 30,  123, 193, 224, 112, 227, 117, 218,
      103, 248, 4,   51,  149, 46,  120, 96,  17,  89,  131, 25,  82,  212, 100, 208, 59,  217, 115, 164, 12,  160, 246,
      153, 75,  117, 193, 163, 63,  113, 164, 207, 9,   173, 23,  96,  106, 88,  54,  238, 196, 29,  56,  140, 182, 89,
      45,  218, 236, 192, 155, 6,   191, 31,  227, 177, 93,  150, 43,  162, 50,  15,  100, 121, 196, 19,  134, 76,  8,
      214, 135, 165, 73,  76,  201, 108, 105, 173, 113, 235, 144, 135, 53,  60,  47,  129, 78,  154, 85,  101, 106, 79,
      5,   222, 20,  68,  225, 73,  112, 145, 141, 151, 46,  211, 81,  48,  21,  130, 102, 30,  150, 29,  182, 92,  99,
      221, 156, 147, 7,   177, 236, 236, 40,  29,  169, 121, 221, 53,  139, 23,  151, 103, 107, 45,  57,  91,  15,  15,
      53,  177, 29,  204, 7,   22,  16,  209, 116, 135, 76,  201, 242, 219, 235, 149, 139, 239, 72,  5,   76,  209, 245,
      129, 139, 194, 191, 16,  46,  2,   252, 92,  55,  175, 12,  10,  176, 110, 225, 110, 251, 113, 4,   204, 143, 72,
      34,  40,  226, 174, 6,   102, 158, 1,   85,  183, 181, 101, 132, 249, 206, 207, 73,  20,  25,  33,  141, 229, 37,
      219, 56,  188, 159, 195, 70,  157, 49,  218, 100, 216, 253, 213, 254, 198, 227, 224, 118, 156, 166, 172, 47,  205,
      15,  157, 185, 244, 132, 61,  254, 115, 214, 88,  132, 212, 230, 220, 157, 172, 58,  210, 48,  143, 151, 196, 87,
      197, 118, 98,  185, 162, 176, 98,  219, 8,   201, 8,   13,  232, 139, 162, 137, 8,   152, 83,  165, 145, 13,  109,
      196, 59,  152, 137, 87,  148, 24,  252, 244, 52,  172, 177, 21,  248, 49,  76,  97,  175, 186, 79,  164, 134, 165,
      155, 126, 51,  25,  191, 128, 69,  55,  58,  218, 192, 153, 157, 137, 47,  26,  250, 195, 141, 229, 153, 11,  132,
      67,  235, 249, 159, 148, 146, 169, 20,  151, 122, 188, 13,  85,  201, 114, 171, 106, 51,  89,  171, 86,  54,  204,
      93,  182, 121, 197, 64,  213, 192, 221, 46,  22,  54,  126, 218, 184, 116, 9,   119, 62,  180, 131, 88,  6,   56,
      122, 253, 196, 121, 107, 2,   141, 16,  73,  252, 56,  156, 142, 94,  251, 161, 219, 214, 245, 38,  143, 162, 247,
      3,   0,   240, 43,  50,  12,  123, 162, 220, 128, 123, 55,  10,  237, 181, 63,  25,  193, 77,  247, 152, 220, 30,
      73,  244, 56,  149, 197, 88,  0,   11,  29,  249, 73,  230, 117, 18,  14,  152, 52,  250, 160, 193, 232, 184, 3,
      251, 234, 68,  223, 110, 181, 228, 189, 26,  140, 164, 250, 77,  38,  1,   246, 209, 144, 8,   24,  151, 89,  228,
      119, 96,  181, 26,  214, 240, 224, 178, 59,  95,  221, 25,  164, 235, 121, 249, 113, 239, 78,  141, 250, 147, 70,
      230, 14,  73,  137, 106, 24,  191, 182, 234, 31,  182, 16,  188, 192, 135, 73,  203, 8,   89,  191, 225, 64,  2,
      50,  96,  141, 89,  73,  165, 84,  82,  194, 1,   148, 58,  10,  149, 199, 216, 69,  185, 228, 31,  118, 124, 24,
      13,  2,   79,  170, 75,  70,  161, 7,   237, 86,  161, 146, 49,  195, 175, 137, 85,  174, 215, 175, 61,  238, 192,
      129, 177, 123, 199, 42,  74,  192, 212, 129, 19,  153, 216, 124, 246, 244, 56,  229, 1,   207, 151, 30,  9,   177,
      144, 121, 229, 221, 36,  147, 157, 241, 73,  151, 96,  42,  250, 29,  45,  177, 111, 103, 178, 189, 17,  84,  218,
      111, 243, 52,  35,  243, 164, 44,  25,  232, 117, 63,  184, 188, 186, 141, 41,  92,  190, 119, 45,  149, 232, 108,
      28,  95,  203, 233, 77,  199, 251, 249, 62,  96,  54,  5,   128, 154, 49,  108, 160, 237, 59,  165, 186, 112, 121,
      227, 211, 174, 25,  87,  37,  217, 153, 67,  156, 144, 128, 130, 183, 62,  144, 70,  123, 147, 163, 203, 44,  254,
      185, 4,   205, 127, 129, 125, 143, 193, 95,  26,  92,  75,  68,  11,  174, 167, 209, 0,   2,   222, 235, 123, 129,
      21,  204, 143, 107, 239, 44,  10,  27,  153, 56,  48,  48,  164, 108, 74,  20,  143, 233, 98,  192, 254, 128, 226,
      46,  0,   166, 183, 15,  106, 144, 254, 65,  54,  204, 249, 29,  105, 253, 134, 148, 38,  16,  9,   3,   28,  132,
      222, 157, 65,  159, 106, 78,  107, 32,  72,  7,   206, 122, 9,   105, 21,  115, 182, 104, 228, 136, 58,  186, 167,
      179, 98,  83,  254, 5,   118, 174, 128, 144, 32,  179, 166, 213, 155, 220, 175, 79,  107, 137, 150, 117, 44,  98,
      58,  76,  7,   29,  193, 165, 124, 171, 218, 3,   210, 191, 62,  252, 98,  249, 70,  205, 149, 74,  107, 208, 57,
      72,  56,  11,  143, 190, 64,  60,  67,  52,  159, 110, 118, 133, 141, 125, 44,  1,   55,  211, 201, 122, 169, 7,
      160, 44,  134, 213, 248, 42,  95,  139, 36,  143, 103, 0,   147, 199, 59,  201, 64,  106, 141, 37,  82,  234, 245,
      76,  253, 149, 75,  131, 211, 209, 4,   68,  87,  121, 186, 216, 240, 172, 69,  19,  31,  5,   125, 177, 155, 207,
      18,  22,  56,  240, 221, 161, 161, 75,  1,   32,  157, 104, 239, 162, 213, 118, 228, 212, 111, 157, 167, 238, 157,
      151, 22,  79,  102, 138, 53,  233, 230, 248, 181, 245, 252, 146, 70,  145, 94,  37,  66,  187, 23,  23,  155, 210,
      135, 62,  241, 166, 88,  67,  60,  180, 11,  13,  238, 198, 161, 207, 133, 126, 174, 113, 118, 247, 159, 223, 191,
      57,  222, 196, 80,  138, 173, 12,  180, 94,  233, 186, 16,  204, 8,   92,  106, 85,  90,  186, 83,  136, 248, 164,
      79,  238, 160, 177, 130, 187, 100, 202, 200, 122, 245, 18,  35,  176, 246, 5,   58,  186, 214, 162, 33,  128, 244,
      199, 152, 111, 236, 226, 113, 92,  138, 92,  68,  157, 191, 218, 41,  218, 190, 244, 184, 189, 159, 167, 11,  71,
      92,  62,  53,  195, 84,  116, 254, 32,  63,  160, 128, 21,  149, 237, 125, 80,  133, 93,  31,  184, 158, 193, 122,
      218, 152, 218, 52,  227, 76,  167, 91,  36,  242, 5,   196, 215, 173, 166, 122, 9,   244, 30,  136, 150, 244, 127,
      221, 28,  65,  252, 237, 40,  243, 162, 205, 13,  32,  125, 126, 156, 228, 244, 208, 73,  177, 148, 167, 72,  217,
      234, 184, 222, 42,  74,  187, 159, 202, 50,  248, 39,  172, 7,   116, 151, 230, 221, 36,  35,  112, 118, 30,  166,
      35,  167, 24,  24,  135, 112, 167, 41,  22,  83,  202, 80,  2,   67,  249, 1,   133, 11,  58,  64,  185, 114, 95,
      16,  53,  188, 66,  1,   59,  230, 194, 167, 25,  47,  144, 245, 59,  192, 206, 213, 50,  131, 226, 233, 87,  242,
      246, 156, 26,  189, 5,   140, 235, 41,  119, 145, 38,  175, 193, 228, 190, 204, 151, 142, 228, 47,  42,  244, 183,
      190, 215, 178, 220, 68,  33,  139, 62,  95,  64,  156, 236, 180, 106, 53,  198, 194, 20,  162, 33,  146, 100, 86,
      134, 119, 198, 40,  225, 54,  19,  220, 51,  18,  83,  45,  120, 1,   164, 146, 174, 179, 170, 247, 151, 151, 125,
      43,  35,  131, 76,  130, 251, 160, 212, 3,   57,  145, 237, 61,  66,  200, 187, 45,  236, 87,  115, 224, 171, 159,
      208, 35,  61,  14,  123, 132, 114, 181, 33,  72,  93,  106, 119, 169, 156, 78,  158, 209, 185, 243, 253, 37,  199,
      192, 90,  121, 87,  65,  161, 106, 65,  187, 81,  53,  249, 37,  182, 115, 203, 134, 169, 32,  153, 251, 223, 183,
      221, 162, 211, 182, 237, 225, 89,  7,   225, 57,  5,   56,  139, 2,   18,  248, 163, 151, 7,   199, 218, 199, 4,
      237, 101, 46,  178, 15,  171, 49,  161, 58,  229, 16,  103, 150, 219, 159, 235, 162, 40,  22,  178, 126, 125, 201,
      253, 99,  60,  165, 195, 192, 148, 184, 72,  46,  151, 97,  206, 13,  89,  7,   145, 217, 39,  3,   226, 93,  182,
      131, 241, 63,  138, 207, 31,  212, 89,  160, 45,  194, 90,  78,  123, 58,  55,  55,  193, 170, 128, 121, 174, 94,
      76,  113, 150, 77,  97,  254, 74,  154, 68,  235, 109, 3,   205, 132, 1,   206, 140, 217, 31,  4,   67,  85,  201,
      44,  79,  223, 156, 172, 39,  242, 215, 113, 127, 33,  65,  203, 100, 213, 81,  75,  14,  236, 197, 16,  15,  49,
      241, 183, 29,  211, 202, 130, 74,  189, 233, 105, 80,  233, 60,  197, 22,  138, 141, 124, 236, 64,  109, 182, 142,
      131, 240, 95,  94,  34,  240, 220, 110, 164, 134, 208, 26,  138, 123, 113, 197, 137, 205, 229, 75,  208, 51,  153,
      168, 23,  2,   245, 198, 151, 172, 126, 56,  241, 41,  34,  109, 29,  36,  175, 63,  233, 78,  203, 19,  43,  155,
      15,  217, 229, 201, 56,  28,  195, 17,  90,  33,  102, 243, 81,  58,  169, 58,  163, 160, 177, 90,  26,  113, 77,
      253, 227, 187, 8,   214, 224, 195, 108, 243, 106, 215, 66,  184, 129, 51,  51,  221, 70,  152, 196, 12,  43,  140,
      111, 58,  239, 27,  158, 104, 226, 196, 37,  161, 233, 46,  9,   198, 53,  198, 114, 11,  103, 81,  119, 9,   215,
      190, 22,  202, 41,  19,  143, 19,  39,  221, 221, 16,  176, 53,  45,  85,  108, 152, 1,   106, 154, 223, 25,  101,
      190, 154, 35,  130, 202, 209, 158, 244, 240, 81,  237, 12,  224, 68,  87,  140, 130, 50,  130, 205, 17,  164, 22,
      93,  86,  41,  7,   201, 220, 254, 242, 188, 41,  105, 119, 168, 202, 95,  238, 151, 240, 73,  56,  197, 48,  28,
      205, 164, 189, 253, 254, 28,  23,  254, 106, 178, 201, 96,  233, 111, 135, 73,  202, 40,  189, 47,  146, 229, 71,
      142, 119, 39,  20,  192, 93,  135, 67,  191, 84,  139, 7,   130, 110, 105, 204, 18,  68,  217, 85,  226, 51,  232,
      112, 191, 32,  214, 133, 149, 58,  68,  177, 181, 73,  24,  134, 206, 95,  144, 187, 157, 130, 142, 85,  138, 115,
      184, 123, 139, 181, 102, 99,  154, 58,  241, 50,  94,  218, 172, 207, 254, 143, 106, 158, 174, 236, 189, 215, 207,
      72,  86,  105, 174, 157, 118, 164, 36,  125, 246, 199, 140, 98,  68,  209, 151, 235, 183, 232, 12,  140, 1,   250,
      178, 245, 85,  177, 17,  203, 29,  40,  97,  56,  186, 111, 252, 224, 86,  140, 249, 127, 180, 243, 239, 161, 252,
      218, 8,   200, 113, 136, 85,  221, 89,  1,   158, 157, 144, 7,   15,  123, 181, 230, 221, 194, 27,  4,   123, 177,
      26,  215, 192, 150, 68,  65,  103, 210, 16,  0,   42,  233, 32,  250, 236, 175, 49,  39,  253, 196, 196, 222, 9,
      32,  249, 9,   110, 63,  20,  16,  128, 177, 155, 48,  191, 190, 211, 36,  48,  27,  66,  100, 152, 138, 14,  12,
      3,   229, 223, 47,  131, 30,  65,  243, 101, 197, 6,   102, 20,  17,  152, 65,  23,  53,  254, 164, 152, 101, 134,
      106, 20,  53,  227, 167, 211, 47,  166, 23,  120, 235, 246, 42,  200, 214, 14,  20,  182, 212, 4,   142, 108, 71,
      167, 104, 98,  158, 223, 24,  220, 63,  177, 61,  143, 60,  55,  0,   214, 124, 137, 228, 232, 239, 59,  113, 186,
      134, 231, 179, 134, 239, 109, 25,  174, 108, 205, 226, 210, 183, 75,  197, 26,  74,  132, 57,  200, 227, 243, 15,
      245, 71,  234, 51,  29,  20,  125, 18,  120, 136, 194, 173, 228, 72,  72,  232, 155, 3,   248, 215, 23,  40,  29,
      121, 65,  148, 107, 126, 250, 150, 198, 58,  177, 29,  167, 192, 244, 19,  65,  250, 200, 159, 167, 151, 58,  249,
      101, 227, 163, 216, 11,  112, 34,  222, 50,  14,  156, 252, 42,  136, 95,  89,  175, 87,  177, 117, 6,   68,  47,
      59,  148, 197, 62,  220, 143, 43,  56,  166, 94,  123, 113, 143, 196, 136, 216, 32,  127, 79,  126, 231, 141, 231,
      229, 2,   44,  154, 107, 200, 191, 47,  213, 144, 64,  70,  37,  126, 208, 229, 39,  132, 235, 198, 56,  196, 73,
      20,  83,  14,  242, 145, 236, 43,  95,  67,  180, 236, 131, 240, 163, 223, 96,  113, 176, 248, 31,  50,  5,   218,
      166, 120, 144, 231, 119, 239, 58,  104, 147, 122, 235, 224, 57,  188, 185, 244, 42,  152, 72,  107, 221, 208, 118,
      167, 221, 146, 139, 127, 110, 196, 111, 148, 133, 74,  55,  138, 253, 27,  207, 130, 33,  243, 64,  14,  196, 162,
      200, 104, 136, 180, 159, 167, 154, 215, 81,  170, 147, 124, 30,  219, 16,  197, 108, 42,  213, 249, 195, 174, 240,
      173, 185, 220, 222, 160, 45,  125, 37,  196, 35,  43,  180, 251, 240, 50,  53,  246, 211, 191, 2,   38,  23,  164,
      185, 179, 124, 208, 230, 24,  131, 242, 46,  11,  211, 211, 152, 188, 230, 34,  80,  241, 71,  19,  96,  40,  222,
      9,   53,  107, 140, 56,  182, 170, 149, 152, 52,  19,  89,  181, 149, 35,  17,  11,  5,   142, 192, 234, 36,  206,
      149, 235, 94,  96,  37,  223, 101, 200, 213, 101, 36,  109, 86,  187, 10,  211, 60,  66,  123, 229, 224, 239, 49,
      173, 212, 5,   96,  128, 241, 234, 76,  2,   9,   170, 175, 84,  77,  130, 181, 193, 179, 63,  71,  83,  24,  105,
      149, 85,  187, 224, 170, 90,  41,  209, 194, 124, 19,  57,  198, 55,  184, 119, 206, 134, 110, 30,  38,  229, 24,
      115, 211, 73,  154, 181, 11,  191, 41,  20,  85,  12,  18,  138, 196, 162, 200, 135, 104, 27,  191, 82,  253, 154,
      123, 82,  21,  119, 196, 175, 142, 254, 49,  121, 249, 8,   31,  227, 160, 134, 217, 36,  221, 161, 153, 4,   182,
      136, 110, 208, 233, 196, 120, 226, 22,  58,  74,  150, 110, 253, 160, 182, 124, 149, 32,  49,  145, 214, 187, 121,
      61,  105, 28,  134, 176, 15,  0,   5,   211, 178, 2,   249, 64,  239, 6,   175, 247, 162, 19,  253, 199, 216, 40,
      39,  92,  8,   4,   14,  2,   130, 214, 155, 199, 186, 215, 24,  130, 215, 199, 217, 223, 158, 83,  49,  60,  79,
      61,  221, 210, 151, 194, 98,  52,  89,  39,  5,   162, 34,  194, 75,  213, 201, 53,  94,  12,  10,  184, 54,  173,
      17,  97,  51,  68,  22,  197, 77,  127, 80,  136, 76,  3,   118, 229, 60,  235, 46,  173, 61,  16,  232, 138, 162,
      148, 179, 122, 172, 238, 126, 227, 21,  5,   180, 235, 4,   78,  37,  10,  221, 129, 209, 93,  86,  198, 192, 134,
      33,  33,  107, 109, 155, 47,  184, 31,  32,  46,  194, 141, 227, 33,  150, 113, 78,  7,   93,  199, 184, 103, 55,
      107, 140, 49,  197, 227, 94,  41,  207, 115, 157, 16,  45,  124, 169, 205, 167, 170, 12,  213, 235, 211, 239, 107,
      42,  42,  9,   55,  22,  9,   91,  151, 177, 171, 23,  123, 241, 48,  77,  179, 178, 33,  105, 245, 121, 221, 143,
      77,  67,  57,  110, 230, 64,  249, 91,  3,   34,  81,  173, 223, 11,  23,  157, 211, 12,  119, 30,  156, 132, 182,
      81,  66,  208, 155, 35,  252, 138, 216, 239, 163, 187, 90,  99,  210, 154, 77,  192, 4,   220, 54,  87,  126, 142,
      185, 84,  135, 97,  9,   70,  10,  12,  177, 199, 140, 112, 107, 175, 127, 227, 136, 245, 236, 58,  132, 83,  36,
      59,  205, 163, 85,  35,  151, 113, 228, 72,  110, 197, 22,  227, 89,  185, 169, 231, 99,  132, 186, 38,  21,  65,
      52,  67,  248, 60,  83,  66,  49,  112, 91,  232, 150, 125, 111, 132, 182, 88,  185, 231, 209, 212, 240, 253, 92,
      180, 206, 173, 22,  226, 3,   58,  56,  21,  162, 95,  138, 227, 193, 219, 167, 51,  62,  109, 214, 247, 136, 218,
      199, 104, 15,  51,  184, 98,  223, 122, 179, 211, 210, 76,  9,   49,  215, 69,  69,  91,  55,  90,  131, 161, 195,
      91,  39,  176, 231, 221, 233, 128, 51,  70,  89,  88,  14,  226, 2,   174, 23,  112, 175, 56,  74,  2,   0,   109,
      25,  241, 72,  116, 170, 41,  32,  160, 228, 23,  191, 244, 186, 46,  76,  129, 205, 40,  188, 210, 15,  125, 118,
      21,  136, 101, 12,  107, 81,  1,   98,  175, 228, 68,  36,  111, 178, 118, 47,  125, 29,  201, 202, 200, 159, 216,
      50,  202, 99,  23,  55,  223, 195, 86,  45,  231, 95,  123, 191, 184, 162, 59,  105, 7,   22,  186, 55,  195, 49,
      215, 32,  186, 197, 33,  91,  204, 138, 186, 72,  145, 18,  1,   206, 47,  186, 212, 91,  97,  18,  196, 156, 240,
      172, 19,  159, 104, 254, 230, 52,  214, 130, 225, 141, 248, 114, 161, 91,  62,  54,  235, 11,  189, 67,  128, 24,
      245, 10,  68,  225, 235, 218, 169, 40,  245, 211, 212, 92,  38,  14,  210, 19,  240, 111, 15,  87,  128, 153, 201,
      19,  160, 216, 116, 175, 216, 109, 98,  209, 220, 61,  132, 252, 228, 7,   252, 234, 252, 54,  21,  166, 116, 88,
      144, 22,  188, 158, 135, 76,  102, 70,  103, 157, 122, 201, 154, 191, 143, 129, 96,  59,  244, 197, 90,  194, 103,
      242, 148, 204, 125, 199, 100, 52,  169, 149, 196, 173, 88,  235, 159, 132, 208, 106, 189, 163, 237, 164, 95,  209,
      115, 180, 164, 182, 68,  78,  26,  108, 135, 214, 153, 249, 59,  14,  66,  242, 18,  45,  119, 43,  49,  75,  235,
      250, 94,  194, 202, 244, 140, 29,  211, 77,  171, 40,  140, 154, 83,  203, 217, 68,  20,  168, 37,  101, 85,  97,
      249, 193, 168, 205, 149, 196, 127, 222, 49,  107, 254, 173, 25,  198, 151, 161, 187, 23,  68,  206, 57,  120, 156,
      223, 87,  169, 127, 77,  40,  56,  182, 145, 156, 56,  110, 199, 45,  248, 121, 216, 245, 134, 134, 4,   105, 221,
      29,  82,  1,   162, 18,  188, 59,  110, 8,   30,  213, 221, 42,  143, 125, 248, 43,  140, 102, 218, 221, 65,  213,
      39,  84,  210, 82,  82,  247, 6,   148, 196, 32,  29,  245, 196, 113, 46,  157, 64,  170, 97,  95,  155, 99,  159,
      34,  160, 92,  248, 29,  155, 218, 154, 46,  16,  16,  47,  243, 16,  124, 159, 183, 123, 196, 200, 28,  180, 188,
      12,  121, 133, 134, 225, 36,  53,  169, 165, 207, 103, 202, 82,  119, 127, 179, 184, 229, 2,   236, 5,   39,  118,
      143, 227, 213, 34,  194, 220, 152, 54,  254, 227, 230, 197, 31,  77,  113, 191, 239, 224, 115, 253, 171, 143, 3,
      61,  250, 164, 84,  136, 16,  103, 163, 163, 93,  77,  180, 64,  103, 11,  248, 141, 177, 123, 33,  185, 118, 56,
      178, 68,  138, 180, 54,  21,  124, 133, 104, 202, 217, 237, 107, 107, 244, 148, 111, 124, 67,  29,  88,  19,  189,
      19,  171, 36,  203, 155, 203, 176, 81,  156, 232, 138, 16,  233, 213, 19,  114, 253, 238, 254, 165, 192, 105, 121,
      194, 206, 244, 149, 12,  198, 125, 195, 171, 235, 38,  169, 51,  50,  43,  10,  213, 224, 236, 196, 184, 222, 131,
      221, 189, 205, 168, 59,  157, 128, 247, 25,  62,  247, 26,  66,  13,  111, 60,  162, 168, 164, 204, 185, 47,  91,
      84,  99,  207, 87,  65,  99,  227, 146, 52,  244, 122, 32,  34,  68,  248, 228, 137, 227, 133, 125, 111, 152, 120,
      176, 54,  110, 194, 250, 69,  100, 219, 243, 37,  46,  24,  204, 102, 115, 206, 29,  176, 16,  206, 71,  166, 89,
      51,  110, 205, 9,   172, 150, 40,  12,  126, 44,  52,  241, 74,  91,  147, 6,   43,  179, 83,  36,  230, 32,  154,
      5,   209, 36,  4,   5,   220, 200, 127, 41,  84,  14,  221, 219, 115, 36,  34,  68,  66,  213, 102, 172, 54,  150,
      186, 52,  1,   46,  130, 46,  143, 181, 248, 234, 113, 142, 248, 43,  149, 21,  66,  111, 41,  173, 224, 63,  62,
      179, 8,   145, 20,  128, 103, 157, 13,  66,  219, 121, 104, 67,  149, 27,  136, 83,  101, 132, 156, 26,  45,  219,
      197, 209, 228, 228, 168, 83,  156, 0,   244, 198, 145, 242, 120, 226, 10,  164, 221, 80,  167, 242, 147, 14,  38,
      52,  110, 27,  235, 175, 58,  64,  20,  71,  74,  183, 202, 171, 134, 230, 19,  9,   120, 67,  231, 10,  14,  172,
      71,  187, 198, 251, 2,   210, 11,  0,   83,  95,  249, 91,  155, 207, 183, 89,  149, 111, 62,  210, 220, 44,  152,
      6,   104, 88,  53,  3,   172, 2,   21,  91,  125, 36,  144, 235, 196, 183, 60,  251, 192, 104, 197, 117, 72,  249,
      249, 68,  71,  212, 88,  247, 69,  87,  167, 126, 227, 233, 229, 105, 240, 183, 234, 170, 67,  248, 142, 22,  91,
      117, 126, 6,   208, 52,  28,  241, 170, 206, 135, 126, 15,  80,  132, 16,  213, 118, 74,  200, 63,  57,  32,  251,
      63,  114, 224, 187, 98,  142, 39,  135, 16,  74,  20,  11,  106, 28,  50,  1,   187, 106, 220, 108, 20,  224, 155,
      150, 51,  203, 242, 61,  95,  243, 124, 211, 237, 253, 16,  135, 246, 32,  46,  227, 132, 60,  25,  212, 190, 3,
      148, 173, 233, 171, 250, 191, 47,  92,  52,  214, 139, 34,  162, 136, 42,  131, 96,  230, 36,  132, 14,  214, 138,
      101, 249, 73,  222, 141, 83,  81,  22,  22,  19,  117, 184, 176, 105, 174, 86,  103, 8,   238, 28,  39,  19,  49,
      59,  146, 128, 172, 244, 215, 86,  158, 218, 116, 52,  124, 75,  30,  102, 92,  103, 141, 68,  125, 18,  74,  220,
      154, 181, 126, 212, 6,   83,  55,  27,  129, 40,  121, 199, 20,  178, 112, 239, 99,  81,  223, 93,  35,  30,  60,
      73,  197, 243, 239, 142, 251, 21,  60,  138, 49,  216, 27,  56,  21,  49,  14,  165, 125, 38,  234, 212, 177, 252,
      102, 94,  192, 125, 223, 42,  106, 116, 50,  202, 37,  223, 33,  125, 225, 40,  145, 180, 159, 84,  81,  198, 28,
      82,  27,  241, 108, 240, 223, 132, 151, 197, 18,  178, 21,  128, 151, 32,  236, 103, 126, 36,  189, 111, 50,  9,
      49,  2,   104, 27,  166, 231, 2,   203, 79,  109, 31,  28,  182, 168, 252, 212, 28,  112, 191, 213, 73,  34,  143,
      239, 36,  94,  181, 46,  115, 186, 100, 212, 247, 45,  134, 93,  146, 65,  74,  104, 89,  117, 244, 187, 132, 103,
      33,  90,  40,  191, 166, 241, 111, 59,  64,  96,  98,  176, 196, 230, 171, 245, 94,  179, 37,  233, 200, 2,   198,
      115, 48,  184, 18,  161, 24,  242, 153, 146, 82,  178, 209, 99,  60,  183, 98,  252, 11,  95,  41,  137, 176, 203,
      14,  185, 226, 79,  227, 218, 143, 164, 177, 97,  56,  231, 251, 13,  170, 211, 167, 181, 106, 17,  220, 9,   58,
      44,  231, 47,  9,   104, 178, 143, 130, 138, 28,  163, 136, 63,  148, 230, 54,  76,  83,  97,  202, 238, 65,  218,
      126, 246, 74,  84,  114, 210, 175, 52,  153, 146, 93,  199, 173, 132, 129, 131, 190, 249, 245, 4,   226, 77,  248,
      242, 221, 34,  199, 79,  57,  39,  159, 19,  79,  251, 3,   53,  74,  84,  190, 171, 227, 95,  194, 113, 216, 145,
      145, 130, 66,  164, 161, 12,  112, 41,  133, 87,  127, 128, 220, 206, 154, 193, 46,  209, 161, 124, 73,  246, 7,
      135, 98,  72,  248, 123, 59,  25,  73,  182, 95,  97,  216, 254, 239, 207, 68,  40,  11,  138, 240, 150, 62,  7,
      164, 168, 237, 10,  151, 9,   64,  108, 172, 214, 35,  43,  178, 85,  198, 222, 41,  121, 121, 60,  233, 86,  207,
      244, 72,  19,  242, 70,  46,  21,  8,   186, 245, 110, 220, 171, 146, 41,  154, 66,  139, 28,  73,  146, 189, 15,
      199, 211, 247, 46,  0,   33,  70,  247, 229, 248, 11,  166, 196, 49,  22,  223, 9,   114, 0,   85,  192, 62,  242,
      88,  234, 45,  84,  1,   152, 154, 59,  229, 226, 131, 145, 94,  38,  210, 207, 96,  62,  242, 144, 123, 214, 5,
      44,  72,  57,  58,  17,  217, 8,   241, 131, 118, 95,  12,  157, 216, 20,  135, 30,  42,  152, 197, 131, 119, 23,
      2,   171, 231, 205, 127, 47,  21,  210, 230, 207, 220, 172, 118, 11,  244, 21,  251, 150, 195, 64,  31,  17,  86,
      83,  13,  112, 232, 40,  165, 233, 4,   73,  169, 99,  252, 142, 186, 133, 106, 27,  83,  253, 123, 210, 69,  201,
      201, 42,  129, 172, 34,  201, 27,  247, 109, 16,  84,  211, 158, 2,   217, 135, 111, 209, 9,   129, 147, 83,  92,
      243, 179, 116, 96,  175, 89,  97,  24,  225, 21,  61,  39,  129, 168, 121, 81,  206, 33,  72,  10,  7,   182, 156,
      130, 198, 199, 172, 132, 29,  32,  19,  155, 108, 34,  200, 48,  3,   45,  191, 190, 107, 229, 229, 115, 194, 91,
      244, 64,  203, 111, 175, 67,  137, 58,  134, 193, 55,  162, 134, 83,  26,  85,  191, 116, 111, 162, 97,  188, 198,
      163, 113, 209, 56,  243, 49,  148, 230, 58,  130, 160, 43,  194, 252, 95,  118, 23,  104, 254, 144, 204, 2,   228,
      100, 199, 232, 188, 178, 5,   142, 70,  44,  33,  72,  114, 252, 249, 192, 48,  237, 240, 59,  46,  154, 10,  180,
      6,   224, 33,  86,  165, 236, 217, 153, 28,  105, 54,  10,  211, 156, 38,  211, 46,  145, 220, 162, 216, 5,   239,
      248, 84,  85,  224, 179, 191, 126, 124, 191, 1,   88,  32,  31,  52,  190, 164, 216, 12,  16,  1,   37,  68,  10,
      157, 66,  176, 161, 196, 188, 55,  228, 216, 213, 132, 113, 102, 101, 100, 129, 250, 155, 157, 101, 136, 225, 35,
      60,  19,  126, 154, 176, 29,  103, 138, 239, 194, 17,  209, 80,  88,  23,  133, 167, 119, 84,  245, 36,  123, 42,
      123, 67,  48,  55,  66,  245, 91,  168, 202, 81,  16,  193, 173, 123, 66,  65,  28,  251, 226, 204, 239, 129, 243,
      93,  156, 6,   56,  98,  120, 85,  99,  133, 184, 117, 236, 41,  50,  159, 199, 249, 42,  56,  146, 80,  170, 170,
      213, 209, 141, 131, 223, 13,  231, 4,   62,  61,  20,  11,  22,  137, 138, 52,  111, 154, 113, 56,  129, 234, 51,
      225, 222, 184, 46,  89,  42,  243, 230, 44,  142, 72,  77,  10,  158, 71,  42,  169, 158, 1,   179, 191, 145, 156,
      126, 170, 75,  9,   127, 40,  183, 72,  240, 196, 74,  121, 87,  37,  130, 3,   117, 113, 30,  117, 31,  16,  179,
      105, 187, 11,  15,  82,  27,  84,  54,  44,  120, 46,  239, 191, 71,  4,   224, 9,   178, 198, 138, 8,   208, 229,
      45,  243, 77,  142, 73,  164, 161, 143, 14,  121, 28,  124, 191, 67,  248, 21,  184, 77,  58,  152, 19,  172, 58,
      214, 200, 208, 37,  59,  219, 188, 57,  225, 131, 85,  106, 5,   192, 33,  161, 162, 170, 130, 165, 116, 25,  254,
      187, 207, 127, 30,  193, 85,  132, 36,  9,   29,  59,  170, 112, 70,  208, 86,  6,   244, 59,  159, 17,  118, 2,
      157, 166, 95,  197, 22,  240, 120, 146, 183, 208, 73,  47,  164, 168, 251, 79,  182, 200, 204, 162, 85,  12,  78,
      167, 211, 111, 67,  213, 166, 55,  176, 158, 70,  219, 3,   110, 23,  14,  115, 218, 165, 115, 231, 222, 245, 33,
      200, 174, 210, 249, 29,  218, 108, 38,  215, 222, 184, 36,  92,  244, 133, 234, 239, 25,  226, 47,  33,  114, 154,
      206, 202, 190, 70,  94,  132, 163, 84,  73,  109, 34,  216, 149, 147, 185, 157, 166, 237, 149, 186, 248, 54,  206,
      216, 62,  52,  243, 11,  223, 45,  26,  169, 22,  237, 20,  4,   176, 73,  187, 130, 138, 44,  18,  191, 34,  162,
      160, 48,  140, 40,  130, 166, 171, 65,  3,   137, 83,  238, 28,  18,  244, 1,   125, 114, 60,  57,  238, 250, 94,
      136, 72,  249, 65,  0,   2,   102, 208, 197, 78,  89,  253, 148, 54,  65,  210, 121, 163, 68,  98,  184, 25,  28,
      13,  61,  167, 2,   192, 206, 59,  220, 183, 88,  145, 207, 242, 168, 101, 79,  105, 179, 98,  51,  231, 95,  64,
      99,  136, 123, 91,  115, 23,  188, 177, 170, 176, 201, 204, 124, 230, 232, 185, 35,  191, 116, 215, 122, 98,  183,
      75,  152, 236, 213, 237, 183, 29,  17,  139, 86,  136, 183, 6,   43,  206, 129, 103, 148, 92,  131, 139, 23,  187,
      35,  71,  230, 30,  29,  124, 158, 224, 29,  131, 1,   3,   79,  0,   178, 205, 87,  74,  44,  221, 25,  137, 180,
      194, 57,  65,  68,  241, 160, 200, 209, 73,  208, 223, 204, 53,  122, 237, 100, 64,  194, 53,  156, 46,  103, 214,
      239, 230, 155, 247, 251, 190, 102, 0,   245, 104, 10,  72,  202, 45,  112, 235, 32,  198, 234, 183, 205, 95,  170,
      145, 214, 10,  245, 25,  53,  74,  214, 201, 7,   240, 79,  44,  79,  149, 46,  203, 28,  40,  24,  222, 232, 70,
      10,  34,  137, 90,  50,  199, 239, 192, 72,  184, 44,  39,  31,  237, 211, 110, 118, 170, 251, 50,  219, 173, 39,
      91,  212, 108, 177, 198, 244, 155, 113, 176, 69,  210, 38,  53,  168, 89,  201, 55,  59,  98,  111, 188, 192, 224,
      231, 181, 243, 236, 71,  185, 241, 186, 3,   232, 119, 0,   187, 121, 82,  3,   115, 58,  35,  189, 44,  121, 152,
      178, 69,  199, 36,  127, 242, 31,  72,  23,  151, 15,  96,  179, 185, 154, 64,  119, 3,   119, 16,  138, 185, 181,
      218, 104, 95,  171, 0,   245, 176, 212, 35,  112, 152, 191, 162, 92,  15,  19,  37,  180, 67,  231, 225, 215, 12,
      199, 43,  14,  213, 228, 161, 119, 161, 215, 5,   125, 58,  159, 113, 28,  190, 36,  68,  24,  193, 78,  229, 125,
      229, 219, 16,  207, 78,  130, 203, 214, 110, 48,  234, 97,  181, 172, 125, 45,  74,  171, 71,  169, 154, 230, 137,
      212, 231, 172, 198, 224, 1,   127, 221, 113, 209, 42,  150, 18,  212, 175, 164, 175, 102, 24,  139, 1,   206, 148,
      252, 45,  15,  128, 55,  237, 20,  147, 175, 171, 17,  222, 50,  217, 165, 15,  15,  231, 56,  107, 221, 125, 80,
      237, 234, 89,  138, 112, 149, 250, 73,  104, 139, 93,  185, 205, 249, 248, 4,   249, 6,   152, 3,   63,  191, 113,
      46,  203, 167, 122, 9,   22,  146, 83,  107, 74,  163, 68,  41,  159, 117, 90,  84,  0,   253, 182, 32,  245, 92,
      177, 149, 148, 80,  245, 154, 84,  130, 134, 103, 215, 225, 169, 11,  112, 1,   172, 236, 69,  204, 56,  135, 244,
      169, 102, 128, 221, 19,  180, 37,  112, 193, 111, 44,  166, 79,  70,  95,  115, 189, 28,  141, 204, 195, 32,  215,
      74,  102, 59,  247, 135, 216, 216, 67,  230, 169, 71,  154, 105, 50,  49,  197, 250, 167, 32,  99,  159, 177, 131,
      167, 242, 53,  205, 2,   31,  169, 128, 9,   99,  47,  16,  141, 207, 68,  152, 102, 135, 151, 141, 178, 122, 68,
      122, 42,  152, 119, 179, 157, 134, 228, 189, 189, 239, 83,  186, 99,  119, 81,  146, 203, 98,  160, 98,  246, 1,
      71,  176, 155, 104, 150, 221, 197, 76,  150, 149, 242, 189, 68,  207, 127, 216, 30,  83,  215, 252, 124, 78,  99,
      150, 139, 224, 89,  144, 57,  7,   219, 47,  111, 185, 254, 34,  182, 107, 190, 158, 212, 98,  32,  113, 73,  52,
      111, 241, 253, 223, 168, 215, 86,  128, 160, 169, 141, 47,  1,   225, 26,  2,   90,  76,  60,  167, 129, 107, 5,
      23,  135, 75,  63,  27,  217, 3,   30,  28,  224, 73,  144, 62,  32,  232, 170, 24,  87,  97,  108, 136, 186, 56,
      48,  249, 158, 157, 145, 78,  2,   154, 4,   5,   144, 74,  253, 254, 77,  201, 67,  72,  170, 66,  230, 102, 85,
      176, 58,  62,  214, 22,  159, 78,  191, 113, 131, 165, 68,  4,   124, 72,  173, 221, 21,  170, 149, 226, 43,  165,
      52,  87,  247, 225, 179, 65,  119, 41,  145, 134, 64,  131, 117, 93,  111, 169, 90,  69,  68,  56,  162, 23,  69,
      66,  172, 135, 74,  160, 192, 130, 112, 94,  41,  143, 147, 207, 4,   253, 66,  66,  210, 113, 195, 43,  95,  4,
      33,  231, 253, 33,  139, 215, 66,  23,  27,  60,  37,  164, 116, 133, 132, 87,  106, 36,  99,  235, 202, 21,  84,
      32,  92,  76,  40,  190, 183, 254, 81,  95,  186, 155, 22,  245, 52,  114, 208, 231, 121, 187, 6,   165, 188, 4,
      155, 138, 118, 21,  2,   239, 248, 121, 104, 140, 102, 177, 86,  133, 221, 201, 147, 208, 224, 71,  167, 225, 137,
      152, 94,  117, 110, 131, 194, 172, 33,  217, 25,  189, 58,  95,  220, 8,   154, 59,  171, 188, 130, 118, 239, 248,
      35,  32,  213, 54,  238, 127, 193, 37,  221, 201, 81,  108, 104, 62,  223, 197, 150, 57,  73,  218, 54,  44,  238,
      54,  188, 153, 27,  166, 159, 146, 230, 218, 117, 42,  177, 49,  115, 24,  121, 176, 35,  245, 34,  248, 152, 219,
      23,  150, 50,  41,  134, 156, 205, 147, 183, 95,  40,  52,  246, 10,  48,  233, 205, 24,  237, 95,  225, 63,  139,
      252, 53,  254, 93,  195, 116, 201, 243, 231, 42,  192, 30,  247, 127, 84,  40,  35,  55,  132, 192, 168, 90,  228,
      49,  174, 155, 58,  14,  61,  96,  105, 198, 146, 107, 196, 61,  176, 15,  241, 27,  140, 92,  250, 163, 234, 251,
      113, 31,  136, 162, 100, 122, 186, 148, 215, 250, 207, 228, 206, 180, 186, 253, 208, 40,  220, 177, 158, 155, 188,
      160, 141, 243, 201, 94,  97,  41,  105, 225, 22,  201, 154, 75,  214, 120, 227, 25,  5,   79,  63,  216, 154, 179,
      202, 213, 39,  2,   53,  224, 22,  72,  141, 47,  168, 112, 229, 159, 97,  219, 138, 199, 219, 82,  36,  148, 175,
      196, 155, 191, 237, 251, 100, 131, 8,   246, 182, 189, 134, 181, 0,   160, 195, 158, 152, 201, 47,  184, 69,  214,
      244, 196, 120, 178, 30,  13,  46,  232, 124, 168, 100, 199, 160, 157, 0,   95,  21,  45,  231, 206, 31,  77,  149,
      208, 78,  246, 4,   40,  246, 205, 39,  153, 47,  147, 123, 106, 124, 234, 153, 146, 208, 244, 23,  99,  94,  87,
      4,   77,  72,  136, 186, 27,  10,  26,  226, 230, 246, 151, 221, 216, 74,  156, 132, 12,  192, 215, 6,   205, 76,
      98,  113, 178, 139, 38,  145, 30,  214, 223, 161, 21,  187, 143, 81,  230, 29,  193, 32,  181, 223, 9,   86,  164,
      190, 36,  148, 66,  33,  205, 31,  96,  115, 144, 154, 45,  238, 249, 234, 1,   150, 37,  120, 44,  224, 76,  224,
      232, 46,  85,  225, 242, 38,  243, 28,  127, 64,  133, 127, 176, 85,  92,  142, 17,  201, 161, 249, 145, 237, 41,
      42,  67,  97,  2,   146, 172, 17,  187, 33,  190, 33,  98,  204, 22,  30,  149, 245, 86,  223, 209, 200, 209, 106,
      173, 55,  113, 221, 145, 254, 101, 54,  174, 19,  240, 14,  42,  80,  12,  18,  237, 127, 72,  42,  39,  209, 239,
      206, 106, 85,  99,  98,  160, 211, 89,  177, 117, 247, 67,  183, 20,  171, 210, 1,   2,   36,  228, 250, 61,  93,
      17,  193, 132, 124, 126, 164, 113, 84,  15,  30,  81,  50,  44,  223, 146, 2,   130, 176, 205, 254, 28,  233, 230,
      56,  189, 228, 23,  246, 97,  102, 120, 254, 238, 102, 177, 97,  167, 26,  158, 49,  126, 81,  154, 214, 56,  191,
      101, 200, 174, 238, 91,  189, 135, 82,  141, 199, 68,  128, 220, 87,  57,  142, 1,   57,  84,  240, 84,  45,  143,
      147, 170, 52,  131, 77,  116, 117, 177, 206, 191, 125, 124, 69,  92,  23,  65,  10,  196, 192, 138, 232, 115, 133,
      199, 62,  122, 206, 243, 72,  161, 219, 232, 82,  248, 203, 60,  144, 176, 138, 93,  154, 76,  233, 127, 129, 38,
      223, 47,  75,  226, 81,  14,  210, 150, 113, 232, 141, 188, 135, 229, 185, 196, 72,  203, 230, 88,  38,  167, 10,
      14,  55,  101, 74,  19,  213, 37,  226, 92,  103, 159, 85,  36,  190, 30,  104, 119, 64,  237, 178, 34,  17,  69,
      185, 88,  34,  150, 154, 39,  127, 86,  195, 151, 142, 108, 233, 53,  163, 148, 97,  246, 124, 82,  78,  72,  14,
      210, 239, 107, 133, 79,  192, 75,  37,  148, 163, 46,  50,  139, 32,  38,  178, 177, 151, 133, 51,  24,  100, 17,
      217, 219, 175, 251, 63,  70,  36,  28,  92,  144, 227, 219, 168, 7,   251, 130, 73,  81,  14,  79,  211, 155, 100,
      97,  216, 200, 121, 93,  31,  156, 8,   154, 210, 245, 216, 24,  94,  11,  31,  202, 221, 136, 45,  167, 14,  123,
      190, 194, 62,  191, 88,  254, 19,  104, 183, 41,  186, 215, 68,  228, 179, 78,  100, 115, 36,  167, 192, 215, 218,
      155, 249, 182, 98,  163, 176, 82,  104, 63,  181, 94,  51,  7,   236, 103, 104, 208, 91,  3,   48,  98,  250, 71,
      73,  119, 93,  113, 131, 16,  75,  197, 245, 51,  196, 104, 241, 219, 63,  201, 172, 12,  234, 202, 80,  6,   158,
      117, 194, 120, 99,  13,  133, 104, 23,  162, 32,  137, 201, 123, 7,   239, 2,   253, 2,   219, 34,  14,  171, 254,
      175, 11,  40,  88,  5,   221, 147, 7,   223, 83,  185, 25,  178, 165, 170, 108, 68,  112, 123, 54,  241, 49,  93,
      176, 100, 91,  180, 102, 0,   162, 191, 239, 111, 191, 165, 239, 119, 69,  187, 44,  200, 201, 60,  51,  132, 165,
      46,  23,  138, 165, 188, 170, 112, 28,  9,   166, 14,  120, 19,  162, 247, 150, 92,  65,  147, 88,  133, 177, 194,
      37,  120, 230, 247, 210, 21,  40,  56,  215, 26,  48,  66,  220, 169, 190, 137, 161, 84,  84,  215, 222, 91,  21,
      162, 190, 139, 6,   132, 87,  49,  92,  209, 97,  103, 211, 73,  48,  250, 55,  40,  63,  94,  254, 103, 191, 65,
      159, 19,  19,  68,  157, 181, 29,  153, 133, 245, 73,  241, 131, 73,  165, 82,  41,  22,  30,  211, 138, 115, 13,
      144, 30,  241, 16,  3,   82,  73,  18,  154, 246, 115, 196, 47,  139, 20,  99,  210, 94,  247, 202, 109, 57,  106,
      28,  55,  15,  245, 242, 241, 136, 56,  148, 254, 249, 153, 202, 132, 174, 233, 213, 212, 237, 10,  78,  218, 137,
      141, 227, 187, 75,  54,  167, 60,  60,  170, 32,  210, 219, 242, 149, 110, 60,  155, 0,   113, 16,  248, 233, 194,
      145, 49,  5,   232, 225, 81,  206, 164, 81,  226, 181, 164, 180, 210, 99,  230, 34,  99,  218, 169, 107, 109, 190,
      59,  117, 245, 5,   179, 139, 137, 94,  145, 202, 90,  194, 69,  105, 2,   170, 108, 108, 203, 50,  17,  222, 71,
      74,  26,  98,  183, 133, 109, 103, 4,   246, 241, 199, 25,  90,  197, 14,  140, 139, 95,  192, 230, 130, 81,  126,
      117, 136, 248, 18,  13,  238, 105, 237, 237, 222, 33,  101, 192, 45,  141, 203, 30,  42,  134, 75,  193, 5,   45,
      63,  37,  137, 254, 102, 80,  144, 156, 48,  213, 163, 214, 30,  121, 178, 187, 120, 153, 98,  157, 32,  250, 196,
      207, 55,  247, 101, 191, 120, 58,  150, 226, 156, 57,  42,  169, 49,  49,  74,  193, 188, 184, 7,   201, 192, 245,
      170, 79,  190, 230, 44,  33,  28,  244, 21,  7,   250, 169, 53,  180, 29,  193, 33,  199, 118, 140, 161, 189, 198,
      70,  89,  190, 178, 42,  173, 112, 61,  87,  137, 74,  162, 217, 166, 108, 70,  64,  212, 239, 61,  163, 77,  19,
      42,  190, 87,  37,  210, 134, 40,  207, 151, 248, 66,  22,  227, 56,  39,  62,  30,  238, 164, 43,  124, 92,  106,
      90,  146, 55,  16,  86,  155, 153, 115, 79,  136, 197, 38,  26,  36,  152, 157, 77,  237, 42,  162, 208, 220, 220,
      105, 227, 79,  100, 25,  238, 183, 227, 147, 74,  201, 242, 108, 242, 19,  143, 152, 82,  73,  203, 139, 39,  219,
      98,  140, 173, 126, 129, 123, 179, 0,   98,  44,  123, 158, 153, 214, 212, 15,  1,   161, 203, 21,  203, 223, 152,
      53,  213, 54,  155, 166, 203, 126, 130, 147, 66,  13,  165, 29,  167, 75,  222, 182, 215, 100, 69,  39,  129, 135,
      229, 31,  147, 239, 230, 97,  83,  102, 236, 242, 108, 106, 132, 230, 106, 73,  88,  9,   8,   142, 196, 79,  194,
      222, 205, 45,  212, 241, 125, 221, 40,  57,  24,  184, 246, 178, 101, 244, 153, 225, 219, 179, 17,  47,  184, 3,
      78,  85,  151, 206, 33,  251, 109, 159, 155, 138, 161, 190, 201, 194, 167, 2,   172, 6,   239, 23,  238, 93,  80,
      101, 33,  66,  169, 91,  62,  200, 53,  237, 60,  13,  153, 239, 113, 130, 89,  39,  31,  135, 171, 216, 15,  27,
      52,  211, 39,  1,   11,  241, 142, 221, 159, 4,   194, 151, 228, 55,  181, 130, 78,  168, 15,  130, 99,  119, 69,
      88,  142, 225, 98,  39,  249, 110, 85,  163, 119, 29,  108, 180, 172, 56,  9,   164, 50,  189, 149, 129, 200, 50,
      67,  237, 172, 47,  139, 53,  67,  24,  20,  88,  35,  51,  190, 64,  69,  209, 169, 193, 10,  219, 226, 151, 160,
      45,  6,   5,   242, 3,   148, 84,  167, 80,  96,  16,  133, 195, 185, 179, 198, 100, 21,  179, 167, 158, 113, 86,
      248, 192, 17,  141, 110, 77,  173, 168, 91,  221, 16,  73,  2,   90,  27,  214, 100, 239, 213, 83,  175, 175, 59,
      63,  60,  98,  133, 120, 201, 179, 100, 171, 143, 140, 62,  140, 10,  229, 36,  52,  56,  125, 13,  239, 24,  177,
      158, 77,  65,  137, 237, 225, 237, 25,  6,   148, 1,   154, 202, 226, 90,  81,  139, 112, 229, 112, 41,  43,  54,
      238, 147, 20,  221, 152, 198, 142, 209, 249, 54,  145, 25,  15,  229, 80,  174, 27,  139, 76,  173, 20,  86,  192,
      63,  159, 160, 117, 168, 168, 236, 181, 107, 44,  203, 195, 251, 185, 18,  249, 154, 192, 18,  152, 248, 80,  93,
      42,  51,  30,  90,  239, 24,  210, 237, 249, 55,  102, 79,  23,  202, 123, 196, 73,  171, 186, 248, 173, 113, 207,
      46,  67,  41,  221, 218, 20,  137, 104, 109, 73,  193, 8,   140, 99,  34,  98,  221, 74,  224, 246, 221, 48,  130,
      9,   53,  56,  16,  109, 210, 247, 9,   124, 7,   219, 218, 225, 138, 177, 44,  46,  4,   30,  236, 101, 227, 41,
      18,  115, 21,  115, 18,  75,  254, 54,  171, 228, 76,  2,   50,  39,  16,  56,  82,  74,  210, 121, 39,  92,  209,
      112, 121, 13,  47,  134, 244, 65,  120, 86,  172, 27,  155, 115, 22,  2,   71,  131, 78,  108, 192, 246, 254, 237,
      74,  22,  37,  158, 31,  81,  203, 127, 146, 194, 210, 148, 158, 78,  231, 170, 122, 57,  178, 112, 253, 75,  228,
      218, 90,  192, 82,  134, 238, 220, 182, 90,  219, 169, 154, 113, 33,  240, 183, 147, 174, 140, 5,   226, 182, 104,
      148, 227, 188, 58,  9,   76,  23,  137, 245, 156, 113, 50,  25,  14,  188, 78,  69,  70,  172, 25,  37,  239, 90,
      204, 142, 127, 97,  254, 172, 106, 222, 165, 158, 226, 212, 160, 137, 23,  140, 234, 157, 201, 126, 160, 243, 229,
      55,  243, 153, 216, 229, 130, 75,  71,  69,  217, 85,  53,  217, 175, 244, 1,   201, 169, 97,  187, 234, 42,  200,
      13,  103, 80,  100, 61,  19,  222, 235, 43,  44,  97,  244, 205, 57,  132, 240, 92,  12,  102, 171, 85,  160, 48,
      94,  232, 29,  106, 129, 170, 160, 225, 90,  95,  89,  25,  248, 45,  101, 62,  207, 219, 97,  119, 27,  84,  119,
      158, 158, 4,   127, 6,   252, 187, 80,  59,  37,  217, 182, 235, 27,  75,  11,  25,  109, 33,  32,  232, 34,  193,
      80,  49,  207, 87,  77,  227, 163, 7,   80,  85,  185, 30,  46,  104, 229, 84,  65,  128, 217, 179, 129, 83,  254,
      149, 53,  139, 242, 157, 180, 190, 151, 43,  192, 109, 125, 234, 239, 196, 19,  100, 213, 154, 147, 120, 129, 184,
      132, 161, 117, 209, 205, 216, 80,  1,   159, 233, 2,   65,  99,  171, 117, 209, 84,  177, 203, 47,  58,  216, 209,
      198, 161, 72,  93,  69,  6,   121, 242, 116, 61,  133, 98,  126, 31,  24,  124, 215, 136, 4,   155, 210, 39,  158,
      216, 145, 73,  100, 177, 125, 92,  232, 219, 147, 69,  38,  43,  110, 122, 88,  101, 141, 51,  73,  49,  117, 35,
      49,  12,  115, 98,  112, 35,  231, 32,  89,  28,  157, 202, 188, 90,  1,   176, 121, 71,  239, 160, 176, 178, 64,
      168, 101, 223, 133, 4,   10,  211, 109, 29,  86,  241, 248, 38,  60,  48,  128, 106, 212, 74,  179, 245, 229, 99,
      23,  200, 232, 43,  139, 228, 37,  1,   199, 101, 222, 11,  127, 172, 54,  58,  62,  64,  32,  81,  24,  1,   163,
      168, 24,  102, 72,  134, 37,  56,  86,  27,  98,  94,  45,  151, 153, 18,  104, 108, 16,  251, 73,  174, 85,  55,
      188, 242, 216, 83,  27,  110, 136, 198, 180, 181, 107, 86,  138, 122, 89,  79,  99,  12,  102, 100, 94,  167, 119,
      8,   26,  254, 232, 157, 87,  76,  240, 155, 48,  47,  159, 233, 27,  193, 38,  188, 220, 194, 25,  33,  181, 120,
      206, 214, 244, 106, 179, 207, 4,   137, 178, 202, 205, 46,  196, 93,  130, 183, 242, 21,  158, 59,  66,  49,  55,
      182, 38,  180, 71,  60,  218, 77,  241, 15,  238, 8,   132, 243, 160, 202, 12,  26,  179, 169, 250, 224, 105, 233,
      124, 24,  155, 102, 254, 200, 165, 4,   210, 177, 156, 153, 70,  151, 251, 174, 215, 28,  215, 181, 143, 75,  8,
      123, 245, 228, 228, 190, 140, 228, 72,  115, 38,  251, 172, 195, 68,  79,  105, 12,  198, 54,  168, 120, 83,  61,
      146, 239, 73,  97,  239, 29,  88,  147, 254, 7,   144, 221, 101, 197, 239, 229, 98,  205, 135, 67,  9,   106, 81,
      187, 186, 163, 199, 144, 124, 73,  64,  239, 123, 181, 184, 205, 186, 93,  180, 197, 27,  53,  156, 7,   40,  4,
      151, 101, 26,  36,  201, 42,  174, 140, 251, 220, 73,  163, 195, 24,  86,  206, 0,   48,  79,  23,  223, 72,  102,
      181, 85,  132, 141, 24,  112, 150, 72,  217, 127, 204, 111, 73,  189, 140, 190, 80,  216, 36,  205, 131, 43,  116,
      251, 57,  166, 54,  37,  94,  230, 157, 119, 37,  166, 87,  93,  31,  166, 60,  37,  191, 136, 101, 240, 246, 24,
      75,  160, 31,  215, 173, 175, 101, 155, 85,  106, 201, 147, 39,  244, 100, 55,  250, 144, 165, 246, 175, 229, 114,
      56,  77,  175, 121, 191, 161, 242, 158, 45,  135, 121, 33,  227, 69,  195, 108, 225, 27,  18,  165, 221, 193, 253,
      153, 119, 79,  191, 222, 190, 8,   231, 95,  126, 151, 34,  125, 138, 42,  67,  32,  65,  128, 37,  157, 12,  108,
      106, 250, 130, 98,  145, 155, 25,  98,  116, 115, 111, 38,  114, 143, 125, 225, 33,  252, 159, 251, 189, 88,  133,
      117, 218, 98,  178, 184, 108, 194, 56,  105, 112, 47,  58,  36,  139, 120, 83,  9,   150, 10,  69,  224, 228, 192,
      68,  9,   198, 150, 170, 129, 249, 91,  214, 143, 181, 91,  75,  74,  176, 67,  207, 148, 21,  173, 239, 189, 52,
      218, 216, 9,   211, 105, 138, 142, 114, 118, 172, 194, 21,  43,  176, 207, 170, 141, 111, 137, 74,  187, 216, 69,
      141, 49,  38,  232, 44,  2,   6,   159, 128, 229, 33,  40,  66,  137, 212, 116, 158, 86,  21,  92,  26,  72,  37,
      61,  239, 131, 55,  84,  228, 12,  160, 210, 244, 163, 148, 114, 208, 184, 213, 62,  57,  14,  65,  149, 155, 244,
      81,  59,  227, 98,  113, 80,  93,  247, 68,  193, 50,  232, 234, 245, 155, 233, 206, 51,  158, 106, 197, 37,  116,
      114, 50,  192, 180, 149, 14,  7,   44,  251, 209, 165, 72,  222, 115, 17,  43,  32,  253, 138, 20,  136, 241, 171,
      220, 199, 86,  172, 59,  143, 140, 123, 222, 145, 179, 145, 111, 100, 136, 95,  40,  169, 18,  202, 0,   72,  246,
      68,  194, 75,  23,  100, 71,  237, 106, 244, 115, 116, 6,   32,  31,  60,  154, 78,  24,  92,  238, 241, 204, 106,
      75,  64,  166, 251, 176, 6,   67,  223, 129, 171, 242, 16,  94,  165, 17,  178, 95,  136, 76,  202, 136, 218, 95,
      173, 211, 38,  11,  156, 40,  32,  142, 158, 150, 117, 173, 218, 82,  137, 120, 187, 222, 14,  196, 114, 253, 99,
      157, 111, 65,  139, 198, 150, 38,  60,  26,  105, 70,  58,  74,  177, 157, 15,  0,   47,  76,  119, 30,  89,  167,
      206, 252, 177, 136, 215, 37,  151, 41,  1,   146, 79,  84,  211, 121, 105, 194, 128, 65,  43,  78,  152, 104, 223,
      121, 12,  203, 158, 7,   218, 126, 79,  124, 101, 215, 152, 20,  243, 137, 89,  219, 155, 46,  135, 154, 218, 122,
      5,   43,  151, 59,  106, 210, 33,  9,   140, 126, 90,  19,  139, 138, 114, 47,  13,  184, 31,  95,  225, 251, 248,
      55,  142, 41,  39,  246, 125, 175, 249, 22,  187, 85,  5,   13,  140, 54,  95,  149, 228, 152, 74,  208, 27,  88,
      11,  254, 206, 164, 169, 55,  80,  75,  57,  23,  80,  192, 250, 43,  32,  58,  74,  252, 46,  208, 53,  167, 125,
      176, 169, 148, 117, 70,  203, 80,  58,  223, 229, 154, 183, 11,  8,   253, 238, 53,  192, 92,  19,  51,  171, 213,
      79,  210, 206, 66,  187, 242, 124, 97,  219, 44,  153, 221, 103, 239, 121, 75,  20,  45,  245, 183, 198, 29,  42,
      131, 248, 215, 227, 10,  122, 117, 98,  54,  130, 4,   40,  147, 52,  52,  3,   228, 160, 167, 201, 47,  155, 254,
      15,  122, 96,  194, 127, 244, 209, 82,  210, 246, 207, 114, 190, 54,  147, 95,  193, 219, 208, 166, 13,  142, 64,
      181, 69,  162, 65,  145, 87,  102, 180, 90,  137, 156, 138, 118, 215, 18,  215, 92,  234, 253, 129, 226, 226, 186,
      176, 142, 42,  49,  154, 66,  246, 7,   202, 140, 245, 16,  21,  246, 53,  67,  234, 58,  251, 155, 142, 115, 190,
      184, 132, 192, 159, 104, 99,  249, 228, 147, 147, 90,  130, 247, 0,   8,   114, 67,  187, 50,  51,  8,   44,  204,
      106, 185, 136, 102, 208, 240, 235, 23,  79,  171, 194, 96,  5,   160, 250, 29,  182, 252, 107, 184, 15,  108, 184,
      237, 208, 247, 131, 31,  184, 97,  126, 57,  235, 132, 64,  183, 108, 146, 104, 135, 17,  160, 193, 104, 16,  120,
      94,  101, 0,   182, 178, 138, 86,  56,  110, 26,  112, 69,  222, 234, 253, 27,  118, 130, 101, 251, 84,  142, 173,
      214, 204, 172, 47,  110, 115, 25,  53,  21,  138, 253, 195, 249, 164, 80,  211, 189, 151, 193, 199, 95,  109, 248,
      241, 21,  192, 146, 32,  226, 101, 133, 75,  153, 159, 102, 114, 6,   159, 153, 127, 96,  91,  206, 130, 153, 167,
      129, 193, 238, 100, 81,  105, 36,  50,  236, 134, 86,  188, 127, 170, 30,  72,  8,   65,  155, 2,   15,  128, 237,
      203, 75,  155, 184, 52,  145, 228, 192, 218, 212, 200, 141, 126, 252, 124, 248, 177, 236, 128, 158, 245, 32,  213,
      24,  199, 217, 223, 252, 235, 82,  26,  31,  19,  134, 222, 31,  208, 180, 53,  218, 19,  201, 247, 221, 73,  173,
      132, 142, 66,  26,  230, 127, 106, 12,  36,  249, 75,  142, 166, 34,  144, 133, 59,  148, 45,  122, 155, 4,   129,
      189, 16,  121, 197, 1,   164, 111, 12,  182, 164, 100, 167, 183, 15,  167, 135, 28,  189, 198, 168, 54,  112, 51,
      5,   248, 147, 85,  130, 161, 153, 33,  107, 64,  81,  71,  119, 93,  29,  14,  59,  207, 18,  196, 99,  207, 141,
      243, 152, 207, 179, 69,  222, 32,  188, 187, 253, 67,  13,  116, 99,  16,  235, 195, 176, 143, 21,  120, 109, 132,
      208, 64,  201, 227, 84,  191, 163, 40,  112, 207, 2,   39,  98,  172, 115, 185, 9,   106, 194, 181, 245, 3,   133,
      145, 24,  192, 124, 105, 214, 44,  114, 78,  160, 254, 156, 153, 242, 143, 211, 199, 192, 221, 187, 227, 56,  246,
      201, 198, 157, 32,  141, 184, 167, 194, 1,   142, 173, 57,  149, 246, 121, 84,  200, 237, 153, 158, 157, 20,  10,
      108, 36,  185, 192, 199, 218, 185, 130, 76,  207, 126, 135, 46,  219, 45,  3,   89,  51,  73,  180, 211, 27,  99,
      99,  252, 30,  250, 12,  50,  145, 145, 252, 87,  198, 40,  188, 89,  126, 224, 125, 139, 81,  18,  76,  123, 82,
      188, 94,  60,  243, 21,  138, 148, 36,  252, 8,   78,  188, 186, 204, 190, 161, 193, 27,  223, 124, 223, 81,  202,
      55,  229, 212, 62,  32,  248, 254, 225, 152, 195, 113, 185, 165, 164, 117, 251, 189, 134, 89,  9,   231, 4,   217,
      54,  194, 89,  236, 194, 34,  144, 169, 171, 31,  23,  52,  220, 79,  236, 179, 190, 1,   63,  142, 34,  55,  203,
      138, 12,  17,  86,  137, 152, 170, 178, 102, 182, 9,   186, 79,  195, 44,  189, 49,  181, 118, 118, 133, 155, 92,
      182, 157, 79,  34,  155, 77,  26,  2,   224, 173, 128, 87,  167, 134, 212, 193, 106, 111, 143, 9,   145, 189, 2,
      147, 209, 80,  225, 247, 67,  9,   92,  73,  252, 129, 162, 176, 171, 204, 168, 115, 253, 173, 128, 217, 151, 211,
      49,  51,  200, 237, 229, 59,  28,  149, 234, 188, 127, 20,  6,   54,  179, 224, 73,  64,  9,   49,  200, 112, 43,
      40,  144, 203, 183, 186, 40,  109, 185, 9,   212, 200, 67,  238, 3,   174, 58,  129, 89,  110, 182, 120, 169, 52,
      173, 23,  157, 8,   83,  9,   28,  161, 223, 158, 1,   134, 169, 121, 2,   173, 178, 8,   220, 195, 139, 164, 57,
      70,  190, 236, 85,  49,  121, 231, 19,  92,  160, 11,  213, 170, 60,  102, 77,  126, 51,  218, 204, 109, 46,  118,
      157, 7,   189, 21,  2,   238, 133, 216, 101, 117, 83,  92,  104, 52,  183, 220, 213, 72,  154, 225, 160, 76,  18,
      28,  35,  157, 19,  2,   32,  91,  228, 250, 11,  187, 81,  96,  57,  47,  156, 51,  43,  12,  193, 26,  26,  162,
      197, 153, 67,  100, 112, 132, 88,  117, 218, 122, 183, 197, 98,  104, 235, 21,  114, 128, 238, 15,  229, 244, 52,
      78,  122, 180, 115, 13,  54,  80,  160, 14,  92,  199, 138, 95,  98,  58,  148, 173, 72,  3,   86,  56,  218, 244,
      63,  60,  26,  233, 129, 6,   54,  153, 109, 54,  18,  184, 207, 232, 63,  157, 200, 22,  163, 24,  18,  161, 206,
      143, 1,   154, 52,  12,  212, 238, 26,  138, 7,   211, 112, 147, 138, 47,  54,  211, 169, 210, 43,  146, 34,  9,
      123, 130, 206, 161, 211, 221, 101, 81,  103, 220, 250, 182, 34,  160, 222, 134, 2,   113, 123, 70,  213, 82,  75,
      42,  228, 184, 49,  174, 60,  223, 243, 28,  66,  176, 10,  92,  195, 227, 124, 249, 105, 53,  93,  153, 173, 164,
      244, 162, 59,  242, 251, 93,  11,  199, 50,  165, 18,  49,  82,  19,  223, 113, 100, 92,  185, 108, 222, 93,  9,
      189, 161, 254, 61,  235, 125, 52,  149, 173, 42,  25,  193, 65,  30,  154, 217, 164, 170, 158, 149, 216, 43,  88,
      179, 16,  254, 220, 119, 70,  243, 40,  221, 129, 124, 110, 187, 251, 149, 56,  142, 216, 44,  248, 75,  113, 112,
      106, 225, 150, 207, 137, 15,  88,  33,  30,  234, 252, 106, 92,  166, 103, 113, 89,  187, 199, 191, 167, 152, 181,
      140, 86,  118, 242, 118, 97,  182, 227, 129, 62,  106, 129, 126, 173, 100, 253, 47,  140, 204, 148, 134, 31,  179,
      75,  238, 188, 227, 217, 212, 176, 127, 251, 69,  68,  220, 187, 230, 25,  35,  175, 233, 15,  132, 128, 85,  192,
      94,  92,  191, 222, 151, 153, 243, 113, 96,  70,  43,  178, 223, 172, 147, 27,  135, 198, 72,  143, 144, 251, 47,
      249, 251, 214, 243, 36,  148, 241, 151, 47,  88,  5,   115, 164, 87,  69,  215, 226, 51,  238, 222, 243, 244, 193,
      84,  238, 41,  254, 55,  36,  196, 6,   65,  148, 14,  232, 237, 47,  216, 233, 129, 88,  245, 157, 2,   55,  181,
      121, 32,  6,   99,  145, 209, 239, 131, 16,  237, 50,  33,  241, 103, 143, 135, 10,  36,  34,  164, 87,  191, 138,
      115, 92,  158, 170, 206, 202, 79,  47,  226, 129, 0,   67,  39,  72,  121, 42,  133, 155, 16,  149, 31,  112, 197,
      169, 59,  218, 68,  39,  199, 84,  35,  147, 17,  76,  208, 33,  37,  167, 167, 92,  17,  67,  210, 251, 216, 150,
      108, 252, 4,   108, 184, 124, 179, 135, 62,  140, 177, 1,   1,   25,  239, 73,  253, 83,  18,  70,  63,  48,  10,
      153, 137, 241, 19,  210, 65,  86,  65,  217, 190, 104, 55,  144, 98,  169, 55,  223, 82,  230, 158, 211, 47,  213,
      228, 111, 66,  7,   104, 91,  222, 242, 10,  198, 112, 18,  194, 145, 3,   123, 16,  195, 76,  100, 17,  7,   176,
      105, 178, 158, 92,  186, 159, 186, 217, 225, 36,  134, 156, 16,  9,   142, 147, 76,  210, 120, 162, 106, 67,  218,
      11,  134, 43,  174, 214, 173, 138, 210, 93,  129, 70,  238, 162, 87,  27,  137, 189, 204, 106, 231, 61,  40,  49,
      44,  143, 10,  75,  7,   225, 167, 150, 126, 251, 224, 171, 99,  161, 98,  21,  96,  224, 179, 38,  172, 188, 180,
      40,  196, 63,  76,  119, 76,  128, 56,  76,  122, 119, 97,  149, 16,  179, 116, 246, 147, 165, 157, 177, 86,  237,
      130, 95,  139, 30,  196, 37,  128, 53,  184, 248, 97,  4,   146, 74,  251, 56,  220, 118, 79,  154, 186, 223, 33,
      130, 179, 88,  102, 178, 149, 203, 119, 160, 220, 129, 137, 37,  215, 43,  24,  228, 62,  245, 90,  187, 98,  89,
      188, 72,  91,  64,  106, 9,   144, 32,  14,  212, 143, 124, 246, 236, 169, 5,   3,   251, 224, 122, 135, 202, 186,
      42,  116, 188, 86,  36,  0,   240, 244, 23,  27,  209, 229, 235, 224, 25,  215, 119, 131, 226, 44,  197, 132, 99,
      26,  48,  38,  217, 16,  231, 32,  223, 239, 103, 74,  29,  198, 90,  159, 191, 103, 186, 116, 217, 191, 92,  74,
      222, 158, 217, 133, 46,  0,   253, 182, 166, 187, 205, 216, 168, 48,  157, 69,  189, 135, 145, 152, 118, 181, 43,
      118, 184, 175, 140, 94,  253, 16,  95,  11,  210, 182, 48,  112, 62,  101, 167, 169, 73,  105, 240, 61,  203, 74,
      25,  212, 169, 78,  76,  113, 8,   201, 183, 80,  207, 63,  164, 100, 116, 236, 221, 193, 85,  117, 36,  75,  179,
      67,  28,  220, 219, 24,  97,  119, 102, 68,  100, 119, 74,  152, 85,  101, 230, 254, 105, 162, 6,   174, 186, 37,
      233, 161, 121, 8,   126, 24,  235, 243, 55,  229, 111, 51,  129, 201, 148, 57,  128, 6,   53,  58,  28,  243, 197,
      221, 130, 232, 6,   22,  248, 203, 64,  183, 248, 212, 87,  72,  82,  200, 201, 87,  77,  51,  164, 149, 179, 51,
      99,  243, 9,   94,  208, 64,  247, 84,  124, 191, 131, 148, 196, 151, 187, 72,  151, 116, 38,  164, 174, 188, 160,
      209, 18,  61,  183, 200, 100, 226, 181, 119, 71,  87,  170, 242, 33,  35,  52,  115, 63,  137, 57,  4,   20,  127,
      250, 246, 53,  185, 202, 59,  16,  229, 159, 106, 208, 6,   8,   33,  182, 151, 36,  222, 243, 113, 20,  150, 103,
      175, 121, 54,  31,  156, 16,  205, 25,  167, 254, 139, 148, 85,  228, 90,  165, 198, 110, 80,  61,  95,  83,  136,
      188, 176, 128, 22,  118, 138, 30,  180, 252, 167, 238, 86,  147, 218, 169, 171, 45,  20,  187, 67,  239, 5,   113,
      107, 79,  182, 252, 49,  244, 175, 169, 37,  102, 178, 5,   80,  130, 5,   202, 106, 202, 40,  66,  68,  72,  214,
      189, 46,  223, 100, 70,  132, 160, 69,  61,  154, 116, 140, 177, 173, 246, 40,  159, 212, 93,  169, 79,  231, 55,
      129, 221, 81,  204, 96,  3,   61,  6,   211, 49,  236, 85,  142, 157, 105, 70,  138, 148, 38,  53,  187, 73,  60,
      212, 54,  19,  91,  113, 66,  242, 33,  60,  55,  243, 76,  219, 210, 147, 191, 88,  146, 117, 250, 27,  212, 219,
      192, 53,  70,  104, 127, 111, 217, 22,  131, 176, 45,  152, 90,  188, 120, 150, 57,  99,  71,  172, 51,  22,  159,
      164, 73,  229, 203, 116, 250, 56,  129, 24,  75,  191, 24,  209, 179, 217, 208, 72,  40,  129, 75,  225, 229, 166,
      182, 182, 185, 129, 173, 79,  200, 54,  152, 135, 5,   32,  157, 58,  89,  68,  172, 164, 101, 43,  36,  134, 126,
      144, 161, 90,  197, 248, 143, 18,  144, 148, 220, 33,  126, 182, 222, 124, 121, 218, 121, 129, 36,  190, 141, 49,
      166, 13,  95,  223, 69,  22,  171, 48,  18,  83,  211, 194, 199, 58,  54,  108, 168, 75,  235, 19,  176, 98,  32,
      51,  123, 41,  195, 197, 48,  196, 211, 249, 151, 136, 175, 221, 184, 121, 85,  64,  235, 10,  202, 166, 51,  203,
      229, 215, 189, 179, 108, 114, 207, 191, 164, 114, 114, 20,  174, 200, 138, 247, 104, 52,  47,  199, 45,  211, 126,
      219, 103, 93,  36,  112, 130, 73,  142, 177, 225, 110, 175, 113, 220, 151, 181, 146, 55,  48,  127, 237, 148, 113,
      80,  69,  51,  52,  232, 118, 247, 97,  110, 93,  180, 98,  214, 126, 0,   127, 75,  160, 153, 248, 246, 85,  57,
      73,  127, 38,  99,  2,   141, 151, 29,  120, 147, 158, 155, 240, 182, 101, 228, 23,  242, 96,  155, 29,  210, 181,
      33,  108, 175, 165, 146, 34,  131, 165, 196, 150, 11,  187, 151, 223, 35,  187, 179, 80,  118, 18,  64,  153, 78,
      150, 84,  94,  145, 165, 62,  84,  165, 223, 97,  85,  53,  70,  19,  246, 149, 163, 88,  213, 249, 191, 160, 97,
      2,   57,  98,  103, 143, 174, 219, 245, 147, 41,  101, 249, 119, 199, 20,  42,  94,  11,  176, 136, 208, 253, 212,
      65,  102, 208, 82,  89,  9,   34,  60,  88,  105, 52,  146, 0,   138, 156, 141, 57,  73,  234, 39,  250, 96,  207,
      56,  56,  148, 67,  160, 48,  254, 250, 28,  105, 37,  232, 27,  8,   202, 3,   111, 194, 126, 15,  79,  197, 135,
      24,  43,  104, 93,  32,  22,  254, 146, 245, 159, 56,  99,  32,  199, 28,  142, 115, 176, 16,  75,  10,  151, 93,
      124, 193, 170, 125, 193, 198, 47,  122, 108, 53,  209, 137, 74,  141, 145, 165, 196, 47,  61,  161, 177, 192, 21,
      20,  144, 98,  215, 26,  111, 28,  140, 234, 117, 155, 164, 238, 124, 86,  180, 200, 228, 208, 97,  173, 52,  149,
      254, 204, 38,  164, 219, 112, 222, 190, 238, 69,  20,  229, 226, 129, 54,  39,  249, 53,  168, 173, 102, 68,  110,
      155, 203, 33,  51,  251, 136, 162, 160, 247, 187, 212, 41,  57,  15,  30,  204, 182, 156, 16,  117, 209, 140, 43,
      250, 124, 170, 20,  82,  61,  224, 132, 112, 199, 246, 116, 197, 171, 220, 50,  250, 83,  159, 244, 141, 204, 42,
      213, 134, 233, 194, 170, 111, 242, 159, 133, 120, 55,  5,   222, 130, 103, 77,  73,  124, 116, 146, 183, 233, 207,
      245, 25,  193, 220, 249, 89,  202, 160, 225, 123, 243, 111, 126, 227, 108, 118, 115, 33,  145, 13,  245, 226, 78,
      254, 4,   65,  58,  38,  67,  229, 246, 222, 155, 62,  46,  151, 111, 238, 213, 132, 214, 101, 22,  63,  121, 177,
      22,  89,  197, 235, 89,  36,  147, 135, 205, 183, 211, 46,  16,  152, 212, 133, 163, 36,  119, 108, 120, 33,  113,
      14,  103, 53,  113, 166, 229, 94,  221, 221, 135, 166, 222, 182, 129, 133, 127, 165, 68,  51,  111, 182, 253, 68,
      19,  172, 218, 168, 132, 203, 95,  2,   245, 203, 214, 197, 37,  63,  199, 127, 32,  89,  150, 62,  214, 89,  161,
      59,  190, 42,  203, 42,  32,  227, 39,  199, 207, 138, 181, 53,  155, 229, 118, 133, 45,  144, 155, 218, 201, 86,
      212, 103, 22,  111, 198, 236, 27,  210, 173, 64,  173, 92,  20,  193, 21,  25,  164, 72,  242, 95,  30,  94,  79,
      195, 217, 88,  55,  24,  221, 48,  149, 110, 224, 158, 30,  212, 223, 88,  93,  2,   78,  149, 109, 146, 103, 101,
      175, 118, 11,  59,  141, 76,  235, 38,  138, 25,  249, 65,  187, 119, 240, 185, 175, 71,  110, 202, 116, 40,  217,
      182, 60,  191, 74,  127, 50,  131, 65,  74,  4,   209, 170, 44,  246, 246, 139, 235, 61,  243, 71,  59,  179, 213,
      62,  203, 194, 150, 106, 45,  177, 91,  30,  63,  202, 6,   231, 216, 81,  57,  79,  21,  224, 185, 57,  151, 209,
      206, 46,  218, 15,  2,   21,  30,  104, 242, 87,  98,  239, 167, 112, 235, 20,  152, 236, 216, 153, 196, 48,  211,
      248, 46,  169, 63,  95,  253, 87,  198, 158, 62,  115, 231, 155, 113, 54,  216, 225, 114, 23,  180, 222, 80,  124,
      109, 104, 184, 219, 73,  226, 37,  10,  15,  90,  162, 153, 60,  217, 104, 31,  31,  104, 207, 40,  26,  185, 238,
      216, 112, 160, 245, 51,  177, 6,   94,  75,  59,  251, 107, 182, 120, 213, 98,  90,  136, 161, 230, 81,  223, 216,
      102, 153, 47,  136, 249, 197, 38,  189, 100, 9,   99,  101, 72,  2,   246, 180, 63,  123, 82,  52,  203, 53,  34,
      34,  7,   201, 153, 6,   92,  148, 193, 186, 183, 137, 145, 193, 194, 199, 154, 204, 251, 163, 16,  150, 136, 139,
      206, 160, 137, 207, 42,  249, 117, 35,  58,  190, 34,  252, 26,  254, 90,  235, 166, 94,  143, 104, 79,  53,  169,
      40,  135, 161, 18,  164, 99,  130, 192, 75,  46,  168, 225, 47,  11,  150, 120, 230, 235, 90,  100, 237, 164, 126,
      69,  122, 95,  29,  202, 121, 245, 75,  191, 1,   95,  143, 57,  139, 142, 186, 12,  57,  227, 250, 33,  13,  214,
      243, 138, 80,  141, 23,  127, 154, 69,  108, 72,  51,  86,  77,  247, 128, 115, 225, 204, 211, 136, 11,  66,  9,
      241, 171, 151, 35,  122, 103, 36,  96,  171, 44,  161, 51,  135, 179, 218, 45,  170, 3,   161, 242, 133, 90,  240,
      11,  240, 58,  194, 3,   82,  87,  110, 132, 226, 87,  95,  44,  11,  239, 130, 17,  175, 16,  220, 61,  153, 110,
      160, 191, 36,  15,  28,  137, 21,  180, 109, 168, 26,  57,  251, 111, 26,  36,  132, 84,  92,  250, 167, 222, 162,
      246, 209, 195, 204, 201, 167, 240, 57,  61,  6,   73,  64,  171, 155, 45,  234, 86,  226, 3,   127, 191, 224, 97,
      247, 176, 13,  199, 200, 218, 148, 163, 188, 170, 163, 36,  157, 177, 87,  44,  228, 6,   79,  113, 38,  194, 190,
      105, 189, 241, 146, 75,  144, 87,  208, 85,  171, 55,  2,   151, 28,  85,  99,  157, 51,  37,  127, 197, 95,  99,
      93,  34,  18,  251, 241, 238, 34,  46,  59,  91,  230, 8,   227, 133, 57,  245, 132, 88,  210, 183, 156, 81,  180,
      236, 141, 0,   21,  185, 247, 86,  243, 245, 6,   52,  50,  14,  106, 206, 43,  2,   80,  72,  227, 154, 85,  224,
      84,  228, 11,  162, 232, 160, 212, 56,  108, 17,  174, 165, 233, 208, 238, 213, 68,  154, 142, 179, 105, 188, 75,
      33,  151, 143, 105, 61,  10,  110, 98,  119, 113, 98,  58,  236, 39,  19,  30,  252, 190, 108, 202, 70,  38,  139,
      26,  175, 55,  160, 47,  10,  201, 133, 176, 212, 236, 76,  236, 84,  217, 101, 73,  18,  85,  84,  6,   82,  140,
      175, 230, 28,  231, 89,  175, 236, 68,  156, 199, 125, 65,  95,  0,   120, 157, 169, 132, 208, 53,  103, 201, 247,
      36,  205, 246, 216, 178, 132, 50,  247, 157, 219, 118, 131, 9,   32,  237, 68,  229, 246, 146, 193, 48,  128, 242,
      157, 150, 101, 195, 83,  167, 94,  212, 30,  78,  15,  93,  37,  203, 191, 164, 142, 181, 241, 34,  142, 229, 93,
      40,  183, 203, 213, 93,  235, 38,  154, 175, 68,  248, 237, 194, 138, 205, 64,  222, 196, 187, 234, 122, 96,  204,
      18,  101, 19,  179, 140, 227, 180, 182, 33,  153, 61,  175, 104, 129, 252, 136, 112, 241, 113, 233, 32,  174, 82,
      171, 202, 218, 97,  1,   96,  45,  181, 11,  89,  192, 189, 234, 198, 145, 146, 215, 82,  8,   99,  32,  48,  47,
      138, 156, 107, 136, 123, 196, 160, 113, 197, 166, 119, 254, 250, 119, 182, 134, 169, 79,  136, 24,  86,  0,   29,
      55,  241, 84,  42,  127, 64,  138, 208, 101, 146, 167, 102, 78,  7,   70,  102, 127, 152, 215, 177, 57,  211, 108,
      235, 152, 241, 202, 210, 13,  38,  83,  94,  186, 206, 80,  11,  13,  220, 19,  152, 219, 181, 47,  233, 217, 188,
      49,  95,  225, 17,  59,  163, 156, 3,   77,  98,  204, 50,  239, 133, 94,  60,  155, 50,  134, 94,  106, 70,  186,
      226, 36,  215, 35,  243, 69,  191, 179, 154, 146, 75,  24,  32,  80,  154, 123, 225, 59,  66,  31,  235, 73,  25,
      89,  40,  123, 191, 151, 243, 8,   209, 185, 203, 200, 252, 39,  98,  236, 185, 67,  65,  124, 78,  11,  235, 191,
      227, 150, 45,  20,  111, 23,  159, 16,  249, 89,  10,  3,   102, 127, 24,  53,  245, 57,  33,  74,  10,  177, 145,
      29,  29,  84,  248, 77,  169, 15,  5,   180, 114, 9,   205, 10,  213, 137, 183, 189, 175, 40,  66,  138, 10,  228,
      170, 149, 63,  202, 111, 238, 53,  169, 125, 196, 127, 251, 87,  226, 217, 51,  55,  107, 245, 172, 190, 202, 139,
      217, 202, 28,  91,  110, 47,  136, 111, 138, 231, 54,  84,  198, 248, 176, 11,  40,  84,  166, 97,  45,  15,  223,
      222, 137, 182, 243, 108, 28,  44,  188, 232, 196, 191, 43,  149, 34,  202, 50,  85,  88,  85,  32,  200, 212, 76,
      92,  92,  177, 50,  41,  70,  37,  144, 226, 32,  208, 187, 112, 196, 205, 182, 36,  162, 47,  43,  16,  206, 211,
      227, 182, 228, 178, 92,  239, 83,  57,  132, 244, 87,  247, 213, 37,  111, 16,  21,  71,  2,   139, 13,  183, 108,
      156, 124, 249, 248, 107, 114, 162, 32,  162, 193, 253, 23,  78,  163, 123, 250, 204, 44,  130, 199, 97,  97,  199,
      103, 118, 29,  124, 198, 164, 41,  163, 134, 181, 200, 143, 95,  150, 183, 130, 224, 26,  101, 243, 22,  3,   50,
      193, 233, 105, 251, 90,  208, 171, 100, 79,  238, 107, 65,  254, 126, 159, 25,  23,  174, 198, 71,  55,  163, 71,
      40,  224, 245, 227, 185, 12,  236, 76,  98,  104, 204, 105, 230, 10,  81,  69,  65,  245, 53,  144, 149, 134, 77,
      139, 246, 126, 29,  47,  48,  86,  144, 84,  190, 249, 183, 115, 139, 178, 19,  136, 106, 225, 227, 189, 223, 185,
      79,  199, 246, 251, 9,   114, 158, 219, 204, 198, 97,  171, 69,  39,  64,  2,   250, 76,  196, 47,  238, 9,   240,
      8,   140, 168, 20,  161, 210, 117, 180, 190, 90,  195, 29,  176, 162, 247, 79,  194, 141, 120, 36,  137, 210, 250,
      57,  153, 184, 247, 58,  57,  46,  184, 61,  203, 170, 172, 63,  206, 219, 125, 33,  102, 16,  179, 118, 198, 7,
      115, 215, 226, 226, 1,   123, 91,  59,  231, 238, 48,  98,  248, 121, 17,  89,  29,  228, 129, 182, 197, 165, 95,
      34,  174, 44,  70,  40,  13,  168, 160, 96,  22,  120, 230, 176, 206, 206, 48,  41,  26,  78,  84,  16,  203, 205,
      197, 229, 70,  126, 143, 152, 206, 162, 115, 159, 19,  40,  17,  231, 253, 53,  208, 121, 201, 142, 230, 21,  8,
      212, 105, 205, 221, 151, 28,  122, 160, 98,  162, 209, 25,  25,  39,  170, 113, 92,  160, 217, 196, 71,  53,  32,
      82,  218, 33,  200, 186, 39,  68,  93,  144, 223, 228, 99,  113, 168, 94,  104, 195, 60,  136, 13,  193, 210, 109,
      101, 103, 156, 250, 214, 6,   204, 167, 109, 135, 245, 90,  118, 101, 236, 208, 244, 0,   47,  3,   142, 198, 247,
      15,  9,   97,  206, 117, 65,  98,  121, 129, 243, 251, 3,   24,  252, 207, 89,  180, 21,  139, 88,  237, 227, 128,
      10,  156, 105, 112, 192, 99,  108, 22,  134, 112, 1,   153, 18,  239, 127, 30,  73,  85,  151, 192, 108, 51,  250,
      56,  110, 127, 92,  106, 48,  186, 203, 123, 23,  135, 128, 104, 180, 114, 60,  253, 125, 184, 27,  235, 62,  144,
      216, 76,  54,  44,  212, 4,   2,   234, 208, 103, 212, 67,  169, 51,  104, 215, 198, 47,  133, 80,  11,  67,  209,
      63,  0,   122, 243, 127, 43,  39,  142, 227, 108, 233, 61,  233, 69,  151, 39,  162, 1,   145, 61,  105, 21,  221,
      53,  247, 57,  3,   108, 40,  21,  54,  22,  95,  59,  5,   225, 82,  105, 35,  67,  154, 102, 254, 23,  208, 141,
      109, 174, 218, 169, 241, 71,  77,  160, 12,  65,  145, 118, 96,  233, 165, 37,  51,  1,   106, 238, 244, 226, 206,
      72,  98,  116, 151, 145, 202, 199, 240, 204, 177, 79,  164, 142, 100, 61,  145, 34,  80,  21,  180, 156, 119, 35,
      242, 81,  97,  18,  239, 27,  3,   249, 158, 70,  216, 175, 248, 93,  79,  175, 241, 207, 100, 251, 239, 178, 12,
      80,  77,  227, 178, 5,   111, 19,  57,  10,  30,  63,  30,  69,  203, 179, 170, 236, 168, 114, 26,  34,  189, 9,
      177, 5,   118, 6,   94,  176, 180, 227, 36,  32,  194, 232, 17,  167, 132, 100, 3,   43,  156, 227, 242, 137, 63,
      72,  3,   4,   173, 230, 166, 174, 197, 88,  34,  192, 130, 246, 98,  56,  188, 145, 189, 188, 145, 145, 99,  192,
      64,  68,  13,  218, 134, 53,  212, 137, 61,  7,   29,  97,  171, 26,  170, 28,  92,  51,  115, 126, 35,  189, 147,
      75,  100, 149, 11,  47,  47,  61,  182, 103, 145, 123, 78,  31,  102, 22,  139, 181, 253, 31,  138, 44,  120, 71,
      41,  79,  248, 227, 219, 157, 132, 13,  112, 152, 238, 43,  231, 118, 180, 254, 167, 75,  20,  44,  219, 191, 48,
      196, 112, 53,  35,  207, 120, 51,  66,  217, 195, 148, 177, 173, 172, 157, 225, 23,  213, 166, 131, 209, 61,  96,
      197, 164, 87,  78,  212, 40,  152, 89,  7,   78,  239, 130, 177, 124, 100, 14,  239, 53,  237, 121, 132, 207, 205,
      226, 46,  246, 139, 44,  241, 249, 142, 94,  52,  138, 90,  197, 204, 79,  240, 8,   62,  78,  97,  187, 88,  50,
      235, 149, 35,  197, 149, 87,  94,  103, 56,  124, 136, 4,   187, 33,  163, 240, 3,   73,  142, 53,  60,  227, 82,
      144, 41,  158, 180, 24,  207, 97,  155, 185, 235, 97,  209, 105, 79,  125, 2,   6,   86,  6,   8,   168, 247, 175,
      66,  53,  197, 231, 130, 168, 79,  159, 117, 69,  80,  82,  221, 56,  175, 137, 6,   73,  115, 122, 211, 160, 133,
      41,  24,  207, 38,  191, 2,   150, 129, 156, 121, 227, 17,  166, 66,  223, 29,  142, 83,  24,  106, 206, 136, 158,
      232, 248, 237, 120, 21,  48,  58,  126, 28,  162, 137, 74,  205, 207, 159, 112, 46,  3,   45,  252, 153, 31,  177,
      129, 186, 157, 66,  18,  67,  185, 165, 0,   195, 107, 7,   237, 202, 238, 238, 5,   72,  238, 42,  181, 145, 187,
      187, 227, 67,  163, 199, 2,   94,  155, 95,  67,  239, 160, 135, 139, 10,  136, 222, 130, 182, 139, 51,  164, 129,
      15,  224, 43,  113, 56,  198, 63,  205, 170, 181, 68,  90,  215, 175, 37,  220, 185, 94,  99,  91,  211, 181, 247,
      5,   231, 131, 64,  28,  196, 65,  136, 214, 193, 213, 148, 18,  44,  76,  242, 173, 174, 150, 169, 221, 16,  65,
      94,  28,  225, 201, 118, 104, 115, 213, 112, 46,  184, 35,  186, 24,  198, 169, 91,  102, 65,  210, 120, 26,  148,
      142, 58,  138, 61,  211, 190, 42,  14,  37,  41,  21,  29,  148, 24,  59,  192, 21,  196, 33,  110, 116, 11,  109,
      196, 141, 83,  131, 48,  183, 200, 31,  70,  228, 158, 79,  9,   222, 239, 125, 248, 143, 29,  183, 235, 219, 237,
      252, 21,  84,  14,  105, 199, 84,  217, 68,  141, 204, 251, 5,   0,   70,  57,  200, 181, 98,  162, 185, 130, 151,
      192, 58,  116, 71,  149, 82,  76,  232, 66,  32,  127, 6,   62,  132, 132, 131, 240, 39,  113, 119, 87,  38,  238,
      79,  17,  208, 126, 174, 79,  103, 253, 189, 236, 178, 232, 68,  123, 226, 2,   130, 154, 191, 72,  90,  236, 186,
      235, 149, 46,  155, 94,  203, 96,  129, 61,  8,   57,  123, 37,  206, 240, 72,  59,  64,  43,  87,  165, 216, 176,
      161, 226, 232, 48,  101, 146, 155, 146, 227, 54,  209, 210, 243, 136, 233, 124, 209, 189, 35,  34,  163, 9,   205,
      68,  250, 166, 172, 91,  143, 247, 127, 97,  45,  122, 119, 97,  102, 77,  192, 4,   133, 192, 116, 128, 14,  90,
      22,  161, 42,  106, 20,  53,  224, 201, 84,  178, 107, 49,  54,  33,  44,  108, 101, 111, 124, 84,  110, 218, 54,
      211, 223, 243, 70,  95,  249, 186, 236, 122, 60,  151, 159, 122, 178, 3,   68,  132, 61,  136, 250, 41,  150, 82,
      99,  87,  164, 113, 36,  143, 3,   28,  123, 147, 80,  15,  218, 90,  68,  235, 28,  69,  19,  223, 68,  154, 78,
      52,  95,  164, 242, 60,  65,  28,  63,  14,  220, 194, 155, 186, 221, 97,  248, 9,   244, 251, 238, 137, 204, 137,
      254, 127, 97,  65,  55,  135, 24,  32,  59,  115, 12,  20,  57,  232, 166, 26,  32,  237, 56,  194, 195, 57,  104,
      242, 153, 240, 212, 226, 37,  170, 93,  125, 74,  28,  106, 104, 125, 157, 250, 229, 157, 47,  219, 245, 122, 54,
      2,   238, 75,  53,  37,  160, 60,  47,  192, 97,  251, 4,   148, 142, 252, 108, 137, 43,  8,   227, 27,  122, 51,
      99,  39,  156, 195, 233, 232, 219, 5,   127, 104, 198, 84,  214, 69,  241, 81,  146, 150, 254, 137, 123, 193, 105,
      131, 81,  16,  106, 88,  10,  51,  143, 249, 37,  233, 165, 28,  133, 208, 100, 207, 212, 228, 236, 118, 103, 246,
      156, 220, 226, 199, 17,  163, 190, 122, 20,  71,  29,  105, 155, 164, 143, 129, 190, 49,  106, 249, 87,  245, 107,
      24,  140, 21,  248, 49,  57,  24,  86,  203, 83,  195, 36,  202, 78,  2,   9,   129, 244, 206, 5,   83,  151, 129,
      179, 45,  103, 129, 164, 10,  75,  52,  127, 224, 168, 211, 218, 185, 90,  167, 115, 49,  74,  129, 184, 243, 30,
      163, 223, 191, 199, 97,  254, 243, 180, 27,  91,  16,  26,  24,  172, 21,  98,  53,  0,   190, 180, 236, 33,  187,
      56,  244, 157, 90,  196, 173, 7,   25,  201, 0,   133, 214, 51,  84,  79,  154, 170, 245, 166, 114, 82,  235, 229,
      31,  3,   127, 216, 134, 7,   197, 86,  216, 211, 198, 226, 62,  153, 32,  216, 243, 217, 95,  185, 179, 51,  80,
      53,  180, 125, 161, 45,  169, 219, 219, 160, 253, 68,  199, 19,  34,  152, 112, 136, 252, 10,  133, 105, 155, 99,
      244, 125, 216, 67,  134, 241, 106, 241, 47,  238, 186, 214, 89,  125, 147, 142, 168, 134, 32,  137, 44,  143, 130,
      93,  120, 21,  207, 120, 139, 100, 94,  133, 248, 85,  150, 24,  183, 194, 164, 149, 46,  2,   206, 203, 60,  160,
      232, 48,  112, 200, 215, 108, 201, 129, 33,  122, 175, 161, 155, 187, 121, 136, 229, 149, 15,  100, 196, 238, 65,
      157, 155, 148, 247, 31,  208, 107, 108, 9,   107, 107, 184, 248, 152, 90,  238, 233, 135, 6,   103, 243, 164, 50,
      3,   45,  193, 36,  244, 242, 90,  63,  152, 127, 54,  212, 71,  16,  247, 185, 196, 52,  145, 80,  162, 171, 197,
      213, 40,  31,  252, 76,  246, 124, 63,  163, 107, 52,  115, 134, 124, 23,  107, 142, 133, 216, 67,  195, 103, 246,
      81,  209, 150, 25,  246, 209, 64,  211, 237, 251, 96,  150, 120, 77,  232, 112, 93,  106, 184, 136, 230, 3,   74,
      4,   78,  197, 209, 54,  246, 113, 186, 140, 71,  130, 253, 122, 55,  233, 154, 243, 41,  211, 191, 170, 111, 12,
      179, 188, 148, 104, 233, 191, 212, 94,  66,  187, 33,  160, 232, 75,  124, 121, 125, 14,  67,  126, 129, 145, 64,
      241, 5,   20,  128, 85,  18,  201, 41,  87,  242, 228, 89,  212, 125, 55,  224, 96,  150, 79,  86,  242, 46,  69,
      26,  145, 162, 253, 145, 101, 163, 138, 248, 138, 71,  62,  178, 45,  14,  14,  53,  110, 42,  15,  50,  111, 6,
      171, 40,  73,  215, 207, 119, 19,  208, 145, 59,  236, 6,   164, 149, 128, 214, 244, 45,  72,  147, 145, 234, 138,
      249, 144, 22,  106, 42,  162, 181, 252, 8,   160, 214, 17,  235, 180, 123, 213, 162, 138, 245, 131, 110, 203, 36,
      38,  205, 1,   157, 56,  245, 83,  41,  116, 4,   158, 15,  89,  35,  207, 87,  203, 120, 162, 48,  138, 143, 91,
      253, 58,  177, 10,  18,  232, 142, 13,  152, 106, 67,  189, 170, 21,  178, 185, 82,  239, 178, 43,  116, 101, 149,
      8,   140, 32,  25,  202, 147, 122, 12,  122, 132, 113, 65,  253, 251, 194, 9,   184, 130, 15,  42,  108, 60,  85,
      127, 5,   169, 58,  73,  58,  231, 27,  19,  157, 239, 239, 121, 116, 33,  39,  170, 225, 82,  176, 17,  84,  66,
      245, 90,  196, 91,  139, 79,  67,  52,  27,  55,  25,  71,  131, 77,  68,  134, 71,  236, 122, 120, 139, 150, 219,
      120, 139, 172, 152, 4,   235, 154, 109, 39,  81,  128, 247, 197, 28,  24,  155, 173, 223, 179, 59,  164, 110, 101,
      93,  57,  207, 119, 185, 57,  5,   91,  104, 84,  248, 185, 227, 243, 50,  214, 146, 115, 8,   62,  131, 105, 84,
      55,  237, 13,  229, 247, 230, 44,  125, 249, 17,  29,  140, 212, 241, 122, 46,  220, 185, 241, 42,  104, 218, 158,
      234, 230, 244, 141, 161, 132, 179, 89,  102, 4,   183, 32,  36,  91,  244, 105, 36,  185, 154, 167, 165, 11,  24,
      123, 30,  136, 242, 151, 57,  77,  86,  1,   203, 215, 187, 58,  224, 100, 245, 189, 155, 50,  212, 24,  85,  217,
      110, 69,  251, 48,  124, 194, 106, 158, 169, 245, 119, 71,  42,  120, 78,  128, 184, 116, 230, 34,  22,  21,  166,
      34,  229, 212, 167, 159, 12,  173, 82,  244, 156, 135, 239, 51,  78,  194, 169, 219, 32,  9,   75,  234, 147, 252,
      51,  214, 75,  182, 193, 45,  150, 30,  27,  20,  51,  66,  20,  212, 204, 232, 174, 87,  87,  154, 48,  199, 26,
      23,  205, 62,  86,  3,   230, 124, 215, 150, 244, 215, 150, 243, 178, 71,  216, 156, 20,  233, 230, 138, 87,  144,
      20,  83,  220, 104, 198, 162, 164, 219, 179, 245, 79,  37,  158, 147, 154, 110, 228, 122, 65,  227, 187, 114, 176,
      21,  36,  77,  232, 137, 71,  24,  108, 70,  193, 54,  113, 121, 245, 157, 64,  202, 41,  216, 151, 9,   96,  198,
      172, 14,  204, 193, 83,  108, 86,  154, 225, 215, 114, 169, 117, 20,  73,  82,  18,  37,  225, 102, 214, 248, 222,
      91,  177, 205, 253, 58,  244, 212, 78,  182, 245, 168, 81,  253, 187, 71,  6,   132, 135, 179, 208, 108, 231, 126,
      18,  195, 65,  211, 134, 16,  104, 196, 109, 173, 86,  150, 94,  116, 76,  180, 53,  75,  160, 223, 103, 250, 216,
      51,  31,  218, 153, 163, 184, 223, 252, 140, 98,  101, 187, 10,  99,  229, 82,  80,  253, 226, 79,  46,  211, 140,
      233, 239, 68,  204, 24,  22,  38,  180, 230, 158, 9,   37,  5,   116, 185, 139, 141, 17,  159, 216, 1,   144, 96,
      5,   131, 201, 37,  223, 108, 107, 107, 238, 8,   83,  4,   230, 244, 86,  219, 99,  112, 243, 82,  225, 209, 91,
      10,  224, 52,  90,  246, 21,  178, 97,  69,  72,  138, 227, 172, 109, 238, 161, 238, 79,  6,   35,  143, 96,  37,
      134, 86,  89,  193, 21,  147, 198, 230, 176, 30,  172, 245, 0,   240, 45,  201, 9,   21,  70,  60,  40,  254, 64,
      206, 96,  159, 18,  154, 184, 107, 51,  238, 57,  170, 92,  165, 84,  36,  239, 27,  47,  31,  140, 159, 53,  105,
      213, 34,  43,  5,   124, 251, 110, 60,  231, 106, 99,  7,   220, 173, 20,  194, 28,  12,  100, 196, 194, 157, 126,
      87,  47,  219, 220, 134, 10,  200, 189, 216, 174, 50,  106, 61,  247, 178, 81,  127, 51,  66,  52,  58,  189, 130,
      212, 188, 61,  71,  102, 127, 94,  157, 56,  161, 143, 188, 109, 119, 246, 132, 131, 99,  60,  88,  190, 202, 237,
      203, 253, 167, 16,  54,  224, 187, 239, 115, 94,  32,  166, 152, 8,   70,  10,  55,  92,  234, 186, 173, 3,   165,
      199, 175, 183, 239, 51,  221, 232, 161, 24,  94,  158, 130, 250, 112, 207, 75,  128, 99,  157, 27,  181, 57,  0,
      157, 31,  97,  83,  128, 173, 40,  33,  190, 168, 138, 210, 87,  51,  219, 218, 64,  72,  241, 126, 153, 225, 69,
      119, 151, 214, 127, 252, 63,  161, 100, 83,  192, 218, 176, 136, 87,  69,  223, 109, 104, 38,  79,  37,  245, 172,
      88,  196, 78,  83,  73,  153, 93,  203, 171, 169, 113, 112, 137, 185, 137, 74,  244, 37,  50,  20,  229, 136, 50,
      10,  54,  73,  78,  167, 247, 127, 160, 9,   74,  248, 159, 194, 215, 170, 224, 175, 234, 84,  23,  55,  245, 223,
      63,  126, 179, 0,   139, 144, 148, 7,   210, 93,  246, 172, 243, 20,  158, 79,  51,  38,  167, 95,  234, 223, 37,
      72,  58,  60,  102, 92,  142, 123, 210, 84,  178, 138, 86,  18,  27,  0,   216, 68,  213, 69,  58,  146, 201, 120,
      7,   62,  73,  87,  189, 20,  134, 199, 86,  184, 44,  4,   65,  164, 242, 118, 250, 213, 79,  107, 98,  173, 76,
      110, 157, 68,  159, 167, 71,  87,  11,  88,  144, 44,  157, 92,  133, 182, 182, 173, 199, 187, 228, 216, 239, 242,
      122, 222, 82,  138, 58,  31,  249, 125, 149, 240, 78,  79,  32,  195, 167, 17,  181, 113, 205, 247, 115, 250, 22,
      47,  112, 141, 37,  187, 219, 134, 79,  204, 119, 68,  187, 9,   84,  102, 95,  227, 9,   98,  230, 36,  199, 112,
      42,  112, 68,  236, 0,   191, 62,  58,  211, 91,  1,   244, 200, 223, 15,  235, 247, 200, 164, 65,  63,  29,  220,
      201, 1,   85,  193, 113, 150, 139, 99,  148, 45,  244, 34,  215, 102, 180, 207, 103, 182, 29,  124, 79,  1,   93,
      142, 65,  228, 144, 133, 62,  192, 233, 252, 170, 159, 19,  251, 50,  172, 19,  73,  46,  71,  74,  190, 69,  16,
      172, 183, 9,   86,  31,  189, 122, 148, 155, 140, 137, 174, 52,  115, 162, 122, 204, 108, 52,  20,  108, 198, 238,
      46,  229, 130, 6,   95,  189, 83,  76,  150, 189, 121, 164, 144, 141, 149, 122, 28,  185, 48,  104, 213, 192, 57,
      188, 155, 235, 215, 194, 26,  207, 233, 90,  39,  117, 8,   75,  132, 0,   195, 160, 117, 173, 214, 140, 127, 156,
      69,  19,  46,  186, 199, 144, 149, 99,  129, 229, 68,  237, 64,  42,  248, 106, 225, 169, 16,  36,  123, 31,  79,
      162, 81,  86,  117, 67,  150, 67,  191, 242, 117, 139, 104, 142, 90,  193, 180, 69,  115, 69,  236, 34,  209, 244,
      246, 58,  246, 74,  48,  116, 116, 163, 53,  157, 139, 198, 251, 166, 73,  36,  127, 226, 40,  97,  250, 40,  180,
      79,  174, 145, 228, 105, 234, 221, 209, 235, 241, 7,   184, 248, 196, 38,  33,  54,  132, 174, 207, 60,  14,  61,
      177, 56,  166, 252, 151, 99,  144, 127, 212, 184, 53,  19,  237, 78,  81,  83,  34,  220, 151, 205, 224, 160, 138,
      206, 233, 4,   247, 136, 130, 151, 219, 46,  66,  249, 213, 187, 216, 92,  35,  141, 53,  151, 135, 53,  82,  186,
      121, 98,  23,  7,   168, 204, 123, 107, 219, 138, 232, 137, 110, 171, 137, 86,  79,  167, 27,  92,  36,  106, 239,
      74,  55,  97,  223, 75,  245, 134, 223, 35,  213, 233, 71,  132, 124, 151, 197, 162, 31,  72,  143, 5,   26,  234,
      15,  106, 133, 32,  9,   212, 40,  189, 51,  186, 60,  46,  4,   84,  212, 32,  161, 191, 59,  10,  52,  29,  2,
      208, 56,  203, 74,  80,  38,  195, 13,  190, 237, 207, 204, 211, 192, 190, 156, 237, 222, 159, 168, 99,  175, 12,
      46,  221, 63,  183, 222, 216, 12,  43,  154, 154, 71,  222, 111, 12,  58,  103, 44,  110, 146, 139, 87,  55,  24,
      43,  9,   188, 143, 92,  58,  250, 21,  187, 45,  109, 147, 196, 178, 57,  65,  51,  133, 176, 177, 173, 137, 13,
      162, 46,  110, 77,  190, 231, 126, 218, 204, 71,  198, 190, 124, 92,  176, 197, 158, 224, 239, 236, 250, 178, 171,
      235, 179, 245, 186, 120, 89,  86,  195, 41,  227, 153, 35,  133, 199, 85,  210, 194, 26,  137, 251, 48,  187, 106,
      124, 42,  33,  145, 17,  24,  163, 100, 193, 41,  172, 21,  114, 16,  132, 3,   176, 86,  216, 34,  140, 181, 175,
      183, 162, 153, 52,  196, 154, 54,  68,  134, 51,  195, 175, 233, 247, 44,  108, 60,  180, 127, 219, 31,  227, 210,
      84,  201, 27,  82,  247, 181, 149, 76,  194, 21,  19,  199, 115, 80,  52,  246, 61,  75,  212, 89,  243, 126, 0,
      68,  109, 98,  14,  59,  79,  147, 250, 218, 225, 185, 90,  60,  116, 154, 170, 241, 54,  216, 137, 233, 89,  186,
      98,  95,  30,  63,  12,  113, 74,  239, 65,  22,  18,  119, 82,  27,  112, 31,  237, 105, 219, 194, 64,  109, 144,
      36,  128, 74,  239, 115, 148, 71,  215, 64,  144, 151, 190, 119, 93,  41,  194, 180, 1,   140, 223, 153, 138, 131,
      147, 73,  33,  51,  211, 192, 234, 9,   78,  143, 62,  237, 160, 18,  125, 152, 137, 192, 121, 23,  113, 242, 70,
      199, 20,  77,  118, 197, 207, 11,  99,  121, 159, 37,  78,  65,  201, 102, 86,  22,  87,  11,  141, 214, 150, 233,
      43,  52,  88,  16,  80,  167, 73,  238, 38,  157, 143, 118, 24,  40,  213, 163, 116, 160, 223, 68,  202, 51,  230,
      69,  87,  182, 72,  85,  44,  99,  198, 165, 165, 25,  237, 137, 115, 152, 180, 7,   88,  241, 121, 182, 112, 93,
      29,  252, 31,  143, 24,  227, 241, 150, 207, 8,   253, 21,  246, 181, 158, 40,  115, 97,  51,  209, 169, 86,  108,
      10,  71,  106, 118, 104, 237, 245, 118, 58,  8,   174, 223, 120, 92,  170, 149, 10,  162, 31,  55,  189, 174, 238,
      37,  53,  223, 183, 131, 20,  181, 96,  238, 201, 167, 88,  70,  150, 190, 5,   174, 210, 160, 40,  168, 111, 132,
      33,  184, 193, 244, 47,  133, 78,  169, 170, 144, 177, 138, 228, 87,  139, 173, 96,  156, 4,   24,  54,  170, 100,
      70,  142, 4,   247, 25,  191, 232, 204, 203, 55,  130, 166, 214, 182, 194, 217, 188, 106, 217, 204, 40,  247, 225,
      250, 98,  142, 200, 100, 33,  123, 231, 221, 214, 190, 100, 229, 21,  139, 206, 250, 254, 77,  233, 112, 51,  174,
      183, 89,  54,  36,  231, 120, 230, 200, 19,  67,  206, 248, 81,  220, 54,  118, 56,  200, 75,  239, 150, 42,  41,
      183, 85,  210, 202, 83,  242, 198, 221, 50,  41,  115, 161, 6,   29,  155, 231, 124, 122, 60,  175, 198, 228, 131,
      128, 75,  143, 241, 83,  84,  70,  16,  2,   148, 183, 41,  71,  207, 216, 149, 245, 85,  25,  166, 18,  140, 53,
      251, 123, 199, 131, 88,  175, 29,  222, 236, 47,  230, 32,  203, 198, 59,  196, 222, 45,  31,  145, 94,  74,  92,
      154, 128, 2,   114, 166, 129, 28,  11,  110, 71,  68,  3,   127, 105, 155, 75,  117, 144, 195, 123, 65,  245, 30,
      17,  145, 192, 208, 97,  54,  159, 196, 83,  4,   64,  71,  30,  72,  178, 64,  100, 8,   156, 206, 5,   186, 13,
      59,  240, 25,  37,  22,  150, 27,  240, 154, 13,  110, 212, 132, 41,  108, 76,  175, 127, 239, 117, 11,  66,  16,
      15,  170, 138, 248, 227, 83,  246, 148, 237, 251, 17,  57,  198, 204, 7,   46,  117, 89,  23,  33,  29,  221, 209,
      33,  203, 207, 199, 24,  148, 49,  67,  175, 108, 182, 42,  233, 124, 249, 63,  7,   167, 48,  135, 208, 157, 188,
      189, 35,  204, 18,  202, 199, 131, 168, 94,  143, 126, 76,  224, 252, 205, 238, 133, 145, 34,  189, 159, 195, 233,
      93,  139, 230, 82,  37,  52,  242, 191, 119, 84,  2,   10,  134, 108, 168, 88,  117, 148, 38,  51,  19,  153, 120,
      234, 56,  90,  167, 211, 47,  252, 61,  242, 32,  66,  17,  5,   69,  176, 24,  66,  40,  184, 181, 73,  14,  36,
      189, 9,   238, 181, 166, 139, 48,  216, 148, 223, 227, 109, 148, 152, 16,  19,  251, 117, 48,  178, 82,  40,  104,
      190, 214, 15,  226, 206, 144, 35,  123, 22,  147, 9,   14,  131, 213, 197, 219, 26,  70,  197, 157, 241, 221, 20,
      193, 250, 173, 84,  79,  150, 138, 206, 241, 102, 173, 226, 31,  239, 237, 248, 88,  202, 78,  163, 178, 19,  66,
      231, 28,  245, 31,  212, 39,  129, 201, 167, 71,  195, 22,  250, 222, 231, 67,  116, 170, 155, 148, 137, 161, 101,
      25,  223, 13,  105, 94,  111, 188, 61,  18,  69,  171, 6,   26,  69,  192, 6,   51,  25,  94,  48,  131, 128, 63,
      62,  243, 157, 13,  43,  61,  72,  39,  87,  102, 160, 174, 140, 211, 212, 49,  169, 211, 163, 232, 82,  87,  82,
      17,  73,  59,  69,  58,  63,  165, 178, 9,   186, 191, 115, 134, 141, 104, 150, 197, 202, 99,  181, 120, 154, 95,
      224, 127, 32,  206, 248, 173, 202, 232, 191, 106, 211, 90,  26,  182, 33,  219, 12,  226, 124, 197, 142, 174, 159,
      37,  68,  63,  109, 250, 153, 235, 227, 198, 229, 22,  46,  238, 55,  47,  117, 79,  194, 116, 103, 27,  248, 24,
      147, 95,  161, 146, 228, 228, 209, 40,  33,  37,  6,   88,  59,  89,  254, 191, 211, 91,  224, 99,  210, 168, 169,
      134, 100, 195, 118, 236, 147, 181, 148, 54,  1,   138, 43,  163, 145, 250, 225, 99,  142, 98,  57,  233, 9,   15,
      200, 222, 243, 241, 214, 104, 241, 176, 136, 246, 129, 39,  251, 65,  165, 7,   141, 139, 146, 110, 119, 68,  110,
      49,  73,  30,  126, 48,  252, 134, 219, 214, 218, 133, 138, 26,  90,  152, 23,  112, 175, 44,  22,  32,  180, 135,
      207, 213, 76,  36,  185, 173, 225, 12,  121, 199, 181, 251, 148, 54,  55,  15,  171, 162, 145, 114, 127, 99,  168,
      182, 72,  190, 27,  56,  250, 193, 174, 31,  229, 89,  83,  133, 179, 13,  179, 242, 91,  19,  161, 163, 54,  177,
      111, 70,  144, 67,  178, 123, 216, 232, 86,  52,  218, 1,   224, 30,  182, 69,  225, 0,   71,  45,  232, 144, 48,
      245, 43,  1,   151, 82,  45,  168, 9,   206, 116, 84,  121, 250, 151, 119, 219, 101, 109, 5,   187, 133, 154, 215,
      208, 235, 114, 100, 66,  133, 127, 66,  133, 7,   1,   152, 23,  65,  249, 116, 219, 225, 150, 163, 30,  162, 115,
      60,  164, 96,  26,  9,   253, 13,  100, 210, 147, 159, 78,  179, 168, 110, 197, 99,  72,  30,  144, 68,  196, 222,
      32,  208, 217, 248, 179, 186, 3,   113, 79,  199, 138, 190, 167, 45,  99,  92,  6,   84,  123, 118, 238, 86,  41,
      59,  14,  226, 82,  146, 83,  198, 155, 175, 20,  119, 80,  251, 210, 101, 231, 253, 245, 208, 77,  195, 155, 138,
      72,  86,  228, 120, 134, 227, 33,  221, 181, 3,   78,  71,  26,  131, 116, 76,  62,  198, 183, 87,  69,  62,  26,
      44,  66,  150, 248, 191, 111, 42,  243, 26,  58,  108, 54,  138, 200, 116, 82,  12,  2,   8,   72,  17,  87,  254,
      205, 66,  216, 203, 47,  103, 202, 21,  181, 139, 169, 78,  1,   24,  210, 88,  215, 197, 43,  174, 174, 225, 128,
      121, 1,   218, 75,  243, 37,  230, 47,  246, 186, 176, 221, 184, 91,  34,  98,  1,   75,  207, 78,  135, 166, 48,
      116, 35,  71,  173, 182, 246, 76,  148, 208, 1,   133, 131, 126, 175, 110, 243, 133, 190, 123, 125, 139, 101, 253,
      96,  68,  94,  220, 60,  222, 4,   65,  122, 118, 0,   147, 87,  63,  51,  95,  109, 240, 97,  184, 65,  218, 136,
      45,  233, 68,  78,  49,  135, 233, 165, 151, 216, 26,  240, 115, 250, 122, 223, 122, 118, 9,   40,  130, 33,  169,
      248, 13,  223, 169, 197, 46,  215, 84,  50,  205, 106, 150, 155, 222, 201, 12,  203, 63,  254, 201, 168, 35,  217,
      41,  11,  108, 51,  252, 86,  66,  244, 26,  251, 130, 188, 5,   136, 154, 232, 22,  67,  147, 225, 54,  26,  221,
      38,  158, 139, 236, 91,  141, 243, 212, 113, 227, 116, 121, 19,  68,  203, 106, 147, 29,  216, 242, 18,  55,  18,
      252, 139, 112, 123, 182, 251, 216, 144, 178, 179, 24,  104, 41,  254, 109, 234, 74,  189, 202, 247, 58,  146, 176,
      125, 176, 167, 211, 32,  171, 93,  220, 204, 17,  27,  36,  10,  16,  38,  207, 144, 18,  219, 1,   247, 92,  54,
      91,  44,  137, 49,  31,  39,  54,  187, 95,  98,  2,   20,  130, 253, 149, 74,  227, 64,  4,   127, 167, 16,  242,
      183, 126, 4,   55,  250, 62,  3,   217, 214, 77,  104, 59,  170, 87,  106, 199, 226, 99,  181, 84,  229, 145, 213,
      220, 26,  89,  187, 216, 210, 206, 6,   239, 54,  153, 15,  133, 140, 222, 155, 36,  191, 76,  167, 2,   124, 57,
      192, 120, 157, 27,  128, 252, 88,  7,   117, 222, 28,  221, 95,  25,  43,  79,  43,  191, 100, 243, 53,  96,  1,
      195, 10,  50,  12,  35,  84,  51,  43,  50,  217, 183, 57,  149, 26,  77,  210, 2,   79,  215, 53,  53,  26,  186,
      47,  151, 180, 23,  180, 22,  20,  54,  101, 159, 150, 7,   31,  18,  212, 214, 123, 56,  117, 145, 207, 82,  40,
      55,  111, 171, 106, 97,  13,  22,  4,   193, 221, 85,  52,  58,  31,  139, 229, 166, 8,   67,  107, 188, 55,  248,
      192, 234, 244, 147, 143, 81,  141, 146, 174, 239, 138, 22,  30,  118, 96,  213, 36,  66,  10,  207, 166, 91,  41,
      201, 195, 41,  144, 101, 81,  167, 100, 25,  15,  225, 251, 118, 167, 175, 233, 196, 155, 130, 87,  48,  176, 198,
      97,  143, 71,  192, 227, 25,  202, 202, 174, 197, 178, 217, 1,   248, 180, 191, 46,  186, 41,  58,  25,  60,  114,
      119, 211, 138, 14,  225, 197, 169, 26,  133, 115, 22,  206, 222, 27,  101, 148, 188, 104, 193, 16,  76,  149, 41,
      185, 87,  37,  83,  121, 252, 5,   29,  75,  67,  181, 50,  133, 50,  191, 41,  36,  16,  0,   55,  79,  240, 53,
      235, 237, 94,  107, 85,  123, 214, 67,  227, 154, 73,  125, 165, 226, 182, 216, 110, 179, 116, 63,  205, 241, 176,
      17,  204, 98,  170, 57,  147, 54,  24,  60,  253, 190, 213, 24,  7,   150, 212, 227, 230, 186, 173, 223, 75,  203,
      59,  223, 135, 238, 22,  133, 66,  157, 239, 206, 64,  51,  174, 230, 17,  54,  123, 133, 179, 118, 249, 57,  229,
      37,  235, 230, 87,  231, 137, 223, 226, 212, 225, 186, 149, 123, 222, 219, 195, 140, 134, 27,  141, 43,  94,  140,
      164, 56,  87,  62,  215, 159, 145, 20,  119, 254, 212, 147, 210, 52,  89,  126, 239, 51,  174, 240, 134, 160, 212,
      179, 79,  154, 247, 70,  144, 199, 177, 226, 54,  185, 16,  243, 40,  182, 134, 39,  168, 122, 38,  228, 22,  181,
      34,  214, 252, 114, 13,  247, 85,  128, 252, 171, 183, 127, 26,  238, 125, 140, 230, 147, 220, 119, 112, 243, 232,
      107, 186, 148, 81,  79,  34,  106, 4,   95,  121, 18,  135, 33,  23,  1,   86,  69,  106, 139, 236, 222, 247, 54,
      22,  135, 159, 73,  83,  55,  75,  76,  117, 54,  31,  106, 170, 234, 4,   89,  222, 70,  11,  246, 208, 131, 100,
      68,  48,  111, 12,  66,  224, 118, 196, 215, 230, 193, 2,   246, 12,  52,  154, 57,  146, 113, 122, 112, 58,  152,
      228, 189, 209, 95,  19,  0,   35,  17,  115,
  };

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data),    FORMAT_FRACTAL_Z, FORMAT_NCHW, std::vector<int64_t>({18, 2, 16, 32}),
      std::vector<int64_t>({32, 64, 3, 3}), DT_INT8};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(ret) / sizeof(ret[0]));
  for (int i = 0; i < sizeof(ret) / sizeof(ret[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<uint8_t *>(result.data.get()))[i], ret[i]);
  }
}

TEST_F(UtestFormatTransferFraczNchw, uint8_2c_2n_pad) {
  uint8_t ret[30 * 40 * 3 * 3] = {
      240, 195, 239, 13,  48,  181, 136, 177, 217, 211, 45,  33,  95,  18,  221, 5,   112, 37,  76,  243, 215, 65,  119,
      101, 119, 96,  145, 44,  108, 73,  72,  170, 49,  80,  168, 160, 3,   74,  18,  163, 51,  140, 154, 65,  185, 198,
      253, 26,  28,  246, 116, 4,   84,  189, 126, 186, 154, 217, 252, 13,  241, 5,   159, 60,  6,   53,  68,  48,  221,
      70,  126, 26,  124, 115, 220, 31,  249, 248, 159, 102, 106, 45,  25,  228, 90,  136, 121, 227, 156, 138, 242, 199,
      109, 114, 156, 205, 2,   175, 109, 246, 9,   64,  236, 157, 207, 238, 133, 0,   105, 146, 3,   33,  210, 222, 104,
      130, 151, 69,  30,  247, 210, 220, 163, 242, 245, 61,  129, 225, 135, 76,  33,  213, 90,  182, 151, 230, 40,  99,
      143, 131, 141, 84,  135, 232, 218, 249, 72,  181, 244, 133, 111, 221, 89,  122, 192, 211, 176, 223, 135, 231, 171,
      181, 49,  134, 90,  83,  212, 129, 89,  41,  85,  224, 87,  45,  45,  54,  132, 60,  132, 52,  200, 202, 197, 61,
      220, 26,  9,   114, 217, 134, 201, 240, 217, 48,  213, 57,  212, 92,  254, 217, 229, 151, 151, 249, 37,  178, 61,
      171, 196, 252, 37,  139, 69,  67,  126, 143, 166, 136, 138, 43,  186, 92,  199, 162, 47,  50,  151, 40,  240, 35,
      2,   38,  0,   204, 153, 244, 230, 105, 168, 124, 232, 187, 108, 189, 19,  166, 149, 201, 99,  235, 71,  147, 241,
      214, 70,  12,  232, 86,  141, 39,  140, 155, 179, 86,  161, 67,  139, 213, 223, 212, 136, 117, 164, 28,  118, 17,
      56,  211, 240, 221, 22,  11,  254, 170, 144, 16,  238, 94,  19,  246, 179, 133, 205, 16,  65,  91,  72,  146, 132,
      22,  88,  228, 212, 26,  177, 129, 65,  223, 46,  96,  152, 92,  246, 207, 126, 18,  201, 192, 9,   6,   209, 117,
      200, 161, 222, 82,  216, 213, 33,  18,  172, 249, 167, 192, 47,  254, 50,  162, 34,  129, 241, 25,  28,  254, 241,
      100, 17,  240, 172, 97,  97,  129, 102, 78,  133, 67,  191, 99,  160, 209, 34,  10,  209, 242, 109, 135, 31,  243,
      244, 3,   209, 71,  186, 200, 245, 123, 157, 152, 160, 89,  30,  252, 101, 118, 175, 112, 6,   87,  144, 21,  196,
      217, 54,  4,   160, 85,  194, 31,  188, 242, 180, 144, 96,  172, 214, 251, 39,  151, 249, 146, 178, 122, 8,   166,
      8,   198, 34,  171, 154, 248, 28,  95,  78,  106, 26,  224, 175, 45,  241, 41,  6,   16,  96,  19,  120, 91,  21,
      153, 208, 34,  205, 154, 14,  215, 194, 170, 16,  176, 141, 104, 115, 81,  46,  212, 245, 120, 135, 83,  121, 84,
      71,  159, 240, 156, 44,  37,  161, 183, 233, 66,  156, 203, 54,  4,   111, 62,  100, 171, 209, 127, 87,  118, 8,
      233, 177, 120, 126, 224, 141, 148, 184, 16,  60,  111, 66,  232, 80,  165, 40,  130, 54,  206, 171, 249, 112, 236,
      216, 117, 166, 211, 42,  71,  68,  168, 197, 52,  99,  151, 231, 208, 231, 37,  38,  55,  164, 214, 202, 13,  212,
      18,  16,  89,  103, 113, 109, 2,   53,  103, 86,  128, 60,  58,  235, 40,  49,  118, 197, 182, 118, 3,   8,   75,
      97,  114, 69,  110, 74,  126, 211, 32,  76,  82,  49,  213, 153, 133, 218, 237, 17,  87,  130, 94,  206, 152, 116,
      252, 221, 52,  57,  225, 168, 203, 245, 85,  174, 250, 17,  114, 223, 66,  22,  229, 2,   157, 55,  12,  131, 16,
      213, 240, 129, 150, 234, 118, 78,  52,  199, 222, 102, 59,  226, 251, 68,  204, 188, 203, 221, 51,  172, 147, 49,
      169, 143, 5,   94,  26,  205, 155, 192, 126, 228, 26,  181, 107, 64,  7,   132, 88,  79,  5,   165, 182, 99,  106,
      39,  104, 2,   82,  141, 246, 136, 56,  229, 248, 42,  30,  238, 194, 15,  154, 20,  3,   8,   231, 119, 235, 179,
      204, 175, 46,  64,  109, 214, 79,  82,  150, 188, 18,  96,  231, 191, 189, 46,  136, 178, 11,  77,  151, 165, 111,
      29,  143, 65,  254, 80,  128, 43,  10,  62,  162, 120, 99,  199, 246, 245, 101, 36,  222, 191, 220, 13,  164, 168,
      195, 141, 20,  229, 23,  240, 195, 200, 52,  83,  225, 85,  56,  53,  211, 121, 119, 123, 224, 225, 196, 63,  12,
      17,  11,  32,  210, 86,  119, 26,  19,  13,  60,  76,  196, 89,  7,   55,  167, 118, 105, 90,  120, 177, 6,   223,
      31,  127, 128, 234, 203, 157, 237, 208, 117, 112, 250, 47,  168, 28,  197, 193, 109, 214, 201, 41,  248, 191, 227,
      124, 163, 71,  13,  5,   125, 50,  184, 120, 226, 171, 114, 95,  29,  207, 194, 111, 232, 157, 80,  214, 8,   206,
      104, 63,  198, 115, 178, 229, 52,  201, 203, 173, 84,  58,  185, 69,  75,  118, 162, 92,  2,   56,  212, 108, 18,
      105, 200, 156, 122, 161, 14,  222, 67,  188, 231, 85,  9,   22,  111, 250, 156, 235, 120, 236, 48,  5,   208, 236,
      150, 84,  224, 67,  81,  166, 110, 4,   194, 70,  196, 200, 160, 172, 177, 8,   31,  42,  198, 62,  30,  230, 7,
      50,  6,   252, 59,  57,  205, 214, 85,  80,  157, 75,  129, 210, 213, 142, 220, 133, 235, 97,  113, 195, 149, 2,
      73,  25,  105, 159, 238, 200, 77,  99,  73,  0,   143, 4,   6,   158, 134, 238, 244, 145, 96,  188, 192, 53,  31,
      189, 178, 181, 31,  241, 218, 112, 155, 125, 92,  146, 179, 119, 191, 248, 210, 223, 123, 116, 171, 57,  199, 204,
      75,  221, 234, 33,  170, 44,  162, 6,   212, 236, 217, 108, 251, 140, 23,  124, 196, 209, 181, 39,  221, 242, 23,
      238, 203, 48,  58,  44,  86,  56,  116, 57,  54,  7,   70,  190, 130, 79,  254, 41,  216, 103, 32,  50,  237, 215,
      93,  252, 40,  219, 253, 18,  123, 185, 113, 42,  53,  97,  246, 241, 63,  89,  162, 248, 114, 63,  182, 233, 175,
      186, 235, 245, 171, 135, 86,  191, 30,  229, 101, 200, 77,  76,  88,  70,  225, 222, 68,  148, 48,  142, 48,  251,
      20,  26,  197, 20,  236, 199, 16,  54,  247, 115, 52,  6,   158, 30,  132, 93,  76,  62,  220, 65,  216, 175, 35,
      47,  151, 152, 3,   58,  149, 20,  223, 163, 44,  202, 37,  65,  54,  11,  21,  134, 49,  183, 35,  72,  92,  65,
      55,  213, 36,  112, 3,   188, 62,  68,  128, 31,  196, 78,  249, 63,  108, 59,  157, 142, 248, 6,   45,  183, 177,
      216, 136, 223, 111, 162, 222, 194, 33,  129, 144, 115, 146, 170, 55,  136, 141, 29,  61,  63,  33,  35,  153, 196,
      150, 173, 80,  84,  207, 174, 40,  137, 82,  171, 68,  106, 184, 191, 90,  210, 253, 148, 102, 77,  175, 158, 194,
      6,   134, 215, 225, 49,  253, 226, 179, 44,  176, 137, 78,  143, 72,  96,  109, 7,   68,  188, 118, 176, 108, 7,
      34,  17,  76,  254, 168, 27,  88,  148, 184, 11,  42,  242, 140, 2,   249, 240, 210, 135, 213, 12,  230, 118, 212,
      31,  239, 234, 133, 104, 72,  171, 111, 18,  169, 120, 27,  119, 165, 204, 90,  83,  54,  196, 229, 37,  52,  19,
      132, 1,   121, 156, 139, 180, 102, 11,  159, 66,  222, 242, 246, 114, 38,  118, 179, 71,  112, 144, 108, 181, 188,
      62,  181, 108, 150, 224, 101, 97,  22,  131, 174, 116, 148, 175, 187, 53,  40,  74,  204, 97,  87,  130, 177, 194,
      137, 94,  57,  79,  225, 38,  222, 152, 199, 132, 208, 25,  31,  222, 130, 242, 20,  25,  212, 135, 234, 71,  116,
      193, 200, 120, 122, 124, 168, 230, 20,  213, 137, 63,  4,   6,   90,  116, 123, 182, 4,   20,  177, 62,  42,  242,
      197, 184, 89,  117, 192, 49,  236, 241, 231, 233, 69,  218, 144, 239, 236, 202, 75,  135, 141, 59,  249, 196, 37,
      246, 185, 181, 51,  42,  62,  56,  62,  193, 47,  210, 26,  95,  84,  151, 236, 60,  119, 66,  252, 75,  231, 96,
      16,  226, 106, 184, 8,   64,  174, 61,  208, 114, 43,  87,  71,  23,  95,  116, 100, 6,   187, 5,   189, 191, 36,
      28,  97,  10,  32,  236, 35,  150, 171, 184, 94,  121, 57,  176, 100, 174, 42,  180, 47,  0,   132, 146, 150, 172,
      84,  83,  6,   223, 109, 86,  87,  181, 113, 107, 54,  36,  53,  1,   130, 214, 19,  122, 30,  191, 183, 151, 252,
      48,  123, 66,  114, 198, 178, 199, 107, 188, 191, 22,  187, 146, 183, 93,  74,  99,  74,  16,  47,  110, 219, 65,
      57,  201, 94,  129, 208, 162, 111, 98,  24,  103, 215, 20,  176, 41,  220, 250, 55,  61,  4,   111, 121, 66,  190,
      127, 170, 175, 237, 59,  115, 136, 21,  87,  26,  53,  168, 247, 55,  10,  135, 31,  97,  204, 171, 151, 127, 206,
      122, 96,  225, 131, 155, 178, 105, 149, 203, 169, 59,  129, 5,   63,  136, 114, 92,  62,  9,   172, 230, 86,  213,
      168, 251, 213, 87,  136, 231, 240, 168, 98,  182, 228, 103, 21,  9,   208, 241, 203, 144, 168, 75,  34,  26,  189,
      220, 175, 187, 179, 217, 107, 177, 29,  235, 185, 210, 177, 155, 151, 114, 3,   163, 27,  123, 155, 207, 204, 36,
      237, 211, 57,  246, 157, 251, 231, 251, 59,  90,  235, 189, 78,  141, 221, 108, 132, 143, 218, 50,  180, 157, 97,
      119, 213, 240, 86,  51,  32,  1,   215, 47,  116, 103, 76,  238, 44,  235, 54,  72,  228, 156, 194, 29,  197, 169,
      149, 219, 29,  146, 144, 136, 102, 26,  168, 20,  217, 22,  47,  207, 244, 250, 107, 232, 169, 209, 3,   30,  240,
      90,  14,  42,  106, 170, 194, 212, 24,  170, 154, 91,  23,  244, 56,  108, 159, 156, 210, 97,  181, 151, 31,  87,
      250, 26,  107, 225, 253, 154, 225, 83,  72,  206, 139, 241, 207, 94,  17,  227, 150, 236, 10,  238, 26,  123, 108,
      133, 200, 204, 247, 175, 184, 15,  42,  87,  92,  9,   210, 9,   77,  200, 189, 6,   204, 41,  82,  110, 24,  63,
      152, 224, 35,  108, 124, 6,   108, 226, 144, 179, 37,  156, 128, 206, 73,  175, 136, 38,  238, 133, 196, 188, 123,
      123, 17,  50,  168, 66,  49,  32,  80,  190, 164, 109, 226, 148, 32,  131, 215, 190, 190, 189, 40,  206, 227, 79,
      190, 242, 17,  234, 1,   7,   161, 237, 63,  24,  67,  150, 161, 192, 18,  253, 0,   222, 114, 159, 22,  61,  17,
      70,  114, 58,  180, 220, 169, 48,  186, 82,  242, 213, 76,  149, 99,  170, 104, 252, 123, 242, 218, 125, 68,  166,
      60,  85,  49,  239, 169, 145, 210, 69,  165, 235, 216, 45,  201, 106, 124, 192, 143, 236, 45,  229, 177, 31,  165,
      76,  147, 21,  0,   52,  98,  195, 109, 192, 94,  66,  223, 47,  8,   146, 81,  41,  127, 159, 46,  111, 67,  94,
      55,  205, 220, 59,  160, 216, 198, 79,  47,  34,  152, 188, 47,  161, 61,  160, 206, 98,  80,  195, 122, 99,  165,
      86,  130, 166, 90,  201, 40,  7,   177, 52,  118, 203, 225, 158, 151, 189, 102, 98,  82,  198, 121, 20,  125, 142,
      113, 190, 123, 13,  228, 206, 124, 58,  108, 26,  90,  18,  240, 78,  161, 172, 172, 21,  66,  54,  122, 146, 96,
      46,  109, 152, 28,  26,  138, 10,  247, 15,  16,  147, 116, 208, 220, 46,  139, 27,  142, 86,  240, 230, 32,  71,
      68,  232, 254, 223, 94,  221, 29,  248, 198, 72,  5,   212, 154, 15,  46,  29,  207, 76,  69,  146, 190, 176, 251,
      123, 126, 232, 134, 46,  29,  23,  236, 22,  240, 76,  46,  194, 27,  237, 179, 234, 211, 194, 164, 230, 237, 14,
      46,  248, 82,  61,  86,  64,  44,  66,  190, 69,  79,  111, 96,  152, 150, 234, 77,  237, 186, 216, 9,   110, 158,
      250, 169, 188, 140, 212, 28,  31,  95,  99,  193, 37,  125, 71,  95,  40,  14,  155, 153, 207, 131, 165, 42,  96,
      108, 127, 30,  217, 98,  181, 26,  172, 223, 174, 165, 81,  68,  233, 65,  192, 178, 78,  143, 170, 68,  12,  61,
      6,   64,  142, 173, 78,  94,  191, 89,  97,  112, 90,  200, 212, 193, 199, 15,  179, 119, 136, 205, 114, 234, 220,
      155, 152, 173, 159, 129, 62,  108, 39,  94,  209, 20,  131, 124, 66,  167, 162, 141, 223, 106, 58,  103, 210, 171,
      153, 222, 157, 11,  81,  172, 224, 25,  242, 144, 84,  45,  40,  225, 192, 13,  18,  157, 250, 89,  166, 71,  130,
      3,   98,  148, 66,  41,  81,  116, 9,   80,  9,   19,  205, 44,  228, 183, 146, 68,  57,  104, 122, 218, 139, 58,
      46,  128, 243, 19,  4,   47,  227, 12,  43,  96,  93,  149, 44,  89,  204, 187, 58,  58,  223, 154, 97,  146, 56,
      221, 118, 86,  169, 253, 48,  136, 40,  219, 142, 185, 14,  77,  168, 78,  187, 202, 124, 245, 250, 108, 12,  241,
      136, 136, 175, 243, 117, 217, 4,   11,  115, 21,  14,  221, 185, 250, 125, 245, 13,  71,  212, 32,  230, 8,   94,
      247, 132, 197, 18,  206, 178, 131, 38,  93,  22,  230, 174, 112, 144, 10,  88,  75,  132, 77,  23,  252, 22,  18,
      101, 130, 88,  180, 18,  244, 253, 145, 67,  206, 30,  109, 141, 2,   11,  64,  15,  74,  16,  201, 114, 40,  17,
      225, 146, 231, 35,  9,   55,  49,  143, 181, 104, 102, 88,  171, 38,  68,  4,   91,  225, 232, 186, 100, 210, 41,
      59,  54,  30,  67,  129, 3,   163, 125, 218, 188, 203, 40,  210, 7,   34,  250, 245, 10,  245, 139, 106, 3,   237,
      202, 72,  61,  143, 79,  213, 17,  53,  73,  203, 129, 231, 229, 133, 216, 186, 188, 251, 187, 244, 58,  136, 116,
      1,   178, 142, 172, 20,  233, 134, 247, 86,  186, 59,  161, 77,  187, 86,  45,  177, 241, 41,  114, 35,  156, 173,
      70,  92,  178, 198, 138, 91,  29,  206, 80,  26,  61,  8,   88,  240, 237, 88,  209, 57,  191, 78,  203, 42,  12,
      154, 105, 165, 52,  52,  29,  70,  172, 52,  210, 154, 190, 205, 26,  123, 223, 66,  108, 4,   45,  191, 162, 58,
      142, 205, 254, 64,  205, 84,  235, 41,  50,  18,  77,  97,  86,  52,  189, 80,  198, 254, 78,  26,  131, 142, 69,
      210, 45,  123, 5,   162, 7,   208, 32,  46,  123, 37,  195, 30,  60,  160, 223, 165, 98,  110, 34,  41,  44,  166,
      156, 204, 121, 75,  120, 24,  193, 5,   31,  125, 203, 39,  229, 11,  192, 90,  243, 142, 135, 199, 165, 143, 140,
      37,  133, 30,  65,  18,  88,  11,  190, 181, 126, 158, 2,   174, 225, 221, 151, 73,  64,  216, 212, 28,  40,  131,
      85,  236, 20,  143, 125, 245, 239, 86,  178, 236, 179, 9,   160, 49,  231, 97,  93,  190, 28,  31,  91,  70,  39,
      64,  96,  28,  218, 154, 62,  214, 84,  19,  153, 2,   193, 100, 23,  185, 182, 102, 66,  206, 222, 149, 247, 247,
      19,  193, 64,  56,  68,  149, 175, 139, 30,  221, 10,  242, 35,  166, 113, 63,  206, 186, 163, 217, 25,  17,  217,
      3,   89,  113, 182, 117, 105, 159, 112, 62,  140, 65,  244, 120, 118, 88,  175, 157, 195, 140, 143, 107, 33,  198,
      11,  8,   241, 59,  225, 190, 135, 228, 19,  34,  228, 125, 196, 107, 50,  245, 204, 73,  9,   81,  126, 199, 86,
      88,  178, 220, 53,  87,  74,  24,  188, 156, 104, 122, 20,  111, 16,  102, 212, 169, 10,  172, 50,  167, 89,  28,
      87,  180, 47,  247, 113, 94,  186, 96,  99,  16,  231, 7,   228, 103, 49,  45,  27,  55,  123, 212, 130, 212, 38,
      70,  118, 57,  205, 7,   156, 146, 79,  35,  57,  9,   182, 32,  214, 32,  23,  85,  154, 3,   61,  60,  240, 96,
      105, 153, 63,  3,   229, 235, 129, 194, 129, 233, 71,  122, 8,   195, 137, 39,  158, 86,  2,   147, 77,  28,  41,
      234, 144, 180, 213, 36,  254, 157, 160, 80,  99,  125, 205, 57,  170, 136, 228, 54,  58,  185, 181, 121, 42,  229,
      155, 233, 247, 66,  20,  91,  34,  212, 197, 99,  157, 10,  45,  249, 81,  250, 32,  192, 232, 167, 102, 205, 211,
      168, 240, 103, 117, 91,  82,  184, 28,  85,  221, 91,  167, 233, 46,  254, 95,  182, 186, 219, 28,  102, 154, 227,
      65,  35,  63,  59,  133, 173, 145, 133, 192, 208, 126, 248, 179, 120, 137, 177, 241, 57,  0,   154, 142, 48,  73,
      69,  166, 192, 96,  175, 162, 14,  253, 31,  67,  71,  150, 49,  24,  85,  93,  156, 20,  243, 141, 127, 185, 56,
      71,  60,  137, 29,  51,  87,  39,  138, 130, 32,  229, 70,  180, 35,  252, 134, 167, 121, 33,  246, 56,  143, 42,
      245, 44,  77,  235, 181, 13,  15,  81,  212, 64,  75,  109, 25,  84,  134, 38,  31,  52,  168, 205, 52,  152, 157,
      231, 198, 161, 117, 102, 15,  247, 59,  213, 48,  57,  29,  170, 58,  211, 174, 40,  54,  239, 188, 1,   240, 174,
      197, 70,  149, 134, 94,  48,  49,  251, 144, 51,  153, 185, 169, 222, 68,  237, 130, 187, 13,  74,  224, 43,  17,
      124, 133, 125, 87,  192, 247, 186, 158, 221, 54,  223, 233, 245, 249, 169, 204, 100, 206, 17,  127, 167, 139, 140,
      82,  248, 223, 92,  193, 158, 120, 240, 89,  195, 139, 218, 248, 248, 228, 2,   21,  184, 193, 120, 233, 42,  37,
      209, 20,  81,  131, 72,  196, 124, 21,  216, 160, 168, 250, 208, 109, 225, 126, 148, 204, 106, 101, 13,  235, 239,
      156, 211, 4,   47,  216, 254, 113, 157, 186, 126, 253, 5,   230, 4,   67,  223, 76,  193, 198, 164, 5,   113, 20,
      197, 160, 29,  119, 72,  15,  221, 150, 159, 179, 113, 65,  121, 46,  156, 230, 111, 230, 187, 2,   246, 77,  240,
      10,  146, 68,  199, 198, 38,  120, 201, 4,   27,  103, 236, 116, 13,  224, 57,  126, 138, 232, 206, 227, 131, 251,
      34,  229, 4,   253, 96,  79,  10,  72,  73,  50,  20,  245, 69,  231, 134, 7,   48,  207, 41,  138, 241, 158, 80,
      160, 55,  140, 3,   201, 105, 12,  182, 49,  154, 91,  5,   22,  233, 92,  231, 40,  47,  215, 16,  197, 82,  175,
      105, 76,  229, 30,  9,   88,  29,  203, 193, 199, 129, 14,  239, 168, 129, 192, 158, 100, 248, 88,  64,  139, 177,
      123, 30,  132, 170, 32,  165, 55,  215, 127, 246, 204, 110, 177, 37,  131, 221, 68,  208, 217, 31,  248, 14,  227,
      97,  123, 100, 196, 119, 41,  206, 110, 224, 192, 165, 235, 142, 116, 241, 98,  222, 69,  246, 89,  108, 113, 229,
      35,  69,  70,  21,  129, 130, 30,  36,  27,  143, 247, 243, 63,  185, 206, 229, 204, 171, 201, 45,  60,  93,  227,
      113, 243, 113, 226, 165, 144, 231, 81,  137, 46,  3,   61,  22,  129, 12,  63,  63,  253, 130, 235, 29,  161, 75,
      178, 214, 152, 203, 80,  181, 193, 146, 141, 80,  98,  130, 194, 93,  118, 134, 207, 69,  146, 74,  248, 46,  125,
      39,  253, 253, 64,  30,  172, 73,  199, 1,   228, 64,  221, 111, 166, 64,  184, 121, 248, 72,  12,  118, 237, 122,
      126, 218, 20,  246, 213, 246, 51,  59,  228, 77,  171, 156, 237, 247, 238, 60,  27,  162, 153, 219, 15,  253, 209,
      48,  179, 202, 191, 62,  100, 219, 193, 243, 246, 178, 43,  99,  30,  238, 122, 134, 153, 118, 67,  238, 243, 112,
      28,  241, 194, 249, 147, 235, 24,  182, 127, 2,   142, 70,  181, 247, 162, 32,  158, 53,  214, 251, 197, 85,  70,
      88,  54,  128, 90,  159, 13,  135, 244, 10,  30,  31,  165, 27,  70,  105, 104, 53,  183, 35,  243, 122, 152, 216,
      98,  141, 196, 169, 222, 223, 143, 245, 210, 212, 35,  171, 140, 66,  198, 20,  245, 70,  140, 190, 106, 7,   183,
      199, 38,  46,  85,  52,  23,  1,   199, 239, 241, 202, 96,  140, 236, 154, 182, 50,  82,  211, 211, 104, 217, 217,
      252, 94,  204, 239, 98,  64,  210, 33,  201, 226, 208, 151, 112, 46,  90,  212, 22,  53,  212, 221, 34,  66,  177,
      204, 205, 99,  8,   79,  134, 141, 236, 157, 183, 192, 123, 212, 234, 39,  208, 105, 185, 215, 235, 34,  95,  158,
      54,  94,  46,  42,  172, 47,  34,  233, 238, 195, 45,  6,   166, 27,  11,  22,  243, 126, 9,   239, 62,  238, 84,
      181, 224, 223, 232, 227, 60,  46,  242, 67,  92,  68,  205, 62,  210, 199, 29,  25,  205, 190, 30,  57,  112, 112,
      183, 114, 223, 105, 213, 118, 113, 86,  152, 121, 22,  91,  167, 37,  135, 221, 124, 192, 73,  45,  213, 8,   37,
      51,  152, 18,  17,  114, 73,  63,  78,  65,  8,   136, 151, 20,  33,  181, 61,  223, 9,   99,  138, 166, 114, 150,
      30,  99,  143, 155, 198, 203, 142, 218, 152, 238, 15,  57,  164, 25,  179, 61,  172, 16,  243, 27,  17,  139, 31,
      9,   237, 128, 224, 247, 112, 165, 178, 203, 102, 129, 31,  38,  62,  242, 121, 166, 241, 241, 172, 26,  222, 91,
      101, 15,  11,  115, 121, 20,  34,  38,  24,  206, 238, 215, 104, 221, 214, 81,  84,  96,  26,  225, 20,  152, 149,
      50,  251, 152, 62,  69,  94,  18,  174, 220, 172, 25,  206, 162, 201, 55,  150, 108, 123, 160, 81,  167, 168, 46,
      118, 243, 12,  15,  218, 125, 152, 216, 79,  175, 79,  65,  114, 244, 175, 127, 163, 90,  202, 68,  127, 80,  135,
      206, 9,   237, 117, 228, 244, 163, 26,  211, 25,  123, 107, 96,  61,  74,  132, 202, 180, 55,  73,  196, 65,  252,
      62,  60,  99,  218, 17,  223, 118, 132, 74,  75,  88,  139, 48,  125, 60,  14,  10,  45,  24,  95,  129, 159, 2,
      90,  174, 204, 18,  184, 240, 56,  210, 34,  93,  191, 143, 165, 38,  139, 23,  36,  49,  140, 181, 252, 148, 245,
      137, 240, 153, 31,  62,  111, 22,  14,  156, 200, 97,  175, 157, 253, 104, 178, 161, 184, 192, 97,  73,  11,  38,
      39,  197, 162, 144, 107, 164, 27,  253, 194, 42,  149, 138, 40,  194, 166, 138, 28,  100, 16,  97,  229, 112, 70,
      163, 229, 69,  164, 33,  72,  49,  232, 172, 207, 80,  23,  150, 241, 25,  216, 249, 120, 70,  236, 203, 81,  29,
      173, 31,  160, 34,  234, 133, 144, 187, 36,  160, 100, 56,  179, 51,  25,  128, 162, 63,  129, 184, 172, 224, 113,
      209, 223, 158, 190, 123, 123, 4,   137, 127, 164, 184, 26,  17,  106, 41,  29,  55,  232, 240, 88,  197, 89,  13,
      75,  254, 163, 201, 61,  135, 179, 184, 69,  41,  70,  136, 104, 211, 82,  123, 70,  188, 51,  247, 246, 5,   163,
      125, 107, 81,  125, 150, 156, 7,   22,  48,  156, 108, 5,   252, 181, 48,  48,  230, 253, 62,  88,  151, 228, 94,
      91,  178, 159, 35,  30,  30,  206, 218, 92,  182, 14,  88,  97,  40,  45,  57,  234, 154, 38,  62,  193, 7,   244,
      111, 43,  140, 96,  210, 130, 189, 189, 42,  35,  92,  57,  202, 236, 67,  254, 243, 189, 129, 135, 129, 171, 124,
      206, 41,  178, 62,  101, 166, 215, 146, 15,  44,  138, 145, 20,  170, 200, 240, 72,  193, 71,  154, 249, 68,  117,
      178, 66,  206, 113, 58,  124, 154, 172, 140, 32,  56,  75,  81,  235, 141, 252, 107, 124, 143, 148, 100, 61,  12,
      79,  133, 129, 77,  0,   142, 187, 251, 16,  57,  251, 105, 141, 91,  77,  89,  132, 141, 114, 198, 168, 17,  244,
      194, 60,  36,  93,  107, 146, 177, 122, 77,  15,  100, 184, 200, 127, 45,  209, 142, 121, 65,  211, 126, 179, 161,
      51,  105, 28,  126, 14,  186, 54,  131, 107, 80,  199, 232, 215, 178, 36,  253, 42,  36,  60,  106, 140, 114, 126,
      198, 248, 147, 44,  5,   78,  21,  34,  241, 61,  223, 92,  31,  249, 239, 242, 84,  80,  26,  9,   80,  36,  81,
      3,   77,  10,  233, 35,  65,  79,  125, 81,  136, 35,  10,  248, 165, 65,  91,  23,  76,  168, 206, 179, 35,  186,
      179, 61,  128, 46,  8,   158, 38,  152, 208, 239, 57,  64,  194, 155, 187, 12,  96,  78,  177, 232, 75,  21,  98,
      115, 69,  191, 240, 82,  9,   185, 129, 227, 115, 132, 141, 169, 254, 116, 152, 38,  64,  175, 148, 219, 210, 61,
      248, 109, 167, 13,  148, 159, 78,  16,  59,  102, 29,  8,   220, 113, 161, 204, 150, 4,   25,  53,  58,  22,  158,
      90,  207, 155, 240, 61,  109, 101, 213, 213, 168, 38,  137, 91,  125, 68,  133, 103, 102, 176, 95,  60,  142, 121,
      159, 177, 148, 2,   65,  249, 36,  215, 171, 89,  142, 238, 125, 78,  201, 1,   175, 50,  120, 129, 177, 216, 76,
      201, 234, 41,  159, 149, 53,  206, 125, 4,   97,  13,  33,  181, 232, 235, 234, 220, 14,  165, 76,  15,  125, 140,
      246, 6,   104, 191, 155, 58,  247, 90,  106, 57,  247, 30,  60,  150, 248, 124, 122, 147, 174, 155, 170, 155, 192,
      50,  108, 43,  189, 209, 138, 188, 171, 223, 32,  17,  134, 231, 205, 205, 136, 199, 225, 220, 183, 6,   161, 53,
      107, 141, 58,  35,  1,   166, 92,  25,  41,  88,  16,  182, 59,  0,   99,  73,  100, 238, 225, 32,  38,  113, 171,
      239, 144, 236, 229, 176, 139, 46,  165, 235, 5,   192, 78,  202, 63,  131, 223, 101, 18,  128, 22,  103, 126, 148,
      114, 251, 86,  166, 194, 150, 14,  2,   68,  67,  177, 190, 203, 10,  75,  52,  151, 4,   36,  77,  72,  229, 87,
      137, 203, 51,  54,  34,  72,  164, 121, 30,  146, 127, 63,  8,   100, 119, 33,  71,  91,  81,  80,  16,  17,  17,
      106, 73,  195, 162, 110, 184, 65,  37,  180, 208, 171, 195, 28,  204, 60,  199, 21,  10,  241, 156, 172, 179, 71,
      229, 180, 235, 238, 143, 150, 93,  164, 219, 230, 84,  46,  190, 123, 251, 237, 154, 88,  196, 147, 132, 74,  113,
      239, 40,  185, 184, 154, 76,  31,  48,  234, 57,  69,  147, 158, 173, 54,  49,  26,  168, 209, 205, 109, 112, 184,
      104, 30,  115, 28,  43,  87,  2,   218, 171, 103, 87,  113, 211, 98,  229, 55,  118, 169, 139, 82,  215, 249, 67,
      222, 122, 121, 110, 196, 231, 201, 125, 239, 59,  139, 79,  157, 162, 93,  121, 154, 92,  28,  232, 200, 35,  44,
      38,  114, 44,  126, 72,  68,  7,   79,  44,  91,  215, 6,   139, 16,  109, 240, 13,  227, 167, 241, 49,  75,  58,
      120, 64,  48,  38,  159, 128, 136, 55,  223, 187, 233, 75,  46,  167, 101, 213, 132, 23,  245, 41,  160, 146, 223,
      44,  253, 28,  214, 82,  220, 205, 118, 95,  113, 206, 0,   111, 6,   3,   74,  238, 217, 37,  164, 76,  246, 45,
      0,   229, 233, 70,  196, 60,  96,  46,  196, 7,   251, 156, 77,  124, 6,   80,  157, 158, 64,  6,   223, 205, 254,
      124, 37,  0,   247, 150, 212, 205, 8,   109, 99,  164, 146, 238, 76,  94,  131, 156, 195, 93,  247, 101, 125, 171,
      145, 87,  130, 213, 227, 6,   207, 151, 84,  9,   60,  15,  141, 55,  30,  14,  173, 26,  115, 247, 169, 239, 3,
      231, 17,  15,  217, 124, 139, 213, 139, 8,   39,  210, 203, 231, 126, 175, 46,  182, 54,  241, 244, 98,  183, 77,
      194, 138, 72,  221, 201, 77,  219, 115, 10,  252, 62,  152, 20,  16,  101, 26,  82,  86,  87,  86,  58,  104, 7,
      116, 102, 6,   216, 87,  204, 132, 192, 199, 86,  11,  111, 77,  86,  130, 89,  15,  31,  179, 107, 44,  100, 180,
      207, 30,  16,  214, 39,  130, 154, 201, 197, 220, 10,  177, 182, 233, 162, 146, 189, 108, 109, 35,  113, 96,  105,
      101, 32,  24,  18,  203, 101, 130, 40,  100, 38,  128, 180, 56,  114, 70,  30,  15,  160, 11,  33,  164, 192, 1,
      157, 234, 104, 202, 202, 172, 69,  221, 252, 203, 75,  95,  5,   99,  230, 11,  220, 44,  113, 227, 224, 234, 69,
      168, 8,   208, 93,  229, 254, 216, 154, 186, 40,  79,  94,  224, 233, 55,  78,  108, 217, 154, 4,   203, 164, 204,
      73,  159, 56,  185, 159, 128, 28,  115, 97,  244, 32,  131, 76,  19,  32,  209, 163, 148, 196, 8,   241, 46,  197,
      92,  70,  103, 212, 233, 31,  79,  44,  160, 96,  53,  97,  208, 120, 81,  251, 253, 90,  47,  102, 103, 174, 91,
      113, 130, 56,  211, 172, 220, 81,  126, 111, 84,  75,  8,   208, 134, 144, 165, 203, 151, 134, 238, 137, 10,  177,
      107, 121, 245, 60,  0,   33,  234, 190, 171, 56,  25,  22,  124, 148, 32,  210, 144, 210, 142, 31,  119, 253, 109,
      129, 103, 17,  103, 194, 140, 36,  123, 47,  180, 220, 5,   64,  232, 251, 5,   15,  87,  45,  87,  40,  45,  161,
      71,  149, 30,  158, 11,  204, 167, 214, 152, 120, 32,  170, 122, 53,  135, 217, 152, 118, 137, 9,   55,  208, 113,
      205, 238, 9,   23,  74,  141, 225, 180, 213, 184, 222, 218, 171, 103, 0,   195, 25,  65,  161, 162, 25,  87,  78,
      202, 12,  37,  13,  58,  172, 94,  132, 77,  134, 228, 117, 45,  183, 45,  81,  80,  151, 58,  173, 225, 42,  6,
      31,  64,  4,   194, 131, 175, 129, 204, 102, 105, 115, 106, 239, 132, 225, 22,  213, 163, 137, 161, 193, 76,  126,
      52,  94,  190, 239, 34,  246, 161, 43,  163, 250, 214, 124, 235, 209, 191, 147, 53,  136, 181, 47,  134, 31,  181,
      32,  24,  211, 158, 83,  254, 122, 64,  252, 245, 236, 0,   211, 73,  191, 45,  175, 239, 137, 43,  62,  228, 108,
      204, 172, 2,   164, 150, 162, 170, 202, 39,  10,  101, 19,  185, 189, 30,  181, 3,   238, 187, 164, 194, 106, 87,
      169, 121, 146, 40,  29,  239, 58,  246, 238, 231, 99,  52,  249, 92,  54,  193, 46,  13,  250, 82,  126, 76,  252,
      3,   185, 69,  74,  42,  150, 74,  36,  51,  88,  113, 136, 231, 101, 106, 234, 163, 43,  10,  172, 26,  222, 59,
      27,  212, 74,  92,  148, 14,  40,  36,  37,  249, 221, 111, 148, 79,  44,  168, 196, 174, 248, 80,  151, 153, 73,
      200, 193, 210, 245, 97,  19,  203, 149, 58,  216, 191, 164, 167, 106, 64,  82,  216, 151, 153, 194, 143, 187, 133,
      117, 252, 33,  23,  203, 28,  167, 30,  214, 167, 204, 18,  79,  112, 216, 231, 189, 219, 118, 194, 115, 126, 51,
      177, 55,  212, 122, 52,  215, 149, 9,   61,  56,  186, 53,  249, 6,   199, 43,  40,  91,  44,  94,  191, 245, 14,
      215, 197, 46,  179, 226, 10,  235, 230, 99,  251, 164, 222, 25,  2,   97,  208, 115, 77,  174, 251, 199, 10,  136,
      72,  137, 48,  113, 3,   96,  247, 57,  0,   211, 168, 254, 201, 24,  94,  226, 205, 44,  71,  177, 227, 248, 242,
      90,  81,  242, 157, 77,  4,   46,  127, 249, 40,  47,  44,  115, 85,  120, 147, 135, 160, 42,  192, 25,  17,  166,
      50,  188, 205, 215, 16,  200, 132, 36,  113, 157, 30,  147, 72,  41,  213, 229, 19,  250, 218, 105, 227, 218, 241,
      35,  95,  246, 41,  126, 96,  60,  62,  95,  63,  230, 21,  101, 230, 66,  46,  117, 173, 114, 195, 148, 23,  122,
      116, 55,  102, 14,  162, 203, 202, 250, 118, 169, 237, 169, 95,  207, 134, 235, 97,  194, 185, 192, 153, 189, 247,
      90,  33,  231, 119, 3,   149, 32,  62,  72,  71,  80,  189, 100, 29,  226, 87,  109, 54,  84,  155, 19,  133, 13,
      144, 62,  237, 236, 30,  78,  164, 25,  24,  89,  63,  174, 207, 198, 58,  208, 111, 93,  192, 109, 252, 201, 145,
      155, 123, 132, 172, 123, 185, 46,  100, 183, 63,  205, 59,  183, 224, 27,  216, 9,   111, 186, 90,  24,  70,  234,
      47,  181, 201, 189, 176, 87,  71,  227, 52,  55,  101, 55,  25,  240, 120, 9,   207, 73,  172, 49,  201, 251, 194,
      249, 197, 232, 95,  224, 240, 245, 216, 160, 67,  157, 27,  104, 251, 182, 132, 48,  6,   150, 228, 219, 162, 78,
      57,  12,  126, 161, 183, 77,  210, 76,  22,  153, 44,  143, 252, 6,   119, 179, 24,  196, 70,  195, 247, 125, 202,
      171, 100, 100, 155, 117, 19,  15,  132, 64,  62,  176, 38,  225, 63,  128, 244, 216, 76,  229, 243, 44,  100, 247,
      227, 44,  175, 47,  251, 3,   240, 78,  209, 126, 56,  152, 94,  10,  149, 137, 99,  97,  35,  152, 244, 18,  119,
      227, 93,  198, 119, 203, 224, 30,  157, 24,  178, 6,   55,  108, 3,   105, 30,  95,  80,  89,  136, 125, 68,  233,
      215, 159, 68,  73,  3,   62,  46,  200, 130, 24,  70,  10,  254, 218, 141, 132, 8,   4,   94,  232, 226, 223, 190,
      164, 6,   254, 94,  110, 72,  221, 37,  135, 172, 67,  108, 194, 105, 146, 81,  47,  166, 11,  33,  190, 131, 218,
      69,  242, 178, 5,   55,  195, 1,   38,  85,  61,  110, 239, 93,  227, 178, 184, 63,  168, 192, 53,  78,  93,  159,
      174, 222, 123, 111, 156, 101, 150, 172, 225, 86,  60,  79,  234, 78,  144, 122, 74,  195, 208, 46,  234, 189, 130,
      51,  37,  242, 104, 153, 193, 243, 71,  172, 14,  93,  241, 243, 117, 66,  199, 56,  223, 25,  125, 44,  112, 63,
      50,  142, 241, 62,  206, 35,  198, 31,  73,  44,  75,  231, 108, 124, 80,  94,  70,  157, 225, 203, 31,  75,  112,
      82,  228, 104, 93,  195, 131, 221, 168, 10,  38,  240, 173, 64,  186, 195, 217, 12,  137, 6,   233, 242, 104, 190,
      114, 243, 52,  17,  40,  234, 157, 201, 94,  147, 46,  237, 58,  64,  52,  226, 228, 228, 203, 161, 229, 59,  225,
      200, 120, 200, 195, 209, 77,  122, 143, 112, 133, 51,  210, 95,  159, 51,  248, 128, 150, 165, 187, 40,  198, 188,
      188, 160, 100, 137, 140, 80,  189, 10,  4,   59,  78,  218, 242, 107, 230, 69,  247, 167, 222, 57,  58,  120, 41,
      25,  33,  141, 188, 184, 7,   236, 75,  221, 86,  108, 246, 146, 54,  15,  10,  168, 183, 232, 199, 129, 167, 93,
      96,  67,  168, 232, 162, 145, 157, 165, 142, 100, 21,  83,  50,  225, 100, 180, 121, 52,  254, 61,  180, 136, 87,
      161, 140, 138, 191, 134, 140, 105, 23,  105, 132, 122, 176, 213, 12,  40,  186, 34,  182, 242, 175, 45,  54,  74,
      230, 218, 64,  136, 165, 249, 84,  191, 143, 172, 174, 138, 194, 96,  20,  146, 206, 232, 120, 137, 93,  112, 18,
      28,  71,  106, 30,  174, 205, 220, 96,  144, 120, 17,  169, 245, 82,  20,  35,  155, 243, 65,  194, 168, 59,  136,
      32,  10,  171, 207, 80,  117, 221, 90,  87,  197, 242, 243, 5,   14,  171, 93,  14,  232, 166, 172, 122, 57,  38,
      140, 243, 199, 26,  41,  123, 53,  31,  51,  233, 73,  246, 59,  33,  108, 197, 178, 229, 252, 92,  187, 148, 96,
      2,   194, 78,  66,  245, 203, 204, 95,  252, 244, 208, 81,  189, 13,  189, 35,  73,  2,   77,  240, 173, 174, 73,
      197, 204, 162, 148, 62,  35,  2,   66,  189, 188, 142, 117, 59,  203, 37,  160, 30,  238, 225, 220, 159, 12,  249,
      117, 95,  94,  52,  191, 177, 32,  95,  160, 201, 51,  164, 136, 252, 73,  202, 8,   232, 235, 135, 101, 51,  35,
      230, 61,  44,  69,  205, 181, 32,  97,  243, 242, 118, 68,  231, 118, 40,  210, 241, 163, 201, 179, 81,  10,  213,
      229, 78,  119, 221, 74,  33,  178, 214, 80,  221, 121, 150, 126, 157, 19,  165, 6,   53,  17,  134, 209, 166, 85,
      241, 185, 94,  126, 52,  200, 110, 21,  52,  58,  128, 113, 186, 9,   8,   198, 76,  215, 55,  33,  213, 170, 163,
      247, 226, 81,  71,  164, 173, 124, 101, 163, 88,  222, 13,  252, 141, 69,  135, 183, 100, 223, 71,  96,  131, 145,
      61,  150, 47,  42,  141, 224, 175, 214, 163, 70,  222, 20,  233, 193, 186, 242, 106, 80,  108, 226, 92,  131, 52,
      133, 206, 149, 86,  20,  218, 43,  196, 30,  168, 141, 245, 225, 129, 62,  37,  207, 164, 254, 93,  17,  16,  214,
      12,  110, 186, 238, 186, 10,  68,  104, 159, 87,  2,   2,   131, 210, 107, 89,  137, 19,  36,  7,   114, 139, 98,
      221, 90,  133, 236, 115, 233, 215, 246, 110, 180, 164, 114, 224, 166, 95,  77,  219, 195, 50,  35,  122, 181, 119,
      107, 150, 45,  98,  126, 186, 70,  83,  145, 144, 171, 251, 251, 188, 183, 151, 115, 59,  133, 210, 141, 16,  164,
      198, 78,  30,  110, 154, 139, 65,  253, 164, 21,  226, 241, 83,  228, 174, 141, 237, 199, 111, 5,   220, 131, 55,
      222, 145, 231, 183, 197, 150, 181, 69,  156, 241, 124, 244, 11,  200, 253, 105, 196, 108, 135, 20,  237, 182, 223,
      35,  128, 14,  241, 42,  96,  152, 181, 89,  93,  141, 242, 167, 220, 225, 44,  58,  120, 30,  189, 91,  245, 244,
      94,  94,  189, 223, 92,  149, 17,  6,   111, 10,  214, 210, 180, 23,  142, 143, 8,   159, 144, 50,  100, 78,  2,
      11,  128, 173, 160, 239, 30,  92,  76,  165, 81,  180, 54,  119, 83,  220, 204, 47,  122, 253, 249, 235, 99,  146,
      127, 237, 185, 20,  30,  3,   128, 146, 126, 52,  52,  31,  251, 103, 247, 185, 42,  243, 208, 134, 191, 119, 226,
      141, 139, 5,   80,  107, 50,  131, 247, 8,   157, 251, 151, 96,  246, 111, 239, 158, 211, 28,  97,  76,  125, 214,
      161, 10,  2,   171, 74,  140, 188, 232, 189, 10,  77,  113, 45,  253, 94,  206, 222, 153, 141, 2,   206, 180, 104,
      166, 207, 0,   192, 149, 8,   46,  104, 174, 171, 39,  152, 176, 192, 132, 6,   2,   44,  70,  241, 32,  176, 172,
      49,  72,  79,  171, 62,  1,   107, 77,  42,  217, 53,  96,  145, 193, 61,  207, 15,  242, 154, 131, 252, 197, 128,
      64,  157, 254, 250, 160, 224, 157, 197, 18,  224, 67,  246, 2,   23,  35,  90,  186, 182, 148, 177, 42,  118, 35,
      146, 104, 27,  233, 207, 140, 224, 103, 180, 166, 49,  226, 75,  87,  84,  165, 160, 122, 116, 182, 134, 49,  250,
      41,  177, 165, 146, 50,  44,  173, 229, 49,  195, 30,  234, 183, 118, 236, 63,  201, 71,  154, 86,  87,  15,  34,
      22,  192, 87,  90,  39,  120, 166, 59,  176, 192, 245, 203, 178, 184, 153, 245, 90,  91,  27,  215, 31,  48,  210,
      125, 252, 30,  164, 229, 137, 43,  214, 244, 226, 214, 77,  157, 217, 239, 237, 205, 146, 37,  11,  68,  179, 93,
      1,   199, 22,  185, 177, 128, 110, 101, 96,  5,   4,   245, 71,  70,  164, 8,   189, 136, 43,  201, 156, 174, 153,
      22,  254, 250, 10,  186, 37,  148, 16,  209, 145, 95,  246, 43,  154, 48,  131, 149, 2,   158, 149, 164, 79,  245,
      64,  180, 78,  99,  113, 20,  10,  236, 249, 99,  69,  93,  78,  194, 217, 87,  156, 20,  251, 126, 77,  139, 28,
      80,  175, 42,  16,  200, 17,  197, 174, 233, 218, 201, 245, 31,  55,  85,  213, 136, 62,  163, 223, 153, 224, 59,
      97,  59,  25,  85,  153, 128, 251, 118, 39,  227, 31,  28,  214, 228, 109, 253, 139, 163, 224, 119, 223, 253, 152,
      214, 65,  73,  88,  15,  16,  73,  82,  210, 141, 147, 167, 236, 88,  211, 26,  210, 16,  218, 84,  134, 35,  208,
      22,  195, 32,  160, 112, 27,  189, 83,  166, 201, 201, 205, 114, 49,  18,  163, 5,   40,  12,  198, 151, 245, 33,
      161, 170, 10,  206, 12,  48,  248, 244, 96,  47,  90,  37,  71,  93,  76,  47,  188, 101, 243, 62,  212, 66,  8,
      160, 30,  176, 254, 222, 118, 69,  236, 51,  146, 51,  148, 12,  48,  201, 19,  84,  206, 37,  165, 252, 56,  23,
      122, 124, 223, 150, 107, 192, 67,  148, 71,  29,  125, 185, 213, 151, 214, 127, 140, 225, 140, 92,  35,  152, 220,
      93,  1,   172, 112, 205, 184, 110, 1,   114, 17,  20,  36,  240, 120, 75,  228, 142, 90,  31,  147, 195, 186, 42,
      95,  74,  188, 26,  3,   196, 53,  217, 126, 18,  100, 96,  68,  66,  145, 249, 48,  117, 248, 21,  130, 8,   12,
      9,   1,   136, 206, 140, 232, 31,  226, 163, 114, 239, 7,   24,  204, 188, 32,  41,  157, 33,  99,  240, 249, 11,
      23,  107, 175, 224, 154, 5,   182, 50,  74,  185, 34,  241, 196, 220, 168, 38,  40,  94,  107, 63,  125, 218, 244,
      43,  176, 5,   93,  144, 82,  64,  83,  59,  160, 104, 6,   179, 38,  106, 163, 164, 188, 102, 17,  183, 199, 152,
      28,  198, 147, 113, 80,  188, 242, 211, 146, 18,  238, 154, 43,  172, 86,  170, 7,   12,  66,  60,  10,  170, 138,
      95,  18,  119, 122, 251, 58,  119, 221, 46,  6,   202, 24,  26,  162, 97,  0,   155, 9,   163, 232, 14,  254, 69,
      241, 156, 101, 19,  153, 87,  156, 56,  30,  43,  21,  158, 207, 77,  208, 220, 23,  133, 11,  229, 205, 59,  80,
      149, 88,  222, 191, 86,  40,  160, 69,  203, 20,  240, 229, 125, 173, 133, 167, 48,  173, 128, 101, 23,  227, 40,
      240, 74,  39,  164, 238, 10,  112, 220, 163, 128, 12,  211, 5,   6,   44,  14,  62,  231, 70,  41,  52,  242, 88,
      40,  192, 63,  27,  109, 70,  124, 240, 237, 192, 223, 202, 105, 77,  189, 43,  70,  109, 171, 221, 64,  196, 218,
      47,  204, 214, 217, 162, 172, 230, 44,  240, 216, 253, 41,  117, 138, 161, 156, 205, 32,  229, 86,  157, 86,  121,
      145, 91,  40,  41,  47,  65,  101, 199, 111, 225, 82,  238, 150, 3,   11,  65,  137, 203, 122, 156, 174, 125, 73,
      20,  104, 164, 157, 227, 44,  69,  231, 72,  146, 175, 45,  198, 167, 105, 228, 240, 236, 222, 250, 151, 211, 28,
      48,  171, 88,  16,  55,  5,   117, 251, 58,  66,  0,   90,  173, 243, 119, 217, 22,  101, 117, 75,  238, 10,  16,
      78,  209, 84,  50,  1,   194, 54,  177, 187, 226, 182, 20,  53,  126, 233, 167, 14,  219, 88,  28,  99,  99,  92,
      77,  27,  196, 240, 115, 180, 216, 141, 25,  201, 119, 213, 27,  163, 14,  159, 163, 47,  183, 77,  216, 99,  110,
      118, 244, 15,  159, 119, 99,  114, 197, 151, 165, 66,  250, 216, 149, 184, 42,  17,  167, 195, 166, 143, 60,  87,
      95,  81,  250, 47,  189, 149, 144, 8,   40,  233, 252, 129, 55,  145, 132, 193, 168, 253, 77,  21,  69,  131, 33,
      143, 179, 178, 100, 69,  99,  177, 161, 189, 221, 191, 216, 23,  16,  109, 110, 46,  107, 253, 90,  88,  216, 90,
      206, 48,  234, 146, 60,  181, 96,  49,  149, 105, 201, 5,   174, 92,  27,  166, 115, 16,  135, 153, 117, 133, 181,
      170, 106, 179, 106, 65,  226, 79,  253, 237, 110, 247, 215, 26,  6,   131, 47,  74,  193, 231, 115, 252, 165, 249,
      92,  160, 40,  161, 45,  241, 59,  224, 233, 154, 211, 77,  211, 96,  15,  135, 193, 238, 52,  196, 76,  139, 131,
      136, 254, 70,  211, 195, 228, 188, 12,  174, 110, 56,  220, 178, 140, 180, 152, 57,  185, 75,  165, 158, 155, 37,
      50,  27,  173, 34,  119, 209, 183, 252, 142, 34,  135, 148, 95,  85,  127, 75,  79,  76,  172, 247, 160, 196, 157,
      103, 163, 69,  47,  169, 159, 229, 58,  92,  240, 9,   43,  134, 98,  137, 121, 100, 40,  178, 57,  190, 140, 199,
      39,  104, 218, 18,  40,  117, 193, 74,  206, 101, 66,  224, 33,  138, 248, 4,   129, 49,  19,  113, 20,  166, 122,
      158, 58,  50,  17,  136, 175, 217, 51,  35,  66,  196, 1,   113, 159, 157, 33,  195, 207, 206, 182, 143, 29,  193,
      177, 243, 251, 109, 104, 119, 19,  47,  148, 219, 186, 51,  98,  210, 164, 92,  131, 194, 38,  52,  117, 12,  147,
      164, 40,  93,  247, 77,  147, 68,  245, 151, 46,  20,  55,  136, 72,  12,  182, 38,  146, 102, 103, 40,  41,  76,
      133, 104, 78,  96,  71,  131, 113, 89,  217, 90,  46,  52,  29,  195, 199, 5,   196, 233, 16,  58,  249, 226, 83,
      119, 97,  140, 101, 151, 177, 137, 158, 249, 56,  84,  81,  24,  101, 17,  4,   214, 148, 12,  185, 54,  144, 121,
      44,  98,  240, 144, 219, 192, 7,   9,   138, 31,  148, 48,  77,  151, 253, 213, 200, 146, 207, 230, 108, 158, 245,
      72,  61,  23,  243, 217, 2,   217, 164, 136, 96,  144, 116, 91,  43,  23,  163, 0,   248, 119, 25,  68,  162, 167,
      205, 39,  37,  124, 171, 71,  109, 43,  134, 224, 97,  244, 82,  69,  38,  215, 134, 111, 56,  145, 22,  236, 146,
      23,  182, 172, 143, 4,   61,  85,  199, 236, 174, 109, 79,  215, 149, 108, 117, 250, 206, 79,  238, 100, 218, 21,
      207, 5,   1,   212, 233, 118, 222, 100, 77,  171, 61,  80,  164, 228, 177, 69,  52,  238, 249, 227, 19,  217, 174,
      156, 185, 55,  109, 203, 221, 148, 197, 186, 89,  207, 120, 134, 187, 180, 78,  173, 126, 22,  46,  138, 167, 228,
      211, 191, 1,   85,  63,  156, 61,  48,  198, 42,  158, 223, 172, 9,   240, 176, 194, 43,  26,  88,  152, 228, 241,
      12,  126, 237, 235, 224, 120, 144, 113, 135, 44,  142, 77,  129, 191, 125, 137, 44,  59,  39,  78,  168, 82,  107,
      6,   230, 50,  69,  43,  171, 102, 135, 220, 50,  39,  253, 90,  70,  85,  242, 243, 254, 217, 222, 51,  87,  70,
      65,  46,  165, 32,  233, 249, 251, 130, 217, 93,  94,  228, 175, 185, 253, 192, 177, 220, 85,  186, 79,  185, 55,
      42,  253, 70,  129, 103, 40,  174, 89,  56,  209, 42,  169, 21,  45,  126, 86,  123, 163, 125, 151, 179, 186, 22,
      140, 152, 30,  125, 132, 17,  13,  22,  32,  119, 76,  174, 137, 186, 168, 21,  57,  168, 91,  95,  52,  78,  195,
      26,  109, 13,  145, 107, 44,  143, 133, 180, 241, 187, 237, 139, 136, 20,  94,  129, 211, 145, 164, 165, 239, 250,
      195, 195, 208, 139, 169, 110, 73,  32,  205, 151, 124, 101, 186, 222, 88,  248, 231, 64,  61,  232, 225, 5,   99,
      61,  79,  160, 161, 170, 68,  244, 7,   68,  115, 230, 115, 48,  76,  18,  167, 106, 195, 81,  90,  86,  164, 207,
      120, 155, 101, 238, 186, 11,  106, 26,  235, 233, 13,  190, 56,  35,  59,  82,  112, 98,  174, 221, 184, 230, 132,
      131, 173, 123, 50,  45,  190, 69,  129, 220, 6,   99,  15,  63,  246, 178, 31,  107, 14,  28,  30,  19,  138, 239,
      43,  49,  73,  248, 220, 37,  177, 26,  231, 97,  224, 102, 57,  203, 190, 140, 225, 60,  27,  215, 162, 116, 187,
      164, 152, 126, 8,   171, 167, 206, 127, 17,  208, 41,  127, 232, 157, 154, 20,  216, 146, 108, 15,  224, 58,  7,
      77,  10,  20,  223, 182, 144, 159, 44,  108, 85,  91,  124, 71,  240, 21,  18,  13,  157, 172, 147, 206, 237, 89,
      205, 28,  57,  101, 114, 220, 188, 214, 147, 104, 76,  96,  138, 49,  241, 202, 29,  78,  97,  2,   92,  132, 126,
      175, 143, 92,  159, 243, 98,  168, 82,  236, 210, 162, 58,  70,  120, 240, 117, 46,  173, 82,  192, 179, 209, 14,
      217, 58,  227, 65,  185, 208, 103, 110, 138, 118, 235, 128, 100, 201, 15,  222, 123, 206, 204, 33,  162, 136, 168,
      34,  159, 18,  15,  43,  168, 234, 196, 187, 80,  136, 217, 224, 153, 216, 189, 124, 225, 91,  185, 62,  248, 127,
      144, 234, 9,   18,  233, 176, 10,  12,  60,  123, 176, 83,  109, 194, 74,  159, 70,  69,  37,  101, 209, 235, 163,
      185, 125, 166, 25,  182, 114, 26,  3,   173, 73,  95,  7,   2,   165, 75,  127, 29,  61,  181, 66,  124, 62,  23,
      169, 195, 75,  80,  102, 57,  208, 156, 205, 120, 4,   248, 161, 130, 118, 226, 33,  139, 176, 227, 236, 147, 180,
      74,  222, 51,  78,  134, 252, 102, 46,  128, 158, 219, 71,  178, 51,  34,  215, 44,  166, 156, 35,  100, 34,  241,
      162, 2,   14,  155, 7,   173, 220, 58,  39,  69,  242, 184, 21,  227, 236, 174, 57,  36,  53,  133, 20,  198, 184,
      118, 82,  121, 212, 88,  185, 129, 193, 251, 84,  8,   188, 193, 54,  107, 149, 109, 130, 110, 151, 71,  238, 225,
      2,   196, 44,  211, 108, 249, 136, 187, 206, 88,  241, 45,  48,  188, 113, 133, 172, 75,  245, 90,  245, 107, 79,
      12,  112, 108, 96,  188, 61,  233, 75,  55,  175, 253, 210, 28,  145, 125, 244, 77,  116, 130, 84,  208, 190, 221,
      92,  171, 217, 197, 10,  28,  54,  214, 153, 39,  114, 128, 57,  110, 149, 162, 227, 141, 228, 186, 103, 10,  198,
      74,  254, 107, 98,  214, 18,  87,  27,  129, 5,   244, 156, 152, 146, 108, 196, 179, 191, 239, 85,  199, 214, 194,
      164, 216, 6,   150, 238, 92,  154, 232, 164, 61,  155, 118, 174, 113, 189, 63,  135, 125, 24,  229, 119, 200, 7,
      41,  62,  63,  113, 160, 69,  147, 155, 103, 80,  27,  183, 228, 178, 62,  94,  16,  126, 238, 223, 44,  127, 8,
      47,  77,  237, 93,  170, 12,  163, 119, 154, 221, 100, 10,  94,  247, 9,   155, 185, 213, 218, 23,  42,  118, 119,
      159, 177, 11,  10,  111, 215, 170, 80,  51,  167, 127, 108, 87,  194, 207, 133, 168, 233, 4,   166, 175, 57,  78,
      65,  241, 6,   246, 77,  51,  195, 218, 147, 208, 33,  42,  146, 25,  86,  233, 201, 53,  74,  98,  88,  99,  162,
      170, 31,  72,  224, 126, 253, 129, 143, 136, 193, 17,  160, 133, 73,  192, 215, 62,  129, 212, 103, 3,   215, 92,
      98,  162, 130, 0,   182, 132, 14,  100, 172, 111, 66,  2,   218, 84,  253, 221, 53,  181, 86,  140, 79,  44,  8,
      146, 116, 94,  159, 135, 247, 64,  161, 236, 192, 239, 167, 77,  104, 137, 42,  135, 212, 111, 188, 132, 231, 104,
      3,   26,  107, 168, 172, 197, 101, 65,  51,  237, 88,  49,  154, 47,  195, 76,  136, 126, 109, 166, 50,  222, 150,
      102, 76,  63,  5,   253, 21,  137, 177, 77,  62,  96,  236, 182, 236, 119, 110, 166, 143, 58,  82,  218, 35,  68,
      224, 22,  48,  5,   32,  88,  98,  120, 93,  172, 138, 129, 148, 161, 25,  238, 254, 85,  217, 130, 235, 49,  122,
      145, 94,  166, 114, 141, 163, 154, 172, 180, 162, 82,  248, 223, 187, 63,  167, 106, 30,  218, 77,  184, 97,  42,
      120, 38,  141, 101, 100, 177, 176, 29,  147, 250, 84,  227, 241, 240, 220, 139, 217, 13,  110, 131, 133, 239, 88,
      209, 108, 176, 3,   184, 15,  179, 20,  234, 60,  163, 222, 69,  212, 229, 205, 254, 237, 45,  8,   58,  152, 230,
      20,  204, 188, 206, 79,  160, 35,  178, 88,  75,  246, 136, 215, 233, 239, 44,  214, 224, 219, 109, 79,  12,  117,
      170, 2,   234, 9,   212, 187, 94,  57,  81,  35,  155, 29,  106, 117, 171, 71,  156, 30,  212, 49,  169, 42,  33,
      85,  115, 9,   54,  161, 186, 55,  76,  93,  95,  191, 245, 206, 46,  215, 206, 250, 70,  92,  111, 199, 62,  251,
      72,  147, 64,  52,  152, 121, 16,  228, 108, 78,  215, 195, 33,  25,  218, 63,  31,  1,   164, 193, 28,  32,  39,
      171, 87,  40,  136, 234, 85,  170, 125, 16,  45,  33,  84,  90,  151, 21,  135, 204, 108, 203, 17,  67,  249, 29,
      10,  32,  8,   132, 55,  41,  215, 50,  53,  173, 138, 154, 70,  30,  186, 34,  163, 31,  1,   182, 47,  247, 74,
      12,  210, 242, 66,  47,  150, 106, 14,  17,  236, 127, 84,  220, 192, 242, 118, 56,  197, 219, 93,  188, 70,  123,
      252, 8,   63,  162, 63,  146, 111, 230, 242, 8,   141, 213, 192, 83,  25,  81,  174, 127, 209, 31,  155, 194, 90,
      237, 121, 129, 222, 232, 2,   240, 185, 38,  0,   85,  243, 247, 184, 132, 148, 200, 143, 55,  82,  205, 233, 234,
      138, 242, 29,  168, 83,  4,   65,  170, 155, 205, 96,  37,  93,  176, 133, 142, 221, 203, 162, 71,  20,  207, 60,
      71,  104, 220, 16,  197, 251, 74,  177, 35,  190, 219, 226, 22,  28,  100, 159, 128, 94,  244, 172, 101, 38,  17,
      107, 205, 104, 186, 67,  44,  218, 50,  144, 122, 63,  135, 24,  63,  77,  214, 183, 199, 173, 163, 27,  65,  183,
      203, 103, 21,  5,   51,  227, 77,  130, 146, 226, 36,  191, 247, 59,  126, 87,  236, 146, 67,  13,  242, 57,  82,
      59,  159, 236, 202, 212, 195, 211, 119, 171, 63,  121, 37,  44,  96,  36,  88,  211, 232, 67,  202, 113, 42,  25,
      32,  216, 99,  133, 76,  166, 115, 108, 130, 155, 227, 211, 99,  64,  228, 81,  155, 113, 220, 62,  122, 100, 21,
      252, 235, 244, 154, 23,  96,  77,  233, 56,  243, 240, 69,  57,  102, 15,  181, 40,  243, 30,  121, 119, 105, 90,
      210, 231, 247, 197, 7,   164, 8,   117, 28,  231, 186, 111, 57,  163, 46,  216, 103, 36,  79,  19,  49,  40,  119,
      68,  56,  157, 46,  53,  7,   51,  27,  33,  25,  192, 176, 43,  225, 141, 36,  73,  147, 27,  13,  174, 76,  72,
      80,  65,  215, 124, 10,  104, 11,  189, 215, 116, 168, 133, 31,  50,  136, 167, 88,  253, 24,  137, 94,  108, 75,
      227, 99,  2,   203, 202, 182, 244, 96,  163, 63,  42,  147, 91,  162, 15,  252, 144, 32,  110, 4,   128, 34,  185,
      38,  151, 1,   142, 81,  80,  180, 126, 30,  208, 56,  92,  105, 49,  38,  65,  246, 159, 18,  29,  155, 114, 86,
      108, 91,  144, 143, 107, 181, 125, 97,  99,  127, 179, 250, 181, 27,  218, 37,  132, 106, 177, 124, 59,  101, 77,
      127, 8,   189, 117, 67,  108, 221, 73,  170, 197, 183, 188, 175, 151, 116, 24,  242, 93,  147, 44,  208, 100, 168,
      50,  132, 21,  100, 50,  177, 13,  37,  204, 50,  247, 153, 245, 172, 225, 52,  95,  112, 142, 218, 58,  74,  192,
      1,   84,  213, 150, 40,  246, 38,  250, 35,  71,  127, 121, 14,  180, 90,  12,  111, 196, 120, 115, 156, 158, 110,
      218, 177, 223, 0,   237, 180, 37,  233, 246, 73,  166, 154, 166, 99,  218, 73,  178, 20,  194, 135, 84,  116, 157,
      41,  137, 159, 52,  249, 25,  51,  220, 9,   170, 74,  230, 240, 171, 95,  28,  201, 75,  173, 193, 12,  76,  9,
      30,  203, 250, 138, 126, 52,  160, 186, 118, 4,   250, 31,  97,  217, 219, 128, 60,  227, 83,  208, 160, 214, 61,
      90,  12,  228, 76,  64,  105, 176, 242, 176, 13,  239, 174, 28,  98,  128, 35,  246, 51,  191, 97,  56,  182, 196,
      77,  209, 84,  167, 89,  5,   102, 249, 211, 9,   42,  80,  27,  90,  184, 42,  223, 120, 214, 63,  216, 250, 196,
      144, 140, 166, 96,  53,  50,  208, 195, 52,  99,  157, 139, 200, 240, 196, 83,  119, 69,  142, 116, 116, 168, 120,
      168, 70,  124, 230, 39,  203, 65,  21,  100, 85,  233, 21,  147, 26,  135, 136, 164, 99,  165, 204, 110, 242, 200,
      234, 125, 125, 78,  173, 33,  50,  134, 251, 44,  31,  164, 106, 159, 211, 185, 240, 249, 31,  121, 85,  194, 102,
      121, 130, 250, 229, 203, 61,  33,  9,   47,  230, 207, 63,  200, 24,  216, 119, 137, 85,  232, 93,  207, 178, 228,
      165, 198, 201, 45,  145, 50,  92,  57,  47,  193, 221, 17,  124, 81,  229, 102, 52,  107, 247, 118, 17,  93,  75,
      66,  224, 214, 131, 110, 49,  161, 12,  206, 237, 116, 196, 161, 201, 130, 31,  158, 138, 192, 250, 177, 3,   5,
      84,  188, 51,  174, 17,  226, 167, 133, 62,  58,  192, 120, 196, 54,  114, 139, 106, 132, 230, 24,  182, 121, 29,
      44,  64,  82,  161, 227, 112, 177, 76,  109, 243, 248, 74,  230, 39,  78,  170, 214, 54,  92,  86,  249, 103, 190,
      70,  101, 30,  155, 144, 58,  167, 21,  103, 174, 248, 148, 236, 154, 107, 181, 22,  13,  101, 204, 111, 135, 23,
      42,  172, 15,  156, 102, 41,  176, 212, 192, 128, 189, 161, 41,  131, 33,  240, 106, 69,  105, 159, 124, 223, 90,
      145, 86,  21,  20,  57,  89,  46,  133, 15,  65,  65,  209, 167, 21,  153, 77,  125, 145, 227, 58,  24,  52,  206,
      118, 12,  205, 215, 218, 22,  19,  227, 74,  232, 136, 92,  87,  119, 0,   121, 12,  126, 248, 114, 244, 94,  118,
      135, 197, 54,  70,  205, 45,  106, 28,  42,  107, 105, 238, 28,  147, 128, 158, 198, 110, 17,  93,  7,   228, 8,
      43,  213, 134, 41,  84,  33,  221, 246, 203, 87,  39,  197, 7,   210, 4,   168, 42,  192, 209, 71,  103, 91,  213,
      181, 32,  27,  20,  191, 13,  115, 29,  55,  19,  186, 93,  169, 137, 247, 127, 161, 166, 177, 124, 191, 189, 68,
      173, 172, 196, 28,  33,  139, 153, 148, 69,  121, 156, 66,  113, 78,  242, 178, 59,  170, 239, 112, 54,  2,   196,
      107, 110, 208, 162, 71,  135, 126, 253, 35,  102, 41,  148, 172, 21,  170, 47,  184, 66,  224, 197, 207, 193, 146,
      248, 43,  5,   235, 232, 198, 66,  130, 196, 81,  133, 138, 81,  222, 204, 248, 59,  214, 203, 97,  147, 20,  159,
      10,  130, 54,  61,  32,  245, 208, 246, 87,  235, 12,  230, 104, 219, 121, 234, 253, 45,  112, 234, 219, 242, 172,
      146, 200, 219, 59,  199, 233, 202, 234, 159, 126, 8,   14,  188, 23,  130, 254, 114, 12,  86,  47,  184, 231, 136,
      230, 50,  83,  234, 152, 221, 145, 203, 252, 171, 190, 63,  199, 2,   68,  169, 247, 9,   105, 61,  26,  209, 9,
      65,  173, 251, 109, 36,  163, 26,  70,  97,  132, 7,   16,  153, 246, 189, 166, 247, 252, 79,  68,  229, 82,  191,
      160, 134, 214, 208, 29,  39,  26,  204, 211, 102, 207, 171, 5,   178, 88,  220, 107, 165, 136, 147, 248, 93,  130,
      61,  10,  132, 88,  130, 77,  161, 230, 237, 138, 125, 129, 154, 97,  113, 181, 61,  71,  223, 148, 233, 136, 126,
      152, 116, 67,  149, 82,  67,  200, 6,   20,  171, 12,  225, 173, 123, 143, 165, 90,  18,  226, 41,  55,  188, 147,
      230, 20,  131, 245, 108, 124, 149, 111, 35,  95,  95,  200, 182, 72,  8,   119, 93,  46,  133, 50,  241, 147, 175,
      193, 225, 85,  236, 239, 111, 55,  41,  187, 33,  148, 247, 194, 61,  79,  92,  82,  167, 244, 46,  254, 219, 28,
      19,  135, 98,  29,  10,  247, 213, 43,  99,  130, 221, 56,  89,  226, 118, 180, 139, 31,  213, 39,  72,  112, 208,
      180, 79,  35,  107, 164, 14,  245, 201, 78,  221, 111, 155, 58,  220, 68,  31,  27,  244, 79,  240, 196, 137, 60,
      139, 43,  171, 49,  50,  20,  227, 166, 119, 24,  139, 193, 115, 53,  72,  90,  1,   220, 237, 199, 78,  201, 249,
      27,  161, 195, 179, 215, 130, 183, 129, 69,  176, 49,  186, 145, 20,  210, 111, 7,   215, 12,  217, 69,  76,  177,
      104, 115, 253, 73,  67,  40,  110, 98,  80,  254, 184, 120, 250, 97,  250, 101, 214, 65,  59,  210, 251, 139, 210,
      104, 52,  29,  36,  48,  120, 224, 147, 235, 157, 253, 104, 249, 178, 31,  15,  9,   232, 128, 88,  38,  59,  156,
      64,  125, 128, 141, 106, 238, 30,  31,  149, 56,  135, 42,  198, 148, 220, 251, 4,   5,   224, 28,  212, 54,  47,
      212, 69,  3,   41,  244, 131, 80,  180, 1,   44,  211, 190, 170, 127, 208, 221, 21,  187, 178, 192, 51,  26,  203,
      80,  213, 153, 127, 245, 245, 106, 60,  84,  162, 159, 208, 213, 124, 135, 127, 69,  211, 44,  78,  179, 127, 196,
      244, 237, 237, 138, 34,  32,  123, 62,  228, 119, 146, 87,  167, 23,  147, 13,  115, 112, 141, 222, 100, 95,  143,
      50,  120, 65,  186, 70,  61,  203, 82,  216, 109, 211, 225, 63,  174, 247, 130, 10,  7,   214, 146, 58,  94,  195,
      173, 71,  185, 15,  83,  251, 15,  50,  194, 2,   217, 59,  229, 231, 11,  47,  85,  124, 112, 60,  152, 217, 214,
      133, 99,  10,  129, 211, 182, 17,  47,  248, 225, 103, 230, 188, 145, 199, 178, 102, 38,  142, 89,  17,  245, 99,
      22,  242, 37,  84,  242, 152, 108, 96,  95,  181, 229, 65,  58,  77,  114, 238, 65,  206, 71,  243, 94,  208, 251,
      6,   120, 124, 164, 234, 158, 123, 104, 214, 125, 150, 230, 183, 12,  215, 91,  247, 118, 195, 204, 106, 233, 215,
      197, 93,  226, 41,  220, 91,  106, 38,  188, 46,  75,  9,   111, 49,  40,  252, 21,  196, 160, 247, 3,   12,  142,
      49,  229, 44,  220, 2,   8,   195, 75,  66,  30,  181, 106, 177, 61,  58,  7,   100, 226, 60,  144, 226, 175, 253,
      17,  90,  75,  102, 157, 178, 183, 161, 220, 121, 72,  192, 96,  205, 27,  171, 128, 114, 253, 177, 38,  120, 54,
      194, 230, 47,  106, 3,   109, 157, 79,  204, 200, 201, 73,  130, 178, 148, 78,  218, 46,  210, 56,  232, 142, 101,
      75,  132, 142, 165, 241, 77,  28,  162, 69,  72,  58,  116, 98,  79,  118, 66,  253, 117, 100, 177, 252, 46,  249,
      49,  79,  108, 102, 204, 171, 101, 13,  179, 40,  218, 242, 173, 155, 221, 113, 61,  39,  124, 54,  64,  153, 163,
      225, 89,  62,  10,  250, 104, 171, 133, 171, 125, 95,  170, 114, 250, 10,  112, 47,  146, 197, 125, 192, 16,  88,
      187, 230, 241, 4,   69,  243, 4,   97,  195, 77,  197, 197, 25,  136, 111, 57,  96,  11,  123, 159, 108, 9,   191,
      178, 35,  164, 121, 122, 174, 74,  228, 250, 19,  224, 238, 184, 71,  133, 78,  145, 108, 165, 147, 111, 141, 234,
      101, 156, 208, 115, 96,  98,  252, 129, 33,  219, 169, 177, 180, 29,  208, 175, 114, 142, 119, 118, 98,  190, 250,
      163, 235, 127, 155, 38,  61,  104, 72,  70,  111, 114, 3,   59,  172, 120, 50,  249, 212, 43,  157, 136, 93,  166,
      208, 130, 136, 125, 184, 96,  188, 159, 220, 193, 9,   232, 37,  171, 105, 10,  71,  102, 45,  67,  26,  12,  183,
      122, 243, 88,  60,  248, 78,  4,   220, 167, 244, 123, 174, 53,  238, 208, 81,  161, 100, 32,  204, 70,  73,  173,
      184, 63,  161, 32,  84,  106, 222, 208, 126, 210, 147, 117, 77,  215, 74,  186, 118, 52,  9,   5,   171, 16,  252,
      241, 91,  168, 127, 121, 246, 221, 13,  12,  202, 167, 143, 232, 167, 134, 106, 67,  31,  186, 198, 134, 59,  39,
      220, 107, 96,  104, 187, 89,  201, 185, 19,  119, 11,  73,  21,  233, 101, 60,  116, 140, 175, 17,  46,  248, 151,
      149, 26,  38,  114, 189, 220, 86,  183, 222, 117, 62,  65,  114, 220, 237, 216, 219, 27,  15,  172, 110, 237, 130,
      117, 80,  166, 135, 187, 99,  226, 59,  9,   140, 137, 164, 167, 241, 172, 236, 216, 149, 74,  177, 185, 125, 190,
      241, 33,  6,   23,  53,  231, 77,  63,  32,  35,  235, 12,  120, 228, 152, 4,   226, 229, 190, 176, 181, 207, 45,
      58,  171, 25,  25,  178, 222, 23,  145, 38,  108, 122, 139, 209, 154, 249, 237, 97,  253, 234, 13,  131, 158, 6,
      63,  32,  254, 93,  82,  8,   8,   2,   138, 164, 92,  95,  48,  3,   95,  105, 126, 167, 123, 225, 224, 170, 233,
      198, 52,  122, 175, 173, 172, 6,   91,  248, 179, 92,  148, 177, 123, 148, 91,  98,  191, 205, 182, 187, 146, 191,
      16,  85,  162, 86,  39,  136, 226, 251, 10,  70,  1,   183, 139, 140, 23,  182, 185, 14,  164, 232, 191, 15,  213,
      128, 185, 215, 148, 46,  154, 47,  237, 220, 249, 168, 226, 105,
  };
  uint8_t data[18 * 2 * 16 * 32] = {
      240, 211, 76,  44,  3,   198, 126, 60,  124, 45,  242, 246, 105, 69,  129, 230, 218, 122, 49,  224, 200, 134, 254,
      171, 166, 50,  153, 189, 241, 155, 136, 221, 34,  3,   160, 87,  194, 251, 8,   106, 96,  154, 115, 84,  233, 171,
      126, 232, 112, 168, 38,  89,  60,  3,   211, 237, 221, 250, 55,  118, 68,  169, 228, 5,   200, 119, 32,  196, 177,
      237, 193, 163, 171, 80,  229, 75,  105, 231, 236, 81,  172, 7,   80,  235, 159, 6,   53,  155, 223, 234, 108, 221,
      56,  254, 252, 53,  65,  128, 142, 111, 170, 153, 137, 253, 215, 78,  176, 88,  240, 239, 120, 229, 180, 38,  62,
      174, 97,  225, 222, 116, 213, 4,   117, 144, 196, 62,  60,  106, 93,  57,  103, 4,   59,  55,  206, 203, 62,  87,
      21,  26,  29,  163, 57,  189, 180, 1,   54,  219, 217, 209, 194, 108, 250, 206, 10,  175, 77,  63,  144, 38,  210,
      192, 147, 66,  46,  216, 61,  86,  118, 198, 228, 78,  96,  15,  142, 223, 154, 176, 236, 234, 82,  111, 9,   31,
      14,  127, 165, 170, 94,  199, 155, 209, 146, 40,  202, 175, 221, 230, 131, 88,  130, 30,  201, 55,  68,  59,  188,
      245, 79,  133, 116, 86,  241, 198, 88,  42,  172, 66,  254, 97,  131, 208, 223, 204, 23,  247, 30,  186, 182, 120,
      33,  228, 204, 220, 20,  167, 186, 45,  118, 9,   61,  235, 137, 234, 99,  185, 20,  249, 211, 85,  186, 59,  179,
      48,  253, 156, 48,  68,  124, 54,  17,  193, 248, 37,  216, 204, 47,  230, 113, 150, 111, 68,  236, 227, 10,  7,
      55,  91,  16,  88,  129, 123, 246, 217, 119, 116, 229, 27,  118, 51,  60,  179, 178, 67,  235, 162, 88,  30,  35,
      222, 66,  183, 239, 82,  239, 112, 66,  236, 105, 46,  6,   62,  46,  29,  114, 22,  45,  73,  181, 30,  162, 168,
      216, 163, 237, 107, 196, 118, 14,  174, 191, 181, 111, 104, 39,  42,  16,  33,  241, 29,  36,  63,  190, 17,  89,
      184, 70,  81,  5,   151, 206, 57,  0,   91,  244, 77,  121, 126, 215, 114, 34,  84,  10,  10,  179, 38,  12,  69,
      132, 148, 159, 161, 90,  168, 176, 65,  78,  76,  4,   14,  191, 60,  155, 171, 77,  72,  119, 106, 208, 241, 143,
      123, 113, 234, 168, 28,  211, 249, 125, 154, 44,  6,   49,  136, 213, 253, 206, 164, 60,  6,   124, 99,  93,  227,
      55,  3,   177, 113, 130, 30,  234, 75,  227, 254, 55,  73,  244, 196, 233, 120, 91,  111, 151, 60,  124, 253, 123,
      15,  30,  170, 55,  225, 195, 12,  228, 173, 175, 225, 40,  249, 76,  36,  163, 74,  111, 151, 203, 82,  252, 204,
      194, 215, 199, 215, 251, 174, 3,   24,  242, 249, 160, 215, 72,  218, 62,  117, 102, 169, 153, 32,  172, 224, 251,
      78,  22,  196, 155, 225, 100, 78,  99,  198, 55,  125, 46,  132, 6,   67,  33,  195, 178, 174, 86,  208, 153, 117,
      63,  73,  157, 93,  64,  104, 108, 199, 145, 100, 161, 132, 242, 165, 96,  18,  144, 243, 207, 5,   57,  31,  178,
      78,  81,  173, 2,   160, 95,  51,  135, 181, 40,  229, 221, 17,  52,  9,   186, 2,   114, 215, 77,  150, 171, 210,
      139, 174, 222, 241, 135, 42,  220, 244, 111, 159, 160, 119, 99,  146, 42,  5,   151, 76,  188, 206, 207, 39,  241,
      1,   71,  90,  178, 48,  214, 205, 22,  245, 156, 148, 131, 180, 69,  126, 17,  85,  97,  227, 224, 15,  88,  208,
      166, 40,  206, 71,  66,  236, 84,  223, 185, 35,  185, 107, 144, 38,  152, 146, 12,  122, 26,  254, 56,  23,  222,
      229, 23,  112, 14,  192, 223, 221, 172, 161, 145, 225, 122, 227, 167, 28,  58,  101, 50,  53,  179, 191, 90,  181,
      27,  170, 110, 231, 45,  96,  131, 174, 185, 34,  95,  196, 58,  100, 218, 224, 20,  217, 33,  243, 186, 52,  147,
      12,  133, 90,  16,  151, 134, 172, 79,  100, 222, 69,  185, 207, 46,  156, 240, 12,  44,  39,  43,  70,  70,  217,
      220, 129, 21,  186, 22,  57,  13,  237, 165, 73,  248, 79,  230, 90,  17,  146, 223, 71,  237, 214, 29,  92,  58,
      179, 103, 222, 159, 136, 185, 176, 74,  185, 73,  181, 102, 130, 180, 128, 166, 155, 21,  198, 193, 109, 44,  45,
      164, 61,  24,  160, 178, 8,   154, 213, 10,  87,  57,  218, 201, 72,  160, 3,   14,  221, 116, 239, 188, 197, 195,
      102, 62,  58,  32,  161, 122, 180, 30,  101, 71,  115, 191, 111, 121, 218, 171, 45,  203, 55,  30,  74,  17,  197,
      162, 192, 194, 185, 200, 29,  37,  20,  74,  159, 205, 63,  163, 227, 126, 59,  63,  67,  27,  73,  215, 133, 94,
      244, 252, 151, 56,  29,  181, 218, 127, 197, 147, 50,  172, 192, 35,  196, 0,   166, 116, 220, 201, 250, 31,  160,
      176, 35,  209, 42,  102, 47,  85,  45,  124, 93,  12,  158, 51,  120, 182, 177, 170, 101, 248, 204, 41,  33,  145,
      65,  227, 218, 119, 118, 42,  110, 41,  7,   91,  29,  161, 196, 121, 146, 126, 86,  152, 2,   9,   97,  252, 208,
      5,   93,  230, 61,  67,  225, 55,  149, 119, 225, 148, 46,  247, 118, 180, 221, 79,  50,  53,  249, 69,  215, 80,
      221, 213, 159, 78,  32,  147, 50,  109, 214, 83,  231, 214, 248, 38,  84,  58,  208, 104, 247, 226, 9,   3,   195,
      7,   90,  72,  177, 109, 148, 75,  72,  9,   228, 145, 208, 177, 98,  104, 50,  130, 9,   67,  78,  208, 184, 210,
      9,   121, 167, 39,  19,  140, 114, 114, 237, 59,  216, 6,   12,  181, 23,  237, 32,  0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   195, 45,  243, 108, 74,  253, 186, 6,   115, 25,  199,
      9,   146, 30,  225, 40,  249, 192, 134, 87,  202, 201, 217, 196, 136, 151, 244, 19,  214, 179, 117, 22,  10,  209,
      89,  144, 31,  39,  198, 26,  19,  14,  81,  71,  66,  209, 224, 80,  236, 197, 55,  103, 58,  8,   32,  17,  52,
      17,  12,  78,  204, 143, 26,  165, 52,  123, 210, 89,  6,   208, 109, 71,  114, 214, 52,  118, 200, 85,  48,  166,
      177, 50,  157, 97,  238, 158, 31,  125, 123, 33,  251, 242, 116, 41,  40,  97,  55,  31,  248, 162, 55,  196, 82,
      148, 225, 143, 108, 148, 210, 234, 27,  37,  102, 118, 181, 116, 87,  38,  130, 193, 137, 20,  192, 239, 37,  193,
      119, 184, 74,  201, 215, 111, 115, 10,  122, 169, 9,   136, 9,   189, 235, 27,  246, 78,  157, 215, 72,  29,  22,
      3,   212, 159, 26,  139, 238, 184, 200, 152, 179, 238, 69,  143, 21,  223, 111, 198, 160, 130, 203, 121, 206, 161,
      46,  16,  86,  94,  15,  251, 22,  211, 61,  96,  110, 95,  155, 30,  81,  68,  191, 15,  152, 20,  56,  219, 124,
      243, 185, 8,   38,  75,  88,  109, 114, 49,  4,   54,  203, 139, 213, 216, 1,   186, 41,  138, 240, 12,  52,  108,
      64,  86,  142, 32,  165, 121, 185, 19,  221, 163, 117, 118, 198, 19,  73,  53,  111, 89,  96,  27,  57,  182, 60,
      129, 39,  144, 125, 181, 91,  81,  168, 221, 219, 133, 120, 73,  31,  20,  49,  237, 133, 223, 127, 158, 228, 209,
      160, 106, 216, 4,   20,  159, 230, 199, 116, 131, 72,  48,  140, 5,   197, 29,  192, 30,  204, 31,  41,  241, 35,
      143, 237, 59,  27,  202, 43,  238, 24,  32,  54,  31,  243, 223, 198, 199, 241, 211, 98,  46,  177, 157, 185, 42,
      166, 238, 242, 25,  223, 91,  213, 63,  61,  99,  201, 46,  79,  90,  117, 96,  65,  132, 10,  204, 143, 252, 22,
      178, 197, 149, 97,  72,  25,  173, 160, 129, 123, 106, 13,  69,  188, 125, 252, 228, 218, 234, 142, 77,  194, 15,
      65,  14,  178, 126, 241, 80,  233, 248, 35,  152, 96,  191, 141, 219, 78,  204, 207, 38,  95,  249, 201, 201, 97,
      165, 155, 150, 192, 223, 72,  164, 33,  73,  171, 156, 150, 251, 239, 57,  209, 43,  98,  67,  239, 92,  126, 139,
      75,  55,  132, 28,  0,   76,  96,  80,  37,  164, 247, 6,   30,  231, 182, 96,  40,  15,  104, 95,  224, 216, 78,
      159, 32,  8,   31,  81,  113, 84,  134, 0,   148, 109, 47,  87,  158, 122, 208, 180, 25,  37,  117, 225, 129, 22,
      29,  92,  252, 51,  43,  92,  148, 153, 149, 216, 33,  18,  115, 149, 43,  197, 164, 251, 96,  94,  90,  40,  42,
      16,  41,  241, 95,  173, 14,  95,  189, 62,  49,  240, 182, 57,  153, 70,  117, 63,  247, 209, 97,  119, 108, 68,
      200, 8,   254, 108, 190, 1,   184, 222, 60,  46,  193, 66,  50,  44,  225, 195, 186, 190, 246, 129, 157, 180, 140,
      122, 175, 249, 20,  28,  120, 65,  80,  14,  38,  51,  229, 66,  189, 174, 66,  30,  94,  164, 101, 32,  210, 78,
      121, 134, 200, 8,   238, 131, 139, 246, 219, 45,  251, 141, 65,  141, 145, 124, 20,  96,  225, 94,  10,  144, 239,
      83,  146, 126, 243, 80,  96,  125, 232, 222, 0,   152, 32,  107, 154, 39,  184, 210, 244, 146, 185, 71,  174, 16,
      149, 78,  93,  77,  197, 213, 59,  31,  119, 16,  211, 22,  201, 12,  12,  93,  8,   51,  206, 150, 213, 152, 34,
      63,  82,  106, 28,  18,  66,  251, 162, 69,  30,  133, 191, 125, 227, 220, 62,  63,  202, 64,  230, 156, 91,  82,
      156, 44,  105, 48,  66,  117, 1,   126, 178, 216, 88,  96,  166, 106, 247, 115, 241, 15,  136, 110, 75,  119, 85,
      157, 92,  40,  18,  33,  166, 51,  195, 251, 51,  117, 68,  182, 104, 46,  58,  177, 111, 143, 215, 218, 100, 52,
      55,  120, 138, 61,  176, 126, 142, 78,  171, 85,  65,  93,  85,  103, 45,  22,  32,  168, 145, 139, 239, 32,  231,
      160, 115, 86,  208, 108, 182, 240, 89,  147, 78,  159, 70,  209, 110, 123, 18,  217, 62,  10,  159, 125, 95,  66,
      57,  118, 74,  158, 156, 7,   227, 184, 251, 130, 211, 48,  216, 155, 229, 69,  62,  47,  221, 218, 111, 194, 78,
      147, 53,  224, 133, 215, 100, 53,  94,  167, 132, 101, 76,  76,  96,  82,  88,  25,  145, 162, 218, 100, 156, 9,
      245, 199, 16,  63,  87,  33,  17,  41,  186, 12,  236, 219, 63,  83,  90,  38,  143, 168, 93,  207, 177, 128, 104,
      135, 27,  77,  87,  159, 121, 202, 33,  147, 124, 31,  108, 96,  144, 1,   92,  155, 125, 37,  8,   183, 44,  177,
      225, 1,   71,  120, 237, 99,  157, 9,   75,  138, 97,  214, 242, 246, 84,  80,  121, 230, 232, 145, 81,  75,  206,
      138, 174, 196, 121, 76,  214, 30,  148, 111, 176, 240, 86,  65,  58,  22,  0,   135, 107, 17,  84,  210, 213, 55,
      166, 28,  234, 200, 8,   47,  221, 68,  65,  132, 79,  29,  178, 130, 237, 71,  149, 173, 188, 111, 93,  85,  247,
      254, 213, 180, 79,  111, 240, 20,  72,  27,  176, 12,  180, 21,  153, 208, 179, 123, 13,  120, 211, 146, 251, 11,
      133, 225, 142, 242, 77,  251, 214, 118, 41,  111, 12,  75,  100, 75,  192, 38,  157, 78,  132, 58,  191, 250, 108,
      115, 180, 190, 72,  249, 136, 232, 26,  4,   81,  63,  147, 5,   246, 134, 220, 119, 175, 189, 220, 130, 9,   149,
      23,  120, 207, 145, 97,  254, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   239, 33,  215, 73,  18,  26,  154, 53,  220, 228, 109, 64,  3,   247, 135, 99,  72,  211, 90,  45,  197, 240,
      229, 252, 138, 40,  230, 166, 70,  86,  164, 11,  209, 71,  30,  21,  188, 151, 34,  224, 120, 215, 46,  159, 156,
      127, 141, 165, 216, 52,  164, 113, 235, 75,  76,  87,  57,  114, 131, 52,  188, 5,   181, 182, 83,  224, 86,  7,
      223, 117, 214, 13,  95,  8,   201, 162, 156, 9,   5,   110, 8,   6,   75,  113, 200, 134, 189, 92,  116, 170, 140,
      23,  57,  216, 219, 246, 213, 196, 6,   222, 136, 150, 171, 102, 49,  72,  7,   184, 135, 133, 119, 52,  11,  179,
      108, 148, 130, 222, 242, 200, 63,  177, 49,  236, 246, 47,  66,  8,   99,  94,  20,  121, 136, 135, 96,  59,  172,
      231, 208, 220, 185, 123, 157, 141, 97,  47,  228, 146, 47,  30,  24,  156, 107, 241, 26,  15,  189, 224, 37,  133,
      165, 236, 0,   47,  67,  79,  206, 166, 225, 20,  124, 172, 109, 147, 240, 221, 46,  123, 240, 194, 86,  152, 158,
      99,  153, 217, 68,  12,  89,  179, 173, 131, 221, 142, 245, 117, 250, 94,  93,  132, 180, 141, 40,  143, 91,  30,
      40,  106, 17,  186, 178, 59,  114, 91,  237, 154, 210, 4,   205, 52,  69,  46,  98,  75,  182, 193, 10,  217, 105,
      88,  11,  34,  9,   87,  16,  28,  99,  55,  205, 32,  240, 194, 158, 180, 205, 121, 34,  250, 240, 91,  28,  173,
      137, 69,  67,  243, 251, 130, 125, 233, 167, 120, 2,   20,  168, 101, 254, 67,  197, 179, 187, 198, 13,  251, 73,
      207, 3,   22,  82,  203, 158, 132, 110, 248, 206, 98,  69,  247, 122, 228, 162, 191, 99,  243, 182, 158, 128, 165,
      122, 143, 20,  38,  202, 211, 64,  90,  204, 183, 215, 172, 27,  84,  67,  205, 105, 167, 8,   78,  223, 143, 55,
      118, 175, 202, 228, 61,  252, 74,  45,  18,  165, 148, 14,  161, 162, 138, 229, 49,  216, 31,  100, 184, 123, 41,
      75,  41,  51,  150, 181, 94,  92,  154, 187, 89,  60,  100, 211, 186, 36,  198, 61,  26,  35,  165, 186, 208, 78,
      240, 169, 210, 16,  150, 155, 137, 60,  36,  1,   234, 13,  76,  58,  248, 50,  32,  229, 121, 71,  195, 195, 172,
      93,  237, 40,  69,  205, 87,  229, 222, 59,  28,  72,  16,  58,  223, 23,  214, 111, 246, 46,  157, 0,   146, 101,
      207, 14,  17,  233, 105, 100, 160, 202, 5,   234, 154, 108, 56,  131, 241, 79,  251, 130, 75,  238, 33,  32,  129,
      180, 45,  11,  53,  113, 213, 65,  13,  45,  42,  204, 213, 239, 54,  3,   88,  10,  148, 79,  73,  58,  151, 23,
      79,  126, 9,   40,  46,  222, 199, 247, 226, 81,  47,  192, 200, 213, 35,  63,  114, 162, 207, 247, 72,  201, 245,
      132, 12,  44,  195, 19,  128, 227, 126, 35,  203, 3,   233, 130, 4,   94,  194, 131, 38,  63,  123, 79,  234, 243,
      199, 142, 75,  203, 131, 195, 114, 146, 167, 165, 121, 138, 176, 45,  84,  146, 71,  17,  194, 117, 171, 140, 233,
      252, 245, 13,  73,  189, 238, 52,  136, 51,  97,  241, 119, 150, 209, 110, 198, 186, 210, 98,  110, 195, 98,  251,
      16,  253, 237, 231, 244, 237, 152, 44,  94,  214, 50,  30,  220, 127, 52,  208, 107, 246, 214, 189, 153, 192, 176,
      176, 77,  86,  120, 153, 125, 226, 37,  177, 70,  153, 209, 2,   99,  78,  139, 174, 136, 25,  28,  223, 73,  26,
      195, 201, 198, 48,  76,  160, 146, 37,  107, 151, 220, 241, 125, 64,  163, 198, 238, 60,  58,  97,  241, 43,  11,
      86,  173, 40,  163, 231, 27,  105, 196, 44,  205, 40,  238, 174, 69,  228, 171, 0,   75,  194, 233, 100, 23,  216,
      49,  115, 179, 215, 252, 59,  135, 254, 56,  165, 209, 127, 103, 240, 178, 40,  138, 122, 35,  207, 109, 98,  12,
      245, 38,  78,  52,  249, 137, 56,  4,   149, 21,  77,  238, 109, 134, 167, 48,  194, 237, 77,  168, 102, 242, 46,
      94,  186, 40,  126, 140, 119, 91,  107, 136, 250, 205, 64,  161, 48,  164, 41,  15,  144, 21,  205, 104, 97,  243,
      120, 14,  138, 206, 15,  224, 248, 12,  70,  166, 7,   124, 208, 226, 222, 219, 35,  173, 236, 118, 84,  110, 108,
      188, 6,   118, 119, 147, 94,  77,  100, 23,  215, 207, 65,  208, 74,  126, 73,  92,  172, 181, 159, 77,  231, 65,
      136, 63,  236, 218, 98,  238, 94,  82,  77,  177, 30,  54,  206, 62,  228, 31,  40,  84,  67,  215, 34,  210, 127,
      93,  146, 25,  237, 0,   55,  83,  176, 60,  35,  94,  186, 24,  65,  130, 236, 236, 37,  113, 25,  27,  10,  50,
      75,  163, 32,  142, 105, 114, 97,  132, 189, 188, 208, 13,  52,  84,  127, 115, 180, 218, 41,  170, 173, 126, 217,
      61,  176, 51,  167, 27,  130, 207, 93,  50,  229, 66,  237, 192, 17,  54,  29,  109, 54,  155, 236, 135, 212, 106,
      21,  209, 24,  19,  121, 197, 105, 93,  33,  4,   181, 19,  177, 33,  253, 219, 14,  184, 145, 169, 173, 7,   68,
      39,  88,  61,  138, 223, 82,  123, 147, 35,  46,  236, 194, 219, 43,  139, 35,  155, 196, 227, 90,  161, 49,  217,
      1,   187, 127, 213, 127, 62,  115, 65,  225, 58,  15,  47,  99,  103, 89,  152, 114, 6,   125, 195, 220, 49,  142,
      66,  226, 102, 96,  120, 79,  218, 142, 116, 178, 19,  165, 96,  29,  250, 70,  212, 125, 37,  12,  220, 161, 161,
      117, 171, 221, 106, 107, 11,  17,  220, 237, 117, 140, 74,  53,  228, 45,  38,  253, 93,  0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   13,  95,  65,  72,  163, 28,  217, 68,  31,  90,
      114, 236, 33,  210, 76,  143, 181, 176, 83,  45,  61,  217, 151, 37,  43,  240, 105, 149, 12,  161, 28,  254, 242,
      186, 252, 196, 242, 249, 171, 175, 91,  194, 212, 240, 203, 87,  148, 40,  117, 99,  214, 109, 40,  97,  82,  130,
      225, 223, 16,  199, 203, 94,  107, 99,  225, 225, 119, 55,  31,  112, 201, 5,   29,  206, 203, 92,  122, 22,  208,
      4,   31,  252, 129, 195, 77,  238, 178, 146, 171, 44,  23,  238, 54,  103, 253, 241, 36,  78,  45,  194, 141, 173,
      68,  77,  253, 96,  34,  11,  213, 104, 165, 19,  159, 71,  150, 175, 177, 152, 20,  120, 4,   62,  236, 202, 185,
      210, 252, 64,  74,  129, 176, 66,  21,  31,  225, 129, 230, 240, 241, 175, 210, 155, 251, 221, 119, 116, 156, 144,
      207, 240, 170, 210, 225, 207, 123, 42,  6,   35,  156, 196, 235, 45,  52,  8,   94,  47,  98,  90,  158, 125, 58,
      172, 152, 116, 230, 29,  29,  126, 76,  164, 64,  150, 250, 193, 207, 98,  233, 61,  97,  119, 159, 124, 118, 185,
      250, 217, 125, 247, 22,  77,  18,  2,   17,  181, 225, 67,  210, 3,   53,  188, 142, 161, 35,  29,  88,  105, 154,
      45,  84,  189, 210, 123, 110, 120, 102, 64,  242, 25,  159, 175, 8,   228, 81,  74,  102, 87,  16,  123, 7,   214,
      96,  129, 86,  213, 57,  42,  212, 32,  103, 167, 102, 145, 177, 166, 71,  141, 144, 187, 87,  245, 139, 240, 21,
      81,  250, 13,  113, 223, 160, 113, 2,   38,  224, 34,  50,  41,  201, 233, 175, 193, 100, 170, 177, 14,  110, 222,
      70,  243, 126, 77,  153, 62,  30,  112, 127, 53,  90,  27,  152, 245, 245, 46,  96,  104, 210, 212, 205, 192, 235,
      47,  11,  181, 92,  190, 213, 37,  37,  65,  9,   155, 150, 243, 79,  68,  244, 74,  62,  75,  24,  184, 38,  245,
      156, 184, 144, 40,  112, 232, 249, 160, 56,  172, 4,   29,  254, 70,  247, 156, 48,  91,  182, 38,  251, 132, 36,
      184, 126, 54,  253, 248, 223, 9,   65,  65,  179, 239, 177, 82,  254, 61,  59,  4,   240, 91,  142, 215, 175, 41,
      33,  15,  247, 124, 108, 17,  87,  30,  91,  162, 28,  179, 164, 154, 185, 147, 109, 2,   55,  122, 139, 232, 68,
      109, 120, 187, 245, 82,  6,   45,  196, 158, 247, 238, 125, 151, 173, 15,  162, 101, 38,  11,  202, 99,  69,  186,
      217, 185, 76,  46,  44,  253, 56,  8,   137, 234, 210, 103, 220, 87,  204, 135, 205, 184, 161, 58,  183, 6,   102,
      163, 58,  193, 185, 113, 172, 14,  44,  200, 216, 153, 203, 112, 51,  61,  91,  179, 25,  10,  57,  205, 242, 44,
      25,  132, 229, 95,  230, 195, 203, 134, 90,  71,  251, 216, 48,  126, 143, 247, 15,  244, 44,  56,  152, 224, 105,
      215, 24,  94,  110, 105, 218, 85,  168, 111, 234, 189, 71,  56,  241, 231, 31,  221, 217, 243, 54,  93,  142, 52,
      191, 213, 54,  191, 206, 106, 169, 168, 221, 93,  243, 73,  92,  203, 189, 197, 188, 225, 191, 252, 35,  243, 163,
      221, 126, 166, 21,  76,  10,  107, 221, 180, 50,  126, 188, 164, 164, 199, 183, 11,  182, 181, 58,  189, 210, 100,
      92,  204, 237, 52,  134, 50,  111, 161, 10,  141, 149, 192, 172, 42,  87,  166, 245, 252, 214, 11,  128, 164, 22,
      145, 158, 113, 194, 28,  233, 62,  85,  214, 253, 82,  210, 32,  205, 151, 248, 47,  30,  51,  165, 192, 214, 93,
      196, 218, 83,  164, 147, 154, 10,  119, 0,   156, 21,  229, 40,  133, 240, 128, 70,  109, 77,  218, 240, 32,  41,
      150, 125, 231, 240, 88,  90,  238, 54,  167, 69,  16,  90,  149, 16,  106, 26,  165, 224, 193, 70,  220, 158, 183,
      75,  163, 9,   57,  117, 248, 158, 66,  206, 104, 210, 147, 151, 146, 96,  29,  226, 158, 145, 61,  108, 207, 171,
      249, 203, 187, 228, 198, 43,  235, 129, 82,  135, 243, 165, 228, 79,  174, 86,  152, 76,  95,  44,  20,  195, 151,
      61,  170, 76,  207, 127, 224, 159, 18,  28,  76,  2,   98,  240, 217, 118, 204, 43,  153, 127, 60,  69,  25,  2,
      62,  156, 33,  51,  71,  100, 220, 174, 82,  8,   151, 249, 113, 150, 174, 200, 155, 16,  237, 10,  42,  170, 133,
      241, 33,  98,  253, 192, 98,  111, 86,  135, 104, 104, 51,  126, 5,   182, 35,  120, 254, 166, 248, 184, 176, 212,
      161, 46,  251, 108, 1,   136, 90,  249, 50,  163, 242, 84,  188, 111, 81,  121, 85,  82,  4,   133, 71,  190, 244,
      67,  63,  183, 146, 146, 202, 44,  42,  192, 13,  104, 136, 227, 63,  110, 81,  49,  86,  99,  106, 117, 175, 100,
      37,  95,  213, 121, 156, 37,  73,  137, 74,  193, 52,  219, 90,  13,  191, 89,  90,  250, 63,  207, 92,  102, 224,
      116, 250, 226, 114, 44,  243, 92,  144, 154, 23,  192, 69,  20,  167, 52,  227, 12,  54,  238, 7,   221, 168, 32,
      186, 124, 139, 45,  59,  188, 231, 203, 247, 251, 16,  229, 26,  220, 10,  125, 148, 67,  143, 230, 95,  133, 239,
      61,  28,  99,  31,  107, 58,  137, 166, 1,   195, 186, 69,  44,  178, 245, 124, 196, 228, 112, 186, 63,  94,  50,
      85,  10,  230, 17,  108, 238, 120, 150, 204, 91,  40,  49,  30,  60,  157, 205, 54,  204, 46,  165, 98,  35,  224,
      147, 98,  208, 163, 111, 43,  184, 171, 183, 167, 100, 32,  77,  16,  13,  67,  96,  73,  46,  86,  216, 80,  137,
      177, 231, 152, 58,  108, 234, 82,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   48,  18,  119, 170, 51,  246, 252, 48,  249, 136, 156, 157, 210, 220, 33,  131, 244, 223, 212, 54,  220,
      48,  151, 139, 186, 35,  168, 201, 232, 67,  118, 170, 109, 200, 101, 217, 180, 146, 154, 45,  21,  170, 245, 156,
      54,  118, 184, 130, 166, 151, 202, 2,   49,  114, 49,  94,  168, 66,  213, 222, 221, 26,  64,  106, 85,  196, 26,
      167, 127, 250, 41,  125, 207, 104, 173, 2,   161, 111, 236, 194, 42,  59,  210, 149, 99,  244, 181, 179, 57,  162,
      124, 203, 7,   32,  18,  63,  112, 249, 183, 33,  29,  80,  106, 175, 226, 109, 17,  42,  12,  72,  204, 132, 66,
      112, 224, 187, 194, 199, 25,  122, 6,   42,  241, 75,  181, 26,  75,  174, 16,  208, 41,  190, 87,  97,  131, 5,
      86,  168, 203, 187, 177, 207, 231, 108, 213, 103, 194, 136, 244, 90,  154, 97,  253, 94,  108, 87,  204, 108, 128,
      188, 216, 229, 98,  146, 55,  34,  80,  201, 151, 142, 108, 21,  28,  208, 32,  248, 207, 232, 46,  230, 44,  234,
      169, 37,  131, 181, 65,  6,   112, 136, 129, 66,  86,  14,  108, 4,   245, 132, 230, 23,  244, 11,  225, 104, 232,
      129, 7,   237, 73,  251, 172, 77,  156, 206, 209, 165, 190, 191, 235, 80,  45,  37,  34,  24,  66,  56,  35,  17,
      112, 157, 241, 125, 126, 24,  212, 180, 231, 212, 156, 32,  105, 233, 2,   36,  170, 229, 197, 192, 117, 233, 154,
      133, 241, 192, 150, 127, 51,  13,  192, 249, 140, 89,  184, 131, 208, 235, 157, 76,  29,  65,  246, 120, 57,  229,
      20,  138, 105, 92,  105, 199, 248, 32,  37,  227, 224, 69,  21,  63,  218, 171, 219, 100, 238, 28,  2,   214, 159,
      70,  216, 210, 70,  85,  140, 217, 33,  22,  99,  123, 34,  34,  22,  224, 68,  30,  118, 135, 51,  8,   99,  198,
      108, 12,  65,  127, 163, 132, 60,  88,  95,  240, 139, 137, 200, 192, 107, 194, 70,  172, 120, 34,  179, 224, 137,
      55,  163, 136, 246, 7,   48,  178, 14,  62,  16,  141, 93,  200, 179, 131, 42,  147, 92,  80,  79,  91,  61,  57,
      232, 9,   116, 248, 102, 25,  61,  125, 121, 171, 50,  159, 181, 125, 90,  122, 43,  134, 137, 146, 81,  110, 204,
      71,  219, 88,  184, 158, 112, 218, 118, 121, 79,  200, 7,   240, 64,  233, 41,  220, 3,   0,   7,   64,  150, 76,
      171, 84,  26,  217, 146, 32,  128, 33,  172, 230, 168, 40,  154, 159, 19,  197, 160, 90,  211, 208, 10,  190, 144,
      17,  5,   40,  167, 217, 238, 222, 162, 172, 45,  31,  105, 137, 246, 46,  69,  136, 26,  40,  168, 193, 191, 194,
      28,  216, 177, 56,  44,  226, 2,   136, 0,   44,  157, 115, 17,  36,  19,  246, 21,  148, 202, 235, 33,  80,  194,
      160, 6,   161, 252, 125, 132, 216, 175, 152, 244, 30,  30,  159, 70,  232, 72,  146, 69,  61,  192, 156, 78,  130,
      172, 223, 62,  108, 75,  168, 12,  52,  15,  96,  100, 254, 134, 12,  74,  143, 232, 30,  245, 59,  90,  14,  199,
      246, 187, 204, 35,  204, 142, 220, 177, 73,  230, 242, 201, 74,  157, 85,  52,  215, 68,  89,  90,  164, 35,  186,
      183, 198, 21,  111, 197, 200, 223, 89,  120, 223, 180, 78,  76,  47,  185, 31,  191, 131, 239, 10,  77,  2,   8,
      132, 49,  217, 15,  59,  90,  30,  77,  68,  110, 8,   254, 95,  149, 20,  217, 80,  218, 163, 153, 228, 152, 210,
      16,  160, 114, 245, 244, 188, 176, 148, 252, 67,  127, 1,   220, 244, 59,  188, 113, 43,  170, 221, 155, 101, 158,
      205, 160, 167, 74,  12,  41,  70,  189, 47,  216, 229, 47,  3,   73,  72,  236, 16,  173, 10,  177, 14,  99,  109,
      206, 105, 135, 65,  6,   249, 233, 238, 211, 178, 155, 252, 79,  69,  43,  190, 193, 4,   58,  196, 182, 119, 164,
      164, 46,  102, 71,  195, 83,  249, 22,  85,  117, 5,   61,  227, 221, 180, 211, 42,  26,  224, 191, 107, 220, 254,
      32,  175, 185, 89,  123, 30,  174, 52,  143, 94,  195, 124, 232, 68,  18,  120, 232, 58,  44,  13,  57,  96,  92,
      168, 117, 58,  235, 33,  168, 216, 144, 123, 37,  182, 165, 23,  205, 139, 78,  178, 34,  58,  57,  121, 188, 71,
      136, 133, 238, 113, 7,   103, 126, 93,  94,  118, 80,  168, 6,   42,  88,  129, 215, 162, 66,  140, 247, 137, 3,
      237, 109, 253, 236, 68,  93,  85,  114, 223, 97,  29,  49,  186, 215, 72,  78,  164, 234, 151, 29,  53,  31,  66,
      220, 70,  230, 174, 129, 243, 205, 65,  142, 104, 219, 172, 44,  77,  203, 226, 67,  212, 96,  25,  176, 174, 11,
      167, 99,  42,  4,   80,  38,  108, 127, 177, 67,  151, 168, 204, 112, 150, 14,  158, 233, 178, 159, 230, 12,  160,
      128, 12,  239, 97,  5,   184, 229, 200, 178, 57,  52,  214, 196, 177, 167, 139, 64,  248, 86,  58,  107, 42,  128,
      105, 57,  21,  206, 74,  126, 70,  28,  228, 246, 42,  27,  93,  191, 153, 112, 199, 23,  136, 252, 9,   109, 153,
      82,  204, 107, 132, 129, 233, 200, 165, 20,  95,  50,  111, 79,  19,  130, 213, 164, 220, 60,  119, 220, 179, 145,
      76,  211, 192, 245, 135, 244, 119, 141, 70,  174, 195, 194, 124, 129, 188, 245, 96,  65,  124, 230, 106, 106, 252,
      229, 181, 144, 178, 27,  194, 200, 210, 241, 79,  164, 238, 111, 252, 175, 235, 114, 157, 96,  105, 122, 244, 32,
      84,  215, 252, 12,  31,  104, 21,  248, 183, 219, 166, 164, 185, 77,  4,   171, 122, 13,  8,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   181, 221, 101, 49,  140, 116, 13,  221, 248,
      121, 205, 207, 222, 163, 213, 141, 133, 135, 129, 132, 26,  213, 249, 69,  92,  2,   124, 99,  86,  139, 17,  144,
      135, 245, 118, 54,  144, 178, 248, 241, 153, 16,  120, 44,  4,   8,   16,  54,  211, 231, 13,  53,  118, 69,  213,
      206, 203, 22,  240, 102, 51,  205, 7,   39,  56,  63,  19,  118, 128, 47,  248, 50,  194, 63,  84,  56,  14,  250,
      150, 70,  198, 57,  213, 2,   73,  145, 31,  119, 199, 6,   196, 48,  70,  50,  123, 89,  3,   63,  177, 129, 61,
      84,  184, 158, 179, 7,   76,  242, 230, 171, 90,  1,   222, 144, 101, 53,  137, 132, 212, 124, 90,  242, 231, 135,
      51,  95,  231, 61,  47,  162, 220, 127, 26,  204, 155, 63,  213, 98,  144, 179, 155, 204, 251, 132, 240, 76,  29,
      102, 250, 14,  91,  181, 154, 17,  133, 92,  41,  124, 206, 123, 45,  177, 195, 81,  205, 152, 195, 40,  189, 113,
      26,  66,  26,  220, 71,  198, 76,  134, 194, 237, 66,  77,  188, 125, 165, 26,  192, 64,  90,  205, 62,  167, 169,
      77,  12,  11,  13,  197, 174, 252, 253, 64,  146, 102, 186, 3,   34,  202, 203, 187, 20,  187, 173, 80,  57,  52,
      205, 162, 41,  198, 123, 195, 41,  193, 206, 68,  166, 217, 62,  195, 59,  196, 199, 188, 169, 47,  7,   130, 146,
      23,  153, 71,  147, 254, 136, 155, 99,  232, 91,  46,  227, 192, 57,  96,  49,  185, 153, 74,  247, 169, 82,  195,
      193, 72,  109, 239, 186, 193, 119, 121, 77,  201, 126, 4,   245, 241, 12,  231, 76,  129, 88,  165, 131, 97,  192,
      246, 129, 185, 20,  156, 15,  219, 122, 241, 142, 251, 13,  105, 98,  212, 140, 52,  236, 217, 201, 53,  8,   212,
      95,  233, 243, 223, 205, 57,  113, 221, 152, 136, 138, 203, 123, 15,  114, 80,  26,  202, 99,  139, 129, 56,  23,
      240, 97,  97,  164, 166, 163, 207, 70,  234, 51,  113, 127, 232, 201, 104, 5,   22,  230, 159, 88,  193, 57,  114,
      107, 127, 161, 107, 36,  44,  31,  36,  125, 23,  128, 64,  75,  185, 152, 109, 29,  53,  109, 68,  159, 89,  120,
      149, 232, 140, 106, 147, 189, 231, 203, 127, 80,  184, 60,  229, 230, 196, 154, 173, 184, 171, 169, 110, 157, 35,
      79,  13,  48,  75,  160, 205, 74,  229, 251, 6,   212, 94,  145, 9,   115, 124, 189, 24,  180, 164, 69,  11,  8,
      79,  4,   128, 32,  92,  96,  47,  172, 134, 177, 171, 210, 103, 64,  45,  214, 152, 9,   218, 25,  94,  81,  64,
      115, 161, 238, 13,  74,  231, 222, 36,  196, 210, 164, 143, 167, 231, 55,  186, 94,  10,  97,  72,  211, 71,  77,
      85,  166, 113, 250, 41,  101, 23,  250, 97,  231, 189, 249, 67,  150, 183, 6,   202, 64,  76,  47,  94,  18,  157,
      95,  68,  10,  226, 221, 81,  242, 110, 53,  101, 144, 51,  14,  25,  206, 124, 112, 10,  137, 17,  10,  67,  21,
      61,  140, 40,  230, 172, 120, 174, 82,  136, 87,  232, 26,  59,  148, 95,  73,  162, 117, 159, 32,  202, 61,  118,
      179, 33,  19,  241, 58,  55,  104, 137, 133, 114, 122, 70,  151, 78,  226, 5,   150, 253, 35,  93,  30,  92,  23,
      2,   165, 122, 20,  251, 119, 247, 158, 2,   113, 206, 46,  6,   72,  53,  34,  176, 91,  164, 157, 179, 101, 189,
      250, 246, 164, 10,  87,  175, 201, 223, 128, 109, 214, 141, 218, 112, 49,  33,  96,  101, 254, 12,  56,  148, 140,
      172, 168, 43,  160, 102, 80,  172, 138, 46,  9,   19,  207, 59,  69,  48,  39,  211, 52,  124, 43,  204, 253, 86,
      65,  11,  20,  146, 222, 55,  243, 16,  187, 219, 177, 110, 48,  201, 153, 226, 131, 92,  154, 52,  195, 140, 37,
      142, 76,  47,  134, 140, 74,  129, 50,  1,   143, 19,  92,  40,  20,  103, 131, 199, 119, 56,  236, 199, 250, 1,
      80,  19,  148, 78,  191, 158, 88,  120, 125, 6,   50,  217, 233, 185, 55,  56,  163, 125, 137, 78,  133, 129, 208,
      101, 225, 244, 167, 155, 157, 7,   108, 157, 101, 138, 132, 82,  46,  227, 128, 162, 234, 189, 234, 176, 101, 114,
      75,  169, 120, 176, 134, 51,  241, 39,  36,  212, 193, 238, 187, 172, 92,  189, 41,  80,  238, 170, 247, 119, 51,
      233, 246, 146, 99,  143, 62,  130, 2,   79,  64,  42,  26,  88,  166, 21,  119, 224, 172, 217, 141, 187, 42,  147,
      169, 55,  206, 147, 215, 193, 85,  21,  10,  173, 1,   47,  192, 123, 242, 127, 222, 247, 233, 170, 221, 220, 226,
      101, 218, 214, 103, 36,  13,  195, 36,  32,  43,  76,  189, 88,  2,   147, 128, 180, 65,  91,  179, 124, 108, 116,
      50,  50,  142, 40,  180, 110, 246, 20,  52,  240, 76,  186, 60,  228, 174, 56,  102, 42,  203, 24,  228, 47,  107,
      131, 161, 3,   133, 106, 82,  74,  249, 167, 181, 172, 189, 159, 89,  153, 118, 232, 248, 205, 147, 8,   203, 192,
      20,  169, 189, 148, 234, 233, 130, 230, 171, 105, 36,  246, 191, 211, 165, 88,  154, 136, 6,   90,  131, 200, 241,
      55,  92,  135, 221, 39,  14,  68,  139, 24,  237, 215, 20,  177, 190, 51,  106, 127, 237, 146, 222, 61,  247, 173,
      2,   112, 211, 145, 99,  95,  206, 164, 183, 233, 38,  21,  44,  106, 226, 183, 171, 230, 201, 56,  77,  118, 121,
      184, 141, 129, 114, 127, 3,   136, 188, 10,  243, 123, 204, 106, 74,  241, 202, 186, 187, 233, 151, 222, 27,  135,
      167, 125, 63,  226, 25,  139, 131, 8,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   136, 5,   119, 80,  154, 4,   241, 70,  159, 227, 2,   238, 104, 242, 90,  84,  111, 231, 89,  60,
      9,   57,  37,  67,  199, 38,  232, 235, 141, 213, 56,  16,  31,  123, 175, 4,   96,  122, 28,  41,  208, 176, 135,
      37,  111, 233, 60,  206, 42,  208, 212, 103, 197, 110, 153, 152, 245, 229, 129, 59,  172, 155, 132, 104, 53,  12,
      13,  105, 234, 168, 191, 184, 111, 198, 58,  212, 222, 156, 84,  196, 62,  205, 142, 73,  0,   96,  241, 191, 204,
      212, 209, 58,  190, 237, 185, 162, 188, 108, 216, 144, 63,  207, 191, 194, 44,  68,  254, 140, 118, 111, 83,  121,
      242, 108, 97,  40,  94,  208, 135, 168, 116, 197, 233, 141, 42,  84,  96,  208, 110, 111, 250, 170, 53,  171, 178,
      136, 168, 182, 168, 217, 151, 36,  59,  143, 86,  238, 197, 26,  107, 42,  23,  151, 225, 227, 200, 9,   82,  6,
      73,  123, 201, 31,  109, 41,  220, 188, 122, 7,   102, 190, 90,  54,  138, 46,  68,  72,  69,  46,  27,  14,  190,
      237, 140, 71,  42,  172, 178, 142, 200, 114, 108, 162, 253, 168, 241, 115, 71,  18,  112, 22,  145, 15,  231, 88,
      100, 163, 250, 72,  129, 244, 233, 86,  70,  26,  191, 52,  26,  58,  50,  254, 5,   30,  44,  5,   222, 149, 113,
      3,   140, 140, 225, 107, 86,  156, 10,  247, 228, 212, 79,  85,  63,  122, 77,  157, 228, 233, 157, 167, 82,  254,
      65,  208, 0,   175, 24,  56,  185, 224, 186, 204, 248, 139, 120, 196, 225, 156, 126, 198, 72,  46,  240, 4,   138,
      253, 69,  158, 182, 40,  229, 14,  64,  55,  221, 123, 165, 89,  130, 206, 246, 237, 253, 193, 134, 194, 70,  197,
      135, 104, 141, 35,  190, 23,  154, 252, 226, 212, 79,  234, 158, 238, 126, 232, 62,  112, 86,  124, 18,  151, 166,
      142, 160, 218, 244, 135, 211, 180, 218, 48,  159, 210, 36,  153, 175, 73,  27,  138, 229, 80,  236, 133, 25,  209,
      164, 240, 61,  211, 163, 48,  253, 35,  97,  7,   251, 198, 146, 45,  51,  80,  60,  5,   249, 81,  81,  76,  46,
      194, 21,  129, 38,  167, 8,   58,  101, 133, 177, 142, 129, 53,  235, 246, 57,  174, 209, 205, 51,  63,  16,  65,
      199, 180, 84,  147, 76,  54,  104, 103, 139, 196, 162, 44,  44,  227, 38,  46,  146, 118, 238, 233, 156, 223, 205,
      131, 87,  60,  247, 139, 108, 18,  56,  192, 221, 220, 208, 94,  203, 28,  209, 70,  53,  102, 220, 144, 107, 56,
      142, 194, 232, 161, 152, 118, 23,  171, 87,  132, 80,  4,   106, 193, 231, 250, 42,  101, 59,  37,  174, 245, 167,
      187, 30,  189, 212, 53,  191, 235, 208, 137, 168, 177, 4,   120, 50,  157, 218, 126, 230, 122, 118, 194, 119, 100,
      197, 157, 228, 77,  119, 171, 62,  229, 251, 10,  119, 24,  80,  73,  254, 223, 37,  47,  178, 239, 78,  150, 122,
      37,  93,  125, 35,  80,  82,  38,  6,   40,  168, 168, 83,  180, 105, 186, 218, 174, 137, 205, 20,  32,  197, 166,
      41,  33,  96,  252, 2,   148, 59,  12,  95,  8,   44,  68,  81,  178, 165, 185, 128, 33,  159, 19,  236, 224, 181,
      83,  115, 30,  241, 220, 181, 105, 128, 141, 189, 149, 142, 11,  81,  253, 30,  103, 226, 8,   211, 171, 45,  180,
      104, 2,   79,  96,  22,  192, 27,  229, 217, 93,  96,  136, 10,  43,  79,  236, 156, 42,  245, 153, 251, 253, 65,
      147, 84,  27,  18,  161, 47,  243, 222, 48,  23,  71,  225, 112, 38,  176, 104, 17,  188, 86,  95,  6,   163, 153,
      77,  80,  203, 173, 164, 5,   242, 240, 70,  214, 41,  157, 101, 65,  104, 175, 250, 5,   119, 78,  226, 88,  161,
      46,  234, 5,   117, 79,  47,  160, 211, 196, 228, 180, 50,  34,  172, 169, 98,  199, 206, 49,  17,  113, 29,  47,
      131, 93,  55,  40,  113, 5,   97,  84,  146, 236, 206, 212, 164, 217, 197, 173, 1,   223, 152, 144, 137, 230, 39,
      222, 249, 253, 42,  209, 125, 132, 186, 195, 180, 211, 139, 186, 5,   7,   106, 101, 154, 77,  85,  172, 114, 49,
      126, 236, 173, 65,  100, 136, 196, 124, 9,   83,  209, 26,  127, 195, 4,   227, 252, 34,  162, 69,  53,  88,  54,
      225, 206, 75,  154, 63,  62,  27,  223, 12,  9,   159, 167, 4,   77,  25,  162, 136, 129, 0,   218, 44,  161, 135,
      107, 49,  50,  137, 110, 22,  138, 130, 163, 63,  120, 250, 42,  76,  250, 64,  195, 28,  170, 135, 32,  138, 182,
      150, 242, 252, 8,   209, 232, 184, 234, 155, 203, 16,  22,  38,  50,  183, 21,  191, 242, 211, 88,  216, 225, 72,
      215, 253, 203, 91,  34,  126, 246, 144, 250, 59,  221, 24,  132, 247, 218, 246, 90,  218, 73,  194, 249, 171, 9,
      118, 227, 76,  28,  182, 249, 223, 61,  216, 165, 193, 247, 110, 201, 5,   62,  132, 161, 230, 103, 21,  22,  15,
      161, 124, 46,  77,  12,  136, 114, 45,  128, 43,  87,  209, 191, 137, 68,  69,  219, 202, 254, 50,  190, 61,  163,
      189, 160, 102, 136, 130, 97,  126, 20,  18,  245, 182, 147, 41,  82,  98,  56,  72,  245, 31,  43,  139, 199, 130,
      210, 104, 170, 26,  60,  69,  237, 87,  100, 203, 130, 71,  217, 60,  182, 199, 22,  181, 71,  234, 12,  215, 188,
      196, 220, 177, 175, 161, 128, 47,  73,  232, 28,  66,  122, 71,  234, 33,  142, 155, 59,  93,  159, 71,  88,  174,
      70,  222, 186, 91,  167, 198, 89,  101, 149, 117, 15,  187, 241, 190, 32,  229, 25,  209, 158, 2,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   177, 112, 96,  168, 65,  84,  5,   126,
      102, 156, 175, 133, 130, 245, 182, 135, 221, 171, 41,  132, 114, 212, 178, 126, 162, 0,   187, 71,  39,  223, 211,
      238, 243, 157, 112, 160, 172, 8,   95,  6,   34,  141, 83,  161, 62,  177, 111, 171, 71,  231, 18,  86,  182, 74,
      133, 116, 85,  2,   150, 226, 147, 192, 88,  2,   211, 17,  60,  90,  203, 28,  227, 120, 232, 115, 185, 108, 67,
      235, 224, 200, 30,  214, 220, 25,  143, 188, 218, 248, 75,  236, 181, 44,  130, 215, 113, 248, 62,  59,  136, 115,
      33,  174, 90,  6,   176, 188, 168, 2,   212, 18,  54,  156, 246, 181, 22,  74,  57,  25,  234, 230, 123, 184, 69,
      59,  62,  151, 16,  114, 219, 98,  55,  175, 168, 151, 105, 114, 251, 228, 75,  107, 114, 237, 90,  218, 51,  44,
      169, 168, 232, 106, 244, 31,  83,  150, 204, 210, 110, 108, 175, 17,  106, 165, 192, 127, 59,  47,  99,  177, 98,
      123, 18,  122, 10,  139, 232, 5,   146, 29,  237, 46,  69,  186, 212, 95,  96,  223, 78,  173, 212, 234, 39,  141,
      48,  78,  136, 21,  212, 206, 144, 18,  67,  74,  35,  171, 210, 125, 245, 61,  231, 58,  134, 45,  92,  61,  78,
      29,  123, 142, 18,  78,  162, 60,  166, 31,  149, 175, 63,  89,  65,  143, 190, 50,  88,  104, 172, 113, 103, 38,
      35,  154, 3,   8,   28,  160, 54,  247, 10,  102, 184, 95,  35,  126, 154, 162, 85,  71,  169, 43,  158, 100, 223,
      218, 233, 124, 126, 211, 253, 164, 15,  156, 10,  27,  232, 96,  231, 80,  49,  47,  30,  239, 139, 215, 68,  100,
      235, 108, 30,  229, 213, 247, 209, 243, 153, 249, 181, 85,  244, 53,  196, 171, 106, 1,   182, 94,  208, 221, 134,
      39,  54,  195, 9,   227, 210, 112, 152, 192, 17,  20,  114, 218, 81,  125, 175, 206, 25,  55,  17,  125, 2,   34,
      49,  31,  157, 11,  253, 28,  69,  23,  203, 144, 128, 223, 184, 88,  135, 82,  125, 156, 62,  30,  40,  244, 105,
      168, 177, 209, 105, 199, 106, 78,  239, 3,   136, 168, 8,   155, 98,  227, 64,  13,  220, 22,  213, 103, 148, 238,
      177, 206, 234, 6,   247, 155, 138, 205, 54,  8,   17,  37,  21,  235, 46,  132, 31,  49,  30,  87,  82,  231, 93,
      38,  91,  167, 159, 167, 223, 95,  217, 70,  77,  205, 8,   156, 130, 15,  169, 213, 109, 203, 114, 1,   252, 44,
      93,  224, 164, 115, 163, 103, 97,  103, 81,  165, 121, 25,  31,  140, 251, 71,  120, 137, 74,  103, 78,  77,  151,
      194, 239, 76,  99,  82,  150, 106, 27,  249, 248, 97,  106, 133, 214, 219, 122, 249, 245, 230, 115, 48,  254, 227,
      46,  147, 188, 30,  105, 96,  66,  116, 169, 185, 3,   29,  232, 27,  219, 210, 179, 100, 176, 243, 3,   149, 227,
      178, 89,  3,   218, 190, 135, 166, 5,   93,  93,  172, 74,  242, 241, 44,  198, 94,  228, 240, 233, 234, 183, 232,
      50,  136, 23,  34,  64,  138, 93,  220, 35,  10,  242, 172, 123, 108, 2,   244, 77,  62,  203, 249, 160, 232, 69,
      231, 10,  214, 6,   94,  113, 213, 87,  36,  115, 166, 119, 145, 59,  110, 83,  131, 69,  196, 14,  242, 91,  17,
      143, 128, 180, 249, 3,   247, 141, 157, 28,  74,  253, 104, 174, 44,  171, 145, 192, 245, 215, 137, 239, 1,   5,
      43,  186, 154, 245, 249, 20,  16,  31,  224, 118, 139, 73,  167, 134, 189, 163, 170, 90,  62,  118, 201, 122, 29,
      140, 205, 40,  5,   6,   183, 242, 170, 18,  202, 232, 87,  208, 149, 20,  128, 238, 6,   88,  237, 109, 217, 117,
      86,  199, 137, 164, 45,  151, 117, 217, 209, 182, 28,  189, 107, 146, 174, 133, 253, 74,  40,  77,  76,  188, 152,
      27,  135, 247, 159, 137, 39,  101, 19,  136, 159, 193, 148, 194, 247, 136, 41,  89,  196, 140, 81,  23,  174, 79,
      233, 228, 174, 186, 126, 85,  172, 228, 113, 44,  50,  253, 51,  251, 192, 253, 42,  151, 17,  168, 26,  241, 145,
      169, 222, 99,  68,  195, 238, 20,  10,  91,  147, 220, 241, 175, 210, 82,  185, 201, 168, 187, 225, 18,  109, 235,
      3,   29,  75,  248, 236, 102, 215, 2,   242, 133, 185, 107, 2,   88,  245, 232, 135, 63,  183, 44,  163, 155, 177,
      127, 166, 51,  86,  170, 193, 212, 182, 84,  8,   236, 212, 168, 154, 222, 177, 166, 48,  129, 235, 154, 167, 38,
      84,  33,  93,  70,  52,  33,  32,  125, 204, 8,   154, 47,  106, 118, 8,   141, 31,  2,   132, 138, 205, 162, 197,
      28,  17,  144, 199, 5,   247, 57,  119, 211, 99,  141, 80,  116, 24,  202, 162, 185, 30,  159, 143, 181, 101, 73,
      242, 21,  153, 58,  38,  12,  177, 166, 135, 25,  95,  30,  4,   83,  64,  98,  196, 211, 120, 33,  119, 198, 221,
      118, 49,  130, 84,  58,  230, 227, 39,  190, 103, 13,  156, 41,  223, 133, 125, 205, 92,  244, 106, 158, 213, 39,
      71,  13,  247, 173, 121, 242, 234, 114, 83,  63,  26,  26,  166, 134, 207, 147, 77,  113, 152, 171, 226, 108, 72,
      175, 187, 167, 29,  89,  112, 201, 27,  171, 193, 78,  183, 111, 115, 127, 203, 84,  211, 138, 167, 95,  82,  10,
      185, 59,  152, 17,  178, 242, 229, 243, 158, 215, 197, 46,  160, 2,   61,  253, 220, 114, 106, 130, 142, 162, 253,
      174, 133, 101, 219, 119, 38,  172, 166, 220, 102, 60,  53,  73,  208, 118, 168, 143, 134, 201, 60,  26,  62,  172,
      99,  172, 241, 35,  190, 178, 154, 6,   138, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   217, 37,  145, 160, 185, 189, 159, 26,  106, 138, 109, 0,   151, 61,  151, 232, 89,  181, 85,
      52,  217, 92,  61,  143, 47,  204, 108, 147, 140, 212, 240, 94,  244, 152, 6,   85,  214, 166, 78,  16,  205, 104,
      121, 183, 100, 120, 66,  249, 68,  37,  16,  128, 118, 126, 218, 252, 174, 157, 234, 251, 49,  126, 79,  82,  121,
      11,  76,  120, 157, 197, 124, 226, 157, 178, 69,  18,  188, 120, 67,  160, 230, 85,  133, 105, 4,   192, 112, 210,
      221, 217, 39,  86,  79,  93,  42,  114, 68,  157, 223, 146, 35,  40,  210, 134, 137, 118, 27,  249, 31,  169, 196,
      139, 114, 188, 131, 204, 79,  31,  71,  20,  182, 89,  218, 249, 56,  236, 226, 43,  65,  24,  61,  237, 247, 127,
      149, 92,  213, 103, 34,  177, 3,   211, 235, 50,  32,  235, 149, 20,  169, 170, 56,  87,  72,  236, 247, 9,   24,
      226, 136, 50,  124, 76,  94,  159, 160, 161, 165, 52,  82,  13,  240, 146, 247, 27,  254, 212, 190, 23,  179, 248,
      79,  216, 28,  40,  108, 174, 143, 78,  193, 220, 94,  223, 136, 187, 136, 14,  32,  178, 10,  101, 206, 16,  9,
      38,  41,  218, 10,  143, 229, 136, 247, 177, 178, 8,   203, 70,  223, 205, 77,  26,  7,   160, 156, 125, 247, 139,
      206, 113, 244, 107, 135, 245, 178, 122, 50,  94,  49,  70,  57,  3,   229, 195, 41,  80,  58,  66,  45,  205, 28,
      182, 63,  248, 142, 14,  93,  60,  222, 17,  221, 206, 92,  248, 42,  21,  148, 4,   5,   5,   221, 230, 146, 103,
      206, 79,  134, 160, 154, 215, 9,   168, 177, 127, 208, 196, 142, 113, 36,  204, 246, 238, 48,  246, 118, 147, 247,
      70,  10,  183, 169, 140, 7,   199, 50,  204, 151, 34,  141, 208, 94,  45,  239, 60,  199, 183, 121, 73,  114, 33,
      150, 152, 167, 152, 127, 9,   123, 73,  223, 60,  90,  93,  140, 62,  253, 38,  194, 100, 164, 150, 81,  187, 162,
      158, 26,  197, 179, 123, 107, 108, 88,  30,  45,  111, 141, 17,  122, 142, 28,  232, 140, 21,  242, 77,  35,  206,
      158, 187, 115, 115, 175, 148, 113, 158, 213, 102, 2,   125, 216, 125, 220, 104, 30,  170, 188, 136, 34,  100, 17,
      180, 10,  238, 190, 74,  48,  26,  115, 113, 215, 201, 121, 114, 215, 241, 128, 101, 44,  113, 37,  196, 124, 254,
      109, 195, 213, 141, 239, 139, 35,  101, 70,  157, 203, 113, 229, 233, 204, 97,  148, 212, 208, 174, 126, 203, 245,
      22,  119, 36,  5,   149, 32,  9,   141, 0,   202, 134, 58,  131, 132, 126, 52,  126, 74,  234, 212, 221, 80,  19,
      64,  117, 167, 118, 52,  6,   14,  99,  77,  113, 201, 248, 127, 135, 205, 147, 227, 60,  46,  55,  237, 192, 149,
      226, 95,  104, 162, 76,  24,  100, 38,  44,  240, 137, 93,  6,   136, 62,  141, 164, 172, 11,  55,  227, 159, 225,
      195, 104, 243, 112, 31,  70,  104, 173, 242, 157, 232, 162, 225, 87,  105, 182, 136, 194, 112, 96,  155, 171, 243,
      122, 53,  197, 194, 208, 240, 35,  37,  117, 201, 235, 205, 118, 213, 80,  53,  126, 186, 170, 2,   7,   233, 95,
      107, 144, 133, 154, 228, 55,  156, 108, 241, 167, 245, 6,   8,   173, 54,  235, 128, 185, 139, 251, 97,  140, 94,
      166, 171, 70,  62,  193, 87,  203, 31,  43,  237, 199, 4,   201, 37,  48,  64,  99,  251, 200, 55,  59,  39,  163,
      88,  236, 35,  83,  5,   10,  37,  212, 69,  19,  124, 125, 92,  184, 94,  93,  179, 199, 211, 7,   119, 24,  14,
      156, 220, 88,  240, 101, 10,  44,  40,  192, 171, 162, 138, 121, 111, 203, 157, 198, 211, 251, 22,  84,  20,  99,
      221, 253, 60,  92,  181, 237, 193, 161, 211, 139, 12,  57,  173, 148, 160, 229, 121, 104, 66,  113, 175, 157, 177,
      219, 38,  77,  72,  76,  217, 233, 101, 24,  182, 109, 238, 118, 177, 156, 89,  22,  63,  9,   241, 135, 59,  69,
      90,  87,  130, 177, 70,  169, 179, 13,  21,  109, 187, 164, 110, 88,  61,  115, 81,  186, 216, 20,  124, 206, 188,
      202, 143, 162, 192, 208, 15,  34,  80,  91,  233, 194, 163, 173, 61,  80,  161, 147, 46,  44,  14,  184, 20,  129,
      149, 196, 241, 90,  164, 125, 113, 228, 127, 119, 185, 11,  108, 175, 195, 233, 31,  17,  103, 132, 253, 146, 192,
      111, 172, 47,  150, 77,  143, 5,   148, 49,  172, 106, 141, 227, 85,  95,  92,  152, 25,  39,  16,  108, 132, 70,
      247, 14,  56,  63,  213, 155, 240, 148, 242, 96,  71,  251, 100, 107, 122, 173, 51,  59,  82,  171, 232, 133, 36,
      65,  168, 137, 182, 15,  38,  208, 18,  107, 27,  77,  170, 93,  100, 245, 74,  250, 111, 223, 154, 84,  51,  28,
      203, 250, 208, 105, 128, 77,  9,   214, 9,   137, 201, 17,  17,  161, 31,  188, 192, 24,  112, 78,  70,  174, 101,
      102, 131, 90,  15,  145, 215, 87,  94,  28,  198, 134, 197, 103, 115, 127, 172, 156, 172, 159, 12,  234, 199, 209,
      70,  247, 214, 171, 248, 161, 181, 116, 12,  41,  124, 8,   193, 33,  244, 10,  226, 208, 78,  244, 49,  115, 201,
      129, 7,   253, 208, 80,  162, 44,  34,  23,  143, 216, 7,   15,  229, 217, 47,  102, 37,  65,  94,  123, 91,  93,
      75,  247, 8,   58,  17,  121, 253, 3,   178, 101, 69,  117, 74,  78,  156, 169, 118, 61,  120, 208, 193, 45,  248,
      238, 173, 126, 52,  127, 232, 59,  185, 116, 38,  65,  110, 226, 236, 33,  235, 176, 222, 249, 63,  164, 0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   19,  146, 65,  18,  222, 192, 28,
      129, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   141, 194, 179, 150, 178, 254, 199, 164, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   63,  86,  70,  20,  115, 220, 58,  54,  0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   87,  189, 150,
      42,  83,  54,  191, 178, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   168, 148, 227, 237, 0,   58,  76,  125, 0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   106, 81,  225, 130, 80,  57,  19,  44,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   203, 199, 88,  221, 85,  236, 28,  154, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   137, 70,  56,  15,  38,  198, 57,  188, 0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   171, 226, 22,  161,
      146, 207, 253, 221, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   238, 243, 247, 62,  91,  24,  96,  62,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   43,  92,  135, 166, 200, 178, 32,  143, 0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      199, 58,  182, 38,  46,  223, 251, 177, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   8,   54,  221, 20,  104, 192, 15,  16,  0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   52,  250, 181, 83,  73,
      108, 39,  238, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   87,  62,  63,  109, 185, 27,  47,  55,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   201, 228, 200, 210, 40,  189, 69,  33,  0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   163,
      163, 100, 42,  233, 131, 196, 207, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   61,  64,  224, 148, 207, 87,  250, 49,  0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   110, 228, 74,  100, 21,  232,
      188, 23,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   99,  141, 163, 15,  250, 143, 144, 193, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   101, 121, 138, 146, 243, 91,  162, 43,  0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   11,  59,
      131, 6,   28,  220, 203, 187, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   245, 233, 244, 171, 114, 186, 18,  108, 0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   241, 239, 20,  254, 188, 136, 79,
      94,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   76,  64,  21,  56,  243, 197, 57,  40,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   63,  50,  196, 168, 85,  165, 173, 159, 0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   66,  54,  126,
      47,  43,  133, 97,  245, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   73,  250, 139, 147, 9,   128, 42,  212, 0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   100, 204, 155, 163, 171, 146, 4,   25,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   92,  225, 172, 148, 16,  70,  164, 46,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   246, 132, 223, 201, 82,  47,  254, 102, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   246, 15,  204, 188, 11,  80,  246, 168, 0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      182, 191, 225, 26,  52,  65,  149, 11,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   71,  191, 171, 180, 6,   36,  183, 199, 0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   66,  32,  79,  63,  222,
      180, 149, 68,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   58,  172, 192, 3,   9,   104, 4,   89,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   39,  165, 11,  151, 236, 179, 31,  62,  0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   29,
      180, 143, 81,  31,  161, 29,  1,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   201, 165, 129, 75,  141, 69,  64,  111, 0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   15,  27,  112, 242, 101, 206,
      26,  69,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   140, 57,  129, 215, 240, 66,  56,  148, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   225, 35,  59,  113, 165, 101, 86,  190, 0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   39,  241,
      201, 16,  7,   199, 31,  214, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   94,  214, 47,  254, 191, 204, 10,  187, 0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   109, 237, 174, 252, 46,  216, 181,
      101, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   94,  228, 195, 95,  198, 10,  247, 141, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   247, 88,  223, 141, 193, 52,  30,  164, 0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   207, 157, 67,
      177, 140, 84,  41,  195, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   1,   142, 188, 96,  130, 31,  32,  107, 0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   92,  25,  47,  159, 216, 60,  8,   168,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   17,  44,  31,  207, 217, 43,  167, 134, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   106, 82,  173, 99,  30,  37,  190, 164, 0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   107, 75,  77,  217,
      128, 103, 87,  196, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   240, 88,  234, 237, 206, 215, 12,  57,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   166, 228, 252, 243, 30,  7,   163, 119, 0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      216, 208, 83,  70,  233, 204, 33,  211, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   113, 2,   253, 184, 5,   138, 147, 208, 0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   67,  97,  210, 235, 232,
      141, 198, 54,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   177, 171, 221, 225, 125, 197, 69,  136, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   95,  224, 6,   91,  85,  1,   232, 154, 0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   179, 22,  46,  192, 216, 254,
      241, 78,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   136, 154, 175, 18,  77,  128, 245, 195, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   233, 30,  222, 197, 6,   216, 20,  21,  0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   23,  36,
      184, 47,  223, 53,  151, 107, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   49,  131, 190, 24,  114, 220, 99,  166, 0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   103, 224, 13,  98,  19,  122, 47,
      204, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   229, 143, 190, 73,  20,  9,   91,  214, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   51,  35,  42,  212, 52,  117, 170, 240, 0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   45,  144, 12,
      178, 80,  146, 30,  166, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   57,  17,  165, 121, 15,  238, 225, 94,  0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   96,  202, 171, 146, 72,  206, 75,  100,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   220, 1,   0,   171, 235, 18,  166, 203, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   210, 244, 77,  101, 116, 86,  179, 39,  0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   190, 124, 134, 122,
      45,  172, 101, 164, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   54,  236, 207, 201, 100, 9,   201, 55,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   147, 203, 209, 159, 188, 4,   167, 188, 0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      226, 222, 71,  224, 186, 133, 168, 254, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   15,  254, 246, 42,  224, 165, 177, 30,  0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   114, 90,  26,  68,  8,
      226, 41,  175, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   77,  201, 183, 119, 149, 87,  40,  253, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   4,   98,  148, 230, 2,   23,  205, 224, 0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   26,
      112, 123, 15,  19,  177, 140, 152, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   79,  55,  116, 197, 57,  10,  27,  179, 0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   220, 209, 60,  45,  79,  233,
      117, 81,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   115, 81,  235, 240, 121, 164, 46,  68,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   250, 195, 119, 124, 21,  110, 50,  185, 0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   78,  196,
      35,  66,  235, 81,  20,  246, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   40,  250, 104, 157, 128, 106, 148, 47,  0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   252, 101, 113, 89,  95,  125, 243,
      111, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   48,  170, 91,  98,  162, 183, 191, 47,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   133, 88,  96,  9,   213, 50,  100, 133, 0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   56,  20,  46,  96,  151, 43,  101, 141,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   175, 229, 68,  20,  158, 175, 223, 134, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   95,  28,  94,  0,   109, 1,   252, 188, 0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   32,  215, 242, 67,
      159, 169, 170, 60,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   210, 25,  18,  148, 205, 218, 227, 187, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   11,  140, 181, 64,  143, 160, 70,  84,  0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      87,  252, 245, 64,  168, 102, 58,  174, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   60,  231, 63,  214, 98,  74,  172, 64,  0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   164, 139, 178, 166, 11,
      215, 20,  18,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   210, 236, 124, 15,  193, 113, 81,  61,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   183, 166, 99,  239, 5,   128, 194, 10,  0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   203,
      98,  219, 26,  102, 11,  107, 130, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   239, 235, 31,  64,  175, 2,   19,  194, 0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   84,  30,  198, 145, 183, 111,
      189, 25,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   46,  161, 77,  51,  188, 59,  222, 184, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   81,  13,  96,  175, 242, 206, 141, 93,  0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   242, 250,
      2,   118, 103, 160, 165, 234, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   17,  31,  3,   66,  12,  163, 157, 224, 0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   27,  119, 77,  99,  184, 95,  233,
      77,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   214, 240, 48,  108, 217, 163, 39,  97,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   235, 98,  50,  63,  138, 26,  225, 126, 0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   12,  175, 130,
      10,  110, 198, 129, 191, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   139, 108, 163, 8,   160, 239, 170, 35,  0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   108, 155, 244, 69,  119, 8,   216, 56,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   196, 52,  69,  230, 147, 242, 134, 240, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   242, 107, 102, 224, 232, 222, 159, 87,  0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   110, 101, 52,  253,
      88,  238, 220, 212, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   46,  13,  61,  62,  170, 192, 4,   57,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   3,   233, 248, 191, 86,  139, 15,  237, 0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   205, 228, 152, 6,   33,
      162, 17,  67,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   229, 3,   64,  231, 165, 10,  36,  20,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   186, 101, 148, 236, 30,  35,  163, 49,  0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   116,
      97,  121, 132, 86,  130, 48,  191, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   80,  190, 17,  150, 22,  48,  104, 85,  0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   171, 242, 157, 66,  44,  139,
      12,  58,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   192, 37,  126, 216, 125, 49,  39,  19,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   39,  134, 44,  75,  205, 15,  211, 197, 0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   93,  81,
      63,  152, 130, 248, 73,  184, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   25,  31,  203, 241, 115, 104, 152, 174, 0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   130, 67,  206, 44,  71,  58,  235,
      12,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   6,   92,  73,  144, 192, 22,  150, 75,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   231, 183, 115, 82,  6,   111, 44,  154, 0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   34,  209, 181,
      252, 239, 164, 185, 106, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   155, 78,  58,  155, 63,  186, 176, 240, 0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   237, 229, 122, 248, 160, 78,  57,  7,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   71,  252, 131, 214, 106, 149, 245, 17,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   154, 160, 23,  35,  180, 122, 146, 183, 0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   20,  147, 196, 145,
      9,   114, 33,  154, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   196, 213, 216, 114, 42,  81,  252, 21,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   148, 144, 77,  158, 164, 0,   37,  244, 0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      233, 174, 45,  246, 239, 231, 60,  8,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   112, 253, 84,  28,  149, 74,  5,   239, 0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   217, 176, 222, 58,  35,
      44,  2,   155, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   130, 113, 154, 57,  105, 117, 103, 157, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   144, 99,  142, 39,  26,  200, 251, 249, 0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   178,
      110, 41,  197, 198, 204, 10,  235, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   98,  214, 29,  104, 38,  30,  251, 69,  0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   249, 179, 39,  10,  114, 16,
      97,  96,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   95,  198, 179, 205, 39,  140, 213, 220, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   16,  212, 92,  209, 18,  34,  240, 191, 0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   248, 8,   109, 191, 111, 62,  222,
      229, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   235, 200, 48,  199, 132, 47,  44,  183, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   100, 10,  57,  146, 87,  214, 123, 22,  0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   190, 190, 234,
      161, 61,  186, 252, 49,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   153, 144, 250, 41,  228, 58,  43,  58,  0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   90,  133, 158, 212, 245, 231, 64,  153,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   138, 167, 77,  109, 52,  247, 174, 70,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   227, 137, 253, 203, 194, 46,  199, 121, 0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   179, 9,   102, 241,
      121, 221, 149, 220, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   189, 254, 41,  138, 154, 124, 141, 79,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   161, 25,  100, 236, 78,  103, 14,  52,  0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      126, 77,  10,  86,  216, 77,  100, 201, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   246, 191, 32,  245, 137, 150, 189, 87,  0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   19,  164, 208, 123, 205,
      90,  87,  120, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   58,  59,  143, 128, 100, 218, 58,  236, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   164, 141, 145, 163, 80,  86,  225, 16,  0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   131,
      224, 35,  146, 166, 116, 50,  118, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   36,  195, 53,  249, 1,   239, 99,  5,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   240, 27,  99,  197, 17,  250,
      129, 69,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   12,  219, 151, 245, 136, 248, 124, 82,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   13,  221, 190, 178, 43,  97,  27,  171, 0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   108, 210,
      208, 54,  162, 254, 244, 85,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   13,  3,   69,  152, 178, 214, 234, 29,  0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   155, 220, 23,  102, 90,  28,  36,
      46,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   140, 157, 116, 203, 135, 234, 44,  31,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   59,  208, 148, 207, 66,  248, 130, 12,  0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   80,  65,  36,
      249, 59,  31,  4,   3,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   49,  40,  124, 250, 250, 88,  195, 11,  0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   105, 52,  92,  182, 136, 23,  128, 249,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   65,  26,  246, 117,
      172, 129, 172, 99,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   42,  231, 214, 189, 29,  162, 191, 23,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   245, 77,  142, 16,  93,  151, 202, 35,  0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      6,   32,  176, 150, 181, 19,  66,  187, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   164, 189, 1,   192, 17,  82,  123, 239, 0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   222, 84,  89,  81,  183,
      46,  96,  223, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   243, 30,  2,   28,  239, 97,  96,  2,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   130, 121, 235, 25,  152, 59,  40,  149, 0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   113,
      46,  130, 80,  93,  125, 1,   248, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   61,  237, 129, 172, 20,  214, 50,  172, 0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   189, 243, 178, 145, 249, 154,
      252, 133, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   53,  41,  238, 229, 202, 126, 2,   151, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   175, 194, 252, 87,  87,  86,  180, 197, 0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   161, 147,
      24,  236, 43,  162, 30,  169, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   133, 25,  111, 132, 59,  24,  71,  9,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   64,  225, 112, 150, 137, 242, 120,
      75,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   173, 69,  61,  70,  108, 20,  129, 214, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   252, 157, 90,  104, 49,  182, 44,  236, 0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   240, 186, 217,
      48,  136, 7,   240, 182, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   115, 163, 110, 151, 167, 47,  55,  131, 0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   185, 192, 253, 72,  96,  119, 171, 69,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   190, 184, 69,  31,  49,  224, 215, 167, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   96,  28,  190, 214, 227, 107, 156, 199, 0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   110, 184, 212, 230,
      88,  224, 9,   106, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   227, 62,  96,  15,  210, 231, 79,  53,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   166, 139, 116, 65,  136, 125, 31,  121, 0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      170, 162, 172, 193, 130, 59,  54,  230, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   254, 59,  48,  178, 156, 149, 5,   41,  0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   79,  218, 54,  104, 10,
      187, 77,  123, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   126, 122, 148, 187, 226, 182, 185, 168, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   91,  177, 207, 200, 249, 241, 97,  160, 0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   30,  119, 79,  46,  143, 120,
      220, 240, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   171, 76,  48,  54,  76,  152, 37,  72,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   187, 236, 100, 172, 113, 122, 114, 146, 0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   109, 40,
      7,   18,  70,  242, 242, 169, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   157, 45,  166, 116, 146, 128, 93,  154, 0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   142, 65,  174, 40,  86,  93,  28,
      193, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   32,  33,  181, 84,  157, 213, 54,  134, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   243, 3,   235, 181, 118, 39,  228, 72,  0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   172, 128, 31,
      26,  34,  81,  251, 25,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   42,  189, 62,  20,  68,  172, 107, 129, 0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   107, 88,  225, 176, 63,  148, 68,  4,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   46,  138, 62,  86,  204, 130, 207, 220, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   43,  53,  211, 0,   62,  170, 181, 121, 0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   13,  24,  93,  172,
      183, 70,  227, 207, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   52,  200, 133, 165, 140, 107, 41,  221, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   124, 135, 150, 222, 226, 218, 62,  12,  0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      197, 197, 186, 27,  226, 134, 173, 63,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   120, 42,  126, 117, 206, 24,  249, 50,  0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   180, 14,  118, 165, 195,
      189, 145, 33,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   54,  7,   213, 61,  144, 25,  71,  38,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   56,  230, 129, 107, 73,  102, 162, 206, 0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   188,
      145, 221, 153, 141, 98,  152, 214, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   131, 15,  229, 20,  75,  219, 212, 117, 0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   211, 122, 77,  181, 231, 186,
      19,  7,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   96,  200, 168, 21,  164, 125, 164, 85,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   239, 71,  21,  146, 196, 214, 61,  104, 0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   184, 210,
      120, 31,  64,  56,  224, 244, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   108, 242, 64,  171, 112, 230, 197, 159, 0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   167, 175, 177, 146, 251, 185, 215,
      226, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   72,  129, 126,
      161, 167, 25,  97,  209, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   238, 235, 82,  136, 65,  99,  13,  195, 0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   135, 88,  251, 247, 62,  3,   65,  92,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   5,   35,  174, 84,  107, 30,  198, 183, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   226, 206, 161, 253, 114, 213, 218, 145, 0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   11,  40,  71,  9,
      68,  243, 149, 97,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   135, 18,  225, 131, 178, 190, 218, 100, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   229, 246, 13,  134, 231, 48,  239, 94,  0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      113, 61,  29,  193, 134, 253, 64,  12,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   16,  224, 38,  222, 38,  84,  152, 206, 0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   35,  129, 101, 170, 117,
      140, 124, 77,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   141, 16,  32,  139, 131, 114, 67,  36,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   182, 72,  152, 58,  132, 89,  30,  10,  0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   163,
      136, 158, 211, 228, 202, 3,   146, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   144, 89,  192, 123, 224, 234, 52,  73,  0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   226, 120, 51,  187, 80,  230,
      25,  86,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   101, 183, 47,  20,  92,  43,  37,  110, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   128, 18,  182, 233, 75,  49,  229, 201, 0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   75,  95,
      18,  248, 140, 204, 11,  74,  0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   216, 159, 244, 66,  166, 149, 132, 143, 0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   144, 9,   200, 23,  116, 68,  109,
      215, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   35,  132, 220, 14,  248, 57,  116, 127, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   61,  125, 92,  39,  228, 214, 146, 194, 0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   133, 179, 205,
      204, 246, 109, 187, 171, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   99,  100, 233, 40,  247, 111, 49,  51,  0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   53,  240, 120, 100, 99,  78,  106, 194,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   112, 135, 170, 248, 81,  203, 32,  219, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   120, 251, 224, 15,  125, 135, 28,  131, 0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   102, 173, 153, 133,
      47,  241, 197, 108, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   123, 173, 123, 191, 10,  14,  148, 105, 0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,   0,
      0,   0,   0,   0,   0,   0,   0,   0,   0,
  };
  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data),    FORMAT_FRACTAL_Z, FORMAT_NCHW, std::vector<int64_t>({18, 2, 16, 32}),
      std::vector<int64_t>({30, 40, 3, 3}), DT_INT8};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(ret) / sizeof(ret[0]));
  for (int i = 0; i < sizeof(ret) / sizeof(ret[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<uint8_t *>(result.data.get()))[i], ret[i]);
  }
}

TEST_F(UtestFormatTransferFraczNchw, fp16_2c_2n_pad) {
  uint16_t ret[31 * 17 * 4 * 4]{
      14402, 14379, 14633, 14774, 13826, 14869, 13703, 15068, 13624, 13537, 14521, 15114, 13281, 14976, 12899, 15129,
      12289, 12461, 12096, 10696, 14558, 14464, 14389, 15323, 15219, 13308, 14041, 13518, 14388, 15320, 14145, 12633,
      13852, 14410, 14115, 11497, 14819, 13048, 15357, 14681, 15248, 14640, 13757, 13980, 14653, 14856, 11482, 12271,
      13540, 12429, 13971, 14888, 12988, 15321, 14609, 13548, 14965, 15148, 15160, 12685, 14235, 14747, 14941, 15335,
      14810, 12870, 11429, 14307, 14957, 14165, 13507, 15335, 15060, 15034, 15311, 13682, 14949, 14545, 14311, 14478,
      11027, 15231, 14833, 10577, 14400, 14471, 12798, 15263, 15010, 14107, 10812, 14272, 14756, 14933, 11793, 13558,
      12387, 14480, 9432,  14724, 13686, 15178, 7127,  14452, 13930, 14178, 14911, 13538, 13327, 12942, 15308, 14348,
      11288, 10750, 10395, 14163, 14618, 14458, 11317, 14390, 15194, 14971, 13180, 15247, 14157, 12746, 13356, 12826,
      14169, 15177, 11319, 14063, 13328, 14065, 12473, 15172, 12830, 15062, 14396, 14919, 8095,  14663, 15285, 11074,
      14805, 13884, 14428, 14944, 14431, 13189, 10181, 15146, 14422, 14689, 15068, 14981, 14198, 14383, 14843, 14488,
      15159, 14158, 14843, 14811, 13607, 15082, 13934, 13595, 14830, 14087, 15157, 14820, 12592, 14315, 14161, 14704,
      13632, 13253, 14232, 11742, 14661, 11794, 14768, 10439, 15237, 14869, 12268, 14814, 14315, 15076, 14144, 8808,
      8595,  13506, 12548, 13451, 14936, 13946, 9166,  11487, 13611, 15071, 14916, 12429, 14425, 12981, 14181, 13510,
      13403, 12498, 7919,  15307, 14149, 15057, 14686, 13501, 13620, 13326, 14896, 15300, 14529, 12639, 13929, 12700,
      9001,  13153, 15088, 14621, 12825, 14106, 14612, 15019, 13647, 11647, 15259, 15148, 14431, 14651, 15262, 15345,
      14821, 14708, 14305, 14696, 13775, 12556, 14553, 14596, 11235, 14732, 14839, 14391, 13207, 12301, 15282, 13496,
      14065, 13863, 15014, 14777, 14892, 12814, 13125, 15044, 14655, 15189, 14590, 12415, 13728, 14926, 14932, 14764,
      15355, 12676, 14813, 11644, 15105, 12800, 14405, 10732, 12721, 15127, 14816, 12449, 15309, 13831, 13569, 15207,
      14719, 9604,  14686, 13284, 14833, 14904, 15088, 14678, 15095, 14809, 13398, 15326, 14636, 13248, 14586, 14772,
      14958, 15172, 14342, 14089, 14258, 13405, 14868, 14431, 12673, 14949, 13859, 10562, 14993, 13561, 13484, 14952,
      10920, 13863, 13174, 12359, 13959, 15042, 14786, 13899, 14605, 14381, 15227, 13005, 13322, 14089, 14018, 13082,
      9389,  15312, 14400, 14499, 14433, 14837, 13899, 14042, 14997, 12504, 14596, 10240, 14044, 7133,  14595, 14483,
      13408, 14436, 14498, 13580, 14944, 14805, 14672, 12729, 13016, 13193, 15092, 12977, 15251, 15107, 12157, 14744,
      12314, 13490, 12616, 12784, 14866, 15288, 13778, 15029, 12163, 13002, 14842, 13699, 14487, 15243, 11768, 13120,
      11907, 15083, 9064,  15080, 13562, 14200, 14107, 14491, 13819, 12799, 14530, 12957, 11736, 14283, 14595, 13042,
      14759, 13908, 12837, 12547, 12613, 14559, 14426, 14984, 14467, 14962, 14260, 14309, 14643, 13685, 15219, 12760,
      14419, 12270, 13379, 15254, 14804, 14015, 11742, 14901, 11807, 13657, 13213, 14718, 14772, 13837, 13311, 13895,
      15124, 12599, 15281, 13663, 13340, 14732, 15192, 13832, 14527, 10077, 15051, 13593, 14514, 13150, 15300, 13191,
      14591, 15223, 14225, 13678, 12922, 13770, 13051, 9329,  12291, 14056, 14141, 13903, 11930, 15097, 14819, 13038,
      13422, 12463, 13027, 13441, 13934, 13548, 15294, 15243, 15243, 12819, 14495, 14501, 14370, 15066, 15228, 12863,
      14465, 14561, 11927, 14072, 15105, 9506,  15066, 12170, 12578, 14915, 14633, 14105, 9144,  15321, 5703,  15248,
      13347, 14493, 14870, 13255, 14797, 13812, 12955, 14658, 15091, 14919, 15262, 13537, 14312, 14754, 15260, 14447,
      14063, 12352, 14120, 15329, 15133, 14514, 14408, 15279, 13726, 12232, 14959, 13399, 14996, 15065, 14424, 15060,
      14792, 12424, 9458,  10585, 14736, 13592, 14694, 14952, 14586, 14830, 12547, 14625, 12400, 14950, 14851, 14374,
      14776, 14537, 14831, 14373, 15333, 14774, 12938, 15357, 13384, 13289, 13207, 15004, 13933, 15145, 13975, 12681,
      14719, 11611, 15038, 14519, 12573, 13596, 13429, 14441, 14376, 14861, 12909, 14600, 14605, 14604, 14023, 9499,
      15329, 12643, 13931, 15027, 12382, 15240, 14812, 13583, 14716, 13851, 7459,  14481, 14132, 9350,  15247, 14706,
      14763, 14922, 13712, 13187, 15063, 14154, 13719, 15282, 12223, 14027, 13892, 11560, 14396, 14595, 11991, 13362,
      13881, 14035, 14883, 14387, 14096, 14763, 14230, 12512, 12235, 14182, 12596, 12340, 12388, 14753, 12462, 14831,
      14609, 11970, 14616, 11512, 14016, 14776, 13673, 14152, 14832, 13716, 12119, 14614, 12102, 11829, 15047, 13935,
      12510, 14910, 14223, 14306, 13757, 11326, 15331, 14992, 12210, 14593, 10983, 14917, 15068, 11989, 14329, 10062,
      11670, 14490, 13614, 13573, 14299, 15072, 15249, 12920, 11776, 10044, 12957, 10587, 13772, 14282, 14918, 14542,
      15296, 14761, 13670, 14483, 12979, 12365, 12350, 11432, 13509, 14614, 14531, 13746, 13114, 15170, 14216, 14510,
      15309, 15061, 14720, 14117, 14479, 14618, 12389, 13789, 15227, 15248, 13544, 14556, 14727, 13365, 14559, 15345,
      14152, 13666, 13492, 12714, 14525, 12291, 13700, 12309, 14195, 14593, 14509, 11536, 13750, 11849, 11484, 13422,
      14890, 14130, 13209, 14456, 14495, 13529, 15096, 11671, 13535, 4916,  15115, 15142, 15137, 15026, 14921, 14830,
      13556, 15296, 11716, 15265, 14297, 14645, 11518, 15117, 15291, 13955, 13785, 13452, 14107, 15063, 14635, 12740,
      12447, 13639, 14816, 14567, 15025, 14881, 12991, 14427, 15102, 12799, 13379, 12608, 10686, 14496, 14301, 12078,
      13171, 15249, 12861, 14418, 14322, 14303, 13629, 13272, 11448, 11306, 13939, 14624, 13506, 14898, 13155, 13737,
      13672, 8727,  14087, 10373, 13354, 12724, 13212, 11178, 15113, 14428, 14480, 15003, 14924, 13614, 14491, 13762,
      13601, 14831, 10198, 14248, 14119, 13918, 11619, 14970, 14298, 13402, 14792, 15078, 14575, 14180, 14120, 13085,
      14945, 10349, 13979, 15274, 15353, 14721, 13903, 14031, 10803, 14533, 15343, 13146, 14572, 14455, 15118, 14374,
      14355, 14741, 14569, 14673, 15071, 15261, 13330, 14744, 10303, 12502, 9636,  14870, 15097, 11490, 14144, 14772,
      12883, 15080, 15068, 14411, 14392, 13305, 14984, 13344, 13793, 15031, 10720, 14150, 14554, 14996, 12558, 14502,
      14015, 14391, 14409, 12520, 13362, 9595,  13370, 14939, 14396, 15285, 13944, 15225, 15329, 13146, 14257, 12734,
      12895, 14718, 14972, 14462, 15358, 13349, 14435, 15199, 15062, 13372, 11941, 14224, 15000, 15316, 12490, 13597,
      14170, 14301, 13483, 12592, 14349, 15174, 14623, 15014, 15117, 14418, 13620, 15215, 15066, 13678, 13843, 14395,
      14818, 14930, 11160, 14885, 15337, 14179, 14816, 11916, 13010, 14099, 12337, 15046, 11758, 12519, 15011, 14447,
      15274, 14129, 12313, 13104, 15269, 15000, 13469, 14529, 15326, 14479, 14093, 15346, 8366,  12904, 13518, 14932,
      13020, 14326, 13228, 11688, 14668, 14621, 14498, 15187, 9846,  12259, 11449, 14393, 15122, 15215, 15244, 15197,
      14222, 13851, 14602, 14095, 15329, 13367, 14406, 14468, 14612, 11665, 14929, 15237, 13639, 10141, 14844, 14258,
      14246, 15293, 14673, 8674,  12969, 14865, 8639,  14870, 14329, 13712, 13804, 15270, 11621, 14359, 14244, 12867,
      12003, 14246, 11905, 14517, 12346, 14095, 12049, 13745, 12856, 13578, 14913, 15042, 10547, 12764, 11510, 15233,
      11254, 12144, 15025, 12548, 14223, 12241, 14655, 15204, 15285, 14538, 14288, 14430, 14410, 14952, 14862, 13123,
      14901, 14809, 10021, 14583, 13431, 14631, 11804, 14232, 15201, 13574, 12117, 14683, 14117, 14945, 10261, 13901,
      13659, 12441, 15117, 15064, 14558, 13502, 15118, 12588, 15345, 14768, 13934, 14882, 14278, 14318, 13492, 14823,
      14984, 13698, 15153, 13103, 13445, 14964, 14408, 12287, 15235, 13431, 14156, 13610, 14783, 12545, 8340,  14414,
      14857, 15236, 14503, 12692, 13097, 13652, 14563, 13493, 13275, 14246, 14919, 13941, 14055, 10566, 13511, 14873,
      14532, 15026, 13462, 13427, 15137, 14076, 14352, 14771, 15001, 13558, 15329, 14700, 12150, 15021, 12616, 12598,
      14496, 10637, 15339, 13123, 10664, 12994, 12358, 14219, 14679, 13296, 10598, 12601, 12072, 13737, 14608, 12265,
      13605, 14544, 14120, 10293, 12592, 12775, 15207, 14873, 13208, 14159, 13388, 15178, 14949, 12487, 14084, 13458,
      14803, 14678, 10311, 13432, 14963, 14553, 9121,  14458, 13688, 7712,  14816, 13571, 12702, 14273, 14340, 13322,
      14524, 12307, 15128, 14966, 14986, 14754, 11664, 13407, 14351, 13937, 14286, 14787, 14558, 15306, 13877, 15095,
      10024, 13911, 11470, 12557, 12031, 12972, 9385,  8415,  12979, 15047, 15191, 15318, 13354, 14737, 14281, 12678,
      14259, 15175, 12619, 14568, 14361, 14178, 13938, 14643, 11526, 6603,  13799, 12600, 13587, 14678, 14336, 14753,
      14899, 15000, 13442, 12131, 15034, 14496, 14900, 13754, 14595, 14638, 15118, 13036, 14553, 13369, 10026, 12056,
      10336, 13365, 13318, 13572, 15017, 15292, 13038, 15236, 13734, 14787, 14854, 14689, 11582, 14656, 12765, 14876,
      14532, 15001, 14462, 12341, 14861, 13452, 11285, 15176, 14689, 15185, 10596, 14263, 12985, 14651, 15152, 14289,
      12566, 14693, 14980, 13878, 14407, 15024, 11678, 13704, 12398, 14661, 14308, 15030, 12942, 11771, 13970, 11887,
      14884, 14497, 13346, 13997, 15341, 14505, 13124, 12266, 14682, 13100, 14981, 13271, 14217, 14753, 14958, 11132,
      10386, 14887, 10339, 14997, 13711, 14176, 14488, 14775, 14396, 14816, 13629, 10435, 11719, 15329, 15249, 14634,
      14316, 12578, 14495, 13166, 14400, 14770, 14949, 14766, 15209, 13365, 14640, 12379, 10286, 11653, 14512, 14333,
      14828, 14367, 13259, 14640, 15313, 12292, 14965, 13609, 11280, 14185, 14988, 12355, 13961, 14861, 13765, 14876,
      14809, 14256, 14816, 14879, 14709, 14789, 15304, 10730, 14779, 14742, 12668, 14546, 14120, 13370, 13705, 14487,
      14208, 12827, 13700, 15161, 15127, 13212, 14851, 15299, 11610, 13495, 14632, 13567, 15011, 14351, 13677, 13749,
      14952, 13042, 12911, 12277, 8940,  14490, 13899, 13740, 13183, 15285, 12333, 14677, 12922, 14876, 14586, 15255,
      14336, 13771, 13893, 10751, 15160, 14727, 12719, 14521, 13543, 13335, 11533, 14471, 15223, 13650, 12913, 13232,
      12187, 13831, 14669, 14927, 10954, 14735, 15145, 15354, 13745, 14104, 13140, 13314, 14233, 14235, 14076, 9419,
      13557, 13797, 15124, 13409, 13437, 14427, 15224, 14516, 15148, 12570, 15177, 15305, 14738, 14345, 15062, 13113,
      15065, 15268, 14457, 14429, 13989, 15345, 12562, 14693, 12792, 14599, 14705, 12584, 14341, 14666, 14704, 15125,
      15203, 15168, 14563, 14919, 14458, 14713, 15133, 10414, 14050, 14137, 13552, 14631, 14852, 11985, 14657, 13709,
      15318, 13592, 13906, 15264, 14346, 14383, 13413, 11588, 15274, 14520, 13724, 11957, 14378, 15343, 13504, 8430,
      12324, 15333, 15166, 14671, 14929, 13684, 15089, 15132, 14644, 11365, 14799, 14793, 13464, 10510, 11789, 14475,
      13915, 13043, 11106, 14425, 10392, 14334, 14601, 13881, 14806, 12667, 13199, 15038, 15071, 14125, 13853, 14072,
      8520,  15348, 14892, 14677, 14398, 12964, 14925, 15354, 15147, 7907,  15279, 8991,  15134, 13014, 12733, 14976,
      13101, 14720, 14597, 13876, 15291, 12167, 14932, 14552, 14733, 14115, 15194, 14965, 13977, 12078, 15155, 13150,
      14928, 14460, 15320, 13591, 12367, 12637, 13333, 14913, 14071, 14945, 13727, 15046, 13730, 14598, 14505, 13565,
      15148, 14043, 12485, 13313, 14418, 15330, 12989, 14830, 11773, 14848, 13375, 13934, 14450, 14707, 10599, 11062,
      12837, 14486, 14134, 11291, 12551, 14924, 14941, 14537, 11541, 11701, 15246, 13647, 14537, 14807, 12417, 15349,
      14358, 10748, 13921, 14313, 14450, 15036, 11314, 15189, 15325, 15022, 12346, 13532, 11988, 14031, 13781, 13720,
      14794, 14262, 12826, 13999, 14488, 14548, 14486, 14608, 14506, 14795, 13047, 14115, 13453, 14574, 14709, 15089,
      11756, 14985, 13921, 14319, 15095, 15048, 9072,  14221, 14687, 13538, 15146, 15349, 10473, 15117, 14543, 14737,
      13444, 13232, 14758, 14107, 14916, 14465, 13839, 14705, 12335, 14895, 15094, 14380, 14846, 14259, 14560, 14169,
      14120, 14590, 14909, 13828, 14416, 11265, 12928, 14484, 14768, 14556, 13587, 13663, 15224, 12994, 12499, 14668,
      13324, 12376, 14365, 15272, 12284, 9254,  14683, 13013, 14983, 14678, 13719, 13892, 14809, 12920, 12124, 14598,
      15242, 14716, 12603, 14484, 13667, 14577, 13609, 13738, 13610, 14957, 14039, 13726, 14926, 14841, 13801, 14397,
      14846, 14813, 13223, 14103, 13726, 14937, 13802, 13998, 15357, 11519, 14223, 13892, 10310, 12705, 13898, 14420,
      14609, 12310, 13127, 13717, 14808, 15135, 13812, 14919, 13141, 14121, 14334, 15344, 13068, 12426, 14684, 14360,
      14722, 13740, 15191, 14366, 14886, 14963, 9200,  14976, 14685, 14699, 9141,  15272, 14659, 10228, 9283,  13381,
      15019, 14201, 14982, 13275, 12011, 10080, 14860, 15234, 14214, 10350, 13058, 14946, 14381, 12535, 13609, 12958,
      13205, 13379, 12878, 14444, 14410, 9216,  15096, 11936, 12797, 14075, 13097, 15288, 13033, 13332, 13356, 13602,
      13693, 13640, 14189, 14996, 14884, 12579, 12380, 13260, 12342, 13752, 14011, 11327, 13457, 14389, 14633, 11956,
      15019, 13591, 14970, 12885, 14709, 12549, 14703, 13261, 13386, 12190, 14735, 12495, 15043, 14773, 14525, 13638,
      13263, 13190, 15115, 13558, 15335, 14339, 14657, 14576, 14986, 12278, 13165, 14350, 14285, 12520, 14352, 15086,
      13500, 14543, 12713, 14974, 12206, 14805, 12983, 14725, 15346, 15205, 14608, 15288, 14822, 14113, 14154, 14589,
      11845, 10802, 13664, 14438, 14610, 13225, 14806, 15335, 15291, 13492, 13898, 13236, 14320, 11590, 14939, 14067,
      12709, 15231, 14020, 12789, 14247, 15103, 12310, 15300, 14202, 14934, 14846, 15099, 14747, 13318, 14956, 11622,
      13293, 11537, 14562, 14954, 14466, 15231, 14985, 14602, 15196, 11438, 15077, 13013, 15336, 14054, 15259, 14921,
      14946, 13782, 14515, 15163, 15209, 14635, 13588, 14531, 13896, 15002, 14128, 12954, 13115, 12727, 14605, 14765,
      12727, 10729, 15122, 13587, 15182, 12600, 12573, 13185, 15246, 13212, 14801, 12867, 12970, 14045, 12805, 13278,
      12279, 13332, 15343, 11654, 14639, 14236, 14206, 13996, 14830, 13807, 14569, 14925, 11143, 15238, 12921, 14506,
      13721, 11635, 11395, 13805, 13531, 12448, 13846, 14956, 14969, 14459, 8896,  11699, 15268, 14760, 14092, 15108,
      14261, 13524, 11082, 15267, 14698, 14538, 15246, 13931, 13360, 12631, 11603, 12644, 14511, 14441, 14138, 11403,
      6931,  11244, 14740, 12551, 13701, 14025, 14010, 14093, 14089, 10293, 14566, 11685, 14287, 13800, 14778, 11014,
      14579, 12494, 12097, 15187, 13845, 14585, 14226, 14721, 12898, 13520, 15100, 14909, 14914, 12903, 11760, 14973,
      15140, 13989, 15132, 12325, 11439, 14354, 13598, 14044, 13476, 15127, 14442, 15264, 12641, 14482, 14176, 12967,
      15252, 15284, 14941, 12852, 14604, 14734, 13837, 14652, 11673, 11775, 13576, 14874, 13277, 14181, 12487, 11848,
      14686, 13862, 14863, 14838, 15190, 12476, 14150, 15086, 10807, 15109, 15088, 11388, 14756, 15116, 14835, 13624,
      13994, 14603, 14428, 14401, 9397,  15154, 15255, 12694, 14838, 14629, 12282, 14071, 12932, 14771, 13389, 14924,
      14934, 13622, 12155, 15213, 13368, 14948, 15338, 12117, 13520, 14842, 14433, 12527, 11826, 14710, 15044, 15303,
      14897, 12856, 9555,  15251, 14343, 14140, 14668, 14414, 13466, 14325, 15299, 12382, 14039, 12219, 13351, 15065,
      14748, 15165, 13747, 14928, 14534, 15315, 13606, 14041, 14693, 14465, 13872, 14128, 13819, 14550, 13906, 11164,
      12729, 13434, 15117, 14501, 14046, 9922,  13786, 15146, 14963, 13558, 13745, 14680, 15342, 15131, 14751, 14361,
      14560, 14572, 9091,  15157, 15277, 15067, 14572, 14382, 11844, 15010, 13441, 11357, 15009, 12956, 11136, 13992,
      14591, 14424, 11442, 10234, 14000, 14953, 11525, 11595, 12761, 13803, 14853, 14657, 11671, 15206, 12328, 14433,
      14318, 15242, 15334, 14105, 13293, 14587, 13374, 13929, 14496, 15010, 13529, 13139, 11919, 14729, 15036, 14897,
      13643, 14869, 14680, 14818, 14977, 13522, 14858, 13828, 14504, 14643, 12404, 15323, 14493, 14452, 14690, 14085,
      15213, 12322, 14657, 12091, 10428, 15237, 15018, 12504, 14724, 10557, 12848, 14513, 15019, 14133, 15177, 13432,
      14626, 14700, 15288, 14329, 14737, 14907, 14807, 15160, 14947, 14667, 14324, 13166, 14550, 14825, 13000, 13679,
      14254, 15336, 14832, 10152, 14760, 13558, 15229, 15236, 13740, 13447, 12418, 15126, 14406, 14782, 13450, 12375,
      14498, 13617, 15327, 14803, 14594, 11370, 15316, 15088, 15234, 14523, 13344, 13490, 14474, 13950, 14235, 14695,
      15019, 10878, 14530, 15328, 13517, 14899, 13853, 13170, 15091, 13688, 13374, 12492, 12501, 14989, 12819, 14953,
      15162, 11790, 14762, 15142, 14970, 14456, 15130, 15254, 7194,  12757, 14484, 14094, 14314, 14822, 14293, 12370,
      14222, 14230, 11369, 10653, 15112, 13350, 14369, 15086, 15065, 14386, 12633, 15206, 14509, 14550, 11417, 12341,
      14382, 13544, 12046, 12672, 13351, 11326, 13994, 11902, 14590, 11731, 14552, 14004, 14271, 14960, 11860, 15181,
      14808, 15139, 14239, 12911, 11824, 14597, 15006, 14226, 14707, 14541, 10802, 14287, 12944, 13546, 13690, 14370,
      15215, 12874, 14444, 14842, 15183, 14251, 14406, 14445, 14749, 14117, 13784, 10214, 7321,  15291, 13293, 13683,
      14977, 14437, 14427, 14444, 13677, 13013, 15073, 15348, 5073,  15042, 12116, 12793, 14959, 15047, 14079, 13684,
      14343, 15094, 13630, 8237,  15109, 14686, 14783, 15218, 14503, 10056, 12288, 13024, 14552, 14721, 14799, 14325,
      14110, 10034, 14440, 14515, 12718, 14832, 14370, 14063, 14474, 12985, 14760, 14627, 13422, 12796, 15218, 13300,
      14728, 13460, 14185, 12908, 9219,  14892, 15032, 13379, 11454, 13616, 12854, 7760,  14685, 13049, 14458, 13569,
      14801, 12359, 15358, 14428, 15136, 14019, 14014, 13255, 14668, 8243,  11920, 15161, 14722, 13984, 13535, 14682,
      14096, 13631, 15096, 9218,  15347, 13791, 10685, 14481, 13793, 13429, 14307, 15253, 14798, 14351, 13315, 13562,
      14190, 13440, 14596, 14433, 11343, 11530, 4925,  14968, 10188, 15344, 15094, 10408, 15248, 15339, 12326, 14868,
      15069, 13762, 13507, 15114, 9799,  14562, 14985, 14077, 13698, 13323, 11698, 12813, 13906, 13726, 14279, 14656,
      14546, 14905, 12182, 14777, 12893, 12312, 14101, 9896,  14965, 14677, 10655, 12688, 15278, 14888, 14893, 14537,
      14639, 12550, 15128, 14666, 14379, 14723, 13339, 15116, 14368, 14744, 11817, 15255, 13789, 13541, 11920, 15188,
      13854, 14582, 15258, 13371, 15119, 13665, 15075, 14713, 13231, 15155, 11286, 13870, 12703, 14481, 14464, 9057,
      13561, 12519, 13328, 14986, 14656, 14479, 13804, 10682, 12985, 14935, 12912, 13770, 12355, 11449, 9670,  13714,
      12404, 14664, 13382, 11184, 14713, 14093, 14547, 13433, 14669, 14482, 15163, 13946, 14521, 15145, 15145, 14465,
      14199, 14553, 14445, 14102, 14743, 14238, 14464, 14524, 14546, 14885, 14920, 15027, 13515, 15314, 11534, 14510,
      12155, 13654, 15311, 13957, 13445, 14112, 14298, 15358, 13316, 13472, 14672, 11218, 14498, 14862, 12507, 12882,
      13052, 15352, 14266, 15101, 15284, 9318,  13471, 15069, 11003, 9380,  14529, 13474, 9958,  14949, 12900, 13514,
      14615, 15101, 14393, 11326, 14671, 15011, 11166, 15167, 15198, 14745, 13075, 14726, 15012, 15320, 14952, 15251,
      8692,  14574, 14805, 13973, 10449, 13937, 14144, 14675, 10440, 15337, 13138, 14544, 15088, 15218, 15346, 14690,
      14454, 14744, 15317, 14778, 14915, 14942, 13830, 15288, 14237, 13099, 13504, 12516, 14775, 13523, 13570, 14821,
      13955, 12251, 15329, 14560, 11675, 15305, 14830, 14473, 12337, 14818, 14404, 13443, 14876, 13857, 10871, 15016,
      14348, 11255, 14980, 14038, 13835, 14487, 14440, 9509,  15004, 11235, 13522, 13627, 14482, 14019, 10659, 14776,
      14669, 14438, 13958, 14695, 15073, 13004, 13469, 10959, 13679, 13413, 14504, 14660, 14983, 14322, 12697, 14634,
      11157, 15262, 13184, 14070, 14294, 14928, 14581, 14592, 14384, 12975, 15314, 13474, 13877, 13756, 14657, 11925,
      14725, 10879, 14426, 14095, 13314, 13728, 14996, 11853, 14118, 15112, 14969, 12360, 15083, 14519, 15218, 14755,
      14013, 14460, 14342, 15143, 14800, 14388, 10658, 13742, 15351, 14182, 11547, 14074, 13026, 13801, 14874, 14577,
      10298, 14690, 14592, 13917, 15093, 13848, 15141, 12502, 15011, 12813, 14734, 14088, 15222, 13015, 13795, 14514,
      13418, 14639, 14623, 14689, 13773, 14156, 15300, 14591, 13250, 14332, 14905, 15354, 13686, 10633, 11730, 14490,
      10995, 15314, 14516, 15251, 14372, 14883, 12828, 14779, 14965, 14958, 15107, 13954, 15189, 14755, 14170, 14143,
      13411, 15079, 15332, 14533, 14477, 12481, 11910, 14374, 13119, 13170, 13908, 13999, 10315, 11690, 14592, 12098,
      14635, 15153, 15033, 14213, 14362, 14363, 15233, 12589, 14695, 15205, 9897,  12297, 12789, 14555, 14551, 15029,
      14876, 14607, 14530, 14681, 15249, 14664, 14074, 14430, 14839, 15230, 13781, 12222, 13710, 14487, 15168, 14748,
      14851, 13996, 14147, 13595, 13173, 13576, 14825, 15011, 13717, 11066, 8014,  14771, 14859, 14256, 12243, 14516,
      13306, 14167, 14488, 14483, 14433, 14540, 10296, 13495, 15021, 15054, 14788, 15322, 11598, 13564, 14472, 14139,
      13907, 14901, 15019, 14818, 15222, 14525, 6472,  13969, 14399, 14337, 14205, 13871, 13848, 12702, 13646, 9531,
      14367, 13147, 15210, 14307, 14375, 12863, 13301, 14488, 14391, 12766, 10000, 13605, 14604, 11090, 14970, 13087,
      13818, 14311, 14461, 14756, 12764, 13897, 14411, 14595, 14016, 15359, 12874, 14143, 9806,  15349, 14591, 15199,
      14603, 12178, 14918, 14408, 10838, 14980, 13930, 13054, 15065, 12418, 13865, 7336,  15131, 13077, 14193, 14901,
      13917, 14388, 15027, 14375, 13635, 12519, 13658, 12588, 11456, 14437, 14373, 14738, 11235, 14269, 15329, 12515,
      13571, 15357, 15086, 14778, 14860, 10133, 13691, 14097, 15173, 14934, 14442, 14951, 14405, 15188, 13948, 15335,
      12292, 10437, 14963, 13320, 12041, 14768, 12774, 14656, 13852, 13624, 11927, 15193, 14651, 15267, 14617, 14867,
      13815, 15084, 14973, 15238, 14606, 14491, 14457, 12787, 15143, 13930, 12877, 13706, 7898,  13954, 14947, 11315,
      14143, 14408, 14397, 15167, 11191, 14279, 13385, 11270, 12510, 14445, 9398,  12842, 14241, 14889, 12265, 15279,
      12729, 14791, 14615, 12734, 13448, 15303, 12395, 11327, 13078, 11758, 15062, 14169, 11721, 14325, 11429, 15326,
      14858, 14912, 14523, 11814, 12967, 14788, 13480, 14427, 13523, 11151, 14598, 13381, 14579, 13875, 12986, 13345,
      15061, 14584, 14954, 12876, 15126, 13938, 15240, 9596,  12696, 15070, 12366, 14768, 13142, 14970, 14938, 13394,
      15024, 13196, 14492, 13531, 14826, 14218, 14183, 13774, 12475, 13375, 13106, 13068, 13406, 13963, 11598, 14553,
      15344, 14388, 14660, 14709, 14981, 13993, 14804, 14524, 15034, 14845, 15232, 13489, 13562, 14442, 14570, 10464,
      14674, 14588, 15128, 11673, 14958, 12747, 13111, 14488, 12986, 14640, 14749, 12370, 14886, 14405, 14835, 13564,
      14804, 14193, 14428, 13621, 15111, 14766, 12161, 15046, 13311, 14535, 14962, 9925,  15058, 13956, 14436, 15067,
      13627, 14915, 13198, 15265, 11041, 14758, 13433, 14663, 15095, 14063, 15242, 14721, 13923, 14661, 14768, 13190,
      13595, 15183, 15025, 14342, 13992, 15134, 13397, 13406, 14617, 12811, 14564, 11967, 12291, 14439, 13390, 15029,
      12462, 14922, 11796, 13895, 11696, 11550, 11153, 14576, 14512, 13791, 15028, 15162, 13886, 14251, 14957, 13106,
      12942, 15118, 13754, 14828, 14786, 13267, 13484, 14093, 14924, 14659, 14269, 14758, 14662, 14473, 13054, 12338,
      14556, 14491, 10344, 15154, 13659, 15176, 13262, 14342, 15033, 14855, 10100, 15157, 14492, 14590, 14693, 15167,
      14479, 13039, 12936, 15105, 15300, 14390, 14721, 15286, 14022, 15269, 15054, 15129, 8785,  15279, 15213, 14579,
      15022, 14974, 13814, 14826, 12961, 13032, 15101, 14650, 14667, 14759, 13584, 14126, 14705, 15312, 13255, 12368,
      14480, 14845, 13778, 15055, 13481, 15199, 14392, 14377, 13732, 12418, 13193, 10893, 14093, 14900, 14875, 14287,
      14651, 13376, 13428, 14856, 13953, 7701,  12967, 13071, 13972, 14590, 13375, 13267, 13319, 15103, 14467, 14676,
      15247, 14354, 12732, 12222, 14445, 12821, 13096, 14216, 10315, 14903, 14611, 14922, 14371, 15129, 4195,  14490,
      14809, 15131, 14407, 14474, 12547, 14612, 15050, 14870, 13957, 15111, 11912, 14078, 13297, 12912, 12705, 14064,
      14470, 12635, 13246, 14890, 14996, 14240, 14258, 11484, 11762, 12959, 14210, 14992, 14752, 14690, 14166, 15245,
      15184, 14944, 14762, 13533, 13247, 13172, 13298, 11818, 14519, 14816, 14297, 15240, 14933, 14609, 14194, 14118,
      14802, 14679, 10268, 14549, 14583, 14106, 13745, 13472, 12738, 15175, 13091, 14910, 13149, 13013, 15188, 14384,
      12978, 13913, 14164, 15012, 15063, 14524, 13910, 14018, 11713, 13615, 15098, 11827, 14319, 14625, 14352, 15154,
      15009, 15109, 13468, 15206, 15331, 11571, 13163, 14590, 13805, 14527, 14903, 14301, 15048, 14788, 15112, 14798,
      15202, 13552, 14436, 6181,  15349, 14674, 13595, 14654, 13776, 14460, 15187, 14121, 10985, 14695, 13565, 15267,
      14529, 14077, 11718, 15025, 15059, 14203, 15296, 14050, 14075, 13587, 14270, 14654, 15077, 13120, 12581, 13168,
      13084, 15065, 14979, 13511, 10628, 14836, 14917, 13534, 13017, 13475, 13106, 14724, 14658, 14112, 14909, 14834,
      11909, 9046,  14861, 14215, 14658, 14982, 15165, 14611, 13693, 13530, 12456, 11465, 15309, 13991, 12447, 14392,
      12829, 14797, 13928, 14497, 14949, 15277, 13876, 13152, 15306, 13578, 14514, 14836, 15038, 14719, 13551, 14985,
      15295, 15152, 14641, 14852, 13345, 14391, 14900, 14750, 13871, 14655, 10900, 14496, 14519, 13317, 12369, 13895,
      14812, 14474, 14702, 14441, 13287, 14544, 13580, 13006, 13714, 14145, 11718, 13914, 14080, 14953, 12856, 13405,
      14878, 11354, 12458, 15245, 6968,  15275, 13654, 15123, 14722, 13091, 15114, 14521, 13682, 14861, 13661, 15033,
      15096, 13724, 14597, 14216, 15168, 15191, 13811, 14961, 13695, 14493, 12913, 10037, 14333, 14541, 14488, 14233,
      13653, 14821, 14620, 12786, 13684, 13172, 15324, 13266, 14558, 15283, 13413, 13585, 11960, 12677, 14803, 14648,
      14935, 8686,  14603, 14495, 15181, 8810,  14953, 14259, 10367, 13338, 13196, 13562, 12674, 14062, 12494, 13778,
      13949, 15155, 10232, 13347, 12586, 13700, 15001, 15053, 14406, 15283, 15113, 13179, 14814, 13603, 14553, 12922,
      15068, 14645, 14343, 12467, 14444, 13263, 13261, 13472, 14442, 14787, 9223,  14937, 13905, 11757, 11957, 9520,
      15269, 14280, 15360, 12646, 15134, 10573, 14929, 13898, 14412, 13546, 14720, 14686, 14884, 13703, 15300, 13830,
      13831, 13560, 12329, 15206, 12724, 13434, 10310, 15258, 13807, 13712, 14567, 13899, 12316, 13908, 10307, 14078,
      13386, 14454, 14416, 14573, 13452, 14279, 15148, 15107, 14965, 11718, 14824, 14493, 14700, 12437, 11859, 13337,
      14766, 12429, 8166,  13597, 14371, 12949, 14903, 12516, 12142, 14405, 14460, 13764, 12754, 9314,  13746, 15191,
      14954, 11976, 11019, 12273, 15022, 15185, 14690, 14833, 12032, 13876, 14432, 13087, 14468, 12550, 15120, 14099,
      14239, 7793,  15011, 13346, 12886, 14524, 13676, 14551, 14047, 15129, 15139, 15281, 14064, 8987,  11608, 14020,
      14171, 13349, 14075, 11284, 14345, 12222, 14771, 11824, 14323, 13036, 14575, 13444, 15130, 14382, 14829, 15230,
      11621, 11970, 14114, 14446, 9328,  9011,  12151, 11692, 14402, 11492, 14995, 13940, 11870, 11060, 15028, 13925,
      15245, 11974, 14414, 12268, 14528, 14474, 12725, 10772, 14032, 14549, 14689, 13250, 13111, 15021, 14252, 14606,
      13899, 11055, 14717, 13935, 13394, 14669, 13397, 12845, 15062, 13751, 13067, 14737, 14785, 14353, 12338, 14374,
      15152, 14006, 14340, 15023, 14302, 13270, 9584,  14651, 9547,  11630, 9113,  14233, 13544, 14740, 13494, 7937,
      15235, 13838, 15139, 12175, 15294, 12139, 13397, 13355, 15305, 15334, 15235, 12423, 14927, 14386, 14231, 15081,
      12208, 12354, 14753, 14930, 13302, 15027, 13815, 15317, 13396, 12925, 14072, 15178, 12934, 15250, 13788, 15094,
      14666, 14891, 15240, 14322, 15292, 14873, 14885, 14194, 14386, 14550, 14282, 13957, 10080, 15262, 12999, 14374,
      13278, 15255, 14814, 12500, 12829, 15351, 14535, 7340,  14355, 13560, 14710, 11705, 14046, 9436,  14335, 15336,
      14746, 13720, 14748, 14429, 15314, 10338, 13590, 12994, 12711, 10352, 15017, 15191, 12328, 14217, 14031, 8978,
      14092, 14502, 14476, 14175, 14816, 11195, 14330, 7635,  12453, 14345, 14294, 13076, 13732, 13844, 12662, 14746,
      13507, 10339, 14488, 14375, 14267, 10468, 13784, 14559, 15122, 12805, 12492, 14108, 14914, 15278, 14483, 14837,
      15111, 14936, 11043, 14225, 12337, 12922, 11299, 14401, 11737, 15336, 13506, 14963, 13920, 13635, 15057, 12443,
      13682, 13696, 14510, 14464, 13664, 12852, 13677, 15101, 9868,  13736, 11041, 14682, 10934, 13995, 13938, 14773,
      14883, 15107, 12898, 10470, 13757, 15329, 13268, 11995, 9404,  14685, 13773, 14005, 15350, 14881, 8440,  11460,
      15186, 11576, 13609, 13521, 15094, 13019, 15053, 14869, 8654,  14177, 13757, 15184, 14796, 15223, 14639, 14236,
      12843, 15216, 12786, 14523, 14433, 15216, 15359, 11906, 13810, 13255, 9798,  14534, 13997, 14603, 14175, 8503,
      14421, 15054, 13935, 13100, 14262, 14025, 13674, 14659, 11623, 14140, 12974, 12839, 14814, 14039, 15089, 14702,
      12372, 14624, 14567, 13429, 15222, 12075, 14833, 14950, 14351, 15219, 14446, 15109, 10346, 14987, 10597, 14511,
      12425, 13412, 12992, 14254, 15285, 14664, 12385, 10345, 14662, 12555, 12195, 14780, 14982, 14338, 13834, 13591,
      13297, 14306, 11696, 12366, 15004, 10601, 15337, 11481, 9434,  14448, 9658,  15302, 12412, 14804, 15083, 13645,
      13970, 10568, 15028, 15032, 14089, 14565, 14208, 11671, 14048, 13703, 13742, 15102, 11890, 15039, 15312, 12096,
      15276, 14587, 14636, 12712, 11922, 14456, 15157, 15243, 14453, 13956, 14931, 14704, 13211, 13950, 14730, 14453,
      11920, 13559, 14499, 15247, 15192, 14549, 14784, 15227, 15195, 14323, 14846, 15217, 14526, 14757, 14421, 14490,
      15006, 15300, 14751, 14670, 14744, 13911, 14540, 13809, 14929, 15171, 14542, 11798, 7608,  15027, 15270, 13910,
      14257, 14269, 15252, 15326, 14025, 15072, 14971, 15075, 13726, 14701, 14320, 14783, 15091, 14131, 15034, 14809,
      13132, 12607, 14273, 8883,  15242, 14912, 14463, 15209, 13899, 14203, 12161, 14588, 14696, 15027, 10214, 9859,
      14866, 14541, 13307, 13942, 14988, 14966, 14055, 11382, 12514, 13357, 15116, 15198, 13887, 14733, 14149, 11043,
      13398, 14415, 13844, 14724, 13491, 12650, 12399, 15318, 13635, 14545, 12761, 15147, 15345, 12575, 12903, 15091,
      15335, 12965, 11071, 14433, 15083, 14579, 14457, 14605, 13615, 14625, 14822, 12612, 13824, 14899, 13272, 12568,
      15000, 15145, 14553, 12196, 13125, 14821, 14490, 10374, 15060, 13522, 13260, 10312, 14880, 15147, 10594, 13063,
      13704, 10865, 12454, 14995, 11604, 14229, 14423, 14656, 12259, 14356, 15196, 13328, 14774, 14539, 15231, 12664,
      15281, 12531, 11382, 14129, 13901, 14870, 6364,  14784, 14948, 13543, 14983, 14922, 14041, 14019, 13343, 12483,
      13112, 14627, 13440, 15098, 10310, 13792, 13283, 14056, 13234, 11589, 15271, 15029, 13713, 14406, 14936, 12531,
      13769, 14380, 14607, 14725, 10571, 15332, 13591, 13719, 13610, 13550, 12995, 10556, 15009, 11221, 14773, 13754,
      14760, 14476, 15161, 14151, 12334, 14001, 13908, 14638, 14556, 14519, 13760, 14484, 15119, 14913, 15295, 15352,
      13475, 13385, 13849, 14591, 13386, 14882, 13477, 14897, 15331, 14323, 11677, 12727, 14820, 15175, 12305, 15089,
      13698, 12942, 14990, 15325, 10272, 14616, 13892, 13245, 12515, 15206, 15335, 13751, 14655, 15293, 14615, 13728,
      14499, 10869, 13858, 10495, 13301, 15217, 10530, 15223, 15072, 15103, 15026, 14369, 12334, 13944, 13113, 11477,
      15030, 11206, 14570, 12650, 15064, 15235, 14196, 14995, 11554, 14940, 15032, 14926, 14429, 9951,  7949,  12352,
      14342, 14110, 14867, 15110, 11879, 13263, 9198,  14179, 11654, 10586, 14363, 14853, 13663, 15132, 15000, 11202,
      15154, 15039, 14131, 15086, 14576, 13333, 14071, 14903, 14101, 14358, 14872, 15310, 14259, 14276, 14883, 12859,
      14952, 9443,  15356, 13387, 10482, 13661, 14759, 14925, 13582, 14954, 13753, 13435, 11618, 15309, 14394, 8557,
      9074,  12794, 12642, 13799, 14958, 14636, 15199, 13112, 14605, 13572, 15275, 10187, 12647, 11477, 15307, 13822,
      12046, 14946, 13217, 14354, 11616, 9404,  15175, 15073, 12639, 13673, 14934, 14401, 15121, 13522, 13544, 13402,
      14860, 14995, 11878, 15048, 14554, 15318, 12198, 15073, 14676, 14714, 15101, 15200, 14091, 9043,  14360, 13676,
      14751, 14134, 14944, 14023, 14794, 15017, 12898, 15021, 14604, 7564,  13562, 12906, 14538, 14880, 14421, 10691,
      13603, 14693, 10750, 15143, 13244, 15227, 15348, 12597, 15274, 9717,  14657, 15279, 13483, 14893, 14897, 15003,
      11537, 14666, 14647, 14346, 14444, 14077, 11350, 15350, 14545, 14139, 11250, 15019, 15153, 10032, 14917, 11804,
      14708, 14989, 13837, 14518, 14136, 11937, 10184, 9024,  14700, 12563, 13622, 13355, 15303, 12896, 11127, 14072,
      12316, 11793, 14591, 14516, 14047, 15058, 15169, 14790, 14400, 14969, 15131, 14651, 11594, 12839, 12531, 11966,
      11701, 12480, 14343, 14706, 14854, 8858,  14391, 12986, 9641,  12041, 15102, 13387, 14800, 13480, 14685, 15007,
      14308, 15113, 12894, 14400, 14623, 14339, 14836, 14229, 14999, 15183, 15186, 14433, 14366, 15326, 12592, 13977,
      11710, 13660, 14669, 13687, 15082, 11505, 15240, 14765, 11556, 14301, 14419, 14551, 11562, 8593,  15233, 14454,
      13911, 13025, 14586, 13830, 13854, 14795, 14585, 13021, 15043, 14382, 14856, 13391, 14498, 10118, 15031, 14459,
      15075, 14645, 14970, 14554, 13204, 15063, 11593, 15236, 11083, 14828, 14402, 14118, 15267, 15130, 14239, 11890,
      14568, 14029, 12816, 14444, 14768, 14568, 13293, 12624, 14104, 15130, 11097, 14722, 14114, 14666, 15016, 14098,
      12946, 14780, 13995, 14353, 14714, 14210, 14839, 14827, 14856, 14383, 15183, 14891, 13669, 12011, 13867, 14407,
      12962, 14497, 11248, 15151, 13330, 14573, 14629, 11126, 14386, 14505, 15235, 13990, 12398, 12737, 15320, 13689,
      12315, 15299, 15353, 14423, 12857, 13913, 11089, 14815, 12375, 14363, 14167, 5843,  14270, 10912, 15334, 14974,
      12828, 15145, 14566, 8487,  14109, 13482, 14511, 13124, 12995, 10760, 9888,  12131, 13029, 14774, 14753, 11621,
      14600, 12294, 14838, 14858, 10505, 8317,  12171, 11843, 13416, 13472, 14476, 15192, 15225, 12419, 12701, 11774,
      15237, 14905, 12767, 13453, 15276, 14624, 14492, 14994, 14927, 12977, 13534, 15148, 14450, 14430, 14847, 12809,
      12665, 14661, 14363, 14810, 13856, 15180, 14863, 10524, 15252, 15338, 12744, 14414, 15071, 14359, 15179, 14080,
      12387, 11443, 12677, 14534, 13383, 15047, 14229, 14922, 15085, 15199, 14564, 14604, 15015, 13123, 14850, 14352,
      15167, 13140, 15029, 13427, 15325, 13804, 14680, 13076, 14730, 14584, 12131, 15079, 13722, 12935, 14042, 12959,
      14544, 14311, 14034, 13325, 14371, 15037, 15245, 13394, 15341, 14846, 15200, 12226, 14481, 15153, 13791, 13544,
      15042, 14003, 10915, 14918, 13744, 14885, 13817, 14614, 14234, 14324, 13018, 7803,  13872, 14394, 14559, 14411,
      12462, 13561, 14965, 15296, 14888, 13830, 14374, 15058, 9350,  12027, 13109, 10866, 11952, 14949, 11986, 12972,
      15284, 15314, 14393, 15080, 13009, 13054, 14489, 14349, 15060, 15346, 13829, 14738, 13001, 14113, 15197, 15266,
      14667, 14892, 13749, 14842, 13484, 15277, 14934, 11430, 14652, 14523, 14142, 14102, 14710, 14684, 14747, 14629,
      14184, 14809, 14888, 14704, 13659, 8931,  12806, 14686, 13643, 13870, 14821, 13970, 14538, 13348, 15275, 15013,
      14449, 13535, 12912, 14026, 14535, 14791, 14636, 14919, 13215, 13064, 14718, 15067, 14735, 15086, 13160, 15166,
      14428, 12842, 14157, 13991, 13033, 14553, 14843, 13633, 10203, 13332, 11350, 14543, 15116, 12414, 14448, 14163,
      14351, 15227, 12477, 10914, 15251, 15165, 12029, 13639, 14742, 14160, 13458, 14071, 12694, 13651, 15170, 13738,
      15348, 14534, 13838, 14816, 13440, 14583, 13870, 14790, 12977, 13485, 14263, 14180, 14141, 13458, 14344, 13829,
      15299, 13468, 12290, 15045, 13479, 13324, 12783, 13423, 15356, 14701, 14819, 15247, 15168, 14803, 15001, 14381,
      14783, 14777, 12232, 14649, 13830, 13881, 14715, 14598, 14228, 13714, 14899, 11606, 14990, 15243, 14447, 14666,
      12810, 12869, 13419, 14902, 14471, 12596, 13506, 14675, 14781, 13091, 11836, 13890, 13575, 12783, 13626, 15249,
      14127, 11685, 14643, 14866, 15209, 14734, 15174, 13694, 14448, 14463, 15002, 11493, 14597, 11518, 14376, 14878,
      14890, 14879, 14877, 14437, 12633, 14694, 14376, 13787, 13935, 14509, 12015, 13911, 10412, 14356, 13635, 14485,
      14184, 13661, 14879, 14425, 13784, 13115, 13423, 12274, 14956, 14562, 13231, 14480, 13787, 14634, 13367, 15023,
      9201,  14314, 13383, 14023, 12862, 14394, 13738, 14260, 14205, 11792, 13408, 13782, 13336, 14364, 13908, 11462,
      12311, 12576, 14649, 15283, 13615, 9514,  12444, 11583, 14368, 15151, 11930, 13709, 15341, 13764, 14136, 14618,
      15125, 13520, 13527, 14573, 13439, 13875, 14558, 13380, 15054, 15324, 13611, 13827, 14384, 14562, 11662, 15282,
      13550, 15102, 14441, 14357, 11794, 15347, 12865, 13426, 15304, 12958, 14337, 13780, 14827, 13552, 13239, 9025,
      15200, 13322, 15160, 11566, 13377, 12064, 14347, 12401, 14781, 13065, 15196, 11400, 13904, 11792, 15069, 10056,
      13236, 12698, 13934, 15055, 13452, 14342, 15335, 11267, 10657, 15243, 13508, 15072, 14229, 13272, 11325, 12482,
      14286, 14268, 13328, 14607, 15347, 11916, 13690, 15311, 14554, 13239, 12894, 12473, 14331, 14166, 14682, 14399,
      14352, 13505, 14586, 12724, 14386, 13715, 13522, 14742, 13900, 15350, 11714, 14587, 14880, 13093, 9847,  11935,
      14444, 15323, 13407, 12836, 14979, 13902, 14786, 15215, 14452, 12792, 15190, 11537, 14209, 14020, 14223, 14529,
      14558, 14348, 13580, 13413, 13897, 12100, 14362, 13759, 13681, 12425, 11316, 12289, 10221, 14624, 14389, 13849,
      9967,  13954, 13475, 13747, 12796, 13801, 14716, 9595,  13863, 13439, 14151, 13455, 14378, 14673, 14973, 15329,
      14572, 14903, 11751, 14717, 12501, 14774, 12440, 13786, 13466, 15105, 15326, 14526, 15247, 11360, 15085, 10444,
      15254, 13785, 15161, 14463, 13963, 12740, 14663, 12602, 12423, 15022, 15287, 15094, 13762, 14245, 15009, 10458,
      15045, 14378, 14851, 15013, 13629, 14707, 14844, 14856, 13130, 11235, 14572, 12745, 15124, 14856, 14663, 15164,
      13801, 13589, 14579, 14579, 15051, 13822, 14773, 14785, 15163, 15296, 14340, 9218,  13800, 14758, 10946, 14283,
      14841, 15077, 12361, 12294, 14902, 15298, 9077,  14738, 11731, 13945, 14503, 15088, 14845, 13310, 12595, 15047,
      12947, 14844, 15143, 14879, 15350, 14155, 12870, 14311, 14533, 14790, 14626, 15056, 14774, 14676, 13998, 15122,
      15065, 14877, 11779, 15221, 12400, 14852, 11268, 12707, 14894, 14394, 15076, 15208, 12305, 12681, 9601,  15061,
      14880, 14023, 14347, 14948, 14509, 14043, 13449, 14501, 12325, 13123, 14068, 14648, 15287, 11917, 14156, 15336,
      14239, 14486, 15340, 15299, 13238, 15250, 14771, 14582, 14494, 15251, 12422, 14387, 14592, 15018, 11295, 13797,
      15183, 14448, 15178, 13466, 14809, 13527, 14844, 13607, 14009, 14424, 14483, 14917, 13862, 12183, 14518, 15325,
      12408, 14507, 14467, 10446, 15239, 14589, 13446, 13323, 15151, 13850, 15212, 15032, 9171,  13263, 12537, 14232,
      14605, 13758, 13465, 15047, 15104, 12808, 13647, 14156, 12925, 15040, 13402, 12307, 13419, 13681, 10824, 15121,
      14977, 15168, 8601,  13726, 13226, 14671, 14683, 15115, 14477, 10156, 14674, 9955,  14593, 14529, 14509, 13841,
      12482, 12245, 14510, 13818, 15030, 12719, 13705, 11588, 14835, 14728, 14647, 12872, 15168, 12440, 14633, 14858,
      14857, 14379, 11762, 15354, 12257, 13862, 13822, 15100, 15222, 14653, 14210, 10061, 12761, 15073, 12603, 14648,
      14744, 14242, 14033, 14450, 14353, 15096, 14537, 14608, 14148, 14565, 15006, 14341, 14892, 14063, 14577, 14813,
      11292, 11658, 13531, 12399, 13550, 14930, 15140, 14805, 14077, 13784, 14530, 14633, 14663, 15030, 14325, 15091,
      11937, 14744, 13428, 15264, 10249, 14423, 12749, 15186, 14784, 13583, 15210, 14675, 14776, 11081, 14560, 12648,
      14706, 11331, 13998, 15216, 14165, 13174, 15121, 14098, 13705, 14000, 12721, 14866, 14976, 14591, 12694, 14668,
      15322, 14951, 11058, 15054, 15016, 14918, 15204, 10408, 13474, 15004, 13927, 15325, 13314, 11958, 11881, 15276,
      13482, 15038, 13442, 14677, 15044, 14393, 13774, 13179, 14610, 15110, 15354, 13700, 15073, 15136, 15345, 10408,
      12420, 14948, 15027, 14375, 15211, 14673, 14670, 13165, 14458, 15241, 12381, 12137, 12674, 13898, 15215, 10465,
      14965, 15088, 11686, 14636, 15160, 13466, 14668, 12900, 14355, 14464, 15243, 14069, 14822, 14443, 12140, 15050,
      12978, 14722, 14933, 13466, 13653, 11384, 14565, 15129, 14702, 12955, 14850, 15208, 14765, 13995, 14573, 14160,
      14939, 15243, 15052, 13989, 12291, 13939, 13758, 14569, 14786, 14795, 15076, 13356, 15289, 14057, 14656, 13669,
      15313, 14382, 15124, 15181, 10955, 11986, 13730, 14500, 12446, 15179, 15261, 13354, 13832, 10504, 13104, 14655,
      14848, 15343, 14766, 14871, 15055, 15010, 14424, 14574, 13750, 15057, 10670, 13682, 14804, 13954, 14137, 15008,
      15123, 14736, 11392, 14873, 8800,  14063, 13741, 14311, 13968, 14548, 13510, 14550, 14411, 13927, 14680, 15149,
      13955, 15104, 15291, 13049, 13594, 13568, 13231, 13574, 13497, 12142, 13559, 9415,  15166, 12744, 14727, 15034,
      15176, 14563, 14762, 15052, 14758, 14347, 13698, 12820, 12855, 14204, 10514, 12219, 14909, 14617, 14107, 14635,
      13256, 14717, 9723,  13506, 15046, 15136, 14174, 13975, 11803, 14339, 14573, 14245, 14581, 14437, 14586, 15302,
      13717, 10431, 14811, 13462, 14590, 14448, 11559, 14707, 14628, 14852, 8620,  14355, 14703, 15311, 14857, 13467,
      12591, 15346, 15062, 13480, 14687, 14993, 13695, 12358, 13082, 15275, 13998, 11465, 15126, 15068, 15337, 11320,
      14557, 15197, 15193, 15236, 13394, 14350, 13599, 14687, 12309, 11507, 10362, 14609, 14984, 15004, 11186, 15279,
      14749, 13735, 10349, 14844, 14336, 14946, 14945, 11370, 14529, 15108, 14784, 14134, 13050, 7602,  15053, 15145,
      8647,  14642, 13312, 13027, 13651, 14852, 10707, 10558, 14802, 13474, 10743, 14551, 15107, 15221, 14947, 13339,
      14672, 14661, 13594, 14467, 15016, 11392, 14937, 15038, 14758, 13761, 14881, 14889, 14908, 15353, 15277, 15047,
      11742, 11502, 14974, 12319, 15041, 13839, 13208, 14426, 13469, 13802, 13378, 14736, 14379, 14844, 13652, 8492,
      12934, 12512, 15282, 12825, 13545, 11065, 15255, 14599, 14231, 14374, 13660, 13331, 15357, 14487, 13831, 13484,
      12885, 15320, 13357, 10661, 15152, 14588, 14871, 14631, 14884, 14918, 15118, 12648, 14384, 15278, 12764, 14358,
      15266, 15152, 13630, 15270, 13812, 13645, 14720, 12818, 14635, 14610, 14459, 11684, 12170, 14855, 14462, 14790,
      15263, 14818, 14902, 14720, 14193, 15071, 13320, 14799, 14696, 13878, 14676, 14757, 15186, 13187, 14954, 14121,
      13137, 10891, 6814,  14841, 12873, 14341, 10999, 14054, 9437,  15338, 14387, 14961, 13363, 11666, 15013, 15256,
      14374, 14429, 14823, 14546, 12956, 14562, 13556, 11866, 12615, 11663, 14999, 14788, 14094, 12805, 14735, 14492,
      15272, 15334, 12348, 13485, 13936, 14347, 14846, 15039, 8456,  13099, 14541, 15216, 15131, 12928, 14759, 14495,
      15169, 12609, 15254, 15219, 14021, 14474, 10317, 12306, 15133, 13066, 13982, 15263, 14898, 12768, 14923, 14379,
      8550,  13661, 11617, 14136, 14434, 14600, 14804, 10307, 14727, 14252, 14125, 14019, 14342, 14658, 8551,  13603,
      14026, 14380, 14628, 13957, 14928, 14828, 14698, 14684, 14499, 15254, 13918, 11677, 15090, 7348,  14570, 15282,
      14688, 14429, 14255, 15203, 15053, 13689, 15150, 11834, 14248, 14581, 13424, 15117, 15226, 13047, 15358, 13101,
      15111, 12389, 13503, 14669, 14978, 10874, 14451, 14905, 11598, 12711, 12457, 12509, 14958, 13835, 15245, 14512,
      14892, 15235, 14730, 14742, 14763, 15183, 13892, 9403,  11706, 13388, 14744, 15101, 14340, 11651, 14816, 14199,
      12049, 12891, 14654, 14070, 14622, 11950, 12602, 11437, 15003, 13708, 10829, 13684, 14612, 12457, 14734, 15169,
      14532, 12516, 14792, 14410, 14896, 11471, 14749, 14648, 13853, 13940, 15230, 14919, 12815, 14692, 14971, 14894,
      12929, 13803, 13820, 14587, 15301, 12894, 15291, 15074, 14528, 12384, 11625, 14932, 13927, 12098, 15208, 14209,
      13731, 14302, 14659, 14104, 15050, 15168, 11603, 14955, 13639, 15192, 13706, 12004, 13914, 14889, 12952, 14965,
      15015, 13926, 13421, 13831, 13302, 15281, 14435, 15200, 15201, 14525, 14871, 13786, 14786, 13845, 12846, 15284,
      10406, 14361, 12091, 14027, 11064, 14024, 14509, 15173, 13439, 15316, 15211, 14691, 14459, 14825, 14985, 14068,
      15283, 15141, 12595, 15156, 14745, 13250, 14731, 13605, 10531, 14766, 14743, 7719,  12572, 15299, 14813, 9544,
      15056, 14492, 14134, 14349, 12485, 14944, 15268, 15262, 13316, 12479, 15015, 15137, 12060, 14087, 8634,  13332,
      14663, 14630, 14844, 14418, 13010, 14714, 13587, 14381, 14663, 15227, 13199, 12794, 10287, 12794, 15070, 15286,
      15197, 13694, 13339, 12925, 14215, 13405, 14042, 14504, 14413, 7297,  11460, 13755, 14665, 15135, 12693, 14710,
      14070, 15018, 11345, 14412, 14190, 14627, 14476, 14706, 11985, 12960, 12476, 14670, 13979, 13823, 14531, 11613,
      15157, 13256, 14182, 13995, 14598, 14129, 14818, 14354, 13562, 15157, 13403, 14884, 15070, 14469, 7817,  15354,
      15176, 14412, 13778, 15287, 11413, 13354, 14939, 14743, 12773, 12697, 15277, 6460,  9430,  14422, 14175, 13708,
      14837, 14184, 14931, 13683, 15059, 9382,  14748, 15337, 15208, 13079, 13169, 11961, 13653, 13172, 15264, 13892,
      12923, 14555, 15042, 15192, 14840, 10852, 13206, 13409, 13928, 15121, 15293, 13286, 11688, 13470, 14279, 13712,
      15067, 15251, 15119, 15000, 12990, 11939, 15068, 14642, 14549, 11424, 14080, 15241, 14436, 11489, 14764, 13608,
      13773, 15016, 13958, 14010, 15262, 10867, 14655, 14268, 14146, 12680, 13554, 14853, 13386, 14456, 6596,  13578,
      14392, 14257, 14068, 15253, 13948, 15345, 12468, 14789, 14797, 14262, 12442, 12137, 14209, 14897, 14706, 14036,
      13864, 14495, 15257, 14249, 12620, 11935, 14429, 14944, 9880,  14308, 14985, 13602, 14062, 14513, 14619, 14619,
      13479, 14148, 15093, 14898, 14478, 10035, 14518, 14568, 11792, 10448, 14066, 14396, 14526, 12444, 14492, 14832,
      15246, 14570, 14847, 13838, 13647, 13619, 13369, 12170, 14764, 14260, 15210, 15333, 13840, 15330, 10603, 12956,
      11879, 13319, 13486, 14861, 12444, 15148, 12260, 13883, 15229, 8657,  14791, 11885, 13974, 13965, 13054, 14467,
      14585, 15169, 13739, 14553, 14592, 11569, 14974, 11447, 14669, 13722, 15323, 15354, 12540, 13733, 13316, 13883,
      13794, 12711, 14538, 13990, 14162, 11920, 15254, 13104, 14314, 14656, 14265, 14870, 14728, 14928, 14601, 11593,
      14499, 14371, 15130, 14401, 14271, 15204, 13466, 10327, 14915, 14369, 14558, 15033, 15181, 13571, 14544, 13838,
      14802, 14993, 14807, 13981, 13440, 14509, 14908, 13366, 14729, 14400, 14069, 15355, 13579, 12734, 12521, 15191,
      12367, 13813, 12499, 13611, 15353, 11898, 13545, 13485, 14855, 14838, 14994, 15008, 15201, 14482, 12599, 14717,
      15129, 12538, 14800, 13579, 13803, 14700, 15052, 15263, 13969, 13624, 15037, 14664, 13679, 14512, 14599, 12866,
      15331, 14997, 10530, 15156, 8540,  11104, 12978, 14420, 11660, 13529, 15105, 15104, 14547, 15152, 14708, 13553,
      14589, 14861, 14715, 13974, 14138, 15085, 14428, 14988, 10718, 8447,  12346, 14911, 10079, 13734, 13769, 11320,
      13571, 15193, 14722, 13554, 14681, 11813, 14879, 14463, 14496, 9413,  14842, 14301, 8373,  12670, 14922, 13374,
      13482, 14736, 14937, 14389, 14784, 13805, 14056, 12294, 14388, 14858, 14255, 11977, 15089, 14767, 14526, 13573,
      14394, 4533,  13305, 15323, 14778, 14439, 15245, 14787, 15065, 14569, 14737, 14393, 13436, 11410, 13328, 14363,
      14457, 14850, 14940, 14737, 13467, 12699, 14791, 14837, 13227, 11964, 12820, 14343, 14652, 12819, 13172, 9403,
      12846, 13029, 15320, 14357, 13517, 14409, 13883, 14633, 14170, 13605, 12654, 12836, 15218, 12825, 13819, 15343,
      14652, 14177, 13399, 14935, 14795, 12223, 8417,  13237, 14723, 12390, 15235, 14766, 13794, 14931, 10934, 13873,
      13441, 15229, 15070, 14470, 14727, 12384, 14086, 14020, 13123, 14508, 14283, 12698, 14695, 13675, 14303, 14781,
      11368, 14818, 14769, 12365, 15021, 13636, 11600, 12941, 10553, 13340, 13623, 14500, 14495, 13686, 14878, 9553,
      15347, 13569, 15077, 13978, 14633, 14366, 14734, 11567, 14819, 15181, 14493, 14934, 11434, 12069, 8124,  13937,
      12779, 13882, 12688, 12607, 11536, 13286, 14842, 12943, 15063, 15235, 12396, 12836, 13729, 13955, 13354, 12206,
      14722, 14725, 14882, 14331, 13062, 12238, 15056, 12383, 14600, 13919, 13549, 14637, 14457, 15206, 14921, 15322,
      13745, 13966, 13834, 12056, 15182, 13361, 13241, 14595, 14569, 14151, 13894, 14471, 14750, 14125, 14543, 14692,
      14396, 14946, 13685, 14406, 15136, 14855, 11687, 11361, 12020, 14204, 14453, 14192, 13953, 13530, 13798, 14836,
      15323, 13791, 6802,  11662, 13729, 13820, 12201, 14394, 13081, 15075, 14552, 15029, 14900, 14171, 15034, 12846,
      14403, 15108, 13742, 12894, 12616, 12140, 15112, 13062, 15275, 15330, 15102, 14811, 14345, 11805, 14823, 14908,
      10843, 11432, 14331, 14534, 14904, 13632, 14901, 14608, 15235, 13339, 12439, 12884, 14743, 14993, 12955, 15352,
      13657, 14445, 15268, 14367, 15062, 14112, 15191, 13039, 13644, 9388,  12825, 14442, 8968,  14774, 14768, 14397,
      14527, 11595, 15315, 14090, 13574, 14398, 13983, 14528, 13407, 15057, 12694, 12308, 13501, 15087, 12814, 14597,
      14543, 13988, 15102, 14528, 15183, 15269, 14748, 14863, 12307, 14477, 13811, 14963, 14411, 14289, 15046, 14143,
      14413, 15102, 13398, 14488, 14613, 14445, 14640, 14237, 13787, 14745, 13997, 14062, 14203, 14745, 14910, 11696,
      11866, 15283, 14687, 14863, 14575, 14416, 13608, 11278, 12354, 14029, 10065, 11497, 15088, 14288, 12731, 14374,
      14673, 13409, 14984, 13964, 15177, 13336, 14968, 15322, 13318, 14185, 14513, 14934, 13765, 11321, 11502, 15102,
      14595, 12407, 13290, 15341, 14324, 14574, 11884, 13976, 14845, 15295, 14812, 12533, 14305, 14504, 15217, 13916,
      12352, 12135, 13514, 14708, 12463, 15311, 14754, 11609, 12191, 12946, 13473, 14256, 13611, 15098, 14096, 14190,
      14620, 15187, 13110, 15256, 14350, 14875, 15298, 13768, 13135, 12571, 12791, 14962, 11454, 13589, 15056, 13450,
      14372, 15339, 14748, 14637, 13541, 13113, 14973, 14365, 14179, 15094, 14328, 13988, 12431, 13506, 13701, 15129,
      13934, 14282, 12588, 12783, 14136, 15139, 13208, 11604, 13987, 14492, 15066, 15160, 13916, 15000, 14933, 14920,
      14477, 15035, 14708, 14807, 14595, 14909, 14392, 13060, 12175, 12319, 15048, 15042, 15239, 14225, 13839, 11822,
      11687, 12978, 12295, 12723, 13447, 13747, 12509, 14791, 13546, 12556, 11463, 13813, 15101, 14552, 13640, 15233,
      15349, 14306, 11914, 14545, 13667, 14339, 15216, 15337, 14929, 15210, 14968, 15163, 14880, 10865, 13472, 14898,
      13762, 14404, 14968, 14097, 12860, 14855, 14594, 12749, 12261, 13587, 11341, 14976, 15032, 13645, 13642, 14593,
      12631, 14859, 15328, 14443, 14883, 13676, 12750, 14914, 12478, 14448, 14479, 14372, 14959, 10572, 15028, 12774,
      11892, 14338, 14022, 13578, 8592,  15212, 12442, 15285, 14068, 14693, 15234, 14674, 13428, 14811, 15075, 14934,
      13379, 13535, 14892, 15235, 10946, 14025, 14654, 10445, 14602, 14812, 14784, 14915, 15156, 13932, 14471, 14347,
      13748, 10517, 14949, 14376, 14763, 14557, 8887,  14254, 11931, 14077, 14348, 15000, 12868, 14066, 14484, 14450,
      13053, 15324, 13417, 10492, 15303, 12556, 15126, 14709, 13139, 4628,  12320, 12889, 14880, 14928, 13619, 15266,
      14172, 14669, 14639, 14227, 6439,  15124, 15348, 14005, 14229, 13329, 14464, 14187, 12725, 10842, 13327, 9268,
      14416, 13911, 14573, 14257, 15345, 14405, 15275, 14777, 14906, 14049, 12833, 14454, 14404, 13085, 14529, 15297,
      9122,  14587, 14701, 14517, 13778, 13626, 15196, 13562, 14997, 15012, 15018, 14337, 10812, 13980, 14757, 9337,
      14190, 14338, 14468, 13074, 13625, 11728, 15133, 15016, 14764, 15272, 12601, 12817, 13679, 10430, 14998, 14273,
      13939, 15074, 7623,  13956, 14840, 14451, 14631, 14809, 12458, 14018, 15296, 15282, 15352, 15270, 13596, 15174,
      14895, 13076, 15360, 15012, 9883,  14560, 12924, 14390, 14770, 13920, 14451, 11888, 13889, 13851, 14590, 9064,
      14580, 14566, 14885, 12834, 14114, 14291, 15214, 14800, 12595, 15039, 14974, 14712, 14393, 14104, 13665, 14082,
      13387, 14621, 14298, 14665, 15209, 13838, 15231, 11768, 14593, 14819, 14719, 12716, 14057, 11861, 15055, 15048,
      14882, 14444, 14867, 15113, 14145, 12243, 12469, 14780, 15092, 15049, 14989, 12968, 14521, 12539, 13691, 15183,
      14436, 14832, 13729, 14075, 15115, 12702, 15163, 10247, 14573, 8219,  14241, 14828, 11904, 13621, 14136, 14733,
      13201, 13622, 12630, 14663, 12188, 14666, 14886, 14129, 12408, 15157, 12347, 11331, 13946, 15241, 14868, 12449,
      15039, 11182, 10966, 11274, 15200, 14894, 13209, 13263, 14569, 14588, 14848, 14553, 13475, 15220, 13456, 14415,
      10837, 14552, 14239, 12808, 13942, 15155, 14857, 15312, 12626, 13018, 10659, 14960, 15140, 9377,  13010, 14596,
      11398, 15308, 15346, 14613, 15005, 13828, 13094, 11437, 13585, 8563,  13709, 11745, 12184, 15000, 13747, 13726,
      15275, 14269, 11740, 13731, 12946, 13482, 14757, 14098, 9730,  14406, 12121, 14426, 12043, 14933, 13788, 15278,
      15245, 14726, 12908, 9269,  14307, 13260, 10118, 9705,  15271, 12439, 15087, 13329, 5618,  11082, 14739, 14017,
      14809, 12637, 14318, 13759, 13028, 15114, 13482, 14547, 14099, 14969, 14941, 15029, 13417, 13061, 14456, 14580,
      14355, 12679, 14383, 14773, 13982, 15356, 14015, 13770, 9833,  15118, 13012, 14753, 13754, 14661, 13381, 14793,
      15359, 13504, 14697, 13818, 14061, 12522, 12457, 14482, 15323, 15124, 15149, 12324, 15095, 14605, 14842, 15231,
      11869, 14826, 13947, 14939, 14396, 12589, 15280, 13844, 14235, 14341, 14270, 15039, 14830, 12090, 15239, 13321,
      10164, 13816, 14003, 13577, 13036, 14429, 15191, 14617, 14347, 15111, 15081, 9540,  14417, 11294, 10946, 13312,
      15226, 13790, 15118, 15076, 14599, 14321, 12659, 11447, 13916, 14712, 14664, 13603, 14349, 14511, 14313, 14997,
      14592, 13480, 15315, 14060, 15145, 11816, 14979, 14689, 13526, 14384, 11446, 13163, 14508, 13747, 15197, 13352,
      13638, 15038, 13408, 14401, 15179, 14720, 10039, 14040, 12626, 15184, 13555, 13497, 12684, 14102, 12329, 15232,
      15138, 15183, 15097, 14793, 11650, 14901, 15302, 9593,  4843,  14832, 13773, 14905, 14265, 14932, 13187, 14729,
      11821, 14511, 15062, 14857, 14069, 15091, 14040, 13796, 13964, 15038, 9525,  13530, 14503, 15220, 15178, 15251,
      14889, 14781, 13466, 14301, 5960,  14708, 15152, 14328, 14462, 15212, 12308, 14697, 14177, 12548, 14432, 13066,
      14319, 14688, 14631, 13712, 12265, 15102, 14613, 15214, 15147, 13771, 15295, 15327, 14799, 14598, 13369, 15096,
      15280, 10941, 15271, 12022, 6008,  13833, 14368, 15107, 13742, 15150, 14422, 13809, 12752, 13892, 14970, 13439,
      11849, 14618, 14338, 15027, 15274, 14262, 15052, 12752, 14465, 14718, 13152, 14862, 15176, 15322, 13321, 13541,
      14801, 13681, 14774, 14451, 14211, 14833, 7702,  13656, 15343, 15113, 12539, 15121, 11621, 15107, 12521, 15199,
      15155, 15059, 13913, 12693, 14426, 13741, 15188, 13108, 13473, 11440, 15244, 13496, 14074, 11542, 14943, 13725,
      14847, 15338, 15192, 13091, 11462, 14068, 12011, 13414, 14725, 12559, 13480, 13204, 14476, 15160, 11130, 14965,
      12413, 13181, 11323, 14102, 13629, 12569, 15295, 14289, 11858, 15188, 14684, 14345, 13631, 10795, 14829, 15104,
      12605, 13779, 13644, 14973, 12487, 15129, 15117, 12943, 13955, 11945, 13586, 15099, 14161, 14857, 11468, 12674,
      14302, 15359, 15193, 14541, 14997, 14959, 12240, 14371, 13906, 14994, 14812, 11908, 14996, 15081, 14687, 14034,
      14236, 15202, 13226, 14394, 14795, 14941, 13213, 14489, 14243, 12731, 14554, 15058, 14413, 14780, 14478, 9564,
      13488, 14087, 12455, 12319, 14372, 12067, 14653, 14929, 11319, 12423, 14730, 11816, 14905, 13869, 14998, 12353,
      15153, 11400, 14878, 14442, 14979, 13117, 14876, 15131, 12066, 14827, 13256, 14434, 15034, 14676, 14565, 14342,
      14586, 14998, 13991, 13945, 14710, 11035, 14278, 8191,  13485, 10994, 13691, 13870, 15041, 12543, 14567, 11883,
      14315, 10675, 14922, 14174, 13970, 14525, 13345, 14299, 13712, 13780, 12453, 13792, 14168, 15249, 13773, 13095,
      14635, 14181, 9979,  15296, 14242, 14822, 15178, 15217, 15240, 15293, 6673,  13100, 12507, 14737, 15297, 12822,
      13609, 13456, 14498, 13737, 14238, 15154, 15162, 12381, 12487, 15005, 14755, 14035, 15127, 13867, 13552, 14708,
      14408, 14632, 13985, 12505, 14428, 12193, 15144, 14814, 14136, 10856, 15323, 13718, 14898, 14420, 14613, 15247,
      14435, 15282, 15328, 12070, 12309, 12353, 14372, 12436, 12509, 14405, 14365, 14104, 11142, 14356, 13802, 14645,
      14729, 11306, 13769, 14032, 14079, 12062, 13511, 14757, 14870, 13899, 14017, 10908, 12775, 15208, 15311, 13155,
      13844, 12694, 13469, 15111, 14731, 15253, 11769, 15321, 13362, 14340, 14530, 11752, 14944, 15003, 10983, 14587,
      13668, 14297, 14753, 14664, 13716, 7524,  14226, 13325, 14005, 13822, 13733, 14814, 14429, 13838, 14762, 14819,
      13404, 15198, 13777, 15028, 12461, 11820, 15322, 15271, 15272, 15261, 14433, 15270, 13475, 14914, 14291, 13913,
      14993, 14827, 13021, 15205, 14103, 15269, 14012, 13998, 14363, 14717, 14073, 14354, 14620, 13949, 8820,  14674,
      14513, 15234, 14817, 15261, 14672, 14204, 14255, 14568, 13714, 14426, 14792, 15227, 15249, 15354, 15110, 13300,
      14499, 13342, 12744, 12545, 14793, 12379, 13956, 12976, 13531, 14448, 14871, 14075, 15342, 14987, 11271, 13757,
      14288, 14529, 14427, 10754, 15270, 14779, 15099, 15015, 14422, 14240, 14318, 14864, 14723, 13796, 15056, 8992,
      14638, 14798, 14805, 13607, 14582, 14729, 15249, 13968, 15141, 14409, 15157, 14910, 11573, 15292, 14468, 14810,
      13946, 12579, 15338, 12081, 11006, 14520, 14689, 15105, 14918, 13899, 14565, 14112, 15098, 13661, 8369,  14241,
      15084, 15092, 14357, 14304, 14795, 14892, 12919, 14117, 11091, 15087, 11757, 14770, 12652, 14913, 14950, 14566,
      14046, 14859, 14424, 14120, 10261, 14716, 13908, 14382, 14222, 15056, 11358, 15135, 14089, 14461, 15161, 14986,
      9750,  14419, 14575, 13452, 15022, 11635, 12514, 14825, 14809, 13786, 14243, 15286, 14582, 14694, 11667, 14578,
      14236, 13541, 13141, 14387, 13456, 12328, 13463, 14978, 13447, 14197, 13404, 14790, 14455, 12464, 14952, 15054,
      14230, 15345, 15068, 12740, 10990, 14415, 11246, 15082, 13807, 14851, 14209, 12491, 14365, 13470, 15068, 15239,
      15334, 14380, 15301, 14213, 14975, 15153, 15298, 13132, 14797, 15074, 14693, 10339, 13275, 14888, 13913, 14988,
      13333, 13978, 13172, 14911, 14182, 14087, 11002, 14496, 13252, 12401, 12953, 10311, 13550, 11147, 12478, 14905,
      13980, 14658, 13389, 10828, 11774, 14244, 14798, 14136, 14985, 15096, 14188, 12666, 15278, 15148, 13105, 14537,
      13381, 15288, 14128, 11950, 14543, 12310, 14681, 12642, 14648, 13457, 14867, 13673, 12935, 12601, 13826, 13648,
      14158, 15008, 13845, 12310, 14296, 11970, 9376,  15169, 14804, 15233, 12148, 13661, 14109, 14848, 12650, 14056,
      13403, 14733, 13948, 14133, 14577, 14913, 15205, 13368, 15332, 14686, 9734,  15197, 14868, 14154, 13910, 14587,
      14360, 15078, 13164, 10991, 12172, 14371, 15054, 13330, 14268, 14366, 10471, 15228, 15056, 10547, 13450, 14932,
      14622, 15102, 6953,  14903, 15197, 14692, 14394, 14832, 12289, 11804, 14921, 13740, 15294, 15327, 13828, 12516,
      14809, 11953, 14881, 14904, 13660, 13445, 14110, 14393, 13912, 14916, 12206, 12822, 11444, 13569, 12719, 14622,
      13975, 14346, 11920, 14665, 14874, 14503, 15204, 10579, 14421, 14755, 11834, 14016, 14365, 14875, 14882, 15270,
      14883, 15077, 14451, 14359, 13630, 12341, 14080, 13145, 14889, 14701, 14153, 14858, 14405, 14520, 14228, 15241,
      14256, 12457, 12594, 14769, 14721, 14057, 10622, 14858, 14216, 14937, 13803, 14433, 11798, 14455, 14588, 14960,
      15203, 14880, 13751, 14239, 14789, 12766, 13402, 13637, 14934, 9540,  12527, 13865, 13938, 13110, 14683, 11623,
      13637, 14468, 14620, 14176, 14983, 14089, 15006, 14285, 14891, 11746, 13412, 14953, 12237, 15166, 15289, 12726,
      13372, 12311, 13028, 14811, 14679, 13462, 15167, 15162, 14272, 14478, 14278, 14904, 15015, 13329, 14730, 14406,
      13785, 14827, 11517, 15169, 8074,  14524, 14702, 13869, 12013, 15083, 14379, 12803, 13686, 15242, 13633, 14502,
      12105, 13801, 14912, 14471, 15054, 10619, 13939, 14862, 14294, 14148, 10916, 15006, 12382, 12645, 14851, 13488,
      15241, 13408, 14340, 14257, 14748, 15141, 14354, 14771, 13501, 14683, 15165, 13634, 14689, 14231, 12290, 14570,
      13515, 13066, 14374, 14658, 14095, 12732, 13869, 14729, 14046, 11412, 14495, 13596, 13340, 9729,  14166, 10964,
      14955, 9924,  13897, 14671, 12765, 14778, 11458, 11524, 13595, 14585, 13531, 13522, 15273, 13808, 10686, 13249,
      10557, 14922, 14445, 10205, 14911, 12784, 13285, 14575, 12407, 14639, 13432, 11651, 11760, 14604, 14612, 10442,
      13692, 11749, 14212, 11993, 15031, 13937, 14710, 14182, 14815, 13010, 14209, 8251,  15173, 12699, 14558, 12306,
      14690, 15297, 14758, 13955, 12726, 10271, 13428, 14030, 13861, 15113, 15262, 14412, 11573, 14319, 14602, 13676,
      13984, 14236, 15026, 15301, 13938, 13962, 15164, 14736, 14917, 12314, 11705, 15116, 11228, 14690, 11166, 15122,
      12329, 14400, 11242, 15308, 14928, 15253, 14765, 14840, 14472, 14873, 12347, 13971, 14127, 14702, 11270, 13435,
      12525, 13642, 15082, 15003, 14699, 12820, 12100, 14302, 14542, 14338, 14005, 10563, 15262, 12763, 15023, 14169,
  };

  uint16_t data[32 * 2 * 16 * 16]{
      14402, 12289, 13852, 13540, 14810, 11027, 12387, 11288, 14169, 14805, 15159, 13632, 8595,  13403, 9001,  14821,
      15355, 14719, 14958, 10920, 9389,  13408, 12314, 11907, 14759, 14419, 15124, 14591, 13422, 14465, 13347, 14063,
      14776, 14719, 15329, 14763, 13881, 14609, 12510, 11670, 15296, 15309, 14152, 14890, 13556, 12447, 13171, 13672,
      14945, 14355, 12883, 14015, 12895, 14170, 14818, 15274, 13020, 14222, 14246, 12003, 11254, 14901, 13659, 14984,
      14532, 14496, 13605, 14803, 14524, 10024, 14259, 14899, 10336, 14532, 12566, 14884, 10386, 14316, 14828, 14809,
      14952, 14336, 12187, 13557, 15065, 15203, 15318, 12324, 13915, 8520,  13101, 14928, 15148, 12837, 14358, 14794,
      13444, 14120, 13324, 15242, 14846, 14609, 14722, 15019, 13205, 13693, 15019, 13263, 13500, 11845, 12709, 13293,
      12727, 12279, 13721, 14261, 6931,  14579, 15140, 15252, 14686, 13994, 14934, 14897, 14748, 12729, 14560, 14591,
      13643, 15213, 14626, 14254, 14498, 15019, 15162, 14222, 14382, 14808, 15215, 14977, 14343, 14110, 14728, 14801,
      14190, 15069, 14546, 14639, 13854, 13561, 12404, 14199, 12155, 13052, 14615, 8692,  14454, 13955, 14348, 14669,
      14725, 14013, 10298, 13418, 10995, 13411, 14635, 14876, 14851, 13306, 13907, 14367, 13818, 14603, 13917, 13571,
      13815, 14143, 12729, 14858, 15061, 15024, 15344, 14674, 14804, 13627, 13595, 12462, 12942, 14556, 14479, 15022,
      14651, 15247, 14809, 14470, 15184, 14802, 12978, 15009, 15202, 14529, 13084, 11909, 12829, 15295, 14812, 14878,
      13653, 14935, 13949, 15068, 15269, 13831, 13386, 14766, 14954, 14239, 14171, 11621, 15245, 13899, 15152, 15235,
      14666, 13278, 14746, 14092, 13507, 15111, 13682, 14883, 15186, 12843, 14421, 12372, 12425, 13297, 13970, 15276,
      15006, 14257, 13132, 14866, 13398, 15335, 15000, 13704, 15281, 13112, 13769, 14760, 13475, 13698, 14499, 15030,
      15154, 14952, 9074,  12046, 14860, 14751, 13603, 11537, 14708, 12316, 11701, 14308, 11710, 13911, 15075, 14568,
      12962, 12315, 12828, 14600, 15237, 12665, 12387, 15167, 14544, 15042, 12462, 15284, 14667, 14184, 14449, 14428,
      15348, 15299, 14783, 12810, 14127, 14890, 14184, 9201,  12311, 15125, 13550, 15200, 13236, 14286, 14352, 14444,
      9967,  14572, 15254, 15045, 13801, 14841, 12947, 15065, 14880, 14239, 15183, 12408, 14605, 14977, 12482, 14857,
      11292, 11937, 14706, 15322, 13482, 12420, 14965, 12978, 14939, 15313, 14848, 15123, 13955, 15176, 13256, 13717,
      14557, 14749, 8647,  14672, 11742, 12934, 12885, 15266, 15263, 13137, 14374, 15272, 15169, 8550,  14026, 14688,
      14892, 12049, 14532, 12929, 13731, 15015, 10406, 15283, 15056, 14663, 15197, 14070, 15157, 15176, 14837, 12923,
      13773, 14392, 13864, 13479, 15246, 11879, 14585, 13794, 14499, 14802, 12367, 15129, 15331, 14589, 13571, 13482,
      14457, 12846, 14652, 13441, 11368, 15347, 12779, 14722, 13745, 14396, 15323, 14403, 10843, 13657, 14527, 14543,
      11866, 14673, 14595, 12352, 14620, 14372, 13934, 14477, 11687, 15349, 13762, 12631, 11892, 13379, 13748, 13053,
      14416, 9122,  14190, 13939, 14895, 14580, 13387, 14882, 14436, 13201, 15039, 10837, 11398, 15275, 15245, 14809,
      15359, 11869, 10164, 15226, 14592, 13638, 15138, 11821, 14889, 14319, 15280, 11849, 14801, 15155, 14847, 12413,
      14302, 14236, 13488, 15153, 14586, 14315, 14635, 13609, 14408, 14435, 14729, 13844, 13668, 13404, 14993, 14513,
      14288, 14638, 13946, 15084, 14046, 9750,  14236, 14230, 15334, 13333, 13980, 13381, 14158, 13403, 14360, 14622,
      13975, 14883, 14256, 15203, 13637, 13372, 13785, 12105, 15241, 13515, 14955, 10557, 13692, 14690, 13984, 12329,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14379, 12461, 14410, 12429, 12870, 15231, 14480, 10750, 15177, 13884, 14158, 13253, 13506, 12498, 13153, 14708,
      12676, 9604,  15172, 13863, 15312, 14436, 13490, 15083, 13908, 12270, 12599, 15223, 12463, 14561, 14493, 12352,
      14537, 11611, 12643, 14922, 14035, 11970, 14910, 14490, 14761, 15061, 13666, 14130, 15296, 13639, 15249, 8727,
      10349, 14741, 15080, 14391, 14718, 14301, 14930, 14129, 14326, 13851, 15293, 14246, 12144, 14809, 12441, 13698,
      15026, 10637, 14544, 14678, 12307, 13911, 15175, 15000, 13365, 15001, 14693, 14497, 14887, 12578, 14367, 14256,
      13042, 13771, 13831, 13797, 15268, 15168, 13592, 15333, 13043, 15348, 14720, 14460, 14043, 14486, 10748, 14262,
      13232, 14590, 12376, 14716, 14813, 12310, 13740, 14201, 13379, 13640, 13591, 13190, 14543, 10802, 15231, 11537,
      10729, 13332, 11635, 13524, 11244, 12494, 13989, 15284, 13862, 14603, 13622, 12856, 15165, 13434, 14572, 14424,
      14869, 12322, 14700, 15336, 13617, 10878, 11790, 14230, 13544, 15139, 12874, 14437, 15094, 10034, 13460, 12359,
      13440, 13762, 14905, 12550, 14582, 12519, 14664, 14553, 13654, 15352, 15101, 14574, 14744, 12251, 11255, 14438,
      10879, 14460, 14690, 14639, 15314, 15079, 15153, 14607, 13996, 14167, 14901, 13147, 14311, 12178, 14388, 15357,
      15084, 14408, 14791, 14912, 14584, 13196, 14388, 14588, 14193, 14915, 15183, 14922, 15118, 14491, 13039, 14974,
      13376, 14354, 15131, 12635, 14944, 14679, 13913, 15109, 13552, 14077, 15065, 9046,  14797, 15152, 14474, 11354,
      14821, 8686,  15155, 14645, 14280, 13560, 14454, 12429, 11976, 7793,  13349, 11970, 11974, 11055, 14006, 13838,
      14891, 15255, 13720, 14502, 10339, 14936, 13696, 15107, 11576, 15216, 15054, 14624, 13412, 14306, 10568, 14587,
      15300, 14269, 12607, 14541, 14415, 12965, 15145, 10865, 12531, 14627, 14380, 14476, 13385, 12942, 10869, 11206,
      15039, 9443,  12794, 14946, 14995, 14134, 14693, 14666, 14989, 11793, 12480, 15113, 13660, 13025, 14645, 14029,
      14497, 15299, 15145, 12294, 14905, 14661, 11443, 13140, 14311, 14003, 13561, 15314, 14892, 14809, 13535, 12842,
      14534, 13468, 14777, 12869, 11685, 14879, 13661, 14314, 12576, 13520, 15102, 13322, 12698, 14268, 13505, 15323,
      13954, 14903, 13785, 14378, 13589, 15077, 14844, 14877, 14023, 14486, 14448, 14507, 13758, 15168, 12245, 14379,
      11658, 14744, 11331, 14951, 15038, 14948, 15088, 14722, 15243, 14382, 15343, 14736, 15104, 14563, 14717, 10431,
      15197, 13735, 14642, 14661, 11502, 12512, 15320, 15152, 14818, 10891, 14429, 15334, 12609, 13661, 14380, 14429,
      15235, 12891, 12516, 13803, 14302, 13926, 14361, 15141, 14492, 14630, 13694, 15018, 13256, 14412, 14184, 14555,
      15016, 14257, 14495, 14148, 14570, 13319, 15169, 12711, 14371, 14993, 13813, 12538, 14997, 14861, 15193, 14736,
      14850, 13029, 14177, 15229, 14818, 13569, 13882, 14725, 13966, 14946, 13791, 15108, 11432, 14445, 11595, 13988,
      15283, 13409, 12407, 12135, 15187, 15339, 14282, 15035, 12978, 14306, 14404, 14859, 14338, 13535, 10517, 15324,
      13911, 14587, 14338, 15074, 13076, 14566, 14621, 14444, 14832, 13622, 11182, 14552, 15308, 14269, 14726, 12637,
      13504, 14826, 13816, 13790, 13480, 15038, 15183, 14511, 14781, 14688, 10941, 14618, 13681, 15059, 15338, 13181,
      15359, 15202, 14087, 11400, 14998, 10675, 14181, 13456, 14632, 15282, 11306, 12694, 14297, 15198, 14827, 15234,
      14529, 14798, 12579, 15092, 14859, 14419, 13541, 15345, 14380, 13978, 14658, 15288, 15008, 14733, 15078, 15102,
      14346, 15077, 12457, 14880, 14468, 12311, 14827, 13801, 13408, 13066, 9924,  14922, 11749, 15297, 14236, 14400,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14633, 12096, 14115, 13971, 11429, 14833, 9432,  10395, 11319, 14428, 14843, 14232, 12548, 7919,  15088, 14305,
      14813, 14686, 14342, 13174, 14400, 14498, 12616, 9064,  12837, 13379, 15281, 14225, 13027, 11927, 14870, 14120,
      14831, 15038, 13931, 13712, 14883, 14616, 14223, 13614, 13670, 14720, 13492, 13209, 11716, 14816, 12861, 14087,
      13979, 14569, 15068, 14409, 14972, 13483, 11160, 12313, 13228, 14602, 14673, 11905, 15025, 10021, 15117, 15153,
      13462, 15339, 14120, 10311, 15128, 11470, 12619, 13442, 13318, 14462, 14980, 13346, 10339, 14495, 13259, 14816,
      12911, 13893, 14669, 15124, 14457, 14563, 13906, 15166, 11106, 14892, 14597, 15320, 12485, 14134, 13921, 12826,
      14758, 14909, 14365, 12603, 13223, 13127, 15191, 14982, 12878, 14189, 14970, 15115, 12713, 13664, 14020, 14562,
      15122, 15343, 11395, 11082, 14740, 12097, 15132, 14941, 14863, 14428, 12155, 9555,  13747, 15117, 9091,  11442,
      14680, 14657, 15288, 14832, 15327, 14530, 14762, 11369, 12046, 14239, 14444, 14427, 13630, 14440, 14185, 15358,
      14596, 13507, 12182, 15128, 15258, 13328, 13382, 14445, 15311, 14266, 14393, 14805, 15317, 15329, 14980, 13958,
      14426, 14342, 14592, 14623, 14516, 15332, 15033, 14530, 14147, 14488, 15019, 15210, 14461, 14918, 15027, 15086,
      14973, 14397, 14615, 14523, 14954, 14492, 14660, 15128, 14428, 13198, 15025, 11796, 13754, 10344, 12936, 13814,
      13428, 12732, 14407, 13246, 14762, 10268, 14164, 13468, 14436, 11718, 14979, 14861, 13928, 14641, 14702, 12458,
      14620, 14603, 10232, 14343, 15360, 12329, 14416, 8166,  11019, 15011, 14075, 14114, 14414, 14717, 14340, 15139,
      15240, 14814, 14748, 14476, 14488, 11043, 14510, 12898, 13609, 12786, 13935, 14567, 12992, 11696, 15028, 14636,
      14751, 15252, 14273, 13307, 13844, 11071, 14553, 12454, 11382, 13440, 14607, 15161, 13849, 14990, 13858, 14570,
      14131, 15356, 12642, 13217, 11878, 14944, 10750, 14647, 13837, 14591, 14343, 12894, 14669, 14586, 14970, 12816,
      11248, 15353, 14566, 14838, 12767, 14363, 12677, 15029, 14034, 10915, 14965, 14393, 13749, 14888, 12912, 14157,
      13838, 12290, 12232, 13419, 14643, 14877, 14879, 13383, 14649, 13527, 14441, 15160, 13934, 13328, 14586, 13407,
      13475, 11751, 15161, 14851, 14579, 12361, 15143, 11779, 14347, 15340, 15178, 14467, 13465, 8601,  14510, 11762,
      13531, 13428, 13998, 11058, 13442, 15027, 11686, 14933, 15052, 15124, 14766, 11392, 15291, 14762, 9723,  14811,
      15193, 10349, 13312, 13594, 14974, 15282, 13357, 13630, 14902, 6814,  14823, 12348, 15254, 11617, 14628, 14255,
      14730, 14654, 14792, 13820, 14659, 13421, 12091, 12595, 14134, 14844, 13339, 11345, 14182, 13778, 14931, 15042,
      13958, 14068, 15257, 15093, 14847, 13486, 13739, 14538, 15130, 14807, 12499, 14800, 10530, 14715, 14722, 14937,
      14940, 15320, 13399, 15070, 14769, 15077, 12688, 14882, 13834, 13685, 6802,  13742, 14331, 15268, 15315, 15102,
      14687, 14984, 13290, 13514, 13110, 14748, 12588, 14708, 12295, 11914, 14968, 15328, 14022, 14892, 14949, 13417,
      14573, 14701, 14468, 7623,  15360, 14885, 14298, 14867, 13729, 12630, 10966, 14239, 15346, 11740, 12908, 14318,
      14697, 13947, 14003, 15118, 15315, 13408, 15097, 15062, 13466, 14631, 15271, 14338, 14774, 13913, 15192, 11323,
      15193, 13226, 12455, 14878, 13991, 14922, 9979,  14498, 13985, 15328, 13769, 13469, 14753, 13777, 13021, 14817,
      14427, 14805, 15338, 14357, 14424, 14575, 13141, 15068, 15301, 13172, 13389, 14128, 13845, 13948, 13164, 6953,
      11920, 14451, 12594, 13751, 14620, 13028, 11517, 14912, 14340, 14374, 13897, 14445, 14212, 14758, 15026, 11242,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14774, 10696, 11497, 14888, 14307, 10577, 14724, 14163, 14063, 14944, 14811, 11742, 13451, 15307, 14621, 14696,
      11644, 13284, 14089, 12359, 14499, 13580, 12784, 15080, 12547, 15254, 13663, 13678, 13441, 14072, 13255, 15329,
      14373, 14519, 15027, 13187, 14387, 11512, 14306, 13573, 14483, 14117, 12714, 14456, 15265, 14567, 14418, 10373,
      15274, 14673, 14411, 12520, 14462, 12592, 14885, 13104, 11688, 14095, 8674,  14517, 12548, 14583, 15064, 13103,
      13427, 13123, 10293, 13432, 14966, 12557, 14568, 12131, 13572, 12341, 13878, 13997, 14997, 13166, 14640, 14879,
      12277, 10751, 14927, 13409, 14429, 14919, 15264, 14671, 14425, 14677, 13876, 13591, 13313, 11291, 14313, 13999,
      14107, 13828, 15272, 14484, 14103, 13717, 14366, 13275, 14444, 14996, 12885, 13558, 14974, 14438, 12789, 14954,
      13587, 11654, 13805, 15267, 12551, 15187, 12325, 12852, 14838, 14401, 15213, 15251, 14928, 14501, 15157, 10234,
      14818, 12091, 14329, 10152, 14803, 15328, 15142, 10653, 12672, 12911, 14842, 14444, 8237,  14515, 12908, 14428,
      14433, 15114, 14777, 14666, 13371, 14986, 11184, 14102, 13957, 15101, 11326, 13973, 14778, 14560, 14038, 14695,
      14095, 15143, 13917, 14689, 15251, 14533, 14213, 14681, 13595, 14483, 14818, 14307, 14756, 14408, 14375, 14778,
      15238, 15167, 12734, 11814, 12876, 13531, 14709, 11673, 13621, 15265, 14342, 13895, 14828, 15154, 15105, 14826,
      14856, 12222, 14474, 14890, 13533, 14549, 15012, 15206, 6181,  15025, 13511, 14215, 14497, 14852, 14441, 15245,
      12786, 14495, 13347, 12467, 12646, 15206, 14573, 13597, 12273, 13346, 11284, 14446, 12268, 13935, 15023, 12175,
      14322, 12500, 14429, 14175, 14375, 14225, 14464, 10470, 13521, 14523, 13100, 13429, 14254, 12366, 15032, 12712,
      14670, 15326, 8883,  13942, 14724, 14433, 12196, 14995, 14129, 15098, 14725, 14151, 14591, 15325, 10495, 12650,
      15086, 13387, 13799, 14354, 15048, 14023, 15143, 14346, 14518, 14516, 14706, 14400, 13687, 13830, 14554, 14444,
      15151, 14423, 8487,  14858, 13453, 14810, 14534, 13427, 13325, 14918, 15296, 15080, 14842, 14704, 14026, 13991,
      14816, 15045, 14649, 14902, 14866, 14437, 14425, 14023, 15283, 14573, 14357, 11566, 15055, 14607, 12724, 12836,
      13747, 14717, 14463, 15013, 14579, 12294, 14879, 15221, 14948, 15299, 13466, 10446, 15047, 13726, 13818, 15354,
      12399, 15264, 15216, 15054, 14677, 14375, 14636, 13466, 13989, 15181, 14871, 14873, 13049, 15052, 13506, 13462,
      15236, 14844, 13027, 14467, 12319, 12825, 10661, 15270, 14720, 14841, 14546, 13485, 15219, 14136, 13957, 15203,
      14742, 14070, 14410, 14587, 14104, 13831, 14027, 15156, 14349, 14418, 12925, 14412, 13995, 15287, 13683, 15192,
      14010, 15253, 14249, 14898, 13838, 14861, 14553, 13990, 14401, 13981, 13611, 13579, 15156, 13974, 13554, 14389,
      14737, 14357, 14935, 14470, 12365, 13978, 12607, 14331, 12056, 14406, 11662, 12894, 14534, 14367, 14090, 14528,
      14863, 13964, 15341, 14708, 15256, 14637, 12783, 14807, 12723, 14545, 14097, 14443, 13578, 15235, 14376, 10492,
      14257, 14517, 13074, 13956, 15012, 12834, 14665, 15113, 14075, 14663, 11274, 12808, 14613, 13731, 9269,  13759,
      13818, 14939, 13577, 15076, 14060, 14401, 14793, 14857, 14301, 13712, 12022, 15027, 14451, 12693, 13091, 14102,
      14541, 14394, 12319, 14442, 13945, 14174, 15296, 13737, 12505, 12070, 14032, 15111, 14664, 15028, 15205, 15261,
      10754, 13607, 12081, 14304, 14120, 13452, 14387, 12740, 14213, 14911, 10828, 11950, 12310, 14133, 10991, 14903,
      14665, 14359, 14769, 14239, 14176, 14811, 15169, 14471, 14257, 14658, 14671, 10205, 11993, 13955, 15301, 15308,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13826, 14558, 14819, 12988, 14957, 14400, 13686, 14618, 13328, 14431, 13607, 14661, 14936, 14149, 12825, 13775,
      15105, 14833, 14258, 13959, 14433, 14944, 14866, 13562, 12613, 14804, 13340, 12922, 13934, 15105, 14797, 15133,
      15333, 12573, 12382, 15063, 14096, 14016, 13757, 14299, 12979, 14479, 14525, 14495, 14297, 15025, 14322, 13354,
      15353, 15071, 14392, 13362, 15358, 14349, 15337, 15269, 14668, 15329, 12969, 12346, 14223, 13431, 14558, 13445,
      15137, 10664, 12592, 14963, 14986, 12031, 14361, 15034, 15017, 14861, 14407, 15341, 13711, 14400, 15313, 14709,
      8940,  15160, 10954, 13437, 13989, 14458, 14346, 14929, 10392, 14398, 15291, 12367, 14418, 12551, 14450, 14488,
      14916, 14416, 12284, 13667, 13726, 14808, 14886, 12011, 14410, 14884, 14709, 15335, 12206, 14610, 14247, 14466,
      15182, 14639, 13531, 14698, 13701, 13845, 11439, 14604, 15190, 9397,  13368, 14343, 14534, 14046, 15277, 14000,
      14977, 10428, 14737, 14760, 14594, 13517, 14970, 15112, 13351, 11824, 15183, 13677, 15109, 12718, 9219,  15136,
      11343, 9799,  12893, 14379, 15119, 14656, 14713, 14743, 13445, 15284, 14671, 10449, 14915, 11675, 13835, 15073,
      13314, 14800, 15093, 13773, 14372, 14477, 14362, 15249, 13173, 14433, 15222, 14375, 12764, 10838, 13635, 14860,
      14606, 11191, 13448, 12967, 15126, 14826, 14981, 14958, 15111, 11041, 13992, 11696, 14786, 13659, 15300, 12961,
      13953, 14445, 12547, 14996, 13247, 14583, 15063, 15331, 15349, 15059, 10628, 14658, 14949, 13345, 13287, 6968,
      13684, 15181, 12586, 14444, 15134, 12724, 13452, 14371, 15022, 12886, 14345, 9328,  14528, 13394, 14302, 15294,
      15292, 12829, 15314, 14816, 14267, 12337, 13664, 13757, 15094, 14433, 14262, 15222, 15285, 15004, 14089, 11922,
      14744, 14025, 15242, 14988, 13491, 15083, 13125, 11604, 13901, 10310, 10571, 12334, 13386, 10272, 13301, 15064,
      14576, 10482, 14958, 11616, 14554, 14794, 13244, 14444, 14136, 14047, 14854, 14623, 15082, 13854, 13204, 14768,
      13330, 12857, 14109, 10505, 15276, 13856, 13383, 15325, 14371, 13744, 14888, 13009, 13484, 13659, 14535, 13033,
      13440, 13479, 13830, 14471, 15209, 12633, 13784, 12862, 13615, 13439, 11794, 13377, 13452, 15347, 14386, 14979,
      12796, 12501, 13963, 13629, 15051, 14902, 15350, 12400, 14509, 13238, 14809, 15239, 15104, 13226, 15030, 12257,
      13550, 10249, 14165, 15016, 15044, 15211, 15160, 13653, 12291, 10955, 15055, 8800,  13594, 14758, 15046, 14590,
      13394, 14336, 13651, 15016, 15041, 13545, 15152, 13812, 14193, 12873, 12956, 13936, 14021, 14434, 14928, 15053,
      14763, 14622, 14896, 15301, 15050, 13302, 11064, 14745, 12485, 13010, 14215, 14190, 14598, 11413, 15059, 14840,
      15262, 13948, 12620, 14478, 13647, 12444, 14592, 14162, 14271, 13440, 15353, 13803, 8540,  14138, 14681, 14784,
      13467, 13517, 14795, 14727, 15021, 14633, 11536, 13062, 15182, 15136, 13729, 12616, 14904, 15062, 13574, 15183,
      14575, 15177, 14324, 12463, 14350, 13541, 14136, 14595, 13447, 13667, 12860, 14883, 8592,  10946, 14763, 15303,
      15345, 13778, 13625, 14840, 9883,  14114, 15209, 14145, 15115, 12188, 15200, 13942, 15005, 12946, 14307, 13028,
      14061, 14396, 13036, 14599, 15145, 15179, 11650, 14069, 5960,  12265, 6008,  15274, 14211, 14426, 11462, 13629,
      14997, 14795, 14372, 14979, 14710, 13970, 14242, 14238, 14428, 12309, 14079, 14731, 13716, 12461, 14103, 14672,
      15270, 14582, 11006, 14795, 10261, 15022, 13456, 10990, 14975, 14182, 11774, 14543, 14296, 14577, 12172, 15197,
      14874, 13630, 14721, 14789, 14983, 14679, 8074,  15054, 14748, 14095, 12765, 14911, 15031, 12726, 13938, 14928,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14869, 14464, 13048, 15321, 14165, 14471, 15178, 14458, 14065, 13189, 15082, 11794, 13946, 15057, 14106, 12556,
      12800, 14904, 13405, 15042, 14837, 14805, 15288, 14200, 14559, 14015, 14732, 13770, 13548, 9506,  13812, 14514,
      14774, 13596, 15240, 14154, 14763, 14776, 11326, 15072, 12365, 14618, 12291, 13529, 14645, 14881, 14303, 12724,
      14721, 15261, 13305, 9595,  13349, 15174, 14179, 15000, 14621, 13367, 14865, 14095, 12241, 14631, 13502, 14964,
      14076, 12994, 12775, 14553, 14754, 12972, 14178, 14496, 15292, 13452, 15024, 14505, 14176, 14770, 12292, 14789,
      14490, 14727, 14735, 14427, 15345, 14713, 14383, 13684, 14334, 12964, 12167, 12637, 15330, 14924, 15036, 14548,
      14465, 11265, 9254,  14577, 14937, 15135, 14963, 10080, 9216,  12579, 12549, 14339, 14805, 13225, 15103, 15231,
      12600, 14236, 12448, 14538, 14025, 14585, 14354, 14734, 12476, 15154, 14948, 14140, 15315, 9922,  15067, 14953,
      13522, 15237, 14907, 13558, 11370, 14899, 14456, 13350, 11326, 14597, 14251, 13013, 14686, 14832, 14892, 14019,
      11530, 14562, 12312, 14723, 13665, 14479, 14093, 14238, 14112, 9318,  15011, 13937, 14942, 15305, 14487, 13004,
      13728, 14388, 13848, 14156, 14883, 12481, 14363, 14664, 13576, 14540, 14525, 12863, 13897, 14980, 12519, 10133,
      14491, 14279, 15303, 14788, 13938, 14218, 13993, 12747, 14766, 14758, 15134, 11550, 13267, 15176, 14390, 13032,
      7701,  12821, 14612, 14240, 13172, 14106, 14524, 11571, 14674, 14203, 14836, 14982, 15277, 14391, 14544, 15275,
      13172, 8810,  13700, 13263, 10573, 13434, 14279, 12949, 15185, 14524, 12222, 9011,  14474, 14669, 13270, 12139,
      14873, 15351, 10338, 11195, 10468, 12922, 12852, 15329, 13019, 15216, 14025, 12075, 14664, 10601, 14565, 14456,
      13911, 15072, 14912, 14966, 12650, 14579, 14821, 14229, 14870, 13792, 15332, 14001, 14882, 14616, 15217, 15235,
      13333, 13661, 14636, 9404,  15318, 15017, 15227, 14077, 11937, 15058, 8858,  14339, 11505, 14795, 15063, 14568,
      14573, 13913, 13482, 8317,  14624, 15180, 15047, 13804, 15037, 14885, 13830, 13054, 15277, 8931,  14791, 14553,
      14583, 13324, 13881, 12596, 14734, 14694, 13115, 14394, 9514,  13875, 15347, 12064, 14342, 11916, 13715, 13902,
      13801, 14774, 12740, 14707, 13822, 15298, 14155, 14852, 14043, 15250, 13527, 14589, 12808, 14671, 12719, 13862,
      14930, 14423, 13174, 14918, 14393, 14673, 13466, 11384, 13939, 11986, 15010, 14063, 13568, 14347, 15136, 14448,
      14350, 14946, 14852, 11392, 13839, 11065, 14588, 13645, 15071, 14341, 14562, 14347, 14474, 14600, 14828, 13689,
      15183, 11950, 11471, 12894, 15168, 15281, 14024, 13250, 14944, 14714, 13405, 14627, 14129, 13354, 9382,  10852,
      10867, 15345, 11935, 10035, 13619, 15148, 11569, 11920, 15204, 14509, 11898, 14700, 11104, 15085, 11813, 13805,
      12699, 14409, 12223, 12384, 13636, 14366, 13286, 12238, 13361, 14855, 13820, 12140, 13632, 14112, 14398, 15269,
      14416, 13336, 14574, 15311, 14875, 13113, 15139, 14909, 13747, 14339, 14855, 13676, 15212, 14025, 14557, 12556,
      14405, 13626, 11728, 14451, 14560, 14291, 13838, 12243, 12702, 14666, 14894, 15155, 13828, 13482, 13260, 15114,
      12522, 12589, 14429, 14321, 11816, 14720, 14901, 15091, 14708, 15102, 13833, 14262, 14833, 13741, 14068, 12569,
      14959, 14941, 12067, 13117, 11035, 14525, 14822, 15154, 12193, 12353, 12062, 15253, 7524,  11820, 15269, 14204,
      14779, 14729, 14520, 14892, 14716, 11635, 12328, 14415, 15153, 14087, 14244, 12310, 11970, 14913, 14371, 14692,
      14503, 12341, 14057, 12766, 14089, 13462, 14524, 10619, 15141, 12732, 14778, 12784, 13937, 10271, 13962, 15253,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13703, 14389, 15357, 14609, 13507, 12798, 7127,  11317, 12473, 10181, 13934, 14768, 9166,  14686, 14612, 14553,
      14405, 15088, 14868, 14786, 13899, 14672, 13778, 14107, 14426, 11742, 15192, 13051, 15294, 15066, 12955, 14408,
      12938, 13429, 14812, 13719, 14230, 13673, 15331, 15249, 12350, 12389, 13700, 15096, 11518, 12991, 13629, 13212,
      13903, 13330, 14984, 13370, 14435, 14623, 14816, 13469, 14498, 14406, 8639,  12049, 14655, 11804, 15118, 14408,
      14352, 12358, 15207, 9121,  11664, 9385,  13938, 14900, 13038, 11285, 11678, 13124, 14488, 14949, 14965, 15304,
      13899, 12719, 15145, 15224, 12562, 15133, 13413, 15089, 14601, 14925, 14932, 13333, 12989, 14941, 11314, 14486,
      13839, 12928, 14683, 13609, 13802, 13812, 9200,  14860, 15096, 12380, 14703, 14657, 12983, 14806, 12310, 14985,
      12573, 14206, 13846, 15246, 14010, 14226, 13598, 13837, 14150, 15255, 15338, 14668, 13606, 13786, 14572, 11525,
      14858, 15018, 14807, 15229, 15316, 13853, 15130, 14369, 13994, 15006, 14406, 15073, 14783, 14370, 15032, 14014,
      4925,  14985, 14101, 13339, 15075, 13804, 14547, 14464, 14298, 13471, 11166, 14144, 13830, 14830, 14440, 13469,
      14996, 10658, 15141, 15300, 12828, 11910, 15233, 14074, 14825, 10296, 6472,  13301, 14411, 13930, 13658, 13691,
      14457, 13385, 12395, 13480, 15240, 14183, 14804, 13111, 12161, 13433, 13397, 11153, 13484, 13262, 14721, 15101,
      12967, 13096, 15050, 14258, 13298, 13745, 13910, 13163, 13595, 15296, 14917, 15165, 13876, 14900, 13580, 13654,
      15324, 14953, 15001, 13261, 14929, 10310, 15148, 14903, 14690, 13676, 14771, 12151, 12725, 13397, 9584,  13397,
      14885, 14535, 13590, 14330, 13784, 11299, 13677, 13268, 15053, 15359, 13674, 14833, 12385, 15337, 14208, 15157,
      14540, 14971, 14463, 14055, 12399, 14457, 14490, 14423, 6364,  13283, 13591, 13908, 13477, 13892, 10530, 14196,
      14071, 14759, 15199, 15175, 12198, 12898, 15348, 11350, 10184, 15169, 14391, 14836, 15240, 14585, 11593, 13293,
      14629, 11089, 14511, 12171, 14492, 14863, 14229, 14680, 15245, 13817, 14374, 14489, 14934, 12806, 14636, 14843,
      13870, 12783, 14715, 13506, 15174, 14376, 13423, 13738, 12444, 14558, 12865, 14347, 15335, 13690, 13522, 14786,
      14716, 12440, 14663, 14844, 14773, 9077,  12870, 11268, 13449, 14771, 14844, 13446, 13647, 14683, 13705, 13822,
      15140, 12749, 15121, 15204, 13774, 14670, 14668, 14565, 13758, 13730, 14424, 13741, 13231, 13698, 14174, 11559,
      13599, 14945, 10707, 14937, 13208, 15255, 14871, 14720, 13320, 10999, 13556, 14846, 10317, 14804, 14698, 15150,
      13892, 12602, 14749, 15291, 11603, 14435, 14509, 14731, 15268, 13587, 14042, 14476, 14818, 14939, 14748, 13206,
      14655, 12468, 14429, 14518, 13369, 12260, 14974, 15254, 13466, 14908, 13545, 15052, 12978, 14428, 14879, 14056,
      14791, 13883, 8417,  14086, 11600, 14734, 14842, 15056, 13241, 11687, 12201, 15112, 14901, 15191, 13983, 14748,
      13608, 14968, 11884, 14754, 15298, 14973, 13208, 14392, 12509, 15216, 14594, 12750, 12442, 14654, 8887,  15126,
      15275, 15196, 15133, 14631, 12924, 15214, 15231, 12469, 15163, 14886, 13209, 14857, 13094, 14757, 10118, 13482,
      12457, 15280, 15191, 12659, 14979, 10039, 15302, 14040, 15152, 14613, 14368, 15052, 7702,  15188, 12011, 15295,
      12240, 13213, 14653, 14876, 14278, 13345, 15178, 15162, 15144, 14372, 13511, 11769, 14226, 15322, 14012, 14255,
      15099, 15249, 14689, 12919, 13908, 12514, 13463, 11246, 15298, 11002, 14798, 14681, 9376,  15205, 15054, 14394,
      15204, 14080, 10622, 13402, 15006, 15167, 14702, 13939, 14354, 13869, 11458, 13285, 14710, 13428, 15164, 14765,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15068, 15323, 14681, 13548, 15335, 15263, 14452, 14390, 15172, 15146, 13595, 10439, 11487, 13501, 15019, 14596,
      10732, 14678, 14431, 13899, 14042, 12729, 15029, 14491, 14984, 14901, 13832, 9329,  15243, 12170, 14658, 15279,
      15357, 14441, 13583, 15282, 12512, 14152, 14992, 12920, 11432, 13789, 12309, 11671, 15117, 14427, 13272, 11178,
      14031, 14744, 13344, 14939, 15199, 15014, 11916, 14529, 15187, 14468, 14870, 13745, 15204, 14232, 12588, 12287,
      14771, 14219, 14873, 14458, 13407, 8415,  14643, 13754, 15236, 15176, 13704, 12266, 14775, 14766, 13609, 10730,
      13740, 14521, 15354, 14516, 14693, 10414, 11588, 15132, 13881, 15354, 14552, 14913, 14830, 14537, 15189, 14608,
      14705, 14484, 13013, 13738, 13998, 14919, 14976, 15234, 11936, 13260, 13261, 14576, 14725, 15335, 15300, 14602,
      13185, 13996, 14956, 13931, 14093, 14721, 14044, 14652, 15086, 12694, 12117, 14414, 14041, 15146, 14382, 11595,
      13828, 12504, 15160, 15236, 15088, 13170, 15254, 15086, 11902, 14226, 14445, 15348, 15218, 14063, 13379, 13255,
      14968, 14077, 9896,  15116, 14713, 10682, 13433, 14524, 15358, 15069, 15167, 14675, 15288, 14473, 9509,  10959,
      11853, 13742, 12502, 14591, 14779, 14374, 12589, 14430, 15011, 13495, 13969, 14488, 14595, 13054, 12588, 14097,
      12787, 11270, 11327, 14427, 9596,  13774, 14524, 14488, 15046, 14663, 13406, 14576, 14093, 14342, 15286, 14650,
      13071, 14216, 14870, 11484, 11818, 13472, 14018, 14590, 14654, 14050, 13534, 14611, 13152, 14750, 13006, 15123,
      13266, 14259, 15053, 13472, 13898, 15258, 15107, 12516, 14833, 14551, 11824, 11692, 10772, 12845, 14651, 13355,
      14194, 7340,  12994, 7635,  14559, 14401, 15101, 11995, 14869, 11906, 14659, 14950, 10345, 11481, 11671, 15243,
      13809, 15075, 15209, 11382, 15318, 14605, 10374, 14656, 14784, 14056, 13719, 14638, 14897, 13245, 15223, 14995,
      14903, 14925, 13112, 15073, 15073, 15021, 12597, 15350, 9024,  14790, 12986, 14229, 14765, 13021, 15236, 12624,
      11126, 14815, 13124, 11843, 14994, 10524, 14922, 13076, 13394, 14614, 15058, 14349, 11430, 14686, 14919, 13633,
      14790, 13423, 14598, 14675, 13694, 13787, 12274, 14260, 11583, 13380, 13426, 12401, 11267, 15311, 14742, 15215,
      9595,  13786, 12602, 14856, 14785, 14738, 14311, 12707, 14501, 14582, 13607, 13323, 14156, 15115, 11588, 15100,
      14805, 15186, 14098, 10408, 13179, 13165, 12900, 15129, 14569, 14500, 14574, 14311, 13574, 12820, 13975, 14707,
      14687, 11370, 10558, 15038, 14426, 14599, 14631, 12818, 14799, 14054, 11866, 15039, 12306, 10307, 14684, 11834,
      9403,  11437, 14648, 15074, 14955, 15200, 15173, 13605, 15262, 14381, 14504, 14706, 14354, 14743, 15337, 13409,
      14268, 14789, 14944, 14568, 12170, 13883, 11447, 13104, 10327, 13366, 13485, 15263, 14420, 14988, 14463, 12294,
      14837, 14633, 13237, 14020, 12941, 11567, 12943, 12383, 14595, 11361, 14394, 13062, 14608, 13039, 14528, 14863,
      11278, 15322, 13976, 11609, 13768, 14365, 11604, 13060, 14791, 15337, 12749, 14914, 15285, 10445, 14254, 14709,
      14777, 13562, 15016, 14809, 14390, 14800, 11768, 14780, 10247, 14129, 13263, 15312, 11437, 14098, 9705,  14547,
      14482, 13844, 14617, 11447, 14689, 14040, 9593,  13796, 14328, 15214, 15107, 12752, 13656, 13108, 13414, 14289,
      14371, 14489, 14929, 15131, 8191,  14299, 15217, 12381, 14814, 12436, 14757, 15321, 13325, 15271, 13998, 14568,
      15015, 13968, 15105, 14117, 14382, 14825, 14978, 15082, 13132, 14496, 14136, 12642, 15169, 13368, 13330, 14832,
      10579, 13145, 14858, 13637, 14285, 15162, 13869, 14862, 14771, 14729, 11524, 14575, 14182, 14030, 14736, 14840,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13624, 15219, 15248, 14965, 15060, 15010, 13930, 15194, 12830, 14422, 14830, 15237, 13611, 13620, 13647, 11235,
      12721, 15095, 12673, 14605, 14997, 13016, 12163, 13819, 14467, 11807, 14527, 12291, 15243, 12578, 15091, 13726,
      13384, 14376, 14716, 12223, 12235, 14832, 12210, 11776, 13509, 15227, 14195, 13535, 15291, 15102, 11448, 15113,
      10803, 10303, 13793, 14396, 15062, 15117, 13010, 15326, 9846,  14612, 14329, 12856, 15285, 15201, 15345, 15235,
      15001, 14679, 13208, 13688, 14351, 12979, 11526, 14595, 13734, 14689, 12398, 14682, 14396, 15209, 11280, 14779,
      13183, 13543, 13745, 15148, 12792, 14050, 15274, 14644, 14806, 15147, 14733, 14071, 11773, 11541, 15325, 14506,
      12335, 14768, 14983, 13610, 15357, 13141, 14685, 14214, 12797, 12342, 13386, 14986, 15346, 15291, 14202, 15196,
      15246, 14830, 14969, 13360, 14089, 12898, 13476, 11673, 10807, 14838, 13520, 13466, 14693, 14963, 11844, 12761,
      14504, 14724, 14947, 13740, 15234, 15091, 7194,  15065, 14590, 14707, 14749, 5073,  14503, 14474, 11454, 14668,
      10188, 13698, 14965, 14368, 13231, 12985, 14669, 14546, 13316, 11003, 15198, 10440, 14237, 12337, 15004, 13679,
      14118, 15351, 15011, 13250, 14965, 13119, 14695, 14839, 13717, 15021, 14399, 14391, 14016, 15065, 11456, 15173,
      15143, 12510, 13078, 13523, 12696, 12475, 15034, 12986, 13311, 15095, 14617, 14512, 14924, 15033, 14022, 14667,
      13972, 10315, 13957, 11762, 14519, 12738, 11713, 13805, 13776, 14075, 13017, 13693, 15306, 13871, 13714, 14722,
      14558, 10367, 14406, 14442, 14412, 13807, 14965, 12142, 12032, 14047, 14323, 14402, 14032, 15062, 9547,  15305,
      14386, 14355, 12711, 12453, 15122, 11737, 9868,  9404,  8654,  13810, 11623, 14351, 14662, 9434,  14048, 14453,
      14929, 13726, 13899, 12514, 13635, 13615, 15060, 12259, 14948, 13234, 13610, 14556, 15331, 12515, 15072, 11554,
      14101, 13582, 14605, 12639, 14676, 14604, 15274, 14545, 14700, 14400, 9641,  14999, 11556, 15043, 11083, 14104,
      14386, 12375, 12995, 13416, 14927, 15252, 15085, 14730, 15341, 14234, 9350,  15060, 14652, 13643, 13215, 10203,
      12977, 15356, 14228, 14781, 14448, 13935, 14956, 14205, 14368, 15054, 15304, 14781, 10657, 14554, 13900, 14452,
      13863, 13466, 12423, 13130, 15163, 11731, 14533, 14894, 12325, 14494, 14009, 15151, 12925, 14477, 14835, 15222,
      14077, 14784, 13705, 13474, 14610, 14458, 14355, 14702, 14786, 12446, 13750, 13968, 13497, 12855, 11803, 14628,
      12309, 14529, 14802, 14758, 13469, 14231, 14884, 14635, 14696, 9437,  12615, 8456,  15133, 14727, 14499, 14248,
      11706, 15003, 13853, 14528, 13639, 15201, 13439, 10531, 13316, 14663, 14413, 11985, 13562, 12773, 15208, 13928,
      14146, 14797, 9880,  11792, 14764, 15229, 14669, 14314, 14915, 14729, 14855, 13969, 11660, 10718, 14496, 14388,
      13227, 14170, 14723, 13123, 10553, 14819, 15063, 14600, 14569, 12020, 13081, 15275, 15235, 13644, 13407, 12307,
      12354, 13318, 14845, 12191, 13135, 14179, 13987, 12175, 13546, 14929, 12261, 12478, 14068, 14602, 11931, 13139,
      14906, 14997, 14764, 12458, 14770, 12595, 14593, 15092, 14573, 12408, 14569, 12626, 13585, 9730,  15271, 14099,
      15323, 14235, 14347, 13916, 13526, 12626, 4843,  13964, 14462, 15147, 13742, 14465, 15343, 13473, 14725, 11858,
      13906, 14243, 11319, 12066, 13485, 13712, 15240, 12487, 14136, 12509, 14870, 13362, 14005, 15272, 14363, 13714,
      14422, 15141, 14918, 11091, 14222, 14809, 13447, 13807, 14797, 13252, 14985, 14648, 14804, 15332, 14268, 12289,
      14421, 14889, 14216, 14934, 14891, 14272, 12013, 14294, 13501, 14046, 13595, 12407, 14815, 13861, 14917, 14472,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13537, 13308, 14640, 15148, 15034, 14107, 14178, 14971, 15062, 14689, 14087, 14869, 15071, 13326, 11647, 14732,
      15127, 14809, 14949, 14381, 12504, 13193, 13002, 12799, 14962, 13657, 10077, 14056, 12819, 14915, 14919, 12232,
      13289, 14861, 13851, 14027, 14182, 13716, 14593, 10044, 14614, 15248, 14593, 4916,  13955, 12799, 11306, 14428,
      14533, 12502, 15031, 15285, 13372, 14418, 14099, 14479, 12259, 11665, 13712, 13578, 14538, 13574, 14768, 13431,
      13558, 13296, 14159, 7712,  13937, 15047, 6603,  14638, 14787, 15185, 14661, 13100, 14816, 13365, 14185, 14742,
      15285, 13335, 14104, 12570, 14599, 14137, 14520, 11365, 12667, 7907,  14115, 14945, 14848, 11701, 15022, 14795,
      14895, 14556, 14678, 14957, 11519, 14121, 14699, 10350, 14075, 13752, 12190, 12278, 15205, 13492, 14934, 11438,
      13212, 13807, 14459, 12631, 10293, 13520, 15127, 11775, 15109, 14629, 14842, 14325, 14465, 13558, 15010, 13803,
      14643, 10557, 14667, 13447, 14523, 13688, 12757, 14386, 11731, 14541, 14117, 15042, 10056, 12985, 13616, 8243,
      15344, 13323, 14677, 14744, 15155, 14935, 14482, 14885, 13472, 9380,  14745, 15337, 13099, 14818, 11235, 13413,
      15112, 14182, 12813, 14332, 14958, 13170, 15205, 15230, 11066, 15054, 14337, 12766, 15359, 12418, 14437, 14934,
      13930, 14445, 11758, 11151, 15070, 13375, 14845, 14640, 14535, 14063, 12811, 13791, 14659, 14855, 15269, 14759,
      14590, 14903, 15111, 12959, 14816, 15175, 13615, 14527, 14460, 13587, 13475, 13530, 13578, 14655, 14145, 13091,
      15283, 13338, 15283, 14787, 13546, 13712, 11718, 14405, 13876, 15129, 13036, 11492, 14549, 13751, 11630, 15334,
      14550, 13560, 10352, 14345, 12805, 15336, 13736, 14685, 14177, 13255, 14140, 15219, 12555, 14448, 13703, 13956,
      15171, 14701, 14203, 13357, 14545, 14625, 13522, 14356, 13543, 11589, 13550, 14519, 14323, 15206, 15103, 14940,
      14358, 14954, 13572, 13673, 14714, 7564,  9717,  14139, 12563, 14969, 12041, 15183, 14301, 14382, 14828, 15130,
      14505, 14363, 10760, 13472, 12977, 15338, 15199, 14584, 14846, 14324, 12027, 15346, 14523, 13870, 13064, 13332,
      13485, 14701, 13714, 13091, 14463, 14509, 14562, 11792, 15151, 15324, 12958, 13065, 15243, 13239, 15350, 12792,
      13439, 15105, 15022, 11235, 15296, 13945, 14790, 14394, 13123, 15251, 14424, 13850, 15040, 10156, 14728, 14653,
      13784, 13583, 14000, 15004, 15110, 15241, 14464, 12955, 14795, 15179, 15057, 14548, 12142, 14204, 14339, 14852,
      11507, 15108, 13474, 13761, 13802, 14374, 14918, 14610, 13878, 15338, 11663, 13099, 13066, 14252, 15254, 14581,
      13388, 13708, 13940, 12384, 15192, 14525, 15316, 14766, 12479, 15227, 7297,  12960, 15157, 12697, 13079, 15121,
      12680, 14262, 14308, 10448, 14260, 8657,  13722, 14656, 14369, 14400, 14838, 13624, 13529, 8447,  9413,  14858,
      11964, 13605, 12390, 14508, 13340, 15181, 15235, 13919, 14151, 14204, 15075, 15330, 13339, 9388,  15057, 14477,
      14029, 14185, 15295, 12946, 12571, 15094, 14492, 12319, 12556, 15210, 13587, 14448, 14693, 14812, 14077, 4628,
      14049, 15012, 15272, 14018, 13920, 15039, 14819, 15049, 8219,  15157, 14588, 13018, 8563,  14406, 12439, 14969,
      15124, 14341, 15111, 14712, 14384, 15184, 14832, 15038, 15212, 13771, 15150, 14718, 15113, 11440, 12559, 15188,
      14994, 12731, 12423, 14827, 10994, 13780, 15293, 15005, 10856, 14405, 13899, 14340, 13822, 15261, 14717, 14426,
      14240, 14409, 13899, 15087, 15056, 13786, 14197, 14851, 15074, 12401, 15096, 13457, 15233, 14686, 14366, 11804,
      14755, 14701, 14937, 9540,  11746, 14478, 15083, 14148, 14683, 11412, 14585, 14639, 13010, 15113, 12314, 14873,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14521, 14041, 13757, 15160, 15311, 10812, 14911, 13180, 14396, 15068, 15157, 12268, 14916, 14896, 15259, 14839,
      14816, 13398, 13859, 15227, 14596, 15092, 14842, 14530, 14260, 13213, 15051, 14141, 14495, 14633, 15262, 14959,
      13207, 12909, 7459,  13892, 12596, 12119, 10983, 12957, 14531, 13544, 14509, 15115, 13785, 13379, 13939, 14480,
      15343, 9636,  10720, 13944, 11941, 13620, 12337, 14093, 11449, 14929, 13804, 14913, 14288, 12117, 13934, 14156,
      15329, 10598, 13388, 14816, 14286, 15191, 13799, 15118, 14854, 10596, 14308, 14981, 13629, 14640, 14988, 12668,
      12333, 11533, 13140, 15177, 14705, 13552, 13724, 14799, 13199, 15279, 15194, 13727, 13375, 15246, 12346, 13047,
      15094, 13587, 13719, 14039, 14223, 14334, 9141,  13058, 13097, 14011, 14735, 13165, 14608, 13898, 14846, 15077,
      14801, 14569, 8896,  11603, 14566, 15100, 14442, 13576, 15088, 12282, 14433, 15299, 13872, 13745, 13441, 14853,
      12404, 12848, 14324, 12418, 13344, 13374, 14484, 12633, 14552, 10802, 13784, 12116, 12288, 14760, 12854, 11920,
      15094, 11698, 10655, 11817, 11286, 12912, 15163, 14920, 14672, 14529, 13075, 13138, 13504, 14404, 13522, 14504,
      14969, 11547, 14734, 14905, 15107, 13908, 9897,  13781, 8014,  14788, 14205, 10000, 12874, 13865, 14373, 14442,
      12877, 9398,  15062, 14598, 12366, 13106, 15232, 14749, 14962, 15242, 14564, 15028, 14269, 10100, 15054, 13584,
      13375, 14611, 11912, 14210, 14297, 13091, 15098, 14903, 15187, 14270, 13106, 12456, 14514, 10900, 11718, 15114,
      13413, 13196, 15113, 9223,  14720, 14567, 14824, 14460, 14432, 15139, 14575, 14995, 14689, 13067, 9113,  15235,
      14282, 14710, 15017, 14294, 12492, 13506, 11041, 13773, 13757, 9798,  12974, 14446, 12195, 9658,  13742, 14931,
      14542, 14320, 12161, 15116, 12761, 14822, 13260, 15196, 14983, 15271, 12995, 13760, 11677, 15335, 15026, 15032,
      14872, 13753, 15275, 14934, 15101, 13562, 14657, 11250, 13622, 15131, 15102, 15186, 14419, 14856, 14402, 11097,
      15235, 14167, 9888,  14476, 13534, 12744, 14564, 12131, 15200, 13018, 13109, 13829, 14142, 14821, 14718, 11350,
      14263, 14819, 14899, 11836, 15002, 12015, 13231, 13408, 11930, 13611, 14337, 15196, 13508, 12894, 11714, 15190,
      14151, 15326, 15287, 14572, 14340, 14503, 14626, 15076, 14068, 12422, 14483, 15212, 13402, 14674, 14647, 14210,
      14530, 15210, 12721, 13927, 15354, 12381, 15243, 14850, 15076, 15261, 10670, 13510, 13559, 10514, 14573, 8620,
      10362, 14784, 10743, 14881, 13378, 13660, 15118, 14459, 14676, 14387, 14999, 14541, 13982, 14125, 13918, 13424,
      14744, 10829, 15230, 11625, 13706, 14871, 15211, 14743, 15015, 13199, 11460, 12476, 13403, 15277, 13169, 15293,
      13554, 12442, 14985, 14066, 15210, 14791, 15323, 14265, 14558, 14069, 14994, 15037, 15105, 12346, 14842, 14255,
      12820, 12654, 15235, 14283, 13623, 14493, 12396, 13549, 13894, 14453, 14552, 15102, 12439, 12825, 12694, 13811,
      10065, 14513, 14812, 13473, 12791, 14328, 15066, 15048, 11463, 14968, 11341, 14479, 15234, 14784, 14348, 12320,
      12833, 15018, 12601, 15296, 14451, 14974, 14719, 14989, 14241, 12347, 14848, 10659, 13709, 12121, 15087, 14941,
      15149, 14270, 15081, 14664, 11446, 13555, 13773, 9525,  12308, 15295, 14422, 13152, 12539, 15244, 13480, 14684,
      14812, 14554, 14730, 13256, 13691, 12453, 6673,  14755, 15323, 14365, 14017, 14530, 13733, 14433, 14073, 14792,
      14318, 15157, 14565, 11757, 11358, 14243, 13404, 14209, 14693, 12953, 14188, 14867, 12148, 9734,  10471, 14921,
      11834, 14153, 13803, 12527, 13412, 14278, 14379, 10916, 15165, 14495, 13531, 13432, 14209, 15262, 11705, 12347,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15114, 13518, 13980, 12685, 13682, 14272, 13538, 15247, 14919, 14981, 14820, 14814, 12429, 15300, 15148, 14391,
      12449, 15326, 10562, 13005, 10240, 12977, 13699, 12957, 14309, 14718, 13593, 13903, 14501, 14105, 13537, 13399,
      15004, 14600, 14481, 11560, 12340, 14614, 14917, 10587, 13746, 14556, 11536, 15142, 13452, 12608, 14624, 15003,
      13146, 14870, 14150, 15225, 14224, 15215, 15046, 15346, 14393, 15237, 15270, 15042, 14430, 14683, 14882, 13610,
      14700, 12601, 15178, 13571, 14787, 15318, 12600, 13036, 14689, 14263, 15030, 13271, 10435, 12379, 12355, 14546,
      14677, 14471, 13314, 15305, 12584, 14631, 11957, 14793, 15038, 8991,  14965, 15046, 13934, 13647, 13532, 14115,
      14380, 13663, 13892, 13726, 13892, 15344, 15272, 14946, 15288, 11327, 12495, 14350, 15288, 13236, 15099, 13013,
      12867, 14925, 11699, 12644, 11685, 14909, 15264, 14874, 11388, 14071, 12527, 12382, 14128, 14680, 11357, 14657,
      15323, 14513, 13166, 15126, 13490, 12492, 14094, 15206, 14004, 14287, 10214, 12793, 13024, 14627, 7760,  15161,
      10408, 12813, 12688, 15255, 13870, 13770, 13946, 15027, 11218, 13474, 14726, 14544, 12516, 13443, 13627, 14660,
      12360, 14074, 14088, 15354, 13954, 13999, 12297, 12222, 14771, 15322, 13871, 13605, 14143, 7336,  14738, 14951,
      13706, 12842, 14169, 13381, 14768, 13068, 13489, 12370, 9925,  14721, 11967, 15162, 14758, 15157, 15129, 14126,
      13267, 14922, 14078, 14992, 15240, 14910, 11827, 14301, 14121, 14654, 14724, 11465, 14836, 14496, 13914, 14521,
      13585, 13562, 13179, 14937, 14686, 13899, 14493, 13764, 13087, 15281, 13444, 13940, 13250, 14737, 14233, 12423,
      13957, 11705, 15191, 13076, 14108, 14963, 14682, 14005, 15184, 14534, 12839, 15109, 14780, 15302, 15102, 14704,
      11798, 14783, 14588, 15198, 15147, 12612, 10312, 13328, 14922, 15029, 10556, 14484, 12727, 13751, 14369, 14926,
      15310, 13435, 10187, 14401, 15200, 12906, 15279, 15019, 13355, 14651, 13387, 14433, 14551, 13391, 14118, 14722,
      13990, 5843,  12131, 15192, 15148, 14414, 14604, 15079, 12226, 7803,  10866, 14738, 14102, 13970, 15067, 14543,
      14180, 15247, 11606, 13890, 11493, 13911, 14480, 13782, 13709, 13827, 13780, 11400, 15072, 12473, 14587, 11537,
      13455, 14526, 15094, 12745, 9218,  15088, 15056, 15208, 14648, 14387, 14917, 15032, 12307, 9955,  12872, 10061,
      14633, 14675, 14866, 15325, 13700, 12137, 14069, 15208, 13356, 13354, 13682, 14550, 9415,  12219, 14245, 14355,
      14609, 14134, 14551, 14889, 14736, 13331, 12648, 11684, 14757, 14961, 14788, 15216, 15263, 14019, 11677, 15117,
      15101, 13684, 14919, 14932, 12004, 13786, 14691, 7719,  15137, 12794, 13755, 14670, 14884, 6460,  11961, 13286,
      14853, 12137, 13602, 14396, 15333, 11885, 15354, 14870, 15033, 15355, 15008, 14664, 15104, 14911, 14301, 11977,
      14343, 12836, 14766, 12698, 14500, 14934, 12836, 14637, 14471, 14192, 15029, 14811, 12884, 14442, 12308, 14963,
      11497, 14934, 12533, 14256, 14962, 13988, 15160, 15042, 13813, 15163, 14976, 14372, 14674, 14915, 15000, 12889,
      14454, 14337, 12817, 15282, 11888, 14712, 12716, 12968, 14828, 11331, 14553, 14960, 11745, 14426, 13329, 15029,
      12324, 15039, 9540,  13603, 13163, 13497, 14905, 13530, 14697, 15327, 13809, 14862, 15121, 13496, 13204, 14345,
      11908, 15058, 11816, 14434, 13870, 13792, 13100, 14035, 13718, 14104, 10908, 11752, 14814, 15270, 14354, 15227,
      14864, 14910, 14112, 14770, 15135, 15286, 14790, 12491, 10339, 10311, 12666, 13673, 13661, 15197, 15228, 13740,
      14016, 14858, 14433, 13865, 14953, 14904, 12803, 15006, 13634, 13596, 13522, 11651, 8251,  14412, 15116, 13971,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13281, 14388, 14653, 14235, 14949, 14756, 13327, 14157, 8095,  14198, 12592, 14315, 14425, 14529, 14431, 13207,
      15309, 14636, 14993, 13322, 14044, 15251, 14487, 11736, 14643, 14772, 14514, 11930, 14370, 9144,  14312, 14996,
      13933, 14605, 14132, 14396, 12388, 12102, 15068, 13772, 13114, 14727, 13750, 15137, 14107, 10686, 13506, 14924,
      14572, 15097, 14554, 15329, 15000, 15066, 11758, 8366,  15122, 13639, 11621, 10547, 14410, 14117, 14278, 14783,
      12150, 12072, 14949, 12702, 14558, 13354, 13587, 14553, 11582, 12985, 12942, 14217, 11719, 10286, 13961, 14120,
      12922, 15223, 14233, 14738, 14341, 14852, 14378, 13464, 15071, 15134, 13977, 13730, 14450, 14537, 11988, 13453,
      14846, 15224, 14809, 14926, 10310, 13068, 14659, 14381, 13033, 13457, 15043, 14285, 14822, 14320, 14747, 15336,
      12970, 11143, 15268, 14511, 14287, 14914, 12641, 13277, 14756, 12932, 11826, 14039, 13819, 15342, 15009, 11671,
      14493, 15019, 14550, 14406, 14474, 12501, 14314, 14509, 14271, 12944, 7321,  14959, 14552, 13422, 14685, 14722,
      15248, 13906, 15278, 13789, 12703, 12355, 14521, 13515, 14498, 9958,  15012, 15088, 14775, 14876, 14482, 14983,
      15083, 13026, 15222, 13686, 15189, 10315, 12789, 13710, 14859, 11598, 13848, 14604, 9806,  15131, 11235, 14405,
      7898,  14241, 11721, 14579, 13142, 13406, 13562, 14886, 15058, 13923, 12291, 13886, 14662, 14492, 8785,  14705,
      13319, 14371, 13297, 14752, 14933, 13149, 14319, 15048, 10985, 15077, 14658, 15309, 15038, 14519, 14080, 13682,
      11960, 12674, 14814, 13905, 14884, 12316, 14700, 12754, 14468, 14064, 15130, 11870, 13111, 14785, 13544, 14927,
      10080, 14046, 12328, 13732, 14914, 13920, 10934, 15350, 14796, 13997, 14814, 10346, 14982, 12412, 11890, 13211,
      7608,  15091, 14696, 13887, 15345, 13824, 14880, 14774, 14041, 13713, 15009, 15119, 14820, 14655, 12334, 14429,
      14259, 11618, 12647, 15121, 14091, 14538, 13483, 15153, 15303, 11594, 14800, 14366, 11562, 14498, 15267, 14114,
      12398, 14270, 13029, 15225, 14450, 15071, 15015, 13722, 14481, 13872, 11952, 13001, 14710, 14538, 14735, 15116,
      14141, 15168, 14990, 13575, 14597, 10412, 13787, 13336, 15341, 14384, 14827, 13904, 14229, 14331, 14880, 14209,
      14378, 15247, 13762, 15124, 13800, 14845, 14774, 12305, 15287, 14592, 13862, 9171,  13419, 14593, 15168, 12761,
      14663, 14776, 14976, 13314, 15073, 12674, 14822, 14765, 15289, 13832, 14804, 14411, 15166, 14909, 14581, 14703,
      14984, 13050, 15107, 14908, 14379, 15357, 14384, 12170, 15186, 13363, 14094, 15131, 14898, 14342, 15090, 15226,
      14340, 14612, 12815, 13927, 13914, 14786, 14459, 12572, 12060, 10287, 14665, 13979, 15070, 9430,  13653, 11688,
      13386, 14209, 14062, 14526, 13840, 13974, 12540, 14728, 15181, 13579, 15201, 13679, 14547, 10079, 8373,  15089,
      14652, 15218, 13794, 14695, 14495, 11434, 13729, 14457, 14750, 13953, 14900, 14345, 14743, 8968,  13501, 14411,
      15088, 13765, 14305, 13611, 11454, 12431, 13916, 15239, 15101, 14880, 15032, 14959, 13428, 15156, 12868, 14880,
      14404, 10812, 13679, 15352, 13889, 14393, 14057, 14521, 11904, 13946, 13475, 15140, 12184, 12043, 5618,  13417,
      15095, 14830, 14417, 14349, 14508, 12684, 14265, 14503, 14177, 14799, 12752, 15176, 11621, 14074, 14476, 13631,
      14996, 14413, 14905, 15034, 15041, 14168, 12507, 15127, 14898, 11142, 12775, 14944, 14429, 13475, 14620, 15249,
      14723, 11573, 15098, 12652, 14089, 14582, 14455, 14365, 13275, 13550, 15278, 12935, 14109, 14868, 15056, 15294,
      14365, 14405, 11798, 13938, 12237, 15015, 13686, 12382, 14689, 13340, 15273, 11760, 15173, 11573, 11228, 14127,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14976, 15320, 14856, 14747, 14545, 14933, 12942, 12746, 14663, 14383, 14315, 15076, 12981, 12639, 14651, 12301,
      13831, 13248, 13561, 14089, 7133,  15107, 15243, 14283, 13685, 13837, 13150, 15097, 15066, 15321, 14754, 15065,
      15145, 14604, 9350,  14595, 14753, 11829, 11989, 14282, 15170, 13365, 11849, 15026, 15063, 14496, 14898, 13614,
      14455, 11490, 14996, 13146, 15316, 13678, 12519, 12904, 15215, 10141, 14359, 12764, 14952, 14945, 14318, 12545,
      15021, 13737, 12487, 14273, 15306, 14737, 14678, 13369, 14656, 14651, 11771, 14753, 15329, 11653, 14861, 13370,
      14876, 13650, 14235, 14345, 14666, 11985, 15343, 10510, 14125, 13014, 12078, 14598, 14707, 14807, 14031, 14574,
      14259, 12994, 12920, 14841, 12705, 12426, 10228, 12535, 13332, 14389, 14773, 12520, 14113, 11590, 13318, 14054,
      14045, 15238, 14760, 14441, 13800, 12903, 14482, 14181, 15116, 14771, 14710, 12219, 14550, 15131, 12956, 15206,
      14452, 14133, 14825, 14782, 13950, 14989, 14822, 14550, 14960, 13546, 15291, 15047, 14721, 12796, 13049, 13984,
      15339, 13726, 14888, 13541, 14481, 11449, 15145, 15314, 14862, 14949, 15320, 15218, 13523, 13857, 14019, 14322,
      14519, 13801, 13015, 10633, 14755, 11690, 14555, 14487, 14256, 13564, 12702, 11090, 15349, 13077, 14269, 15188,
      13954, 14889, 14325, 13875, 14970, 13963, 14442, 14405, 13956, 14661, 14439, 14251, 14473, 14590, 15279, 15312,
      15103, 15129, 12912, 14690, 14609, 13013, 14625, 14788, 14695, 13120, 14112, 13991, 14719, 13317, 14953, 14861,
      12677, 14062, 13603, 11757, 13703, 13908, 12437, 9314,  12550, 8987,  14382, 11060, 15021, 14353, 14740, 14386,
      15262, 9436,  14217, 13844, 15278, 13635, 13995, 14881, 15223, 14603, 14039, 14987, 14338, 14804, 15039, 13950,
      15027, 14131, 15027, 14733, 12575, 14899, 15147, 14539, 14019, 14406, 11221, 14913, 15175, 15293, 13944, 9951,
      14276, 15309, 11477, 13522, 9043,  14880, 14893, 10032, 12896, 12839, 13480, 15326, 8593,  10118, 15130, 14666,
      12737, 10912, 14774, 12419, 14430, 14359, 13123, 12935, 15153, 14394, 14949, 14113, 14684, 13348, 15086, 12414,
      13458, 14803, 15243, 12783, 11518, 14356, 14634, 14364, 13764, 14562, 13552, 11792, 13272, 14166, 13093, 14020,
      14673, 11360, 14245, 14856, 14758, 13310, 14676, 12681, 11917, 15018, 12183, 13263, 13681, 14529, 12440, 15073,
      15030, 11081, 14591, 11958, 15136, 13898, 14443, 13995, 14057, 10504, 13954, 13927, 12744, 14617, 14437, 15311,
      15004, 7602,  15221, 15353, 14844, 14487, 15278, 14855, 13187, 11666, 12805, 12928, 12768, 14658, 7348,  13047,
      11651, 12457, 14692, 12098, 14889, 13845, 14825, 15299, 14087, 12794, 15135, 13823, 14469, 14422, 13172, 13470,
      14456, 14897, 14513, 12444, 15330, 13965, 13733, 14928, 13571, 12734, 14482, 14512, 15152, 13734, 12670, 14767,
      12819, 12825, 14931, 13675, 13686, 12069, 13955, 15206, 14125, 13530, 14171, 11805, 14993, 14774, 15087, 14289,
      14288, 11321, 14504, 15098, 13589, 13506, 15000, 14225, 14552, 10865, 13645, 10572, 14811, 13932, 14066, 14928,
      13085, 13980, 10430, 15270, 13851, 14104, 11861, 12539, 13621, 15241, 15220, 9377,  15000, 14933, 11082, 13061,
      14605, 12090, 11294, 14511, 13747, 14102, 14932, 15220, 12548, 14598, 13892, 15322, 15107, 11542, 15160, 10795,
      15081, 14780, 13869, 14676, 12543, 15249, 14737, 13867, 14420, 14356, 15208, 15003, 13838, 14914, 13949, 15354,
      13796, 15292, 13661, 14913, 14461, 14694, 12464, 13470, 14888, 11147, 15148, 12601, 14848, 14154, 10547, 15327,
      14875, 14520, 14455, 13110, 15166, 13329, 15242, 12645, 14231, 9729,  13808, 14604, 12699, 14319, 14690, 14702,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12899, 14145, 11482, 14941, 14311, 11793, 15308, 13356, 15285, 14843, 14161, 14144, 14181, 13929, 15262, 15282,
      13569, 14586, 13484, 14018, 14595, 12157, 11768, 14595, 15219, 13311, 15300, 14819, 15228, 5703,  15260, 14424,
      13975, 14023, 15247, 11991, 12462, 15047, 14329, 14918, 14216, 14559, 11484, 14921, 14635, 14301, 13155, 14491,
      15118, 14144, 12558, 14257, 12490, 13843, 15011, 13518, 15244, 14844, 14244, 11510, 14862, 10261, 13492, 8340,
      12616, 14608, 14084, 14340, 13877, 14281, 14336, 10026, 12765, 15152, 13970, 14958, 15249, 14512, 13765, 13705,
      14586, 12913, 14076, 15062, 14704, 14657, 13504, 11789, 13853, 12733, 15155, 14505, 10599, 12417, 13781, 14709,
      14560, 12499, 12124, 13801, 13898, 14684, 9283,  13609, 13356, 14633, 14525, 14352, 14154, 14939, 14956, 15259,
      12805, 12921, 14092, 14138, 14778, 11760, 14176, 12487, 14835, 13389, 15044, 13351, 13906, 14751, 11136, 12328,
      14690, 15177, 13000, 13450, 14235, 12819, 14293, 11417, 11860, 13690, 13293, 14079, 14799, 15218, 14458, 13535,
      12326, 14279, 14893, 11920, 14464, 9670,  15145, 11534, 12507, 12900, 14952, 15346, 13570, 10871, 10659, 12697,
      15218, 14874, 13795, 11730, 14170, 14592, 14551, 15168, 12243, 14472, 13646, 14970, 14591, 14193, 15329, 13948,
      14947, 12265, 11429, 12986, 14938, 11598, 14570, 14835, 14436, 14768, 13390, 14957, 13054, 14693, 15213, 13255,
      14467, 4195,  12705, 14166, 14194, 15188, 14352, 15112, 13565, 12581, 14909, 12447, 13551, 12369, 12856, 13661,
      14803, 12494, 14553, 11957, 15300, 10307, 11859, 13746, 15120, 11608, 14829, 15028, 14252, 12338, 13494, 14231,
      12999, 14335, 14031, 12662, 14483, 15057, 13938, 8440,  14639, 14175, 15089, 10597, 13834, 15083, 15312, 14730,
      15270, 15034, 10214, 14149, 12903, 13272, 10594, 15231, 13343, 14936, 14773, 15295, 12305, 14615, 13113, 7949,
      14883, 14394, 15307, 13544, 14360, 14421, 14897, 14917, 11127, 12531, 14685, 12592, 15233, 15031, 14239, 15016,
      15320, 15334, 14753, 12701, 14847, 15179, 14850, 14042, 13791, 14559, 11986, 15197, 14747, 15275, 13160, 14448,
      14344, 15001, 14447, 13626, 14376, 13635, 13367, 13908, 14136, 11662, 13239, 15069, 11325, 14682, 9847,  14223,
      14973, 15085, 15009, 14663, 10946, 12595, 13998, 9601,  14156, 11295, 14518, 12537, 10824, 14509, 14633, 12603,
      14325, 14560, 12694, 11881, 15345, 15215, 12140, 14573, 14656, 13104, 14137, 14680, 14727, 14107, 14586, 14857,
      11186, 15053, 14947, 15277, 13652, 13831, 12764, 14462, 14954, 15013, 14735, 14759, 14923, 8551,  14570, 15358,
      14816, 14734, 14971, 15208, 12952, 12846, 14985, 14813, 8634,  15070, 12693, 14531, 7817,  14175, 15264, 14279,
      6596,  14706, 14619, 14492, 10603, 13054, 13316, 14601, 14544, 12521, 12599, 14599, 14708, 13769, 14922, 14526,
      13172, 13819, 10934, 14303, 14878, 8124,  13354, 14921, 14543, 13798, 15034, 14823, 12955, 14768, 12814, 15046,
      12731, 11502, 15217, 14096, 15056, 13701, 14933, 13839, 13640, 13472, 13642, 15028, 15075, 14471, 14484, 13619,
      14529, 14757, 14998, 13596, 14590, 13665, 15055, 13691, 14136, 14868, 13456, 13010, 13747, 13788, 14739, 14456,
      14842, 15239, 10946, 14313, 15197, 12329, 13187, 15178, 14432, 13369, 14970, 13321, 12521, 14943, 11130, 14829,
      14687, 14478, 14998, 14565, 14567, 13773, 15297, 13552, 14613, 13802, 15311, 10983, 14762, 14291, 8820,  15110,
      15056, 14468, 8369,  14950, 15161, 11667, 14952, 15068, 13913, 12478, 13105, 13826, 12650, 13910, 13450, 13828,
      14882, 14228, 14588, 14683, 15289, 14730, 13633, 14851, 12290, 14166, 10686, 14612, 14558, 14602, 11166, 11270,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15129, 12633, 12271, 15335, 14478, 13558, 14348, 12826, 11074, 14488, 14704, 8808,  13510, 12700, 15345, 13496,
      15207, 14772, 14952, 13082, 14483, 14744, 13120, 13042, 12760, 13895, 13191, 13038, 12863, 15248, 14447, 15060,
      12681, 9499,  14706, 13362, 14831, 13935, 10062, 14542, 14510, 15345, 13422, 14830, 12740, 12078, 13737, 13762,
      14374, 14772, 14502, 12734, 13597, 14395, 14447, 14932, 15197, 14258, 12867, 15233, 13123, 13901, 14823, 14414,
      12598, 12265, 13458, 13322, 15095, 12678, 14753, 12056, 14876, 14289, 11887, 11132, 14634, 14333, 14876, 14487,
      15255, 13232, 9419,  13113, 15125, 13709, 8430,  14475, 14072, 14976, 13150, 13565, 11062, 15349, 13720, 15089,
      14169, 14668, 14598, 14397, 14420, 14360, 13381, 12958, 13602, 11956, 13638, 15086, 14589, 14067, 11622, 14921,
      13278, 14506, 15108, 11403, 11014, 14973, 12967, 11848, 13624, 14924, 15303, 15065, 11164, 14361, 13992, 14433,
      14085, 13432, 13679, 12375, 14695, 14953, 12370, 12341, 15181, 14370, 13683, 13684, 14325, 13300, 13569, 14682,
      14868, 14656, 14537, 15188, 9057,  13714, 14465, 14510, 12882, 13514, 15251, 14690, 14821, 15016, 14776, 14634,
      14755, 14577, 14514, 14490, 14143, 12098, 15029, 14748, 14516, 14139, 9531,  13087, 15199, 14901, 12515, 15335,
      11315, 15279, 15326, 13345, 13394, 14553, 10464, 13564, 15067, 13190, 15029, 13106, 12338, 15167, 14579, 12368,
      14676, 14490, 14064, 15245, 14118, 14384, 15154, 14798, 15267, 13168, 14834, 14392, 14985, 13895, 13405, 15033,
      14648, 13778, 12922, 9520,  13830, 14078, 13337, 15191, 14099, 14020, 15230, 13925, 14606, 14374, 7937,  15081,
      14374, 15336, 8978,  14746, 14837, 12443, 14773, 11460, 14236, 8503,  14702, 14511, 13591, 13645, 12096, 14453,
      13910, 14809, 9859,  11043, 15091, 12568, 13063, 12664, 12483, 12531, 13754, 15352, 15089, 13728, 11477, 12352,
      12859, 8557,  13822, 13402, 13676, 10691, 15003, 11804, 14072, 11966, 15007, 13977, 14454, 14459, 11890, 14098,
      13689, 14974, 11621, 11774, 12809, 14080, 14352, 12959, 13544, 14411, 12972, 15266, 14629, 15013, 15166, 14163,
      13829, 14381, 14666, 15249, 14878, 14485, 15023, 11462, 14618, 15282, 9025,  10056, 12482, 14399, 11935, 14529,
      15329, 10444, 10458, 15164, 14283, 15047, 15122, 15061, 15336, 13797, 15325, 14232, 15121, 13841, 14858, 14648,
      15091, 12648, 14668, 15276, 10408, 10465, 15050, 14160, 13669, 14655, 15008, 15149, 15034, 14635, 15302, 13467,
      15279, 15145, 13339, 15047, 8492,  13484, 14358, 14790, 14121, 15256, 14492, 14495, 14379, 13603, 15282, 13101,
      14199, 15169, 14894, 14209, 14965, 15284, 14068, 9544,  13332, 15286, 14710, 11613, 15354, 13708, 13892, 13712,
      13578, 14036, 14619, 14832, 12956, 14467, 13883, 11593, 13838, 15191, 14717, 12866, 13553, 11320, 13374, 13573,
      9403,  15343, 13873, 14781, 9553,  13937, 12206, 15322, 14692, 14836, 12846, 14908, 15352, 14397, 14597, 14143,
      14374, 15102, 13916, 14190, 13450, 15129, 14920, 11822, 15233, 14898, 14593, 12774, 14934, 14347, 14450, 15266,
      15297, 9337,  14273, 15174, 9064,  14082, 15048, 15183, 14733, 12449, 14415, 14596, 13726, 15278, 14017, 14580,
      15231, 13321, 13312, 14997, 13352, 15232, 14729, 15251, 13066, 15096, 13439, 13541, 15199, 13725, 14965, 15104,
      14034, 9564,  12353, 14342, 11883, 13095, 12822, 14708, 15247, 14645, 13155, 14587, 14819, 13913, 14674, 13300,
      8992,  14810, 14241, 14566, 14986, 14578, 15054, 15239, 14988, 14905, 14537, 13648, 14056, 14587, 14932, 12516,
      15270, 15241, 14960, 11623, 12726, 14406, 14502, 13488, 14570, 10964, 13249, 10442, 12306, 13676, 15122, 13435,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14065, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14792, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13601, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14857, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14208, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11756, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14946, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14318, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14096, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11157, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12292, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14480, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15096, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12208, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11920, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14342, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12946, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14351, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14558, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14744, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12591, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15111, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15067, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14394, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14413, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14172, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14355, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12605, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14499, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14809, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12525, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13863, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12424, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14831, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15236, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12827, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14985, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13782, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15242, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13631, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15262, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10437, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14845, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13724, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12354, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13559, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14110, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14780, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15227, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14348, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14242, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15346, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12389, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15251, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      4533,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15102, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14669, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12679, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13779, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13342, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11953, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13642, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15014, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      9458,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10198, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14503, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13700, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13921, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14515, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15334, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15096, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13184, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14963, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13778, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14597, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14753, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14499, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14867, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13995, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12477, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13580, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14033, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15062, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13503, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15119, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13305, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13398, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14639, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14383, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13644, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12744, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14881, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15082, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14777, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10585, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14248, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12692, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15161, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14319, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15163, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14105, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      9218,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14070, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13320, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15055, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14216, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14930, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15247, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15110, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14353, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10914, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13413, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14450, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13480, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14669, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15000, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15323, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14488, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14227, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14773, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14973, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12545, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14904, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15003, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14892, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14736, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14119, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13097, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15127, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15095, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15209, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13293, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15347, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14294, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12041, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13481, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15168, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13302, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15192, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11879, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14714, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15251, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13897, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14353, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14687, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14978, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12990, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14778, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14613, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      6439,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13982, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12487, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14793, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13660, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14699, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12814, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13592, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13918, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13652, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13212, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15048, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14635, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14587, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13791, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14928, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14768, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15199, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15191, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15027, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14549, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13263, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14210, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15165, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12100, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15096, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14993, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10874, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11939, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14439, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14445, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15124, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15356, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15129, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12379, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13445, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12820, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13125, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14694, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11619, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14563, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14851, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      9072,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13588, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13374, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10685, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14581, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12774, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14392, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13811, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13815, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14784, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      9198,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14839, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12029, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14362, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14537, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13695, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14451, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15068, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15245, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14640, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15348, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14015, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15117, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13956, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14110, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12100, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15044, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14952, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14970, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13493, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15299, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14221, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14531, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13929, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14481, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14592, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14656, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14377, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14961, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15317, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15227, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14179, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14827, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13639, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13759, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14608, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12358, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14905, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14642, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14787, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14237, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14005, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13770, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12943, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12976, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14393, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14302, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14655, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14586, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14298, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13275, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11610, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14687, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13896, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14496, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13793, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14384, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13852, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13732, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13695, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13396, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15195, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11654, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14856, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14742, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13681, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14148, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13082, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11598, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14549, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15065, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13787, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14229, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      9833,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13955, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13531, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13912, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14542, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15189, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14830, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13402, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14246, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13495, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13538, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15002, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15010, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13429, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12975, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13624, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12418, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14493, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12925, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14323, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10586, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14383, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14160, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12425, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14565, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15275, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12711, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11424, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14569, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14745, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13329, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15118, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11945, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14448, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14916, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14338, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14590, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12547, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14792, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14919, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14632, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15146, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14128, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13529, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14307, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15314, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11927, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13193, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12913, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14072, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14846, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14363, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15183, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13458, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11316, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15006, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13998, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12457, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14080, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14737, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13997, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14464, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13012, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13586, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14871, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12206, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14005, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12415, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14625, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15078, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13941, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13567, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15349, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12954, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13139, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15253, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13474, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15193, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10893, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10037, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15178, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15217, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14853, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14891, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14071, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12289, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14341, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11465, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12509, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15241, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14393, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14062, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14187, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14753, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15099, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14075, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12822, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10563, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13728, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12400, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14575, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14055, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15011, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10473, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13115, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11919, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14798, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13877, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14651, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14093, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14333, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12934, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14526, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13663, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13669, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12694, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10221, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14892, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15126, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14958, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14436, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13436, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14203, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12725, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13754, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14161, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15342, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11444, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15262, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14926, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14950, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14180, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10566, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14351, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15117, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12727, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14729, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14351, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13756, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15267, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14900, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14541, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15250, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14757, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15132, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12011, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13651, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14624, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14063, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15068, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13835, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11489, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11410, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14745, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      10842, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14661, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14857, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14987, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13569, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12763, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14932, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14851, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14120, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13511, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13677, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14543, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14605, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15036, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13315, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14657, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14617, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14875, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14488, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13788, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14421, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15000, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13867, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15170, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14389, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14577, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15337, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15245, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14764, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13328, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14910, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13327, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13381, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11468, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11271, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12719, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15023, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14764, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14374, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13085, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14873, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13749, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14737, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14765, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14897, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13562, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11925, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14867, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14287, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14233, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      15094, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14490, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11202, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14407, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13738, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13849, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14813, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11320, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14512, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13608, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14363, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      11696, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      9268,  0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14793, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      12674, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      13757, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14622, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      14169, 0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
      0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,     0,
  };

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data),    FORMAT_FRACTAL_Z, FORMAT_NCHW, std::vector<int64_t>({32, 2, 16, 16}),
      std::vector<int64_t>({31, 17, 4, 4}), DT_FLOAT16};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, 31 * 17 * 4 * 4 * 2);
  for (int i = 0; i < 31 * 17 * 4 * 4; ++i) {
    EXPECT_EQ((reinterpret_cast<uint16_t *>(result.data.get()))[i], ret[i]);
  }
}

TEST_F(UtestFormatTransferFraczNchw, fp32_1) {
  float ret[1 * 1 * 1 * 1] = {0.8329063818189503};
  float data[1 * 1 * 16 * 16] = {
      0.8329063818189503,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
      0.0,
  };

  FormatTransferFracZNchw transfer;
  TransArgs args{
      reinterpret_cast<uint8_t *>(data), FORMAT_FRACTAL_Z, FORMAT_NCHW, {1, 1, 16, 16}, {1, 1, 1, 1}, DT_FLOAT};

  TransResult result;
  EXPECT_EQ(transfer.TransFormat(args, result), SUCCESS);
  EXPECT_EQ(result.length, sizeof(ret) / sizeof(ret[0]) * 4);
  for (int i = 0; i < sizeof(ret) / sizeof(ret[0]); ++i) {
    EXPECT_EQ((reinterpret_cast<float *>(result.data.get()))[i], ret[i]);
  }
}
}  // namespace formats
}  // namespace ge